@s Global extern
@s png_structp int
@s png_infop int
@s png_color_8 int
@s png_bytep int
@ Write the contents of the frame buffer to
a portable network graphics bitmap.
@(savepng.h@>=
#ifdef MATHGL_MAIN
#  define Global
#else
#  define Global extern
#endif

@ The main code of the module.
@c
#undef MATHGL_MAIN
@<Header@>@;
@<Static data@>@;
@<Write a PNG file@>@;
@<Save a EPS bitmap of the frame buffer@>@;
@<Save a image as EPS bitmap@>@;
@<Save a PNG bitmap from a given bit array@>@;
@<Save OpenGL's depth buffer@>@;
@<Interpolate the $z$-values@>@;
@<Stereogram creation@>@;
@<Make a single image stereogram from the depth buffer@>@;
@<Stuff to clip the image@>@;
@<Utilities for textures@>@;

@ Save the image in the frame buffer as portable network graphic.

@(savepng.h@>=
int PNGSaveFrameBuffer(
      const char *filename,
      int x,int y,
      int width, int height,
      int interlace,
      double gamma);

@ @<Write a PNG file@>=
int PNGSaveFrameBuffer(
      const char *filename,
      int x,int y,
      int width, int height,
      int interlace,
      double gamma)
{
   FILE *fp;
   int bit_depth;
   GLubyte* image;
   png_structp png_ptr;
   png_infop info_ptr;
   int i,png_color_type;
   png_color_8 sig_bit;
   png_bytep *row_pointers;

   @<Create the storrage for the image@>@;
   @<Open the target file@>@;
   @<Initialize the writing of |png_struct|@>@;
   @<Get the memory for the image information@>@;
   @<Set the error handling@>@;
   @<Initialise the input control@>@;
   bit_depth=8*sizeof(GLubyte);
   png_color_type=PNG_COLOR_TYPE_RGB;
   sig_bit.red=sig_bit.green=sig_bit.blue=(unsigned char)bit_depth;
   @<Set the image information@>@;
   /* Optional gamma chunk is strongly suggested
    if you have any guess as to the correct gamma of the image. */
   png_set_gAMA(png_ptr, info_ptr, gamma);
   /* Write the file header information. */
   png_write_info(png_ptr, info_ptr);
   @<Setup the transformations@>@;

   png_write_image(png_ptr, row_pointers);
   png_write_end(png_ptr, info_ptr);
   @<Free the memory of the image@>@;
   /* clean up after the write, and free any memory allocated */
   png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
   /* close the file */
   fclose(fp);
   return GL_TRUE;
}

@ Open the binray file for writing.
@<Open the target file@>=
   fp = fopen(filename, "wb");
   if (fp == NULL) {
      return GL_FALSE;
     }

@ Create and initialize the |png_struct| with the desired error handler
  functions.  If you want to use the default |stderr| and |longjump| method,
  you can supply |NULL| for the last three parameters.  We also check that
  the library version is compatible with the one used at compile time,
  in case we are using dynamically linked libraries.

@<Initialize the writing of |png_struct|@>=
   png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
      (void *)NULL, NULL, NULL);
   if (png_ptr == NULL)
     @<Return in case of an error@>@;

@ Allocate/initialize the memory for image information.
@<Get the memory for the image information@>=
   info_ptr = png_create_info_struct(png_ptr);
   if (info_ptr == NULL)
     @<Return in case of an error@>@;

@ Set error handling if you are using the |setjmp|/|longjmp| method (this is
  the normal method of doing things with libpng).  REQUIRED unless you
  set up your own error handlers in the |png_create_read_struct()| earlier.
@<Set the error handling@>=
   if (setjmp(png_ptr->jmpbuf))
     @<Return in case of an error@>@;

@ Set up the input control if you are using standard C streams
@<Initialise the input control@>=
   png_init_io(png_ptr, fp);


@ Allocate the memory to hold the image using the fields of |info_ptr|.
@<Create the storrage for the image@>=
   row_pointers=calloc(height,sizeof(png_bytep));
   image=(GLubyte *)calloc(width*height*3,sizeof(GLubyte));
   for(i=0; i<height; i++)
     row_pointers[i]=image+3*width*(height-1-i);
   glPixelStorei(GL_PACK_ALIGNMENT, 1);
   glFlush();
   glFinish();
   glReadPixels(x, y, width, height, GL_RGB, GL_UNSIGNED_BYTE, image);

@ @<Free the memory of the image@>=
if(row_pointers) {
  free(image);
  free(row_pointers);
 }


@ Set the image information.  Width and height are up to $2^31$,
|bit_depth| is one of 1, 2, 4, 8, or 16, but valid values also depend on
the |color_type| selected. |color_type| is one of |PNG_COLOR_TYPE_GRAY|,
|PNG_COLOR_TYPE_GRAY_ALPHA|, |PNG_COLOR_TYPE_PALETTE|, |PNG_COLOR_TYPE_RGB|,
or |PNG_COLOR_TYPE_RGB_ALPHA|.  |interlace| is either |PNG_INTERLACE_NONE| or
|PNG_INTERLACE_ADAM7|, and the |compression_type| and |filter_type| {\it must}
currently be |PNG_COMPRESSION_TYPE_BASE| and |PNG_FILTER_TYPE_BASE|.
@<Set the image information@>=
 png_set_IHDR(png_ptr, info_ptr,
   width, height, bit_depth,
   png_color_type,
   interlace ? PNG_INTERLACE_ADAM7 : PNG_INTERLACE_NONE,
   PNG_COMPRESSION_TYPE_DEFAULT,
   PNG_FILTER_TYPE_BASE);
 png_set_sBIT(png_ptr,info_ptr,&sig_bit);

@ Set up the transformations you want.  Note that these are
  all optional.  Only call them if you want them.
@<Setup the transformations@>=
   /* pack pixels into bytes */
   png_set_packing(png_ptr);

@ @<Return in case of an error@>=
return GL_FALSE;

@ Since we will save the textures as PNG files for the VRML output we
need a general function for saving a bitmap. |PNGSaveBitmap| will work
with the image data supplied in |image|.
@ @(savepng.h@>=
int PNGSaveBitmap(
      const char *filename,
      int width, int height,
      int oglType,
      GLubyte *image,
      int interlace,
      double gamma);

@ @<Save a PNG bitmap from a given bit array@>=
int PNGSaveBitmap(
      const char *filename,
      int width, int height,
      int oglType,
      GLubyte *image,
      int interlace,
      double gamma)
{   FILE *fp;
   int bit_depth;
   int byte_per_pixel;
   png_structp png_ptr;
   png_infop info_ptr;
   int i,png_color_type;
   png_color_8 sig_bit;
   png_bytep *row_pointers;

   @<Determine color type@>@;
   @<Set the |row_pointers| to its positions@>@;
   @<Open the target file@>@;
   @<Initialize the writing of |png_struct|@>@;
   @<Get the memory for the image information@>@;
   @<Set the error handling@>@;
   @<Initialise the input control@>@;
   bit_depth=8*sizeof(GLubyte);
   if(oglType==GL_RGB || oglType==GL_RGBA)
     sig_bit.red=sig_bit.green=sig_bit.blue=(unsigned char)bit_depth;
   else
     sig_bit.gray=(unsigned char)bit_depth;
   if(oglType==GL_LUMINANCE_ALPHA || oglType==GL_RGBA)
     sig_bit.alpha=(unsigned char)bit_depth;
       
   @<Set the image information@>@;
   /* Optional gamma chunk is strongly suggested
    if you have any guess as to the correct gamma of the image. */
   png_set_gAMA(png_ptr, info_ptr, gamma);
   /* Write the file header information. */
   png_write_info(png_ptr, info_ptr);
   @<Setup the transformations@>@;

   png_write_image(png_ptr, row_pointers);
   png_write_end(png_ptr, info_ptr);
   free(row_pointers);
   /* clean up after the write, and free any memory allocated */
   png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
   /* close the file */
   fclose(fp);
   return GL_TRUE;
}

@ @<Set the |row_pointers| to its positions@>=
   row_pointers=calloc(height,sizeof(png_bytep));
   for(i=0; i<height; i++)
     row_pointers[i]=image+byte_per_pixel*width*(height-1-i);

@ @<Determine color type@>=
switch (oglType) {
  case GL_LUMINANCE:
    byte_per_pixel=1;
    png_color_type=PNG_COLOR_TYPE_GRAY;
    break;
  case GL_LUMINANCE_ALPHA:
    byte_per_pixel=2;
    png_color_type=PNG_COLOR_TYPE_GRAY_ALPHA;
    break;
  case GL_RGB:
    byte_per_pixel=3;
    png_color_type=PNG_COLOR_TYPE_RGB;
    break;
  case GL_RGBA:
    byte_per_pixel=4;
    png_color_type=PNG_COLOR_TYPE_RGB_ALPHA;
    break;
  default :
    return GL_FALSE;
}


@ The color palette from 
      \.{http://www.chromatek.com/color_palettes.html}
for ChromaDepth pictures.
@d MAX(a,b) ((a)>(b) ? (a) : (b))
@d MIN(a,b) ((a)<(b) ? (a) : (b))

@d CDRed(range) ((range)>0.75f || (range) <0.f ? 
                        0.                :
	      (((-2.13f*(range)-1.07f)*(range)+0.133f)*(range)+0.0667f)*(range)+1.f)
	      
@d CDGreen(range) ((range)<=0.5f ? (1.6f*(range)+1.2f)*(range) :(3.2f*(range)-6.8f)*(range)+3.6f)

@d CDBlue(range)  ((range)>0.5f ?((-4.8f*(range)+9.2f)*(range)-3.4f) : 0.0f)


@ Save the depth buffer as PNG file. The depth buffer can be saved as
  gray level or with ChromaDepth (http://www.chromatek.com) colors.
@(savepng.h@>=
#define DEPTH_ENC_GRAY  (0)
#define DEPTH_ENC_CDRGB (1)
#define DEPTH_ENC_CDCMY (2)
int PNGSaveDepthBuffer(const char *filename,
      int x,int y,
      int width, int height,
      int depthEncoding,
      int interlace,
      double gamma, double scale);


@ We will save the depth buffer as gray scale, or ChromaDepth
  image. The |depthEncoding| varable determines the type of the image.
  The depth buffer data are rescaled to get the maximum resolution
  of the depth data using the minimum distance |min| and the true
  depth |delta|.
  
@<Save OpenGL's depth buffer@>=
int PNGSaveDepthBuffer(const char *filename,
      int x,int y,
      int width, int height,
      int depthEncoding,
      int interlace,
      double gamma, double scale)
{ GLubyte *bimage,*bptr;
  GLfloat *dbuff,c;
  GLfloat oned,fmin,dmax,dmin;
  int oglType, i;
  long int size;
  int res;

  size=width*height;
  oglType=(depthEncoding>0 ? GL_RGB : GL_LUMINANCE);
  bimage=(GLubyte *)calloc((depthEncoding ? 3 : 1)*size,sizeof(GLubyte));

  @<Get the depth buffer in |dbuff|@>@;
  @<Unproject the $z$ values@>@;
  @<Scaling of the depth buffer@>@;
  @<Color transformation of the depth buffer@>@;  
  res=PNGSaveBitmap(filename,width,height,oglType,bimage,interlace,gamma);
  free(bimage);
  free(dbuff);
  return res;
}

@ @<Get the depth buffer in |dbuff|@>=
  dbuff=(GLfloat *)calloc(size,sizeof(GLfloat)); 
  glPixelStorei(GL_PACK_ALIGNMENT, 1);
  glReadPixels(x,y,width,height,GL_DEPTH_COMPONENT,GL_FLOAT,dbuff);


@ Since the projection is nonlinear we have to unproject the $z$
  values to get the true distance.
@<Unproject the $z$ values@>=
{ GLint i,j;
  GLint viewport[4];
  GLdouble mvmatrix[16], projmatrix[16];
  GLdouble x,y,z,wx,wy,wz;
  GLdouble vpx,vpy,vpz;
  
  glGetIntegerv(GL_VIEWPORT,viewport);
  glGetDoublev(GL_MODELVIEW_MATRIX,mvmatrix);
  glGetDoublev(GL_PROJECTION_MATRIX,projmatrix);
  gluUnProject(0.5*width,0.5*height,0.,
    mvmatrix,projmatrix,viewport,
    &vpx,&vpy,&vpz);
  for(i=0; i<width; i++) {
    x=(GLdouble) i; 
    for(j=0; j<height; j++) {
      y=(GLdouble) j;
      z=dbuff[width*j+i];
      gluUnProject(x,y,z,mvmatrix,projmatrix,viewport,&wx,&wy,&wz);
      wx-=vpx; @+ wy-=vpy; @+ wz-=vpz;
      z=sqrt(wx*wx+wy*wy+wz*wz);
      dbuff[width*j+i]=z;
     }
   }
}  

@ @<Scaling of the depth buffer@>=
  dmin=dmax=0.0f;
  for(i=0; i<size; i++)
    if(dbuff[i]>dmax) dmax=dbuff[i];
    else if(dbuff[i]<dmin) dmin=dbuff[i];
  dmin=dmax;
  for(i=0; i<size; i++)
    if(dbuff[i]>0.0 && dbuff[i]<dmin) dmin=dbuff[i]; 
  if(dmax!=dmin) {
    oned=1.f/(dmax-dmin); 
    for(i=0; i<size; i++) {
      dbuff[i]=scale*(dbuff[i]-dmin)*oned;
      if(dbuff[i]>1.0) dbuff[i]=1.0;
     }
   }   

@ @<Color transformation of the depth buffer@>=
  bptr=bimage;   
  switch (depthEncoding) {
    case  DEPTH_ENC_GRAY :
      for (i=0; i<size; i++)
        *(bptr++)=(GLubyte )(255.f*dbuff[i]);
      break;
    case  DEPTH_ENC_CDRGB  :
      for (i=0; i<size; i++) {
        c=CDRed(dbuff[i]);
	c=MAX(MIN(c,1.f),0.f);
        *(bptr++)=(GLubyte )(255.f*c);
        c=CDGreen(dbuff[i]);
	c=MAX(MIN(c,1.f),0.f);
        *(bptr++)=(GLubyte )(255.f*c);
        c=CDBlue(dbuff[i]);
	c=MAX(MIN(c,1.f),0.f);
        *(bptr++)=(GLubyte )(255.f*c);
       }
      break; 
    default  :
      for (i=0; i<size; i++) {
        c=CDRed(dbuff[i]);
	c=1.f-MAX(MIN(c,1.f),0.f);
        *(bptr++)=(GLubyte )(255.f*c);
        c=CDGreen(dbuff[i]);
	c=1.f-MAX(MIN(c,1.f),0.f);
        *(bptr++)=(GLubyte )(255.f*c);
        c=CDBlue(dbuff[i]);
	c=1.f-MAX(MIN(c,1.f),0.f);
        *(bptr++)=(GLubyte )(255.f*c);
       }
      break; 
   }   	

@ The depth buffer is read and the original depth
values are obtained by unprojection. The data fo the
height field are passed to the stereogram calculation.
The stereogram is returned in |bimage|

@(savepng.h@>+=int PNGSaveSIRDS(const char *filename,
      int x,int y,
      int width, int height,
      int dpiOut,
      double trueWidth,
      double eyeSeparation,
      double mu, int oversmpl, double scale
      );

@ @<Make a single image stereogram from the depth buffer@>=
int PNGSaveSIRDS(const char *filename,
      int x,int y,
      int width, int height,
      int dpiOut,
      double trueWidth,
      double eyeSeparation,
      double mu, int oversmpl, double scale
      )
{ GLubyte *bimage,*bptr;
  GLfloat *dbuff;
  GLfloat oned,fmin,dmax,dmin;
  int oglType, i;
  long int size;
  int iw,ih;
  int res;

  size=width*height;

  @<Get the depth buffer in |dbuff|@>@;
  @<Unproject the $z$ values@>@;
  @<Scaling of the depth buffer@>@;
  for(i=0; i<size; i++)
    dbuff[i]=1.0-dbuff[i];
  CreateStereogram(
    &bimage,&iw,&ih,
    dbuff,width,height,
    dpiOut, /* dpi */
    trueWidth, /*wd in */
    eyeSeparation, /* eye separation */
    mu,oversmpl);  
  oglType=GL_LUMINANCE;  
  res=PNGSaveBitmap(filename,iw,ih,oglType,bimage,GL_FALSE,1.0);
/*  res=EPSSaveBitmap(filename,0,0,iw,ih,1,bimage,trueWidth);*/
  free(bimage);
  free(dbuff);
  return res;
}


@* Textures.
Extend the header.
@(savepng.h@>+=
@<The data structure to store a texture@>@;
@<Macros for the texture data@>@;

@ @<Utilities for textures@>=
@<Check |i| for $i=2^m$@>@;
@<Create a texture@>@;
@<Count the length of the texture list@>@;
@<Clear all textures@>@;
@<Load a PNG file as texture@>@;


@ @<The data structure to store a texture@>=
typedef struct TextureList {
          int textureID;
          char *texIdentifyer;
          char *texSaveName;
          int texType;
          int texWidth,texHeight;
          GLubyte *texPixels;
          GLuint texName;
          struct TextureList *next;
         } TextureList;

@ @<Macros for the texture data@>=
#define NotBounded (-1)
#define GrayTextureQ(t)      ((t->texType)==GL_LUMINANCE)
#define GrayAlphaTextureQ(t) ((t->texType)==GL_LUMINANCE_ALPHA)
#define RGBTextureQ(t)       ((t->texType)==GL_RGB)
#define RGBATextureQ(t)      ((t->texType)==GL_RGBA)
#define TextureID(t)    (t->textureID)
#define TextureName(t)  (t->texIdentifyer)
#define TextureFile(t)  (t->texSaveName)
#define TextureType(t)  (t->texType)
#define TextureWidth(t) (t->texWidth)
#define TextureHeight(t)(t->texHeight)
#define TextureGLName(t)    (t->texName)
#define TextureData(t)  (t->texPixels)

@ OpenGL allows only powers of two as texture sizes.
To get valid image coordinates we choose the power of two
that is smaller than the given |i| if |i| is not of the
form $2^m$. This value is returned in |*lower|.
@<Check |i| for $i=2^m$@>=
int PowerOf2Q(int i, int *lower)
{ int p=1;

  *lower=i;
  while(p<=i && p<INT_MAX) {
    if((p & i) == i) return GL_TRUE;
    p<<= 1;
   }
  *lower=p >> 1;
  return GL_FALSE;
}

@ The textures inside the program are stored in the global variable
  |Textures|.

@(savepng.h@>+=
Global TextureList *Textures;
Global int TextureCount;
@ We will store the texture data here but will will not bound the
  texture to the scene or the mesh.
@(savepng.h@>+=
TextureList *CreateTexture(TextureList *lst,char *name,
                  int ttype,int w,int h, GLubyte *pix, long tsize,
                  int allowscale);

@ @d MAXTEXNAME 256
@<Create a texture@>=
TextureList *CreateTexture(TextureList *lst,char *name,
                  int ttype,int w,int h, GLubyte *pix, long tsize,
                  int allowscale)
{ long psize,lsize;
  int w1,h1, validWidth,validHeight;
  TextureList *head;

  head=(TextureList *)malloc(sizeof(TextureList));
  if(!head) return NULL;
  head->next=lst;
  TextureGLName(head)=-1;
  TextureType(head)=ttype;
  TextureName(head)=(char *)calloc(MAXTEXNAME,sizeof(char));
  TextureFile(head)=(char *)calloc(MAXTEXNAME,sizeof(char));
  strcpy(TextureFile(head),"");
  TextureID(head)=TextureCount;
  if(!*name || (0==strcmp(name,"Automatic")))
    sprintf(TextureName(head),"MVTexture2D[%d]",TextureCount++);
  else
    strcpy(TextureName(head),name);
  @<Get the pixel size form the OpenGL constants@>@;
  @<Copy the input data to the head of the texture list@>@;
  return head;
}


@ @<Get the pixel size form the OpenGL constants@>=
switch(ttype) {
  case GL_LUMINANCE:
    psize=1;
    break;
  case GL_LUMINANCE_ALPHA:
    psize=2;
    break;
  case GL_RGB :
    psize=3;
    break;
  case GL_RGBA :
    psize=4;
    break;
  default :
    psize=0;
 }

@ The bitmap is scaled with the |gluScaleImage| function if
width and height are not powers of 2. The |gluScaleImage| function
works only when the OpenGL libray is initalized. This make it
impossible to scale a commandl ine PNG image because the command line
is processed before the windows are initialized. That is the reason
why the |allowscale| parameter is introduced.

@<Copy the input data to the head of the texture list@>=
  validWidth=PowerOf2Q(w,&w1);
  validHeight=PowerOf2Q(h,&h1);
  if(!(validWidth && validHeight)) {
    int res;
    if(!allowscale) @<|CreateTexture| fails@>@;
    lsize=(TextureWidth(head)=w1)*(TextureHeight(head)=h1)*(psize+4);
    TextureData(head)=(GLubyte *)calloc(lsize,sizeof(GLubyte));
    if(!TextureData(head)) @<|CreateTexture| fails@>@;
    glPixelStorei(GL_UNPACK_ALIGNMENT,1);
    res=gluScaleImage(ttype,w,h,GL_UNSIGNED_BYTE,pix,
                            w1,h1,GL_UNSIGNED_BYTE,TextureData(head));
    if(res) @<|CreateTexture| fails@>@;                        
   }
  else {
    lsize=(TextureWidth(head)=w)*(TextureHeight(head)=h)*psize;
    TextureData(head)=(GLubyte *)calloc(lsize,sizeof(GLubyte));
    if(!TextureData(head)) @<|CreateTexture| fails@>@;
    memcpy(TextureData(head),pix,lsize*sizeof(GLubyte));
   }

@ @<|CreateTexture| fails@>=
{
      free(TextureName(head));
      free(head);
      return NULL;
}

@ Length of a texture list.
@(savepng.h@>+=
int LengthTextureList(TextureList *lst);

@ @<Count the length of the texture list@>=
int LengthTextureList(TextureList *lst)
{ int len=0;

  while(lst) {
    lst=lst->next;
    len++;
   }
  return len;
}

@ Remove the textures form the memory.
@(savepng.h@>=
TextureList *RemoveTextureList(TextureList *lst);

@ @<Clear all textures@>=
TextureList *RemoveTextureList(TextureList *lst)
{ TextureList *tail;
  GLuint *texNames;
  int len,i;

  if(!lst) return NULL;
  @<Clean OpenGLs texture memory@>@;
  while(lst) {
    tail=lst->next;
    free(TextureData(lst));
    free(TextureName(lst));
    free(TextureFile(lst));
    free(lst);
    lst=tail;
   }
  return NULL;
}

@ @<Clean OpenGLs texture memory@>=
  tail=lst;
  len=LengthTextureList(tail);
  texNames=(GLuint *)calloc(len,sizeof(GLuint));
  i=0;
  while(tail) {
    if(glIsTexture(TextureGLName(tail)))
      texNames[i++]=TextureGLName(tail);
    tail=tail->next;
   }
  if(i) glDeleteTextures(i,texNames);
  free(texNames);

@ Read a texture from a Portable Network Graphic file.
@(savepng.h@>=
TextureList *ReadPNGTexture(TextureList *lst,char *fname,double screen_gamma,
               int allowscale);

@ @<Load a PNG file as texture@>=
TextureList *ReadPNGTexture(TextureList *lst,char *fname,double screen_gamma,
               int allowscale)
{
  double image_gamma;
  png_structp png_ptr=NULL;
  png_infop info_ptr=NULL;
  png_uint_32 width, height,row_size;
  int bit_depth, color_type, interlace_type;
  png_bytep *row_pointers, trans=NULL;
  png_byte *image;
  int num_trans=-1;
  png_color_16 *trans_values=NULL;
  FILE *fp;
  TextureList *head;

  if ((fp = fopen(fname, "rb")) == NULL)
    @<Return in case of an error@>@;
  @<Initialize reading of the |png_struct|@>@;
  @<Get the memory for the image information@>@;
  @<Set the error handling@>@;
  @<Initialise the input control@>@;
   /* The call to |png_read_info()| gives us all of the information from the
     PNG file before the first IDAT (image data chunk). */
   png_read_info(png_ptr, info_ptr);

   png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
       &interlace_type, NULL, NULL);
   png_get_tRNS(png_ptr, info_ptr, &trans, &num_trans,&trans_values);
   /* tell libpng to strip 16 bit/color files down to 8 bits/color */
   png_set_strip_16(png_ptr);
   /* expand paletted colors into true RGB triplets expand
      images to the full 8 bits from 1, 2, or 4 bits/pixel */
   if (color_type == PNG_COLOR_TYPE_PALETTE || bit_depth<8)
      png_set_expand(png_ptr);

   /* expand paletted or RGB images with transparency to full alpha channels
     so the data will be available as RGBA quartets */
   if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
      png_set_expand(png_ptr);

   /* Tell libpng to handle the gamma conversion for you.  The second call
     is a good guess for PC generated images, but it should be configurable
     by the user at run time by the user.  It is strongly suggested that
     your application support gamma correction. */
   if (png_get_gAMA(png_ptr, info_ptr, &image_gamma))
      png_set_gamma(png_ptr, screen_gamma, image_gamma);

  /* optional call to gamma correct and add the background to the palette
     and update info structure.  REQUIRED if you are expecting libpng to
     update the palette for you (ie you selected such a transform above).*/
   png_read_update_info(png_ptr, info_ptr);
   @<Create the storrage for the texture@>@;
   png_read_image(png_ptr, row_pointers);

   @<Store the PNG data in a texture@>@;
   /* read rest of file, and get additional chunks in |info_ptr| */
   png_read_end(png_ptr, info_ptr);

   /* clean up after the read, and free any memory allocated */
   png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
   @<Free the memory of the image@>@;
   /* close the file */
   fclose(fp);
   return lst;
}

@ Create and initialize the |png_struct| with the desired error handler
  functions.  If you want to use the default stderr and longjump method,
  you can supply NULL for the last three parameters.  We also supply the
  the compiler header file version, so that we know if the application
  was compiled with a compatible version of the library.

@<Initialize reading of the |png_struct|@>=
   png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
      (void *)NULL, NULL, NULL);
   if (png_ptr == NULL)
     @<Return in case of an error@>@;

@ @<Store the PNG data in a texture@>=
{ int oglType;

  switch(color_type) {
    case PNG_COLOR_TYPE_GRAY:
      oglType=GL_LUMINANCE;
      break;
    case PNG_COLOR_TYPE_GRAY_ALPHA :
      oglType=GL_LUMINANCE_ALPHA;
      break;
    case PNG_COLOR_TYPE_PALETTE:
    case PNG_COLOR_TYPE_RGB :
      oglType=GL_RGB;
      break;
    case PNG_COLOR_TYPE_RGB_ALPHA:
      oglType=GL_RGBA;
      break;
   }
  head=CreateTexture(lst,"",oglType,width,height,image,height*row_size,allowscale);
  if(head)
    lst=head;
}

@ @<Create the storrage for the texture@>=
{ unsigned int i;

  row_pointers=calloc(height,sizeof(png_bytep));
  row_size=png_get_rowbytes(png_ptr, info_ptr);
  image=(GLubyte *)calloc(height,row_size);
  for(i=0; i<height; i++)
    row_pointers[i]=image+row_size*(height-1-i);
}


@ Clip the border color of an RGB image.
@d RGBBytePerPixel 3
@d RGBPixelPtr(i,j,w,img)  (img+(j+w*i)*RGBBytePerPixel)
@<Stuff to clip the image@>=
@<Test a single pixel@>@;
@<Test a single row of the image@>@;
@<Transpose the bitmap@>@;
@<Reverse the rows of the image@>@;
@<Remove a color border arround a RGB picture@>@;

@ @<Test a single pixel@>=
int SameRGBColorQ(GLubyte *clipc,int i, int j, int w,GLubyte *img)
{ GLubyte *pos;

  pos=RGBPixelPtr(i,j,w,img);
  return (*pos==*clipc && pos[1]==clipc[1] && pos[2]==clipc[2]);
}

@ @<Test a single row of the image@>=
int SameRGBRowQ(GLubyte *clipc, int i, int w, GLubyte *img)
{ int j;
  int isSame;

  isSame=SameRGBColorQ(clipc,i,0,w,img);
  j=1;
  while(isSame && j<w) {
    isSame= isSame && SameRGBColorQ(clipc,i,j,w,img);
    j++;
   }
  return isSame;
}

@ @<Transpose the bitmap@>=
void TransposeImage(int *w, int *h, GLubyte *img)
{ GLubyte **row,**tRow,*tImage;
  int i,j;

  tImage=calloc((*w)*(*h)*RGBBytePerPixel,sizeof(GLubyte));
  row=calloc(*h,sizeof(GLubyte *));
  tRow=calloc(*w,sizeof(GLubyte *));
  for(i=0; i<*h; i++) {
    row[i]=RGBPixelPtr(i,0,*w,img);
   }
  for(j=0; j<*w; j++) {
    tRow[j]=RGBPixelPtr(j,0,*h,tImage);
   }
  for(i=0; i<*h; i++)
    for(j=0; j<*w; j++) {
      memcpy(tRow[j]+RGBBytePerPixel*i,
             row[i]+RGBBytePerPixel*j,
             RGBBytePerPixel*sizeof(GLubyte));
     }
  memcpy(img,tImage,(*w)*(*h)*RGBBytePerPixel);
  i=*h;
  *h=*w;
  *w=i;
  free(tRow);
  free(row);
  free(tImage);
}


@  Reverse the rows of an image with |bpp| bytes per pixel,
 |w| and height |h|.
@(savepng.h@>=
void ReverseImageRows(int w, int h, int bpp,
                      GLubyte *image);
@ @<Reverse the rows of the image@>=
void ReverseImageRows(int w, int h, int bpp,
                      GLubyte *image)
{ GLubyte *tmpRow;
  int rowSize;
  int i, n2;

  tmpRow=(GLubyte *)calloc(rowSize=w*bpp,sizeof(GLubyte));
  n2=h/2;
  for(i=0; i<n2; i++) {
    memmove(tmpRow,image+i*rowSize,rowSize*sizeof(GLubyte));
    memmove(image+i*rowSize,image+(h-1-i)*rowSize,rowSize*sizeof(GLubyte));
    memmove(image+(h-1-i)*rowSize,tmpRow,rowSize*sizeof(GLubyte));
   }
  free(tmpRow);
}

@ The clipping of the border is done in two steps. At first
we clip along the upper and the lower edge. Than the bitmap
is transposed and again the upper and the lower edge are cutted.
Finaly the bitmap is transposed again.

@(savepng.h@>+=
void ClipRGBImageBorder(int w, int h, GLubyte *image,
                        int *wout, int *hout, GLubyte **imgout);

@ @<Remove a color border arround a RGB picture@>=
void ClipRGBImageBorder(int w, int h, GLubyte *image,
                        int *wout, int *hout, GLubyte **imgout)
{ GLubyte clipColor[RGBBytePerPixel];
  int i;

  *wout=w;
  *hout=h;
  *imgout=image;
  clipColor[0]=image[0]; @+ clipColor[1]=image[1]; @+ clipColor[2]=image[2];
  @<Clip the along lower edge of the image@>@;
  @<Clip the along upper edge of the image@>@;
  TransposeImage(wout,hout,*imgout);
  @<Clip the along lower edge of the image@>@;
  @<Clip the along upper edge of the image@>@;
  TransposeImage(wout,hout,*imgout);
}

@ @<Clip the along lower edge of the image@>=
  i=*hout-1;
  while(i>=0 && SameRGBRowQ(clipColor,i,*wout,*imgout)) {
    (*hout)--;
    i--;
   }

@ @<Clip the along upper edge of the image@>=
while((*hout)>=0 && SameRGBRowQ(clipColor,0,*wout,*imgout)) {
  (*hout)--;
  *imgout+=RGBBytePerPixel*(*wout);
 }

@ Since PostScrip saving of the 3d data does not support textures
  the saving of the bitmap as encapsulate PostScript is also
  possible.
@(savepng.h@>+=
int EPSSaveFrameBuffer(
      const char *filename,
      int x,int y,
      int width, int height);

@  @<Save a EPS bitmap of the frame buffer@>=
int EPSSaveFrameBuffer(
      const char *filename,
      int x,int y,
      int width, int height)
{
  GLubyte* image;
  FILE *eps;
  int i,k,mu;

  image=(GLubyte *)calloc(mu=width*height*3,sizeof(GLubyte));
  glPixelStorei(GL_PACK_ALIGNMENT, 1);
  glReadPixels(x, y, width, height, GL_RGB, GL_UNSIGNED_BYTE, image);

#if MACINTOSH_MATHLINK
  eps=MacCreateFile(filename,'TEXT','gsVR');
#else    
  eps=fopen(filename,"w");
#endif  
  if(!eps) {
    free(image);
    return GL_FALSE;
   }
  fprintf(eps,"%%!PS-Adobe-3.0 EPSF-3.0\n"@|
              "%%%% Creator MathGL3d (using frame buffer)\n"@|
              "%%%%BoundingBox: %d %d %d %d\n"@|
              "%%EndComments\n\n",x,y,x+width,y+height);
  fprintf(eps,"gsave\n");
  fprintf(eps,"/picstr %d string def\n",3*width);
  fprintf(eps," 0  0 translate\n %d %d scale\n",width,height);
  fprintf(eps," %d %d 8\n [%d 0 0 %d 0 0]\n",width,height,width,height);
  fprintf(eps,"{currentfile picstr readhexstring pop}\nfalse 3\n colorimage\n");
  k=0;
  while(k<mu) {
    for(i=0; i<36 && k<mu; i++)
      fprintf(eps,"%02X",image[k++]);
      fprintf(eps,"\n");
   }
  fprintf(eps,"grestore\n");
  fprintf(eps,"showpage");
  fclose(eps);
  free(image);
  return 0;
}

@  @d PSPointPerIn (72)
@<Save a image as EPS bitmap@>=
int EPSSaveBitmap(
      const char *filename,
      int x,int y,
      int width, int height, int channels,
      GLubyte *image, double trueWidth)
{
  FILE *eps;
  int i,k,mu;

  mu=width*height*channels;
  trueWidth*=PSPointPerIn;
#if MACINTOSH_MATHLINK
  eps=MacCreateFile(filename,'TEXT','gsVR');
#else    
  eps=fopen(filename,"w");
#endif  
  if(!eps) 
    return GL_FALSE;
  fprintf(eps,"%%!PS-Adobe-3.0 EPSF-3.0\n"@|
              "%%%% Creator MathGL3d\n"@|
              "%%%%BoundingBox: %lg %lg %lg %lg\n"@|
              "%%EndComments\n\n",(double)x,(double) y,
                  x+trueWidth,y+height*trueWidth/width);
  fprintf(eps,"gsave\n");
  fprintf(eps,"/picstr %d string def\n",channels*width);
  fprintf(eps," 0  0 translate\n %lg %lg scale\n",trueWidth,height*trueWidth/width);
  fprintf(eps," %d %d 8\n [%d 0 0 %d 0 0]\n",width,height,width,height);
  fprintf(eps,"{currentfile picstr readhexstring pop} %s\n",
              3==channels ? "\nfalse 3\n colorimage" :" image ");
  k=0;
  while(k<mu) {
    for(i=0; i<36 && k<mu; i++)
      fprintf(eps,"%02X",image[k++]);
      fprintf(eps,"\n");
   }
  fprintf(eps,"grestore\n");
  fprintf(eps,"showpage");
  fclose(eps);
  return 0;
}

@* This implements the algorithm for the creation of single image
random dot stereograms. The essential reference is the article
``Displaying 3D Images: Algorithm for Single Image Random Dot Stereograms
by Harold W. Thimbleby, Stuart Inglis and Ian H. Witten.

The function will return the calculated image of width |w| and height |h|
  in |*sirds|. The depth values of width |width| and height |height| 
  are stored in |z|. The output resolution is passed as |dotPerIn|.

@d Separation(z) round((1.-mu*(z))*E/(2.-mu*(z)))  

@d PutPixel(x,y,pv) *((*sirds)+(y)*(*w)+(x))=pv

@ @<Stereogram creation@>=
#if !MACINTOSH_MATHLINK 
#  define round(x) ((int)((x)+0.5))  
#endif
void CreateStereogram(
       unsigned char **sirds, int *w, int *h,
       float *z, int width, int height,
       int dotPerIn, double widthIn,
       double eyeSeparation,
       double mu, int oversample
       )
{ double E,zt,d;
  int *same,*pix;
  int x,y;
  int left,right; /* x-vlaues for left and right eyes */
  int s;          /* Stereo seperation at the point $(x,y)$ */
  int visible,t,l;
  int rowSize;


  @<Initialize the memory for the SIRDS image@>@;
  for(y=0; y<(*h); y++) { /* Loop over rows ... */
    for(x=0; x<rowSize; x++)
      same[x]=x;  /* Link each pixel with itself. */ 
    @<Make one line of the stereogram@>@;
    @<Set the pixels in the |y| line@>@;
   } /* end of |for(y=0;;)| */
  @<Free the dynamic memory of the stereogram@>@;
}       
       
       
@ @<Make one line of the stereogram@>=       
for(x=0; x<rowSize; x++) {
  d=Depth(x,y);
  s=Separation(d);
  left=x-(s/2);
  right=left+s;
  if(0<=left && right< rowSize) {
    @<Hidden surface removal@>@;
   }
}
        

@ @<Hidden surface removal@>=
{ int k;                
  for(k=same[left]; k!=left && k!=right; k=same[left])
    if( k < right ) 
      left = k;
    else {
      left = right;
      right = k;
     }
    same[left] = right; 
}    

@ @<Set the pixels in the |y| line@>=
for(x=rowSize-1; x>-1; x--)
  pix[x]=(same[x]==x ? 255*(rand() & 1) : pix[same[x]]);
for(x=0; x<rowSize; x+=oversample) {
  int k,pval;
  for(pval=0, k=x; k<x+oversample; k++)
     pval+=pix[k]; 
  pval/=oversample;
  PutPixel(x/oversample,y,pval);
 }   

@ @<Initialize the memory for the SIRDS image@>=
E=eyeSeparation*dotPerIn;
ImgWidth_local=*w=dotPerIn*widthIn;
ImgHeight_local=*h=(*w)*(width/height);
rowSize=*w*oversample;
ImgWidth_local*=oversample;
*sirds=(unsigned char *)calloc((*w)*(*h),sizeof(unsigned char));
same=(int *)calloc(rowSize,sizeof(int));
pix=(int *)calloc(rowSize,sizeof(int));
/* assign the global values for interpolation */
DepthData_local=z;
DepthWidth_local=width;
DepthHeight_local=height;

@ @<Free the dynamic memory of the stereogram@>=
DepthData_local=NULL;
free(pix);
free(same);       
       
@ @<Static data@>=
static int DepthWidth_local, DepthHeight_local;
static int ImgWidth_local, ImgHeight_local;
static float *DepthData_local;

@ @d depth(i,j) DepthData_local[(j)*DepthWidth_local+(i)]

@<Interpolate the $z$-values@>=
static float Depth(register int i,register  int j)
{ register float x,y;
  float d,d1;

  x=(float) i/((float) (ImgWidth_local));
  y=(float) j/((float) (ImgHeight_local));
  i=x=DepthWidth_local*x;
  j=y=DepthHeight_local*y;
  if(i>(DepthWidth_local-2)) i=DepthWidth_local-2;
  if(j>(DepthHeight_local-2)) j=DepthHeight_local-2;
  x-=(float)i;
  y-=(float)j;
  d = (1.0f-x)*depth(i,j)+x*depth(i+1,j);
  d1= (1.0f-x)*depth(i,j+1)+x*depth(i+1,j+1);
  return (1.0f-y)*d+y*d1;
}

@ We include \.{mathlink.h} here to get the |MACINTOSH_MATHLINK| variable.
@<Header@>=
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#ifdef __WIN32__
#  include <windows.h>
#endif
#include "mathlink.h"
#include "GL/gl.h"
#include "GL/glu.h"
#include "GL/glut.h"
#include "png.h"
#include "savepng.h"
#if MACINTOSH_MATHLINK
FILE * MacCreateFile(const char *,unsigned long, unsigned long);
#endif

