#define BANNER "GL Viewer for Mathematica"
#define VERSION "2.2" \

#define TWO_PI (2.0*3.14159265358979324) 
#define PI 3.14159265358979324
#define SQRT3 (1.73205080756887729) 
#define MaxLSAngle (75.0) 
#define MENU_ID_SAVE_BASE (100) 
#define MENU_ID_SAVEPOVRAY (101) 
#define MENU_ID_SAVE_3DMF (102) 
#define MENU_ID_SAVE_VRML (103) 
#define MENU_ID_SAVEPOSTSCRIPT (104) 
#define MENU_ID_SAVE_BITMAP_EPS (105) 
#define MENU_ID_SAVE_DXF (106) 
#define MENU_ID_SAVE_OFF (107) 
#define MENU_ID_SAVE_PNG (108) 
#define MENU_ID_LOAD_XXX (121)  \

#define MENU_ID_SPIN_X (201) 
#define MENU_ID_SPIN_Y (202) 
#define MENU_ID_SPIN_Z (203) 
#define MENU_ID_SPIN_SLOW (204) 
#define MENU_ID_SPIN_FAST (205) 
#define MENU_ID_SPIN_TRACKBALL (206) 
#define MENU_ID_STOP (207)  \

#define MENU_ID_ROTATE (501) 
#define MENU_ID_ROTATE_X (502) 
#define MENU_ID_ROTATE_Y (503) 
#define MENU_ID_ROTATE_Z (504)  \

#define MENU_ID_SHOW_SMOOTH (301) 
#define MENU_ID_SHOW_FLAT (302) 
#define MENU_ID_SHOW_WIRE (303) 
#define MENU_ID_TOGGLE_BACK (304) 
#define MENU_ID_SHOW_OUTLINE (305)  \

#define MENU_ID_VIEWCONTROL (401) 
#define MENU_ID_LIGHTCONTROL (402) 
#define MENU_ID_AXISCONTROL (403) 
#define MENU_ID_ABOUT (404) 
#define MENU_ID_RESET (405) 
#define MENU_ID_FLIP_FACE (406) 
#define MENU_ID_ZOOM_IN (407) 
#define MENU_ID_ZOOM_OUT (408) 
#define MENU_ID_CLEAR (409) 
#define MENU_ID_QUIT (666)  \

#define ROTATE_X 2
#define ROTATE_Y 4
#define ROTATE_Z 8 \

#define MVPACKAGE_NAME "OpenGLViewer.m"
#define Line3D (1) 
#define Polygon3D (2) 
#define NOT_FOUND (-1) 
#define INIT_POOLSIZE 8192
#define POOL_INC 1024
#define TINY (0.1e-8) 
#define MachineEpsilon (2.22044604925031308e-16) 
#define TEXTURE_MAP_PLANAR (2) 
#define TEXTURE_MAP_CYLINDRICAL (4) 
#define TEXTURE_MAP_SPHERICAL (8) 
#define TEXTURE_MAP_TORUS (16) 
#define TEXTURE_MAP_MESH_ST (128) 
#define TEXTURE_MAP_MESH_TS (256)  \
 \

#define TILE_TEXTURE_S (2) 
#define TILE_TEXTURE_T (4)  \

#define SceneTextureID (SceneTextureMap->tex->texName)  \

#define InQuadrant1 1
#define InQuadrant2 2
#define InQuadrant3 4
#define InQuadrant4 8
#define MESH_OPEN (0) 
#define MESH_UCLOSED (1) 
#define MESH_VCLOSED (2) 
#define MESH_VERTEX_COLOR (4) 
#define MESH_FACE_COLOR (8) 
#define MeshID(m) (m?(m->meshID) :-1) 
#define MeshIndexExpanded(m,i,j) ((j) +(m->dimV) *(i) ) 
#define MeshIndex(m,i,j) (m->mesh[(j) +(m->dimV) *(i) ]) 
#define MeshCoord(m,i,j) m->mPoints[m->mesh[(j) +(m->dimV) *(i) ]]
#define MeshNormal(m,i,j) m->mNormals[m->mesh[(j) +(m->dimV) *(i) ]]
#define MeshCoordPtr(m,i,j) (m->mPoints+m->mesh[(j) +(m->dimV) *(i) ]) 
#define MeshNormalPtr(m,i,j) (m->mNormals+m->mesh[(j) +(m->dimV) *(i) ]) 
#define MeshTexCoordPtr(m,i,j) (m->mTexturePoints+(j) +(m->dimV) *(i) ) 
#define MeshColorPtr(m,i,j) (m->mpolyColors+(j) +(m->cDimV) *(i) ) 
#define MeshTexMap(m) (m->mTexture) 
#define MeshTextureQ(m) (m->mTexture!=NULL) 
#define MeshTextureID(m) (m->mTexture->tex->texName) 
#define MeshTextureFile(m) (m->mTexture->tex->texSaveName) 
#define MeshTextureType(m) (m->mTexture->tex->texType) 
#define MAX_LIGHT (8) 
#define sign(a) (a>0.0?1.:-1.)  \

#define sqr(a) ((a) *(a) ) 
#define IncPoint3D(p,inc) {p.x+= inc.x;p.y+= inc.y;p.z+= inc.z;}
#define AddPoint3D(res,a,b) {res.x= a.x+b.x;res.y= a.y+b.y;res.z= a.z+b.z;} \

#define AssignMinMax(min,max,d) if((d) <min) min= (d) ;else if((d) >max) max= d; \

#define PolygonPoint(t,n,k) ((t->vpoly[(n) ]) ->points[(k) ]) 
#define PolygonPoints(t,n) ((t->vpoly[(n) ]) ->points) 
#define PolygonPointNo(t,n) ((t->vpoly[(n) ]) ->pno)  \

#define MAX_STRING (256) 
#define ToDegree (180.0/3.1415926) 
#define ToRadian (3.1415926/180.0) 
#define MVNearPlane (1.0) 
#define MVFarPlane (7.0) 
#define MVViewAngle (45.0)  \

#define MAX(a,b) (a>b?a:b) 
#define SET4VEC(arr,col) {arr[0]= col.r; \
arr[1]= col.g; \
arr[2]= col.b; \
arr[3]= 1.0;} \

#define INHERITE_COLOR (1L) 
#define PointSphereSegmentsPhi 18
#define PointSphereSegmentsTheta 9
#define CrossXPlusAxis(p) (((p->pTexQuadrants&InQuadrant1) ==InQuadrant1) && \
((p->pTexQuadrants&InQuadrant4) ==InQuadrant4) )  \
 \

#define MeshSmoothPoint(i,j) glNormal3dv((GLdouble*) MeshNormalPtr(lst,i,j) ) ; \
glVertex3dv((GLdouble*) MeshCoordPtr(lst,i,j) ) ; \

#define MeshSmoothTexPoint(i,j) glNormal3dv((GLdouble*) MeshNormalPtr(lst,i,j) ) ; \
tp= MeshTexCoordPtr(lst,i,j) ; \
glTexCoord2d(scaleS*tp->x,scaleT*tp->y) ; \
glVertex3dv((GLdouble*) MeshCoordPtr(lst,i,j) ) ; \

#define EPS_GOURAUD_THRESHOLD 0.1 \

#define EPS_SMOOTH_LINE_FACTOR 0.06 \

#define POVMeshCount (haveColors?&ord:NULL) 
#define POVMeshColor (haveColors?&k:NULL) 
#define TrueMeshTriangle1(lst,i,j) (MeshIndex(lst,i+1,j) !=MeshIndex(lst,i,j) && \
MeshIndex(lst,i+1,j) !=MeshIndex(lst,i,j+1) && \
MeshIndex(lst,i,j) !=MeshIndex(lst,i,j+1) )  \

#define TrueMeshTriangle2(lst,i,j) (MeshIndex(lst,i+1,j) !=MeshIndex(lst,i,j+1) && \
MeshIndex(lst,i+1,j) !=MeshIndex(lst,i+1,j+1) && \
MeshIndex(lst,i,j+1) !=MeshIndex(lst,i+1,j+1) )  \

#define Camera 2.33
#define VRLastPoint(i,lst) ((NULL==lst->next) &&((i) ==lst->pno-1) ) 
#define VRBool(t) (t?"TRUE":"FALSE") 
#define POLY_LAYER 1
#define LINE_LAYER 2
#define POINT_LAYER 3
#define MESH_LAYER 4 \

#define DXF_LINESIZE 1024
#define DXFGetLine(l) {do{ \
fgets(l,DXF_LINESIZE,dxf) ; \
RemoveSpaces(l) ; \
}while(!*l&&!feof(dxf) ) ; \
} \

#define DXFGetInteger(i,l) {do{ \
fgets(l,DXF_LINESIZE,dxf) ; \
RemoveSpaces(l) ; \
}while(!*l&&!feof(dxf) ) ; \
i= atoi(line) ; \
}
#define DXFGetDouble(d,l) {do{ \
fgets(l,DXF_LINESIZE,dxf) ; \
RemoveSpaces(l) ; \
}while(!*l&&!feof(dxf) ) ; \
d= atof(line) ; \
} \
 \

#define DXFEndQ(line) strstr(line,"EOF") 
#define DXFEntitiesQ(line) strstr(line,"ENTITIES") 
#define DXFPointQ(line) strstr(line,"POINT") 
#define DXFPolyLineQ(line) strstr(line,"POLYLINE") 
#define DXFVertexQ(line) strstr(line,"VERTEX") 
#define DXF3DFaceQ(line) strstr(line,"3DFACE") 
#define DXFEndSeqQ(line) strstr(line,"SEQEND") 
#define DXFUpdateBBox(x,y,z) {if((x) <xMin) xMin= (x) ; \
if((x) >xMax) xMax= (x) ; \
if((y) <yMin) yMin= (y) ; \
if((y) >yMax) yMax= (y) ; \
if((z) <zMin) zMin= (z) ; \
if((z) >zMax) zMax= (z) ; \
}
#define DXFNone (0) 
#define DXFPoint (2) 
#define DXFLine3D (4) 
#define DXFPoly3D (8)  \

#define OFF_LINESIZE 1024 \

#define ALL_OBJECTS (0) 
#define SCENE_OBJECTS (1) 
#define TransformSceneQ(type) ((ALL_OBJECTS==type) ||(SCENE_OBJECTS==type) ) 
#define TransformMeshQ(type,meshid) ((ALL_OBJECTS==type) ||(type==meshid) ) 
#define BoxUpdateQ(up) (1==((up) &1) ) 
#define DisplayUpdateQ(up) (2==((up) &2) )  \

#define MLPutList(n) MLPutFunction(stdlink,"List",n) 
#define STRINGLEN 256
#define BEGIN_COMMENT '#'
#define SPACE ' '
#define NUMBER_DELIMITER " ,:=\n" \

#define CurrentColor LCWindow->LightColor[LCWindow->sLight]
#define CurrentPos LCWindow->LightPosition[LCWindow->sLight] \

#define NoAxis 0
#define CrossAxis 1
#define BoxAxis 2
#define PlaneAxis 4
#define NTICK (5) 
#define AxisArrowBase (0.05*scale) 
/*1:*/
#line 18 "MathView3D.w"

#ifdef __WIN32__
#  define ARCH "Win32"
#elif defined(__linux__)
#  define ARCH "Linux"
#elif defined(__sgi)
#  define ARCH "Silicon Graphics"
#elif MACINTOSH_MATHLINK
#  define ARCH "Macintosh"
#else
#  define ARCH "Solaris"
#endif

#ifdef __WIN32__
#  define MVSleep(ms) Sleep(ms)
#elif MACINTOSH_MATHLINK
# define MVSleep(ms)     MacintoshSleep((UInt32) ms)
#else
#  define MVSleep(ms) usleep(ms)
#endif

#if !MACINTOSH_MATHLINK
#  define HAVE_GLUT_SUBWINDOW 1
#endif
#if MACINTOSH_MATHLINK
#  define DBL_FMT "%Lg"
#else
#  define DBL_FMT "%lg"
#endif

/*895:*/
#line 14595 "MathView3D.w"

#define MATHGL_MAIN
#include <assert.h>
#include <stdlib.h>    
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <float.h>
#include "GL/glut.h"
#include "mui/mui.h"
#include "mathlink.h"
#include "cmdln.h"
#include "savepng.h"
#if !(defined(__WIN32__) && defined(__BORLANDC__))
# include <unistd.h>
#endif
#ifdef __WIN32__
#  include <commdlg.h>
#  include <dos.h>
#  include <alloc.h>
#endif
#if MACINTOSH_MATHLINK
#  include <StandardFile.h>
#endif
/*:895*/
#line 48 "MathView3D.w"

#if MACINTOSH_MATHLINK
#  define RGBColor RGBColor__
#endif
#ifndef MAX
#  define MAX(a,b) ((a)>(b) ? (a) : (b))
#endif
#ifndef MIN
#  define MIN(a,b) ((a)<(b) ? (a) : (b))
#endif
#define MAX3(a,b,c) ((a)>(b) ? MAX(a,c) : MAX(b,c))
#define MIN3(a,b,c) ((a)<(b) ? MIN(a,c) : MIN(b,c))

/*62:*/
#line 1513 "MathView3D.w"

/*63:*/
#line 1522 "MathView3D.w"

typedef struct Point2D{double x,y;}Point2D;

/*:63*/
#line 1514 "MathView3D.w"

/*64:*/
#line 1525 "MathView3D.w"

typedef struct Point3D{double x,y,z;}Point3D;

/*:64*/
#line 1515 "MathView3D.w"

/*68:*/
#line 1549 "MathView3D.w"

typedef struct RGBColor{double r,g,b;}RGBColor;

/*:68*/
#line 1516 "MathView3D.w"

/*69:*/
#line 1552 "MathView3D.w"

typedef struct PointList{
RGBColor color;
double size;
int point;
struct PointList*next;
}PointList;

/*:69*/
#line 1517 "MathView3D.w"

/*76:*/
#line 1632 "MathView3D.w"

typedef struct LinePolyList{
short int what;
RGBColor color;
int aux1;
double aux2;
int*points;
Point3D*pTexturePoints;
int pno;
int pTexQuadrants;
struct LinePolyList*prev;
struct LinePolyList*next;
}LinePolyList;

/*:76*/
#line 1518 "MathView3D.w"

/*139:*/
#line 2726 "MathView3D.w"

typedef struct Mesh3DList{
char name[256];
int meshID;
int dimU,dimV;
int mu;
int mPointNo;
Point3D*mPoints;
Point3D*mNormals;
Point3D*mTexturePoints;
int*mesh;
int cDimU,cDimV;
RGBColor*mpolyColors;
int attrib;
int aux1;
double aux2;
RGBColor mColor;
struct TextureMapping*mTexture;
struct Mesh3DList*next;
}Mesh3DList;

/*:139*/
#line 1519 "MathView3D.w"

/*210:*/
#line 3892 "MathView3D.w"

typedef struct Quaternion{
double q0,q1,q2,q3;
}Quaternion;

/*:210*/
#line 1520 "MathView3D.w"


/*:62*//*94:*/
#line 1927 "MathView3D.w"

typedef struct PointOctTree{
int point;
int polyNo;
LinePolyList**vpoly;
struct PointOctTree*leafs[8];
}PointOctTree;

/*:94*//*112:*/
#line 2238 "MathView3D.w"

typedef struct TextureMapping{
int type;
int tileTexture;
Point3D*translate;
Point3D*axis;
double angle;
Point3D*scale;
double scaleS,scaleT;
TextureList*tex;
}TextureMapping;


/*:112*//*274:*/
#line 5119 "MathView3D.w"

typedef struct DepthIndex{
GLfloat*ptr;
GLfloat depth;
}DepthIndex;

/*:274*//*281:*/
#line 5303 "MathView3D.w"

typedef struct Feedback3DColor{
GLfloat x;
GLfloat y;
GLfloat z;
GLfloat red;
GLfloat green;
GLfloat blue;
GLfloat alpha;
}Feedback3DColor;


/*:281*//*766:*/
#line 12608 "MathView3D.w"

typedef struct ViewCntrlWin{
int WinID;
int PreviewWinID;
int muiID;
int CWinX;
int CWinY;
muiObject*ApplyButton,
*HideButton,
*QuitButton;
muiObject*FilePrefixBox,
*ViewPointBox,
*Load3ScriptBox,
*LoadDXFBox,
*BoxRatioBox;
muiObject*RSlider,
*PhiSlider,
*ThetaSlider,
*XScaleSlider,
*YScaleSlider,
*ZScaleSlider;
char ViewPointStr[256];
char BoxRatioStr[256];
char FilePrefix[256];
char Load3Script[256];
char LoadDXF[256];
double PreviewR,
PreviewTheta,
PreviewPhi;
Point3D PreViewPoint;
Point3D PreviewBox;
}ViewCntrlWin;

/*:766*//*808:*/
#line 13171 "MathView3D.w"

typedef struct LightCntrlWin{
int WinID;
int PreviewWinID;
int muiID;
int CWinX;
int CWinY;
muiObject*NewLSButton,
*DelLSButton,
*ApplyButton,
*HideButton,
*QuitButton;
muiObject*CubeRButton,
*DodecahedronRButton,
*TorusRButton,
*SphereRButton;
muiObject*LTextList,
*PositionBox,
*ColorBox;
muiObject*RSlider,
*PhiSlider,
*ThetaSlider,
*RedSlider,
*GreenSlider,
*BlueSlider;
char*LNameN[MAX_LIGHT+1];
char*LNames[MAX_LIGHT+1];
char PositionStr[256];
char ColorStr[256];
double PreviewR,
PreviewTheta,
PreviewPhi;
Point3D LightPosition[MAX_LIGHT];
RGBColor LightColor[MAX_LIGHT];
int LNo;
int sLight;
}LightCntrlWin;

/*:808*//*850:*/
#line 13839 "MathView3D.w"

typedef struct AxisData3D{
Point3D Origin;
int AxisType;
int NumericLabels;
double FontScale;
char XLabel[128],
YLabel[128],
ZLabel[128];

}AxisData3D;

typedef struct AxisCtrlWin{
int WinID;
int PreviewWinID;
int muiID;
int CWinX;
int CWinY;
int VObject;
muiObject*ApplyButton,
*HideButton,
*QuitButton;
muiObject*NoneRButton,
*CrossRButton,
*BoxRButton,
*PlaneRButton;

muiObject*NumLabelRButton;
muiObject*XLabelBox,
*YLabelBox,
*ZLabelBox,
*OriginBox;
muiObject
*XOriginSlider,
*YOriginSlider,
*ZOriginSlider,
*FontScaleSlider;
Point3D Origin;
char OriginStr[512];
int AxisType;
int NumericLabels;
double FontScale;
char XLabel[128],
YLabel[128],
ZLabel[128];
}AxisCtrlWin;


/*:850*/
#line 61 "MathView3D.w"

void CreateObject(void);
/*15:*/
#line 444 "MathView3D.w"

char ThreeScriptName[256]= "";
char DXFInputName[256]= "";
char OFFInputName[256]= "";
char TextureName[256]= "";
char TexMapType[256]= "planar";
int SmoothShading= GL_FALSE;
int WireFrame= GL_FALSE;
int PolyOutline= GL_FALSE;
int COMTriangulation= GL_FALSE;
Point3D ZoomMove= {0.,0.,0.};
Point3D CenterOfScene= {0.0,0.0,0.0};
Point3D ScaleOfScene= {1.0,1.0,1.0};
Point3D BoxRatios= {1.0,1.0,1.0};
const Point3D XUnitVector= {1.0,0.0,0.0};
const Point3D YUnitVector= {0.0,1.0,0.0};
const Point3D ZUnitVector= {0.0,0.0,1.0};
double MaxScaleOfScene= 1.0;
int MakeNewView= GL_FALSE;
double RadialZoom= 1.0;

/*:15*//*16:*/
#line 466 "MathView3D.w"

char POVOutName[256]= "";
char DXFOutName[256]= "";
char QD3dOutName[256]= "";
char VRMLOutName[256]= "";
int UseVRMLExtrusion= GL_TRUE;
int NoDisplay= GL_FALSE;
int MakePackage= GL_FALSE;

/*:16*//*17:*/
#line 483 "MathView3D.w"

double LineTubeSize= 0.0;
int BlackBackground= GL_TRUE;
int MouseInsideMain= GL_TRUE;
int TubeSegments= 12;
int LineColor= 1.0;
double LineSegmentAngle= 30.;
double CosSkipCont;
double PointSphereSize= 0.0;
const RGBColor Black= {0.0,0.0,0.0};
const RGBColor White= {1.0,1.0,1.0};
RGBColor ClearColor;
Point3D XAxis= {1.0,0.0,0.0};
Point3D YAxis= {0.0,1.0,0.0};
Point3D ZAxis= {0.0,0.0,1.0};
const Point3D ZeroPoint= {0.0,0.0,0.0};

/*:17*//*18:*/
#line 500 "MathView3D.w"

int MakeNewModel= GL_FALSE;
int MVFaceFlip= GL_FALSE;
int DoReset= GL_FALSE;
int DoTracking= GL_FALSE;
int DoTrackSpin= GL_FALSE;
int DoRotate= GL_FALSE;
int GetMouseTrack= GL_TRUE;
int ScreenX,ScreenY;
int ScreenXStart,ScreenYStart;
Point3D TrackAxis;
double TrackAngle,DTrackAngle;
int ScreenWidth,ScreenHeight;

/*:18*//*19:*/
#line 556 "MathView3D.w"

int menuIDMain;
int menuIDFiles;
int menuIDSpin;
int menuIDRotate;
int menuIDAppear;

/*:19*//*40:*/
#line 999 "MathView3D.w"

char WriteFilePrefix[256]= "mview3d";


/*:40*//*65:*/
#line 1532 "MathView3D.w"

Point3D*PointPool= NULL;
Point3D**NormalPool= NULL;
int PoolSize= 0;
int MaxPoolSize= 0;

/*:65*//*106:*/
#line 2134 "MathView3D.w"

PointList*Points= NULL;
PointOctTree*OctalPointTree= NULL;
LinePolyList*Lines= NULL;
LinePolyList*Polygons= NULL;

/*:106*//*113:*/
#line 2255 "MathView3D.w"

TextureMapping*SceneTextureMap= NULL;

/*:113*//*140:*/
#line 2749 "MathView3D.w"

Mesh3DList*Meshes= 0;
static int MeshCount= 2;
/*:140*//*157:*/
#line 2976 "MathView3D.w"

Point3D ViewPoint= {1.3,-2.4,2.0};
Point3D ViewVertical= {0.,0.,1.};
RGBColor AmbientColor;
Point3D BoundP1,BoundP2;
Point3D LightPosition[MAX_LIGHT];
RGBColor LightColor[MAX_LIGHT];
int LightNo= 0;
int PolygonEdge= 1;

/*:157*//*209:*/
#line 3886 "MathView3D.w"

double PolyReduceFactor= 0.0;
int PolyReduceIterations= 0;

/*:209*//*243:*/
#line 4515 "MathView3D.w"

int ObjID= -1;

/*:243*//*282:*/
#line 5315 "MathView3D.w"

GLfloat pointSize;
GLint DoPSOutput= GL_FALSE;

static char*gouraudtriangleEPS[]= 
{
"/bd{bind def}bind def /triangle { aload pop   setrgbcolor  aload pop 5 3",
"roll 4 2 roll 3 2 roll exch moveto lineto lineto closepath fill } bd",
"/computediff1 { 2 copy sub abs threshold ge {pop pop pop true} { exch 2",
"index sub abs threshold ge { pop pop true} { sub abs threshold ge } ifelse",
"} ifelse } bd /computediff3 { 3 copy 0 get 3 1 roll 0 get 3 1 roll 0 get",
"computediff1 {true} { 3 copy 1 get 3 1 roll 1 get 3 1 roll 1 get",
"computediff1 {true} { 3 copy 2 get 3 1 roll  2 get 3 1 roll 2 get",
"computediff1 } ifelse } ifelse } bd /middlecolor { aload pop 4 -1 roll",
"aload pop 4 -1 roll add 2 div 5 1 roll 3 -1 roll add 2 div 3 1 roll add 2",
"div 3 1 roll exch 3 array astore } bd /gouraudtriangle { computediff3 { 4",
"-1 roll aload 7 1 roll 6 -1 roll pop 3 -1 roll pop add 2 div 3 1 roll add",
"2 div exch 3 -1 roll aload 7 1 roll exch pop 4 -1 roll pop add 2 div 3 1",
"roll add 2 div exch 3 -1 roll aload 7 1 roll pop 3 -1 roll pop add 2 div 3",
"1 roll add 2 div exch 7 3 roll 10 -3 roll dup 3 index middlecolor 4 1 roll",
"2 copy middlecolor 4 1 roll 3 copy pop middlecolor 4 1 roll 13 -1 roll",
"aload pop 17 index 6 index 15 index 19 index 6 index 17 index 6 array",
"astore 10 index 10 index 14 index gouraudtriangle 17 index 5 index 17",
"index 19 index 5 index 19 index 6 array astore 10 index 9 index 13 index",
"gouraudtriangle 13 index 16 index 5 index 15 index 18 index 5 index 6",
"array astore 12 index 12 index 9 index gouraudtriangle 17 index 16 index",
"15 index 19 index 18 index 17 index 6 array astore 10 index 12 index 14",
"index gouraudtriangle 18 {pop} repeat } { aload pop 5 3 roll aload pop 7 3",
"roll aload pop 9 3 roll 4 index 6 index 4 index add add 3 div 10 1 roll 7",
"index 5 index 3 index add add 3 div 10 1 roll 6 index 4 index 2 index add",
"add 3 div 10 1 roll 9 {pop} repeat 3 array astore triangle } ifelse } bd",
NULL
};

/*:282*//*752:*/
#line 12400 "MathView3D.w"

MLINK LinkToMath;
#if !MACINTOSH_MATHLINK
int MathLinkMode= GL_FALSE;
#else
int MathLinkMode= GL_TRUE;
#endif
int MathLinkActive= GL_FALSE;
int DoMathematicaCall= GL_FALSE;

/*:752*//*773:*/
#line 12746 "MathView3D.w"

int MVDrawWindow;
ViewCntrlWin*VCWindow= NULL;

/*:773*//*809:*/
#line 13209 "MathView3D.w"

LightCntrlWin*LCWindow= NULL;

/*:809*//*851:*/
#line 13887 "MathView3D.w"

AxisCtrlWin*AxisWindow;
static AxisData3D Axis= {{0.0,0.0,0.0},NoAxis,GL_FALSE,0.25,"x","y","z"};

/*:851*/
#line 63 "MathView3D.w"

/*25:*/
#line 651 "MathView3D.w"

FILE*MacCreateFile(const char*filename,unsigned long ftype,unsigned long creator);

/*:25*//*28:*/
#line 692 "MathView3D.w"

void MacGetFullPath(char*fullPath,
long int dirID,
int vRefNum);

/*:28*//*31:*/
#line 725 "MathView3D.w"

int MacQueryLoadFilename(char*);

/*:31*//*33:*/
#line 748 "MathView3D.w"

int MacQuerySaveFilename(char*cprompt,char*filename);

/*:33*//*107:*/
#line 2145 "MathView3D.w"

int AddToPointPool(double x,double y,double z);
/*:107*//*647:*/
#line 10795 "MathView3D.w"

void mvClearContents(void);

/*:647*//*670:*/
#line 11183 "MathView3D.w"

void mlWritePOVray(const char*,const char*,const char*);

/*:670*//*672:*/
#line 11200 "MathView3D.w"

void mlWrite3DMF(const char*,int,int);

/*:672*//*675:*/
#line 11235 "MathView3D.w"

void MVLoad3Script(const char*fname);

/*:675*//*677:*/
#line 11265 "MathView3D.w"

void mlWriteDXF(const char*,double,int,double,double);

/*:677*//*679:*/
#line 11297 "MathView3D.w"

void MVLoadDXF(const char*fname);

/*:679*//*688:*/
#line 11397 "MathView3D.w"

void mvWriteOFF(const char*fname);

/*:688*//*690:*/
#line 11414 "MathView3D.w"

void MVLoadOFF(const char*fname);

/*:690*//*692:*/
#line 11434 "MathView3D.w"

void mlWritePostScript(const char*fname,int bitmapImage);

/*:692*//*694:*/
#line 11451 "MathView3D.w"

void mlRawWriteVRML(const char*fname,int headlight,int useextrude,
const char*ptexpre,const char*mtexpre);

/*:694*//*696:*/
#line 11470 "MathView3D.w"

void mlRawWritePNG(const char*fname,double gamma,int interlace);

/*:696*//*755:*/
#line 12451 "MathView3D.w"

int ProcessMathLink(MLINK*mlp);

/*:755*//*767:*/
#line 12642 "MathView3D.w"

ViewCntrlWin*CreateViewControl(void);

/*:767*//*769:*/
#line 12677 "MathView3D.w"

void CopyToVCWindow(void);

/*:769*//*771:*/
#line 12712 "MathView3D.w"

int CopyFromVCWindow(void);

/*:771*//*775:*/
#line 12763 "MathView3D.w"

void DisplayPreview(void);

/*:775*//*777:*/
#line 12803 "MathView3D.w"

void VCReshape(int w,int h);

/*:777*//*788:*/
#line 12947 "MathView3D.w"

void QuitPressed(muiObject*obj,enum muiReturnValue r);

/*:788*//*790:*/
#line 12957 "MathView3D.w"

void HidePressed(muiObject*obj,enum muiReturnValue r);

/*:790*//*792:*/
#line 12985 "MathView3D.w"

void ApplyViewPoint(muiObject*obj,enum muiReturnValue r);

/*:792*//*794:*/
#line 13013 "MathView3D.w"

void RSliderMove(muiObject*obj,enum muiReturnValue ret);
void PhiSliderMove(muiObject*obj,enum muiReturnValue ret);
void ThetaSliderMove(muiObject*obj,enum muiReturnValue ret);
void XSliderMove(muiObject*obj,enum muiReturnValue ret);
void YSliderMove(muiObject*obj,enum muiReturnValue ret);
void ZSliderMove(muiObject*obj,enum muiReturnValue ret);

/*:794*//*813:*/
#line 13288 "MathView3D.w"

void DisplayLightPreview(void);

/*:813*//*817:*/
#line 13364 "MathView3D.w"

void LCReshape(int w,int h);

/*:817*//*852:*/
#line 13892 "MathView3D.w"

AxisCtrlWin*CreateAxisControl(void);

/*:852*//*856:*/
#line 13941 "MathView3D.w"

void DisplayAxisPreview(void);

/*:856*//*859:*/
#line 13991 "MathView3D.w"

void RenderString(char*label);

/*:859*//*861:*/
#line 14001 "MathView3D.w"

double expt(double a,register int n);

/*:861*//*863:*/
#line 14018 "MathView3D.w"

double nicenum(double x,int round);

/*:863*//*865:*/
#line 14045 "MathView3D.w"

void loose_label(double min,double max,
Point3D*p1,Point3D*p2,double fntscale,
double alpha,Point3D*axis,
Point3D*offset);

/*:865*//*867:*/
#line 14098 "MathView3D.w"

void DrawAxis(int type,Point3D*origin,
Point3D*p1,Point3D*p2,
char*xlabel,char*ylabel,char*zlabel,double fntScale,int numlabel);

/*:867*/
#line 64 "MathView3D.w"

/*758:*/
#line 12502 "MathView3D.w"

/*760:*/
#line 12520 "MathView3D.w"

int FrontToken(char**l,char*delimiter,char*token)
{register char*lline= *l;
register unsigned int i,c;

while((*lline)&&((*lline==SPACE)||strchr(delimiter,*lline)))
lline++;
i= 0;
while((lline[i])&&(!strchr(delimiter,lline[i])))
i++;
if(i>strlen(lline))
return GL_FALSE;
for(c= 0;c<i;c++)
token[c]= lline[c];
token[c]= '\0';
(*l)= &lline[c];
return GL_TRUE;
}

/*:760*/
#line 12503 "MathView3D.w"

/*761:*/
#line 12539 "MathView3D.w"

double ScanDouble(char**lline,double default_val,int*fail)
{char token[25];
double value;

if(FrontToken(lline,NUMBER_DELIMITER,token)){
*fail= sscanf(token,"%lf",&value)<1;
return!(*fail)?value:default_val;
}
*fail= GL_TRUE;
return default_val;
}

/*:761*/
#line 12504 "MathView3D.w"

/*762:*/
#line 12552 "MathView3D.w"

int ScanInt(char**lline,int default_val,int*fail)
{char token[25];
int value;

if(FrontToken(lline,NUMBER_DELIMITER,token)){
sscanf(token,"%d",&value);
*fail= sscanf(token,"%d",&value)<1;
return!(*fail)?value:default_val;
}
*fail= GL_TRUE;
return default_val;
}

/*:762*/
#line 12505 "MathView3D.w"

/*763:*/
#line 12571 "MathView3D.w"

void PrepareLine(char commentchar,char**newline,char*lline)
{char*endp;

(*newline)= lline;
while(**newline==SPACE)
(*newline)++;
if(**newline==commentchar){
*newline= &lline[strlen(lline)];
return;
}
endp= &lline[strlen(lline)];
while(endp!=(*newline))
if((*(--endp))==commentchar){
*endp= '\0';
break;
}
}


/*:763*/
#line 12506 "MathView3D.w"


/*:758*/
#line 65 "MathView3D.w"

/*158:*/
#line 2988 "MathView3D.w"

/*159:*/
#line 3002 "MathView3D.w"

void SwapPoint3D(Point3D*p,Point3D*q)
{Point3D tmp;

tmp= *p;
*p= *q;
*q= tmp;
}

/*:159*/
#line 2989 "MathView3D.w"

/*160:*/
#line 3011 "MathView3D.w"

double Dot(Point3D*p1,Point3D*p2)
{
return p1->x*p2->x+p1->y*p2->y+p1->z*p2->z;
}

/*:160*/
#line 2990 "MathView3D.w"

/*161:*/
#line 3017 "MathView3D.w"

void Cross(Point3D*c,Point3D*a,Point3D*b)
{
c->x= -a->z*b->y+a->y*b->z;
c->y= a->z*b->x-a->x*b->z;
c->z= -a->y*b->x+a->x*b->y;
}

/*:161*/
#line 2991 "MathView3D.w"

/*162:*/
#line 3025 "MathView3D.w"

double Norm(Point3D*p)
{double n2;

n2= fabs(Dot(p,p));
return sqrt(n2);
}

/*:162*/
#line 2992 "MathView3D.w"

/*164:*/
#line 3057 "MathView3D.w"

void SphericalFromCartesian(Point3D*p,
double*r,double*theta,double*phi)
{double arg;
*r= Norm(p);

if((fabs(p->x)<TINY)&&(fabs(p->y)<TINY))
*phi= atan2(p->y,TINY);
else
*phi= atan2(p->y,p->x);
if(*r==0.0)
*theta= 0.0;
else{
if(fabs(*r)<TINY)*r= TINY;
arg= p->z/(*r);
if(arg>=1.0)arg= 1.0-TINY;
if(arg<=-1.0)arg= -1.0+TINY;
*theta= acos(arg);
}
}

/*:164*/
#line 2993 "MathView3D.w"

/*165:*/
#line 3078 "MathView3D.w"

void CartesianFromSpherical(double r,double theta,double phi,
Point3D*p)
{double st;

st= r*sin(theta);
p->x= cos(phi)*st;
p->y= sin(phi)*st;
p->z= r*cos(theta);
}

/*:165*/
#line 2994 "MathView3D.w"

/*163:*/
#line 3033 "MathView3D.w"

int SurfaceNormal(Point3D*nrm,
Point3D*p1,Point3D*p2,Point3D*p3)
{Point3D v1,v2;
double n;

v1.x= p2->x-p1->x;
v1.y= p2->y-p1->y;
v1.z= p2->z-p1->z;
v2.x= p3->x-p2->x;
v2.y= p3->y-p2->y;
v2.z= p3->z-p2->z;
Cross(nrm,&v1,&v2);
n= Norm(nrm);
if(n!=0.0){
nrm->x/= n;
nrm->y/= n;
nrm->z/= n;
return GL_TRUE;
}
return GL_FALSE;

}

/*:163*/
#line 2995 "MathView3D.w"

/*166:*/
#line 3093 "MathView3D.w"

void OrthogonalSet(Point3D*v1,Point3D*v2,Point3D*v3)
{double a,b,c,n,*p;
int i;

a= v1->x;b= v1->y;c= v1->z;
p= (double*)v1;
i= 0;
if(fabs(p[i])<fabs(p[1]))i= 1;
if(fabs(p[i])<fabs(p[2]))i= 2;
switch(i){
case 0:
/*168:*/
#line 3132 "MathView3D.w"

n= sqrt(fabs(a*a+c*c));
if(n>0.0){
v2->x= -a*b/n,
v2->y= n;
v2->z= -b*c/n;

v3->x= -c*sign(a)/n;
v3->y= 0.;
v3->z= fabs(a)/n;
}
else{
v2->y= 1.0;v2->x= v2->z= 0.0;
v3->z= 1.0;v2->x= v2->y= 0.0;
}

/*:168*/
#line 3105 "MathView3D.w"

break;
case 1:
/*169:*/
#line 3148 "MathView3D.w"

n= sqrt(fabs(b*b+c*c));
if(n>0.0){
v2->x= n;
v2->y= -a*b/n;
v2->z= -a*c/n;

v3->x= 0.0;
v3->y= -c*sign(b)/n;
v3->z= fabs(b)/n;
}
else{
v2->z= 1.0;v2->x= v2->y= 0.0;
v3->x= 1.0;v2->y= v2->z= 0.0;
}


/*:169*/
#line 3108 "MathView3D.w"

break;
case 2:
/*167:*/
#line 3116 "MathView3D.w"

n= sqrt(fabs(b*b+c*c));
if(n>0.0){
v2->x= n;
v2->y= -a*b/n;
v2->z= -a*c/n;

v3->x= 0.0;
v3->y= fabs(c)/n;
v3->z= -b*sign(c)/n;
}
else{
v2->x= 1.0;v2->y= v2->z= 0.0;
v3->y= 1.0;v2->x= v2->z= 0.0;
}

/*:167*/
#line 3111 "MathView3D.w"

break;
}
}

/*:166*/
#line 2996 "MathView3D.w"

/*170:*/
#line 3165 "MathView3D.w"

void GramSchmidt(Point3D*e1,Point3D*e2,Point3D*e3)
{double d12,d13,d23,n;

d12= Dot(e1,e2);
e2->x-= d12*e1->x;
e2->y-= d12*e1->y;
e2->z-= d12*e1->z;
n= Norm(e2);
e2->x/= n;e2->y/= n;e2->z/= n;
d13= Dot(e1,e3);
e3->x-= d13*e1->x;
e3->y-= d13*e1->y;
e3->z-= d13*e1->z;
n= Norm(e3);
e3->x/= n;e3->y/= n;e3->z/= n;
d23= Dot(e2,e3);
e3->x-= d23*e2->x;
e3->y-= d23*e2->y;
e3->z-= d23*e2->z;
n= Norm(e3);
e3->x/= n;e3->y/= n;e3->z/= n;
}

/*:170*/
#line 2997 "MathView3D.w"

/*171:*/
#line 3190 "MathView3D.w"

void ConePoint(double r,double phi,Point3D*u,Point3D*v,Point3D*p)
{double cphi,sphi;

cphi= r*cos(phi);
sphi= r*sin(phi);
p->x= (cphi*u->x+sphi*v->x);
p->y= (cphi*u->y+sphi*v->y);
p->z= (cphi*u->z+sphi*v->z);

}

/*:171*/
#line 2998 "MathView3D.w"

/*178:*/
#line 3357 "MathView3D.w"

double DistancePointCoord(Point3D*p,double x,double y,double z)
{double dx,dy,dz;

dx= p->x-x;
dy= p->y-y;
dz= p->z-z;
return sqrt(dx*dx+dy*dy+dz*dz);
}

/*:178*/
#line 2999 "MathView3D.w"

/*179:*/
#line 3369 "MathView3D.w"

void EstimateViewDepth(double*nearp,double*farp,
Point3D*vp,Point3D*p1,Point3D*p2)
{double d;

*nearp= *farp= DistancePointCoord(vp,p1->x,p1->y,p1->z);
d= DistancePointCoord(vp,p2->x,p1->y,p1->z);
AssignMinMax(*nearp,*farp,d)
d= DistancePointCoord(vp,p2->x,p2->y,p1->z);
AssignMinMax(*nearp,*farp,d)
d= DistancePointCoord(vp,p1->x,p2->y,p1->z);
AssignMinMax(*nearp,*farp,d)
d= DistancePointCoord(vp,p1->x,p1->y,p2->z);
AssignMinMax(*nearp,*farp,d)
d= DistancePointCoord(vp,p2->x,p1->y,p2->z);
AssignMinMax(*nearp,*farp,d)
d= DistancePointCoord(vp,p2->x,p2->y,p2->z);
AssignMinMax(*nearp,*farp,d)
d= DistancePointCoord(vp,p1->x,p2->y,p2->z);
AssignMinMax(*nearp,*farp,d)
*farp*= SQRT3;
}


/*:179*/
#line 3000 "MathView3D.w"


/*:158*/
#line 66 "MathView3D.w"

/*95:*/
#line 1937 "MathView3D.w"

/*96:*/
#line 1945 "MathView3D.w"

PointOctTree*InsertPointOctTree(int pIndex,PointOctTree*root)
{PointOctTree*tree= root,*next;
Point3D*tp,*p;
int index;

if(!root){
/*97:*/
#line 1976 "MathView3D.w"

{int i;

next= (PointOctTree*)malloc(sizeof(PointOctTree));
for(i= 0;i<8;i++)
next->leafs[i]= NULL;
next->point= pIndex;
next->polyNo= 0;
next->vpoly= NULL;
}

/*:97*/
#line 1952 "MathView3D.w"

return next;
}
p= PointPool+pIndex;
while(tree){
tp= PointPool+tree->point;
index= 0;
/*99:*/
#line 2001 "MathView3D.w"

if(p->x>tp->x)index|= 1;
if(p->y>tp->y)index|= 2;
if(p->z>tp->z)index|= 4;


/*:99*/
#line 1959 "MathView3D.w"

next= tree->leafs[index];
if(next)
tree= next;
else{
/*97:*/
#line 1976 "MathView3D.w"

{int i;

next= (PointOctTree*)malloc(sizeof(PointOctTree));
for(i= 0;i<8;i++)
next->leafs[i]= NULL;
next->point= pIndex;
next->polyNo= 0;
next->vpoly= NULL;
}

/*:97*/
#line 1964 "MathView3D.w"

tree->leafs[index]= next;
break;
}
}
return root;
}


/*:96*/
#line 1938 "MathView3D.w"

/*98:*/
#line 1987 "MathView3D.w"

int InOctTreeQ(Point3D*p,PointOctTree*tree)
{Point3D*tp;
double d;
int index= 0;

if(!tree)return NOT_FOUND;
tp= PointPool+tree->point;
d= fabs(p->x-tp->x)+fabs(p->y-tp->y)+fabs(p->z-tp->z);
if(d<TINY)return tree->point;
/*99:*/
#line 2001 "MathView3D.w"

if(p->x>tp->x)index|= 1;
if(p->y>tp->y)index|= 2;
if(p->z>tp->z)index|= 4;


/*:99*/
#line 1997 "MathView3D.w"

return InOctTreeQ(p,tree->leafs[index]);
}

/*:98*/
#line 1939 "MathView3D.w"

/*100:*/
#line 2007 "MathView3D.w"

PointOctTree*RemovePointOctTree(PointOctTree*tree)
{int i;

if(!tree)return NULL;
for(i= 0;i<8;i++)
if(tree->leafs[i])
RemovePointOctTree(tree->leafs[i]);
if(tree->vpoly)free(tree->vpoly);
free(tree);
return 0;
}


/*:100*/
#line 1940 "MathView3D.w"

/*101:*/
#line 2026 "MathView3D.w"

int InsertPolyInOctTree(PointOctTree*root,LinePolyList*poly,int pIndex)
{Point3D*p,*tp;
int index;

if(!root)return NOT_FOUND;
if(pIndex==root->point){
/*102:*/
#line 2043 "MathView3D.w"

if(root->vpoly){
LinePolyList**newvp;

newvp= (LinePolyList**)realloc(root->vpoly,(root->polyNo+1)*sizeof(LinePolyList*));
if(!newvp){
newvp= (LinePolyList**)calloc((root->polyNo+1),sizeof(LinePolyList*));
if(!newvp){
fprintf(stderr,"Panic can't allocate memory for polygon vertex link.");
return NOT_FOUND;
}
memcpy(*newvp,*(root->vpoly),root->polyNo*sizeof(LinePolyList));
free(root->vpoly);
}
newvp[root->polyNo]= poly;
(root->polyNo)++;
root->vpoly= newvp;
}
else{
root->vpoly= (LinePolyList**)malloc(sizeof(LinePolyList*));
*(root->vpoly)= poly;
root->polyNo= 1;
}


/*:102*/
#line 2033 "MathView3D.w"

return pIndex;
}
p= PointPool+pIndex;
tp= PointPool+root->point;
index= 0;
/*99:*/
#line 2001 "MathView3D.w"

if(p->x>tp->x)index|= 1;
if(p->y>tp->y)index|= 2;
if(p->z>tp->z)index|= 4;


/*:99*/
#line 2039 "MathView3D.w"

return InsertPolyInOctTree(root->leafs[index],poly,pIndex);
}

/*:101*/
#line 1941 "MathView3D.w"

/*103:*/
#line 2070 "MathView3D.w"

int RemovePolyFromOctTree(PointOctTree*root,LinePolyList*poly,int pIndex)
{Point3D*p,*tp;
int index;

if(!root)return NOT_FOUND;
if(pIndex==root->point){
/*104:*/
#line 2087 "MathView3D.w"

if(root->vpoly){
LinePolyList**newvp;
int i,k;

k= 0;
while(k<root->polyNo&&poly!=root->vpoly[k])k++;
if(k==root->polyNo){
fprintf(stderr,"Can't find the polygon in tree.\n");
return pIndex;
}
for(i= k;i<root->polyNo-1;i++)
root->vpoly[i]= root->vpoly[i+1];
(root->polyNo)--;
if(0==root->polyNo){
free(root->vpoly);
root->vpoly= NULL;
return pIndex;
}
newvp= (LinePolyList**)realloc(root->vpoly,(root->polyNo)*sizeof(LinePolyList*));
if(!newvp){
newvp= (LinePolyList**)calloc((root->polyNo),sizeof(LinePolyList*));
if(!newvp){
fprintf(stderr,"Panic can't allocate memory for polygon vertex link.");
return NOT_FOUND;
}
memcpy(*newvp,*(root->vpoly),root->polyNo*sizeof(LinePolyList));
free(root->vpoly);
}
root->vpoly= newvp;
}


/*:104*/
#line 2077 "MathView3D.w"

return pIndex;
}
p= PointPool+pIndex;
tp= PointPool+root->point;
index= 0;
/*99:*/
#line 2001 "MathView3D.w"

if(p->x>tp->x)index|= 1;
if(p->y>tp->y)index|= 2;
if(p->z>tp->z)index|= 4;


/*:99*/
#line 2083 "MathView3D.w"

return RemovePolyFromOctTree(root->leafs[index],poly,pIndex);
}

/*:103*/
#line 1942 "MathView3D.w"

/*105:*/
#line 2121 "MathView3D.w"

void StripPolygonsFromOctTree(PointOctTree*tree)
{int i;

if(!tree)return;
if(tree->vpoly)free(tree->vpoly);
tree->vpoly= NULL;
tree->polyNo= 0;
for(i= 0;i<8;i++)
if(tree->leafs[i])StripPolygonsFromOctTree(tree->leafs[i]);
}


/*:105*/
#line 1943 "MathView3D.w"


/*:95*/
#line 67 "MathView3D.w"

/*73:*/
#line 1599 "MathView3D.w"

/*70:*/
#line 1560 "MathView3D.w"

PointList*PrependPointList(PointList*lst,int index,
double size,
double red,double green,double blue)
{PointList*head;

head= (PointList*)malloc(sizeof(PointList));
head->next= lst;
head->point= index;
head->size= size;
head->color.r= red;
head->color.g= green;
head->color.b= blue;
return head;
}

/*:70*/
#line 1600 "MathView3D.w"

/*71:*/
#line 1576 "MathView3D.w"

PointList*RemovePointList(PointList*lst)
{PointList*head;

while(lst){
head= lst;
lst= lst->next;
free(head);
}
return NULL;
}

/*:71*/
#line 1601 "MathView3D.w"

/*:73*//*89:*/
#line 1844 "MathView3D.w"

/*77:*/
#line 1649 "MathView3D.w"

LinePolyList*PrependLinePolyList(LinePolyList*lst,
short int what,
int*lpoints,
int pno,
double r,double g,double b)
{LinePolyList*head;
int tri[3],*newpoints,nold;
int i,j;

if(lst&&Line3D==what&&lpoints[0]==lst->points[lst->pno-1]){
/*83:*/
#line 1754 "MathView3D.w"

nold= lst->pno;
newpoints= (int*)calloc(nold+pno-1,sizeof(int));
memcpy(newpoints,lst->points,nold*sizeof(int));
memcpy(newpoints+nold,lpoints+1,(pno-1)*sizeof(int));
free(lst->points);
lst->points= newpoints;
lst->pno+= pno-1;
return lst;


/*:83*/
#line 1660 "MathView3D.w"

}
if(Polygon3D==what&&pno>3){
if(COMTriangulation)
/*85:*/
#line 1774 "MathView3D.w"

{Point3D cntr= {0.0,0.0,0.0};
int i;

head= lst;

for(i= 0;i<pno;i++){
cntr.x+= PointPool[lpoints[i]].x;
cntr.y+= PointPool[lpoints[i]].y;
cntr.z+= PointPool[lpoints[i]].z;
}
cntr.x/= pno;
cntr.y/= pno;
cntr.z/= pno;

tri[0]= AddToPointPool(cntr.x,cntr.y,cntr.z);

for(i= 1;i<pno;i++){
tri[1]= lpoints[i-1];
tri[2]= lpoints[i];
head= PrependLinePolyList(head,what,tri,3,r,g,b);
}
tri[1]= lpoints[pno-1];
tri[2]= lpoints[0];
return PrependLinePolyList(head,what,tri,3,r,g,b);
}

/*:85*/
#line 1664 "MathView3D.w"

else
/*84:*/
#line 1765 "MathView3D.w"

{int i;

head= PrependLinePolyList(lst,what,lpoints,3,r,g,b);
for(i= 1;i<pno-1;i++)
lpoints[i]= lpoints[i+1];
return PrependLinePolyList(head,what,lpoints,pno-1,r,g,b);
}

/*:84*/
#line 1666 "MathView3D.w"

}
head= (LinePolyList*)malloc(sizeof(LinePolyList));
if(!head){
fprintf(stderr,"Panic ... can't allocate memory for LinePolyList.");
return lst;
}
head->prev= NULL;
head->next= lst;
if(lst)lst->prev= head;
head->what= what;
head->pTexturePoints= NULL;
head->pTexQuadrants= 0;
head->pno= pno;
head->points= (int*)calloc(pno,sizeof(int));

head->points[0]= lpoints[0];
for(i= j= 1;i<pno;i++)
if(lpoints[i-1]!=lpoints[i])
head->points[j++]= lpoints[i];
if(j!=pno)head->pno= j;
head->color.r= r;
head->color.g= g;
head->color.b= b;
head->aux1= -1;
head->aux2= -1.0;
return head;
}

/*:77*/
#line 1845 "MathView3D.w"

/*78:*/
#line 1699 "MathView3D.w"

LinePolyList*DeleteSinglePolygon(LinePolyList*lstBegin,LinePolyList*lst)
{LinePolyList*head,*tail;

if(!lst)return lst;
if(NULL==(lst->prev)){
tail= lst->next;
/*80:*/
#line 1730 "MathView3D.w"

free(lst->points);
if(lst->pTexturePoints)free(lst->pTexturePoints);
free(lst);


/*:80*/
#line 1706 "MathView3D.w"

if(tail)tail->prev= NULL;
return tail;
}
head= lst->prev;
tail= lst->next;
head->next= tail;
if(tail)tail->prev= head;
/*80:*/
#line 1730 "MathView3D.w"

free(lst->points);
if(lst->pTexturePoints)free(lst->pTexturePoints);
free(lst);


/*:80*/
#line 1714 "MathView3D.w"

return lstBegin;
}

/*:78*/
#line 1846 "MathView3D.w"


/*:89*//*114:*/
#line 2258 "MathView3D.w"

TextureMapping*CreateTextureMapping(
int how,
int repeat,
Point3D*move,
Point3D*raxis,
double alpha,
Point3D*scale,
double scS,double scT,
TextureList*which)
{TextureMapping*tm;

/*118:*/
#line 2346 "MathView3D.w"

if(!which)return NULL;
tm= (TextureMapping*)malloc(sizeof(TextureMapping));
if(!tm)return NULL;
tm->type= how;
tm->tileTexture= repeat;
tm->tex= which;

/*:118*/
#line 2270 "MathView3D.w"

/*119:*/
#line 2354 "MathView3D.w"

if(move){
tm->translate= (Point3D*)malloc(sizeof(Point3D));
*(tm->translate)= *move;
}
else
tm->translate= NULL;

/*:119*/
#line 2271 "MathView3D.w"

/*120:*/
#line 2362 "MathView3D.w"

if(raxis){
tm->axis= (Point3D*)malloc(sizeof(Point3D));
*(tm->axis)= *raxis;
tm->angle= alpha;
}
else
tm->axis= NULL;

/*:120*/
#line 2272 "MathView3D.w"

/*121:*/
#line 2371 "MathView3D.w"

if(scale){
tm->scale= (Point3D*)malloc(sizeof(Point3D));
*(tm->scale)= *scale;
}
else
tm->scale= NULL;
tm->scaleS= scS;
tm->scaleT= scT;

/*:121*/
#line 2273 "MathView3D.w"

return tm;
}

/*:114*//*115:*/
#line 2277 "MathView3D.w"

TextureMapping*DefaultTextureMapping(TextureList*which)

{const int repeat= TILE_TEXTURE_S|TILE_TEXTURE_T;
const double scT= 1.0,scS= 1.0;
int how;
TextureMapping*tm;
Point3D*scale,*move;
char*c;

/*116:*/
#line 2298 "MathView3D.w"

c= TexMapType;
scale= malloc(sizeof(Point3D));
move= malloc(sizeof(Point3D));
while(*c)*(c++)= tolower(*c);
switch(*TexMapType){
case'p':
how= TEXTURE_MAP_PLANAR;
scale->x= (BoundP2.x-BoundP1.x);
scale->y= (BoundP2.y-BoundP1.y);
scale->z= (BoundP2.z-BoundP1.z);
*move= BoundP1;
if(scale->x!=0.0)scale->x= 1.0/scale->x;
if(scale->y!=0.0)scale->y= 1.0/scale->y;
if(scale->z!=0.0)scale->z= 1.0/scale->z;
break;
case'c':
how= TEXTURE_MAP_CYLINDRICAL;
/*117:*/
#line 2341 "MathView3D.w"

move->x= 0.5*(BoundP1.x+BoundP2.x);
move->y= 0.5*(BoundP1.y+BoundP2.y);
move->z= 0.5*(BoundP1.z+BoundP2.z);

/*:117*/
#line 2316 "MathView3D.w"

move->z= BoundP1.z;
scale->x= scale->y= 1.0;
scale->z= BoundP2.z-BoundP1.z;
if(scale->z!=0.0)scale->z= 1.0/scale->z;
break;
case's':
how= TEXTURE_MAP_SPHERICAL;
/*117:*/
#line 2341 "MathView3D.w"

move->x= 0.5*(BoundP1.x+BoundP2.x);
move->y= 0.5*(BoundP1.y+BoundP2.y);
move->z= 0.5*(BoundP1.z+BoundP2.z);

/*:117*/
#line 2324 "MathView3D.w"

free(scale);
scale= NULL;
break;
case't':how= TEXTURE_MAP_TORUS;
/*117:*/
#line 2341 "MathView3D.w"

move->x= 0.5*(BoundP1.x+BoundP2.x);
move->y= 0.5*(BoundP1.y+BoundP2.y);
move->z= 0.5*(BoundP1.z+BoundP2.z);

/*:117*/
#line 2329 "MathView3D.w"

free(scale);
scale= NULL;
break;
default:
how= TEXTURE_MAP_PLANAR;
free(scale);
free(move);
scale= move= NULL;
break;
}

/*:116*/
#line 2287 "MathView3D.w"

/*118:*/
#line 2346 "MathView3D.w"

if(!which)return NULL;
tm= (TextureMapping*)malloc(sizeof(TextureMapping));
if(!tm)return NULL;
tm->type= how;
tm->tileTexture= repeat;
tm->tex= which;

/*:118*/
#line 2288 "MathView3D.w"

/*119:*/
#line 2354 "MathView3D.w"

if(move){
tm->translate= (Point3D*)malloc(sizeof(Point3D));
*(tm->translate)= *move;
}
else
tm->translate= NULL;

/*:119*/
#line 2289 "MathView3D.w"

/*121:*/
#line 2371 "MathView3D.w"

if(scale){
tm->scale= (Point3D*)malloc(sizeof(Point3D));
*(tm->scale)= *scale;
}
else
tm->scale= NULL;
tm->scaleS= scS;
tm->scaleT= scT;

/*:121*/
#line 2290 "MathView3D.w"

tm->axis= NULL;
tm->angle= 0.0;
if(scale)free(scale);
if(move)free(move);
return tm;
}

/*:115*//*122:*/
#line 2381 "MathView3D.w"

TextureMapping*DeleteTextureMapping(TextureMapping*tm)
{
if(!tm)return NULL;
if(tm->translate)free(tm->translate);
if(tm->axis)free(tm->axis);
if(tm->scale)free(tm->scale);
free(tm);
return NULL;
}


/*:122*//*142:*/
#line 2756 "MathView3D.w"

Mesh3DList*PrependMesh3DList(Mesh3DList*lst,
int ndim,int mdim,
Point3D*data2d,
RGBColor*c,
int nc,int mc,
double*colors)
{Mesh3DList*head;
Point3D*p,normij;
int n,m,i,j,found;
int*count;

head= (Mesh3DList*)malloc(sizeof(Mesh3DList));
if(!head)return lst;
head->next= lst;
head->aux1= 0;
head->aux2= 0.0;
head->attrib= MESH_OPEN;
head->mTexture= NULL;
if(NULL==c)
head->mColor.r= head->mColor.g= head->mColor.b= 1.0;
else
head->mColor= *c;

head->meshID= MeshCount;
sprintf(head->name,"MVMesh3D[%d]",MeshCount++);

head->cDimU= nc;
head->cDimV= mc;
if(colors){
head->mpolyColors= (RGBColor*)calloc(nc*mc,sizeof(RGBColor));
memcpy(head->mpolyColors,colors,3*nc*mc*sizeof(double));
if(nc==ndim&&mc==mdim)
head->attrib|= MESH_VERTEX_COLOR;
else
head->attrib|= MESH_FACE_COLOR;
}
else
head->mpolyColors= NULL;

head->dimU= ndim;
head->dimV= mdim;
head->mu= ndim*mdim;

head->mPoints= (Point3D*)calloc(head->mu,sizeof(Point3D));
head->mNormals= (Point3D*)calloc(head->mu,sizeof(Point3D));
head->mTexturePoints= (Point3D*)calloc(head->mu,sizeof(Point3D));
head->mPointNo= 0;
head->mesh= (int*)calloc(head->mu,sizeof(int));
/*143:*/
#line 2815 "MathView3D.w"

for(m= 0;m<head->mu;m++){
found= GL_FALSE;
for(i= 0;i<head->mPointNo;i++){
p= head->mPoints+i;
if((fabs(data2d[m].x-p->x)+
fabs(data2d[m].y-p->y)+
fabs(data2d[m].z-p->z))<TINY){
found= GL_TRUE;
break;
}
}
if(!found){
head->mPoints[head->mPointNo]= data2d[m];
head->mNormals[head->mPointNo]= ZeroPoint;
i= head->mPointNo++;
}
head->mesh[m]= i;
}

/*:143*/
#line 2805 "MathView3D.w"

/*144:*/
#line 2835 "MathView3D.w"

count= (int*)calloc(head->mu,sizeof(int));
for(i= 0;i<head->mPointNo;i++)
count[i]= 0;
n= head->dimU-1;
m= head->dimV-1;
/*146:*/
#line 2872 "MathView3D.w"

for(i= 0;i<n;i++)
for(j= 0;j<m;j++){
SurfaceNormal(&normij,MeshCoordPtr(head,i,j+1),
MeshCoordPtr(head,i,j),
MeshCoordPtr(head,i+1,j));
/*145:*/
#line 2866 "MathView3D.w"

MeshNormal(head,i,j).x+= normij.x;
MeshNormal(head,i,j).y+= normij.y;
MeshNormal(head,i,j).z+= normij.z;
count[MeshIndex(head,i,j)]++;

/*:145*/
#line 2878 "MathView3D.w"

}

/*:146*/
#line 2841 "MathView3D.w"

/*147:*/
#line 2881 "MathView3D.w"

for(i= 1;i<head->dimU;i++)
for(j= 0;j<m;j++){
SurfaceNormal(&normij,MeshCoordPtr(head,i-1,j),
MeshCoordPtr(head,i,j),
MeshCoordPtr(head,i,j+1));
/*145:*/
#line 2866 "MathView3D.w"

MeshNormal(head,i,j).x+= normij.x;
MeshNormal(head,i,j).y+= normij.y;
MeshNormal(head,i,j).z+= normij.z;
count[MeshIndex(head,i,j)]++;

/*:145*/
#line 2887 "MathView3D.w"

}

/*:147*/
#line 2842 "MathView3D.w"

/*148:*/
#line 2890 "MathView3D.w"

for(i= 1;i<head->dimU;i++)
for(j= 1;j<head->dimV;j++){
SurfaceNormal(&normij,MeshCoordPtr(head,i,j-1),
MeshCoordPtr(head,i,j),
MeshCoordPtr(head,i-1,j));
/*145:*/
#line 2866 "MathView3D.w"

MeshNormal(head,i,j).x+= normij.x;
MeshNormal(head,i,j).y+= normij.y;
MeshNormal(head,i,j).z+= normij.z;
count[MeshIndex(head,i,j)]++;

/*:145*/
#line 2896 "MathView3D.w"

}

/*:148*/
#line 2843 "MathView3D.w"

/*149:*/
#line 2899 "MathView3D.w"

for(i= 0;i<n;i++)
for(j= 1;j<head->dimV;j++){
SurfaceNormal(&normij,MeshCoordPtr(head,i+1,j),
MeshCoordPtr(head,i,j),
MeshCoordPtr(head,i,j-1));
/*145:*/
#line 2866 "MathView3D.w"

MeshNormal(head,i,j).x+= normij.x;
MeshNormal(head,i,j).y+= normij.y;
MeshNormal(head,i,j).z+= normij.z;
count[MeshIndex(head,i,j)]++;

/*:145*/
#line 2905 "MathView3D.w"

}

/*:149*/
#line 2844 "MathView3D.w"

for(i= 0;i<head->mPointNo;i++)
if(count[i]){
double absn;
head->mNormals[i].x/= count[i];
head->mNormals[i].y/= count[i];
head->mNormals[i].z/= count[i];
absn= Norm(head->mNormals+i);
if(absn!=0.0){
absn= 1.0/absn;
head->mNormals[i].x*= absn;
head->mNormals[i].y*= absn;
head->mNormals[i].z*= absn;
}
else{
head->mNormals[i].x= 0.0;
head->mNormals[i].y= 0.0;
head->mNormals[i].z= 1.0;
}
}
free(count);

/*:144*/
#line 2806 "MathView3D.w"

/*150:*/
#line 2908 "MathView3D.w"

found= GL_TRUE;
for(j= 0;j<head->dimV&&found;j++)
found= found&&(MeshIndex(head,0,j)==MeshIndex(head,head->dimU-1,j));
if(found)head->attrib|= MESH_UCLOSED;

/*:150*/
#line 2807 "MathView3D.w"

/*151:*/
#line 2914 "MathView3D.w"

found= GL_TRUE;
for(i= 0;i<head->dimU&&found;i++)
found= found&&(MeshIndex(head,i,0)==MeshIndex(head,i,head->dimV-1));
if(found)head->attrib|= MESH_VCLOSED;

/*:151*/
#line 2808 "MathView3D.w"

return head;
}

/*:142*//*154:*/
#line 2952 "MathView3D.w"

Mesh3DList*RemoveMesh3DList(Mesh3DList*lst)
{Mesh3DList*tail;

while(lst){
tail= lst->next;
if(lst->mpolyColors)free(lst->mpolyColors);
if(lst->mTexture)DeleteTextureMapping(lst->mTexture);
free(lst->mPoints);
free(lst->mNormals);
free(lst->mTexturePoints);
free(lst->mesh);
free(lst);
lst= tail;
}
return lst;
}


/*:154*//*156:*/
#line 2972 "MathView3D.w"

/*108:*/
#line 2147 "MathView3D.w"

int AddToPointPool(double x,double y,double z)
{int pn,i;
Point3D newp,*pos,**nvec;

if(!PointPool){
PointPool= (Point3D*)calloc(MaxPoolSize= INIT_POOLSIZE,sizeof(Point3D));
NormalPool= (Point3D**)calloc(INIT_POOLSIZE,sizeof(Point3D*));
for(i= 0;i<MaxPoolSize;i++)
NormalPool[i]= NULL;
PoolSize= 0;
}
pn= PoolSize;
pos= PointPool;
newp.x= x;
newp.y= y;
newp.z= z;
i= InOctTreeQ(&newp,OctalPointTree);
if(NOT_FOUND<i)return i;
if(pn==MaxPoolSize){
pos= (Point3D*)calloc(MaxPoolSize+POOL_INC,sizeof(Point3D));
if(!pos){
fprintf(stderr,"Panic ... failed to reallocate PointPool.\n");
exit(1);
}
memcpy(pos,PointPool,MaxPoolSize*sizeof(Point3D));
free(PointPool);
PointPool= pos;

nvec= (Point3D**)calloc(MaxPoolSize+POOL_INC,sizeof(Point3D*));
if(!nvec){
fprintf(stderr,"Panic ... failed to reallocate NormalPool.\n");
exit(1);
}
memcpy(nvec,NormalPool,MaxPoolSize*sizeof(Point3D*));
free(NormalPool);
NormalPool= nvec;
MaxPoolSize+= POOL_INC;
for(i= PoolSize;i<MaxPoolSize;i++)
NormalPool[i]= NULL;
}

PointPool[pn].x= x;
PointPool[pn].y= y;
PointPool[pn].z= z;
OctalPointTree= InsertPointOctTree(pn,OctalPointTree);
/*109:*/
#line 2197 "MathView3D.w"

if(BoundP1.x>x)BoundP1.x= x;
if(BoundP2.x<x)BoundP2.x= x;
if(BoundP1.y>y)BoundP1.y= y;
if(BoundP2.y<y)BoundP2.y= y;
if(BoundP1.z>z)BoundP1.z= z;
if(BoundP2.z<z)BoundP2.z= z;

/*:109*/
#line 2193 "MathView3D.w"

return PoolSize++;
}

/*:108*/
#line 2973 "MathView3D.w"


/*:156*/
#line 68 "MathView3D.w"

/*88:*/
#line 1836 "MathView3D.w"

/*72:*/
#line 1588 "MathView3D.w"

int LengthPointList(PointList*lst)
{
int len= 0;

while(lst){
len++;
lst= lst->next;
}
return len;
}
/*:72*/
#line 1837 "MathView3D.w"

/*79:*/
#line 1718 "MathView3D.w"

LinePolyList*RemoveLinePolyList(LinePolyList*lst)
{LinePolyList*head;

while(lst){
head= lst->next;
/*80:*/
#line 1730 "MathView3D.w"

free(lst->points);
if(lst->pTexturePoints)free(lst->pTexturePoints);
free(lst);


/*:80*/
#line 1724 "MathView3D.w"

lst= head;
}
return NULL;
}

/*:79*/
#line 1838 "MathView3D.w"

/*81:*/
#line 1736 "MathView3D.w"

LinePolyList*JoinLinePolyList(LinePolyList*l1,LinePolyList*l2)
{LinePolyList*head;

head= l1;
while(head->prev)head= head->prev;
while(l1->next)l1= l1->next;
while(l2->prev)l2= l2->prev;
l1->next= l2;
l2->prev= l1;
return head;

}

/*:81*/
#line 1839 "MathView3D.w"

/*87:*/
#line 1813 "MathView3D.w"

void ReversePolygons(LinePolyList*lst)
{int n,tmp;
Point3D texTmp;
while(lst){
if(Polygon3D==lst->what){
n= lst->pno;
if(3==n){
tmp= lst->points[2];
lst->points[2]= lst->points[0];
lst->points[0]= tmp;
if(lst->pTexturePoints){
texTmp= lst->pTexturePoints[2];
lst->pTexturePoints[2]= lst->pTexturePoints[0];
lst->pTexturePoints[0]= texTmp;
}
}
}
lst= lst->next;
}
}


/*:87*/
#line 1840 "MathView3D.w"

/*141:*/
#line 2752 "MathView3D.w"

/*152:*/
#line 2921 "MathView3D.w"

Mesh3DList*FlipMeshNormals(Mesh3DList*lst)
{int i,n;
Point3D*normi;
Mesh3DList*head= lst;
while(lst){
n= lst->mPointNo;
for(i= 0;i<n;i++){
normi= lst->mNormals+i;
normi->x= -normi->x;
normi->y= -normi->y;
normi->z= -normi->z;
}
lst= lst->next;
}
return head;
}

/*:152*/
#line 2753 "MathView3D.w"

/*153:*/
#line 2940 "MathView3D.w"

int LengthMesh3DList(Mesh3DList*lst)
{int n= 0;

while(lst){
n++;
lst= lst->next;
}
return n;
}

/*:153*/
#line 2754 "MathView3D.w"


/*:141*/
#line 1841 "MathView3D.w"

/*211:*/
#line 3897 "MathView3D.w"

/*212:*/
#line 3908 "MathView3D.w"

Quaternion*MakeQuaternion(Quaternion*q,double theta,Point3D*v)
{double st,ct;
double vn;

vn= Norm(v);
if(vn!=0.0){
vn= 1.0/vn;
theta*= 0.5;
st= sin(theta);
ct= cos(theta);
vn*= st;
q->q0= ct;
q->q1= v->x*vn;
q->q2= v->y*vn;
q->q3= v->z*vn;
return q;
}
return NULL;
}

/*:212*/
#line 3898 "MathView3D.w"

/*213:*/
#line 3930 "MathView3D.w"

Point3D*GetQuaternionRotation(double*theta,Point3D*axes,Quaternion*q)
{double st;

*theta= acos(q->q0);
st= sin(*theta);
if(st!=0.0){
axes->x= q->q1/st;
axes->y= q->q2/st;
axes->z= q->q3/st;
*theta*= 2.0;
return axes;
}
return NULL;
}

/*:213*/
#line 3899 "MathView3D.w"

/*214:*/
#line 3952 "MathView3D.w"

Quaternion*QuaternionTimes(Quaternion*qp,Quaternion*q,Quaternion*p)
{
qp->q0= p->q0*q->q0-p->q1*q->q1-p->q2*q->q2-p->q3*q->q3;
qp->q1= p->q1*q->q0+p->q0*q->q1+p->q3*q->q2-p->q2*q->q3;
qp->q2= p->q2*q->q0-p->q3*q->q1+p->q0*q->q2+p->q1*q->q3;
qp->q3= p->q3*q->q0+p->q2*q->q1-p->q1*q->q2+p->q0*q->q3;
return qp;
}

/*:214*/
#line 3900 "MathView3D.w"

/*215:*/
#line 3964 "MathView3D.w"

Quaternion*InverseQuaternion(Quaternion*q)
{
q->q1= -q->q1;
q->q2= -q->q2;
q->q3= -q->q3;
return q;
}

/*:215*/
#line 3901 "MathView3D.w"

/*216:*/
#line 3976 "MathView3D.w"

void QuaternionRotate(Point3D*rp,Quaternion*q,Point3D*p)
{Quaternion conjQ,v,result,tmp;

v.q0= 0.0;v.q1= p->x;v.q2= p->y;v.q3= p->z;
conjQ= *q;
InverseQuaternion(&conjQ);
QuaternionTimes(&result,QuaternionTimes(&tmp,q,&v),&conjQ);
rp->x= result.q1;
rp->y= result.q2;
rp->z= result.q3;
}

/*:216*/
#line 3902 "MathView3D.w"


/*:211*/
#line 1842 "MathView3D.w"


/*:88*/
#line 69 "MathView3D.w"

/*110:*/
#line 2224 "MathView3D.w"

/*123:*/
#line 2409 "MathView3D.w"

int UpdateTextureCoordinates(TextureMapping*map,
Point3D*tp,Point3D*p,int*quad,int n)
{int mtype;
int i;
Point3D*move,*axis;
Point3D*scale;
double nx,ny,nz,st,ct;
double ox,oy,oz,x,y,z;
double rmat[3][3];
int Quadrant,xyQuadrant;

if(!map)return 0;
Quadrant= 0;
/*126:*/
#line 2484 "MathView3D.w"

mtype= map->type;
move= map->translate;
axis= map->axis;
scale= map->scale;
if(axis){
nx= map->axis->x;
ny= map->axis->y;
nz= map->axis->z;
ct= cos(map->angle);
st= sin(map->angle);
}

/*:126*/
#line 2423 "MathView3D.w"

if(axis)
/*667:*/
#line 11101 "MathView3D.w"

{
rmat[0][0]= nx*nx+(1.-nx*nx)*ct;
rmat[0][1]= nx*ny*(1.-ct)+nz*st;
rmat[0][2]= nx*nz*(1.-ct)-ny*st;

rmat[1][0]= nx*ny*(1.-ct)-nz*st;
rmat[1][1]= ny*ny+(1.-ny*ny)*ct;
rmat[1][2]= ny*nz*(1.-ct)+nx*st;

rmat[2][0]= nx*nz*(1.-ct)+ny*st;
rmat[2][1]= ny*nz*(1.-ct)-nx*st;
rmat[2][2]= nz*nz+(1.-nz*nz)*ct;
}

/*:667*/
#line 2425 "MathView3D.w"

for(i= 0;i<n;i++,p++,tp++){
/*127:*/
#line 2498 "MathView3D.w"

if(move){
ox= p->x-move->x;
oy= p->y-move->y;
oz= p->z-move->z;
}
else{
ox= p->x;
oy= p->y;
oz= p->z;
}
if(axis)
/*666:*/
#line 11094 "MathView3D.w"

{
x= rmat[0][0]*ox+rmat[0][1]*oy+rmat[0][2]*oz;
y= rmat[1][0]*ox+rmat[1][1]*oy+rmat[1][2]*oz;
z= rmat[2][0]*ox+rmat[2][1]*oy+rmat[2][2]*oz;
}

/*:666*/
#line 2510 "MathView3D.w"

else{
x= ox;
y= oy;
z= oz;
}
if(scale){
x*= scale->x;
y*= scale->y;
z*= scale->z;
}


/*:127*/
#line 2427 "MathView3D.w"

/*124:*/
#line 2451 "MathView3D.w"

if(x>=0.0&&y>=0.0)
xyQuadrant= InQuadrant1;
else if(x<0.0&&y>=0.0)
xyQuadrant= InQuadrant2;
else if(x<0.0&&y<0.0)
xyQuadrant= InQuadrant3;
else if(x>=0.0&&y<0.0)
xyQuadrant= InQuadrant4;
if(quad)quad[i]= xyQuadrant;
Quadrant|= xyQuadrant;

/*:124*/
#line 2428 "MathView3D.w"

switch(mtype){
case TEXTURE_MAP_PLANAR:
/*128:*/
#line 2523 "MathView3D.w"

tp->x= x;
tp->y= y;
tp->z= z;

/*:128*/
#line 2431 "MathView3D.w"

break;
case TEXTURE_MAP_CYLINDRICAL:
/*129:*/
#line 2528 "MathView3D.w"

{double r,phi;

r= sqrt(x*x+y*y);
phi= atan2(y,(x!=0.0?x:TINY));
if(phi<0.0)phi+= TWO_PI;
tp->x= phi/TWO_PI;
tp->y= z;
tp->z= r;
}


/*:129*/
#line 2434 "MathView3D.w"

break;
case TEXTURE_MAP_SPHERICAL:
/*130:*/
#line 2540 "MathView3D.w"

{double r,th,phi;

r= sqrt(x*x+y*y+z*z);
phi= atan2(y,(x!=0.0?x:TINY));
if(phi<0.0)phi+= TWO_PI;
th= r>0.0&&fabs(z)<r?acos(z/r):(z>0.0?0.0:PI);
tp->x= phi/TWO_PI;
tp->y= 1.-th/PI;
tp->z= r;
}

/*:130*/
#line 2437 "MathView3D.w"

break;
case TEXTURE_MAP_TORUS:
/*132:*/
#line 2567 "MathView3D.w"

{double r1,rho,r,th,phi;

r1= (rho= x*x+y*y)+z*z;
if(rho<TINY)rho= TINY;
r= 0.5*r1/sqrt(rho);
phi= atan2(y,(x!=0.0?x:TINY));
if(phi<0.0)phi+= TWO_PI;
th= (z>0.0?1.0:-1.0)*acos(r1>TINY?-1.0+2.0*rho/r1:1.0)+PI;
tp->x= phi/TWO_PI;
tp->y= th/TWO_PI;
tp->z= r;
}


/*:132*/
#line 2440 "MathView3D.w"

break;
default:
/*128:*/
#line 2523 "MathView3D.w"

tp->x= x;
tp->y= y;
tp->z= z;

/*:128*/
#line 2443 "MathView3D.w"

break;
}

}
return Quadrant;
}

/*:123*/
#line 2225 "MathView3D.w"

/*125:*/
#line 2464 "MathView3D.w"

void UpdatePolygonTextureCoord(TextureMapping*map,LinePolyList*lst)
{Point3D*p;
int i;

while(lst){
p= (Point3D*)calloc(lst->pno,sizeof(Point3D));
if(!lst->pTexturePoints)
lst->pTexturePoints= (Point3D*)calloc(lst->pno,sizeof(Point3D));
for(i= 0;i<lst->pno;i++){
p[i]= PointPool[lst->points[i]];
lst->pTexturePoints[i]= ZeroPoint;
}
lst->pTexQuadrants= 
UpdateTextureCoordinates(map,lst->pTexturePoints,p,NULL,lst->pno);
free(p);
lst= lst->next;
}
}

/*:125*/
#line 2226 "MathView3D.w"

/*133:*/
#line 2583 "MathView3D.w"

void UpdateMeshTextureCoordinates(Mesh3DList*mesh)
{int mtype;
Point3D*tp;
int i,j;
double n,m;

while(mesh){
if(mesh->mTexture){
mtype= mesh->mTexture->type;
if((TEXTURE_MAP_MESH_ST==mtype)||(TEXTURE_MAP_MESH_TS==mtype))
/*137:*/
#line 2665 "MathView3D.w"

{
n= mesh->dimU-1;
m= mesh->dimV-1;
for(i= 0;i<mesh->dimU;i++)
for(j= 0;j<mesh->dimV;j++){
tp= MeshTexCoordPtr(mesh,i,j);
tp->z= 0.0;
if(TEXTURE_MAP_MESH_ST==mtype){
tp->x= ((double)i/(double)n);
tp->y= ((double)j/(double)m);
}
else{
tp->y= ((double)i/(double)n);
tp->x= ((double)j/(double)m);
}
}
}

/*:137*/
#line 2594 "MathView3D.w"

else
/*134:*/
#line 2610 "MathView3D.w"

{
Point3D*mp;
int*xyQuadrants,quad;
int i,j,mu;

mp= (Point3D*)calloc(mu= mesh->dimU*mesh->dimV,sizeof(Point3D));
xyQuadrants= (int*)calloc(mu,sizeof(int));
for(i= 0;i<mu;i++)xyQuadrants[i]= 0;
for(i= 0;i<mesh->dimU;i++)
for(j= 0;j<mesh->dimV;j++)
mp[MeshIndexExpanded(mesh,i,j)]= MeshCoord(mesh,i,j);
UpdateTextureCoordinates(
mesh->mTexture,
mesh->mTexturePoints,
mp,xyQuadrants,mu);
/*135:*/
#line 2631 "MathView3D.w"

if(mesh->mTexture->type>TEXTURE_MAP_PLANAR){
Point3D*p;

for(i= 0;i<mesh->dimU-1;i++)
for(j= 0;j<mesh->dimV-1;j++){

quad= xyQuadrants[MeshIndexExpanded(mesh,i,j)]|
xyQuadrants[MeshIndexExpanded(mesh,i+1,j)]|
xyQuadrants[MeshIndexExpanded(mesh,i+1,j+1)];
if(((quad&InQuadrant1)==InQuadrant1)&&
((quad&InQuadrant4)==InQuadrant4)){
p= MeshTexCoordPtr(mesh,i,j);
/*136:*/
#line 2662 "MathView3D.w"

p->x= (fabs(p->x)<0.5)?1.0+p->x:p->x;

/*:136*/
#line 2644 "MathView3D.w"

p= MeshTexCoordPtr(mesh,i+1,j);
/*136:*/
#line 2662 "MathView3D.w"

p->x= (fabs(p->x)<0.5)?1.0+p->x:p->x;

/*:136*/
#line 2646 "MathView3D.w"

p= MeshTexCoordPtr(mesh,i+1,j+1);
/*136:*/
#line 2662 "MathView3D.w"

p->x= (fabs(p->x)<0.5)?1.0+p->x:p->x;

/*:136*/
#line 2648 "MathView3D.w"

}

quad= xyQuadrants[MeshIndexExpanded(mesh,i+1,j+1)]|
xyQuadrants[MeshIndexExpanded(mesh,i,j+1)]|
xyQuadrants[MeshIndexExpanded(mesh,i,j)];
if(((quad&InQuadrant1)==InQuadrant1)&&
((quad&InQuadrant4)==InQuadrant4)){
p= MeshTexCoordPtr(mesh,i,j+1);
/*136:*/
#line 2662 "MathView3D.w"

p->x= (fabs(p->x)<0.5)?1.0+p->x:p->x;

/*:136*/
#line 2657 "MathView3D.w"

}
}
}

/*:135*/
#line 2626 "MathView3D.w"

free(mp);
free(xyQuadrants);
}

/*:134*/
#line 2596 "MathView3D.w"

}
mesh= mesh->next;
}
}

/*:133*/
#line 2227 "MathView3D.w"


/*:110*/
#line 70 "MathView3D.w"

/*217:*/
#line 3990 "MathView3D.w"

/*218:*/
#line 3998 "MathView3D.w"

#define KEYWORDNO    11
#define BBOX         0
#define VPOINT       1
#define LSOURCE      2
#define POINT        3
#define LINE         4
#define POLY         5
#define AMBIENT      6
#define COLOR        7
#define SURFACECOLOR 8
#define MESH         9
#define COLORMESH   10

char*Keyword[KEYWORDNO]= {"boundingbox",
"viewpoint",
"lightsources",
"point",
"line",
"polygon",
"ambient",
"color",
"surfacecolor",
"mesh",
"colormesh"};

/*:218*/
#line 3991 "MathView3D.w"

/*219:*/
#line 4024 "MathView3D.w"

int SearchKeyWord(char*cline)
{int i;
for(i= KEYWORDNO-1;i>-1;i--)
if(cline==strstr(cline,Keyword[i])){
return i;
}
return-1;
}


/*:219*/
#line 3992 "MathView3D.w"

/*234:*/
#line 4231 "MathView3D.w"

int LUDecomp(int n,double a[][5],int ip[],int*ier)
{
int nm1,k,kp1,m;
register int i,j;
register double t;

*ier= 0;
ip[n]= 1;
if(n!=1){
nm1= n-1;
for(k= 1;k<=nm1;k++){
kp1= k+1;
m= k;
for(i= kp1;i<=n;i++){
if(fabs(a[i][k])>fabs(a[m][k]))
m= i;
}
ip[k]= m;
t= a[m][k];
if(m!=k){
ip[n]= -ip[n];
a[m][k]= a[k][k];
a[k][k]= t;
}
if(t==0.0){
*ier= k;
ip[n]= 0;
return 1;
}
t= 1.0/t;
for(i= kp1;i<=n;i++)
a[i][k]= -a[i][k]*t;
for(j= kp1;j<=n;j++){
t= a[m][j];
a[m][j]= a[k][j];
a[k][j]= t;
if(t!=0.0){
for(i= kp1;i<=n;i++)
a[i][j]+= a[i][k]*t;
}
}
}
}
k= n;
if(a[n][n]==0.0){
*ier= k;
ip[n]= 0;
}
return 0;
}


/*:234*/
#line 3993 "MathView3D.w"

/*235:*/
#line 4284 "MathView3D.w"

void LUSolution(int n,double a[][5],double b[],int ip[])
{
register int k,m,i;
register double t;

if(n!=1){
for(k= 1;k<n;k++){
m= ip[k];
t= b[m];
b[m]= b[k];
b[k]= t;
for(i= k+1;i<=n;i++)
b[i]+= a[i][k]*t;
}
for(k= n;k>=2;k--){
b[k]/= a[k][k];
t= -b[k];
for(i= 1;i<k;i++)
b[i]+= a[i][k]*t;
}
}
b[1]/= a[1][1];
}

/*:235*/
#line 3994 "MathView3D.w"

/*236:*/
#line 4313 "MathView3D.w"

void Transform3ScriptLights(Point3D*viewpnt,Point3D*viewvert,Point3D*viewcntr)
{Point3D x,y,z,vp2;
double n;
double m[5][5],b[5];
int i,ip[5],iter;

vp2.x= -0.5*viewpnt->x;
vp2.y= -0.5*viewpnt->y;
vp2.z= -0.5*viewpnt->z;
z.x= viewpnt->x-viewcntr->x;
z.y= viewpnt->y-viewcntr->y;
z.z= viewpnt->z-viewcntr->z;
n= Norm(&z);
if(n!=0.0){
z.x/= n;
z.y/= n;
z.z/= n;
}
y= *viewvert;
n= Norm(&y);
if(n!=0.0){
y.x/= n;
y.y/= n;
y.z/= n;
}
Cross(&x,&y,&z);
Cross(&y,&z,&x);
m[1][1]= x.x;m[1][2]= x.y;m[1][3]= x.z;m[1][4]= Dot(&x,&vp2);
m[2][1]= y.x;m[2][2]= y.y;m[2][3]= y.z;m[2][4]= Dot(&y,&vp2);
m[3][1]= z.x;m[3][2]= z.y;m[3][3]= z.z;m[3][4]= Dot(&z,&vp2);
m[4][1]= 0.0;m[4][2]= 0.0;m[4][3]= 0.0;m[4][4]= 1.0;
if(0==LUDecomp(4,m,ip,&iter)){
for(i= 0;i<LightNo;i++){
b[1]= LightPosition[i].x;
b[2]= LightPosition[i].y;
b[3]= LightPosition[i].z;
b[4]= 1.0;
LUSolution(4,m,b,ip);
LightPosition[i].x= b[1];
LightPosition[i].y= b[2];
LightPosition[i].z= b[3];
}
}
}

/*:236*/
#line 3995 "MathView3D.w"

/*220:*/
#line 4036 "MathView3D.w"

int Read3Script(char*fname)
{char cline[MAX_STRING];
FILE*infile;
int key,i;
int plBuffer[2048];
double x,y,z,red,green,blue,sred,sblue,sgreen;
double psize;

infile= fopen(fname,"r");
if(!infile)return 1;
strcpy(cline,"");
red= green= blue= 1.0;
sred= sgreen= sblue= 1.0;
psize= 0.001;
while(!feof(infile)){
if(!*cline)
fgets(cline,MAX_STRING,infile);
key= SearchKeyWord(cline);
if(key<0){
strcpy(cline,"");
continue;
}
switch(key){
case BBOX:
/*224:*/
#line 4129 "MathView3D.w"

fscanf(infile,DBL_FMT" "DBL_FMT" "DBL_FMT"\n",
&(BoundP1.x),&(BoundP1.y),&(BoundP1.z));
fscanf(infile,DBL_FMT" "DBL_FMT" "DBL_FMT"\n",&(BoundP2.x),&(BoundP2.y),&(BoundP2.z));
strcpy(cline,"");
CenterOfScene.x= 0.5*(BoundP1.x+BoundP2.x);
CenterOfScene.y= 0.5*(BoundP1.y+BoundP2.y);
CenterOfScene.z= 0.5*(BoundP1.z+BoundP2.z);
ScaleOfScene.x= (BoundP2.x-BoundP1.x);
ScaleOfScene.y= (BoundP2.y-BoundP1.y);
ScaleOfScene.z= (BoundP2.z-BoundP1.z);
MaxScaleOfScene= MAX3(ScaleOfScene.x,ScaleOfScene.y,ScaleOfScene.z);

/*:224*/
#line 4061 "MathView3D.w"

break;
case VPOINT:
/*230:*/
#line 4214 "MathView3D.w"

fscanf(infile,DBL_FMT" "DBL_FMT" "DBL_FMT"\n",
&(ViewPoint.x),
&(ViewPoint.y),
&(ViewPoint.z));
strcpy(cline,"");

/*:230*/
#line 4064 "MathView3D.w"

break;
case LSOURCE:
/*237:*/
#line 4359 "MathView3D.w"

{Point3D vc= {0.5,0.5,0.5};

fgets(cline,MAX_STRING,infile);
i= 0;
do{
sscanf(cline,DBL_FMT" "DBL_FMT" "DBL_FMT"",
&(LightPosition[i].x),&(LightPosition[i].y),&(LightPosition[i].z));
fgets(cline,MAX_STRING,infile);
sscanf(cline,DBL_FMT" "DBL_FMT" "DBL_FMT"",
&(LightColor[i].r),&(LightColor[i].g),&(LightColor[i].b));
fgets(cline,MAX_STRING,infile);
LightNo++;
i++;
if(i>=MAX_LIGHT)break;
}
while(SearchKeyWord(cline)<0);
Transform3ScriptLights(&ViewPoint,&ViewVertical,&vc);
}

/*:237*/
#line 4067 "MathView3D.w"

break;
case POINT:
/*223:*/
#line 4123 "MathView3D.w"

fscanf(infile,DBL_FMT" "DBL_FMT" "DBL_FMT"\n",&x,&y,&z);
i= AddToPointPool(x,y,z);
Points= PrependPointList(Points,i,psize,red,green,blue);
strcpy(cline,"");

/*:223*/
#line 4070 "MathView3D.w"

break;
case LINE:
/*222:*/
#line 4111 "MathView3D.w"

i= 0;
fgets(cline,MAX_STRING,infile);
do{
sscanf(cline,DBL_FMT" "DBL_FMT" "DBL_FMT"\n",&x,&y,&z);
plBuffer[i]= AddToPointPool(x,y,z);
i++;
fgets(cline,MAX_STRING,infile);
}
while(SearchKeyWord(cline)<0&&!feof(infile));
Lines= PrependLinePolyList(Lines,Line3D,plBuffer,i,red,green,blue);

/*:222*/
#line 4073 "MathView3D.w"

break;
case POLY:
/*221:*/
#line 4099 "MathView3D.w"

i= 0;
fgets(cline,MAX_STRING,infile);
do{
sscanf(cline,DBL_FMT" "DBL_FMT" "DBL_FMT"\n",&x,&y,&z);
plBuffer[i]= AddToPointPool(x,y,z);
i++;
fgets(cline,MAX_STRING,infile);
}
while(SearchKeyWord(cline)<0&&!feof(infile));
Polygons= PrependLinePolyList(Polygons,Polygon3D,plBuffer,i,sred,sgreen,sblue);

/*:221*/
#line 4076 "MathView3D.w"

break;
case AMBIENT:
/*238:*/
#line 4379 "MathView3D.w"

fgets(cline,MAX_STRING,infile);
sscanf(cline,DBL_FMT" "DBL_FMT" "DBL_FMT"",&AmbientColor.r,&AmbientColor.g,&AmbientColor.b);
strcpy(cline,"");

/*:238*/
#line 4079 "MathView3D.w"

break;
case COLOR:
/*231:*/
#line 4221 "MathView3D.w"

fscanf(infile,DBL_FMT" "DBL_FMT" "DBL_FMT"\n",&red,&green,&blue);
strcpy(cline,"");

/*:231*/
#line 4082 "MathView3D.w"

break;
case SURFACECOLOR:
/*232:*/
#line 4225 "MathView3D.w"

fscanf(infile,DBL_FMT" "DBL_FMT" "DBL_FMT"\n",&sred,&sgreen,&sblue);
strcpy(cline,"");

/*:232*/
#line 4085 "MathView3D.w"

break;
case MESH:
/*228:*/
#line 4177 "MathView3D.w"

{int n,m,k,mu;
double dx,dy;
Point3D*data;
char*s;
RGBColor c;
char line[MAX_STRING];
char*cutline;

/*225:*/
#line 4145 "MathView3D.w"

c.r= sred;
c.g= sgreen;
c.b= sblue;
s= cline;
while(*s&&(isspace(*s)||isalpha(*s)))s++;
sscanf(s,"%d %d",&n,&m);
data= (Point3D*)calloc(m*n,sizeof(Point3D));
dx= (BoundP2.x-BoundP1.x)/(n-1);
dy= (BoundP2.y-BoundP1.y)/(m-1);

/*:225*/
#line 4186 "MathView3D.w"

/*226:*/
#line 4156 "MathView3D.w"

mu= n*m;
for(k= 0;k<mu;){
fgets(line,MAX_STRING,infile);
PrepareLine('%',&cutline,line);
if(!*cutline)continue;
sscanf(cutline,DBL_FMT"\n",&data[k].z);
data[k].x= BoundP1.x+((int)(k/n))*dx;
data[k++].y= BoundP1.y+((int)(k%n))*dy;
}

/*:226*/
#line 4187 "MathView3D.w"

Meshes= PrependMesh3DList(Meshes,n,m,data,&c,0,0,NULL);
free(data);
strcpy(cline,"");
}

/*:228*/
#line 4088 "MathView3D.w"

break;
case COLORMESH:
/*229:*/
#line 4193 "MathView3D.w"

{int n,m,k,mu;
double dx,dy;
Point3D*data;
RGBColor*colors;
char*s;
RGBColor c;
char line[MAX_STRING];
char*cutline;

/*225:*/
#line 4145 "MathView3D.w"

c.r= sred;
c.g= sgreen;
c.b= sblue;
s= cline;
while(*s&&(isspace(*s)||isalpha(*s)))s++;
sscanf(s,"%d %d",&n,&m);
data= (Point3D*)calloc(m*n,sizeof(Point3D));
dx= (BoundP2.x-BoundP1.x)/(n-1);
dy= (BoundP2.y-BoundP1.y)/(m-1);

/*:225*/
#line 4203 "MathView3D.w"

colors= (RGBColor*)calloc((m-1)*(n-1),sizeof(RGBColor));
/*226:*/
#line 4156 "MathView3D.w"

mu= n*m;
for(k= 0;k<mu;){
fgets(line,MAX_STRING,infile);
PrepareLine('%',&cutline,line);
if(!*cutline)continue;
sscanf(cutline,DBL_FMT"\n",&data[k].z);
data[k].x= BoundP1.x+((int)(k/n))*dx;
data[k++].y= BoundP1.y+((int)(k%n))*dy;
}

/*:226*/
#line 4205 "MathView3D.w"

/*227:*/
#line 4167 "MathView3D.w"

mu= (n-1)*(m-1);
for(k= 0;k<mu;){
fgets(line,MAX_STRING,infile);
PrepareLine('%',&cutline,line);
if(!*cutline)continue;
sscanf(cutline,DBL_FMT" "DBL_FMT" "DBL_FMT"\n",&colors[k].r,&colors[k].g,&colors[k].b);
k++;
}

/*:227*/
#line 4206 "MathView3D.w"

Meshes= PrependMesh3DList(Meshes,n,m,data,&c,n-1,m-1,(double*)colors);
free(data);
free(colors);
strcpy(cline,"");
}


/*:229*/
#line 4091 "MathView3D.w"

break;
}
}
fclose(infile);
return 0;
}

/*:220*/
#line 3996 "MathView3D.w"


/*:217*/
#line 71 "MathView3D.w"

/*93:*/
#line 1876 "MathView3D.w"

void AveragedNormals(LinePolyList*lst)
{int i,j1,j2,j3;
Point3D normal;
int*cnt,res;
double n;

if(!PoolSize)return;
cnt= (int*)calloc(PoolSize,sizeof(int));
for(i= 0;i<PoolSize;i++)cnt[i]= 0;
while(lst){
for(i= 0;i<lst->pno;i++){
/*92:*/
#line 1870 "MathView3D.w"

j1= lst->points[0==i?lst->pno-1:i-1];
j2= lst->points[i];
j3= lst->points[lst->pno-1==i?0:i+1];

/*:92*/
#line 1888 "MathView3D.w"

res= SurfaceNormal(&normal,PointPool+j1,PointPool+j2,PointPool+j3);
if(!res)continue;
if(!NormalPool[j2]){
NormalPool[j2]= (Point3D*)malloc(sizeof(Point3D));
NormalPool[j2]->x= NormalPool[j2]->y= NormalPool[j2]->z= 0.0;
}
NormalPool[j2]->x+= normal.x;
NormalPool[j2]->y+= normal.y;
NormalPool[j2]->z+= normal.z;
cnt[j2]++;
}
lst= lst->next;
}
for(i= 0;i<PoolSize;i++)
if(cnt[i]&&NormalPool[i]){
NormalPool[i]->x/= cnt[i];
NormalPool[i]->y/= cnt[i];
NormalPool[i]->z/= cnt[i];
n= Norm(NormalPool[i]);
if(n>0.0){
NormalPool[i]->x/= n;
NormalPool[i]->y/= n;
NormalPool[i]->z/= n;
}
else{
free(NormalPool[i]);
NormalPool[i]= NULL;
}
}
free(cnt);
}


/*:93*/
#line 72 "MathView3D.w"

/*86:*/
#line 1801 "MathView3D.w"

long int LengthLinePolyList(LinePolyList*lst)
{
long int len= 0;

while(lst){
len++;
lst= lst->next;
}
return len;
}

/*:86*/
#line 73 "MathView3D.w"

/*180:*/
#line 3394 "MathView3D.w"

/*189:*/
#line 3525 "MathView3D.w"

double DistancePointToLine3D(int i,int i1,int i2)
{Point3D*p,*p1,n;
Point3D dp,cp;
double d;

p= PointPool+i;p1= PointPool+i1;n= PointPool[i2];
n.x-= p1->x;n.y-= p1->y;n.z-= p1->z;
dp.x= p->x-p1->x;
dp.y= p->y-p1->y;
dp.z= p->z-p1->z;
Cross(&cp,&dp,&n);
d= Dot(&n,&n);
return sqrt(Dot(&cp,&cp)/d);
}

/*:189*/
#line 3395 "MathView3D.w"

/*188:*/
#line 3511 "MathView3D.w"

void Normalize2D(Point2D*p)
{double len;

len= sqrt((p->x)*(p->x)+(p->y)*(p->y));
if(len>0.0){
len= 1.0/len;
(p->x)*= len;
(p->y)*= len;
}
}


/*:188*/
#line 3396 "MathView3D.w"

/*190:*/
#line 3542 "MathView3D.w"

#define LeftSideLine   (1)
#define RightSideLine  (-1)
#define Collinear (0)
int SideOfLine2D(Point2D*p,Point2D*p1,Point2D*p2)
{Point2D u,v;
double test;

u.x= p2->x-p1->x;
u.y= p2->y-p1->y;
v.x= p->x-p2->y;
v.y= p->y-p2->y;
Normalize2D(&u);
Normalize2D(&v);
test= u.x*v.y-u.y*v.x;
if(test>MachineEpsilon)
return LeftSideLine;
if(test<-MachineEpsilon)
return RightSideLine;
return Collinear;
}

/*:190*/
#line 3397 "MathView3D.w"

/*191:*/
#line 3564 "MathView3D.w"

int PointInPolygon2D(Point2D*pnew,Point2D*p,int np)
{int i,inpoly,test,all;

inpoly= SideOfLine2D(pnew,p+np-1,p);
for(i= 0;i<np-1;i++){
test= SideOfLine2D(pnew,p+i,p+i+1);
if(test==inpoly)all= GL_TRUE;
if(!all)return all;
}
return all;
}

/*:191*/
#line 3398 "MathView3D.w"

/*184:*/
#line 3462 "MathView3D.w"

void RotateIntegersLeft(int*arr,int n)
{int first;
int i,n1;

n1= n-1;
first= arr[0];
for(i= 0;i<n1;i++)
arr[i]= arr[i+1];
arr[n1]= first;
}

/*:184*/
#line 3399 "MathView3D.w"

/*185:*/
#line 3474 "MathView3D.w"

void RotateIntegersRight(int*arr,int n)
{int first;
int i,n1;

n1= n-1;
first= arr[n1];
for(i= n-1;i>0;i--)
arr[i]= arr[i-1];
arr[0]= first;
}


/*:185*/
#line 3400 "MathView3D.w"

/*186:*/
#line 3487 "MathView3D.w"

void ReverseIntegers(int*arr,int n)
{int i,tmp,n2;

n2= n/2;
for(i= 0;i<n2;i++){
tmp= arr[i];
arr[i]= arr[n-1-i];
arr[n-1-i]= tmp;
}
}

/*:186*/
#line 3401 "MathView3D.w"

/*187:*/
#line 3499 "MathView3D.w"

void RotatePoint2DRight(Point2D*pnts,int n)
{Point2D p;
int i,n1;

n1= n-1;
p= pnts[n1];
for(i= n-1;i>0;i--)
pnts[i]= pnts[i-1];
pnts[0]= p;
}

/*:187*/
#line 3402 "MathView3D.w"

/*197:*/
#line 3676 "MathView3D.w"

void SwapPolyPtr(LinePolyList**p1,LinePolyList**p2)
{LinePolyList*tmp;

tmp= *p1;
*p1= *p2;
*p2= tmp;
}

/*:197*/
#line 3403 "MathView3D.w"

/*198:*/
#line 3687 "MathView3D.w"

int InsertAdjacentPolygon(PointOctTree*tree,int j)
{int k,npoly;

npoly= tree->polyNo;
k= j+1;
while(k<npoly){
if(PolygonPoint(tree,j-1,2)==PolygonPoint(tree,k,0)){
SwapPolyPtr(tree->vpoly+j,tree->vpoly+k);
return GL_TRUE;
}
k++;
}
return GL_FALSE;
}


/*:198*/
#line 3404 "MathView3D.w"

/*199:*/
#line 3707 "MathView3D.w"

void FindPlaneEquation(Point3D*pn,double*d,
int*poly,int n)
{Point3D cntr,*p;
Point3D nvect,v;
double nn;
int i;

cntr.x= cntr.y= cntr.z= 0.0;
for(i= 0;i<n;i++){
p= PointPool+poly[i];
cntr.x+= p->x;
cntr.y+= p->y;
cntr.z+= p->z;
}
cntr.x/= n;
cntr.y/= n;
cntr.z/= n;
nvect.x= nvect.y= nvect.z= 0.0;
for(i= 1;i<n-1;i++){
SurfaceNormal(&v,PointPool+poly[i-1],PointPool+poly[i],PointPool+poly[i+1]);
nvect.x+= v.x;
nvect.y+= v.y;
nvect.z+= v.z;
}
SurfaceNormal(&v,PointPool+poly[n-2],PointPool+poly[n-1],PointPool+poly[0]);
nvect.x+= v.x;
nvect.y+= v.y;
nvect.z+= v.z;
nn= Norm(&nvect);
if(nn>0.0){
nn= 1.0/nn;
nvect.x*= nn;
nvect.y*= nn;
nvect.z*= nn;
}
*d= -Dot(&nvect,&cntr);
pn->x= nvect.x;
pn->y= nvect.y;
pn->z= nvect.z;
}


/*:199*/
#line 3405 "MathView3D.w"

/*192:*/
#line 3577 "MathView3D.w"

LinePolyList*TessPolygon3D(int*poly,int pno,
Point3D*planen,
double r,double g,double b)
{Point3D tnormal;
Point3D n1,n2,*p,*p1,*p2,pc;
Point2D*p2d,test;
int i,found,imax,tri[3],sameDirection;
LinePolyList*plst,*lst;

plst= NULL;
p2d= (Point2D*)calloc(pno,sizeof(Point2D));
/*193:*/
#line 3617 "MathView3D.w"

p1= PointPool+poly[0];p2= PointPool+poly[1];
n1.x= p1->x-p2->x;
n1.y= p1->y-p2->y;
n1.z= p1->z-p2->z;
p1= PointPool+poly[2];
n2.x= p2->x-p1->x;
n2.y= p2->y-p1->y;
n2.z= p2->z-p1->z;
n1.x= 0.5*(n1.x+n2.x);
n1.y= 0.5*(n1.y+n2.y);
n1.z= 0.5*(n1.z+n2.z);
Cross(&n2,planen,&n1);
GramSchmidt(planen,&n1,&n2);

/*:193*/
#line 3589 "MathView3D.w"

/*194:*/
#line 3632 "MathView3D.w"

for(i= 0;i<pno;i++){
p= PointPool+poly[i];
p2d[i].x= Dot(p,&n1);
p2d[i].y= Dot(p,&n2);
}


/*:194*/
#line 3590 "MathView3D.w"

while(pno>3){
for(imax= 0;imax<pno;imax++){
/*195:*/
#line 3644 "MathView3D.w"

tri[0]= poly[(imax-1)<0?pno-1:imax-1];
tri[1]= poly[imax];
tri[2]= poly[(imax+1)==pno?0:imax+1];
p1= PointPool+tri[0];
p2= PointPool+tri[2];
pc.x= 0.5*(p1->x+p2->x);
pc.y= 0.5*(p1->y+p2->y);
pc.z= 0.5*(p1->z+p2->z);
test.x= Dot(&pc,&n1);
test.y= Dot(&pc,&n2);

/*:195*/
#line 3593 "MathView3D.w"

if(PointInPolygon2D(&test,p2d,pno)){
found= GL_TRUE;
break;
}
}
if(found)
plst= PrependLinePolyList(plst,Polygon3D,tri,3,r,g,b);
for(i= imax+1;i<pno;i++){
RotateIntegersRight(poly,pno);
RotatePoint2DRight(p2d,pno);
}
pno--;
}
plst= PrependLinePolyList(plst,Polygon3D,poly,pno,r,g,b);
/*196:*/
#line 3660 "MathView3D.w"

lst= plst;
sameDirection= GL_TRUE;
while(lst){
SurfaceNormal(&tnormal,
PointPool+lst->points[0],
PointPool+lst->points[1],
PointPool+lst->points[2]);
sameDirection= sameDirection&&(Dot(&tnormal,planen)>0.25);
lst= lst->next;
}



/*:196*/
#line 3608 "MathView3D.w"

if(!sameDirection){
RemoveLinePolyList(plst);
plst= NULL;
}
free(p2d);
return plst;
}

/*:192*/
#line 3406 "MathView3D.w"

/*183:*/
#line 3447 "MathView3D.w"

void ReducePolygons(double tol,PointOctTree*root,PointOctTree*tree)
{int i;

if(!tree)return;
for(i= 0;i<8;i++)
if(tree->leafs[i])ReducePolygons(tol,root,tree->leafs[i]);
if(tree->polyNo>=3){
/*200:*/
#line 3757 "MathView3D.w"

{int center,npoly,j,k;
int closed,adjacent= GL_TRUE;
int colin= GL_FALSE;

center= tree->point;
npoly= tree->polyNo;
/*201:*/
#line 3771 "MathView3D.w"

for(j= 0;j<npoly;j++){
for(k= 0;k<PolygonPointNo(tree,j);k++){
if(PolygonPoint(tree,j,1)==center)break;
RotateIntegersLeft(PolygonPoints(tree,j),PolygonPointNo(tree,j));
}
adjacent= adjacent&&(PolygonPoint(tree,j,1)==center);
}

/*:201*/
#line 3764 "MathView3D.w"

/*202:*/
#line 3786 "MathView3D.w"

for(j= 1;j<npoly&&adjacent;j++){
if(PolygonPoint(tree,j-1,2)!=PolygonPoint(tree,j,0)){
adjacent= adjacent&&InsertAdjacentPolygon(tree,j);
}
}
closed= PolygonPoint(tree,0,0)==PolygonPoint(tree,npoly-1,2);
if(!closed)
colin= DistancePointToLine3D(center,PolygonPoint(tree,0,0),PolygonPoint(tree,npoly-1,2))<tol;


/*:202*/
#line 3765 "MathView3D.w"

if((closed||colin)&&adjacent){
/*203:*/
#line 3803 "MathView3D.w"

{double d,dist;
double envdist;
Point3D*p,planeNormal;
int*newpoly;
int ppNo;
LinePolyList*newPolygons,*plst;

p= PointPool+center;
ppNo= npoly;
if(!colin){
newpoly= (int*)calloc(ppNo,sizeof(int));
for(j= 0;j<ppNo;j++)
newpoly[j]= PolygonPoint(tree,j,0);
}
else{
ppNo++;
newpoly= (int*)calloc(ppNo,sizeof(int));
for(j= 0;j<npoly;j++)
newpoly[j]= PolygonPoint(tree,j,0);
newpoly[npoly]= PolygonPoint(tree,npoly-1,2);
}
/*204:*/
#line 3845 "MathView3D.w"

ReverseIntegers(newpoly,ppNo);
FindPlaneEquation(&planeNormal,&d,newpoly,ppNo);
dist= fabs(Dot(&planeNormal,p)+d);
for(envdist= 0.0,j= 0;j<ppNo;j++)
envdist+= fabs(Dot(&planeNormal,PointPool+newpoly[j])+d);
envdist/= ppNo;
dist= dist+envdist;


/*:204*/
#line 3825 "MathView3D.w"

if(dist<=tol){
newPolygons= TessPolygon3D(newpoly,ppNo,&planeNormal,1.0,1.0,1.0);
if(newPolygons){
/*205:*/
#line 3855 "MathView3D.w"

plst= newPolygons;
while(plst){
for(j= 0;j<plst->pno;j++)
InsertPolyInOctTree(root,plst,plst->points[j]);
plst= plst->next;
}
Polygons= JoinLinePolyList(newPolygons,Polygons);

/*:205*/
#line 3829 "MathView3D.w"

/*206:*/
#line 3864 "MathView3D.w"

for(j= 0;j<npoly;j++){
RemovePolyFromOctTree(root,tree->vpoly[j],PolygonPoint(tree,j,0));
RemovePolyFromOctTree(root,tree->vpoly[j],PolygonPoint(tree,j,2));
}

/*:206*/
#line 3830 "MathView3D.w"

/*207:*/
#line 3870 "MathView3D.w"

for(j= 0;j<npoly;j++)
Polygons= DeleteSinglePolygon(Polygons,tree->vpoly[j]);

/*:207*/
#line 3831 "MathView3D.w"

free(tree->vpoly);
tree->vpoly= NULL;
tree->polyNo= 0;
}
}
free(newpoly);
}

/*:203*/
#line 3767 "MathView3D.w"

}
}

/*:200*/
#line 3455 "MathView3D.w"

}
}


/*:183*/
#line 3407 "MathView3D.w"

/*181:*/
#line 3411 "MathView3D.w"

int SimplifyPolygones(int iter,double tol,PointOctTree*tree)
{LinePolyList*plst;
int i,it,len1,len2,len3;

len1= len2= LengthLinePolyList(Polygons);
for(it= 0;it<iter;it++){
/*182:*/
#line 3434 "MathView3D.w"

plst= Polygons;
while(plst){
for(i= 0;i<plst->pno;i++)
InsertPolyInOctTree(tree,plst,plst->points[i]);
plst= plst->next;
}


/*:182*/
#line 3418 "MathView3D.w"

ReducePolygons(tol,tree,tree);
StripPolygonsFromOctTree(tree);
len3= LengthLinePolyList(Polygons);
if(len3==len2)break;
len2= len3;
}
AveragedNormals(Polygons);
return len1-len2;
}


/*:181*/
#line 3408 "MathView3D.w"



/*:180*/
#line 74 "MathView3D.w"

/*53:*/
#line 1310 "MathView3D.w"

void SetBackgroundColor(int isBlack);
void SetBackgroundColor(int isBlack)
{
if(isBlack){
LineColor= 1.0;
ClearColor.r= ClearColor.g= ClearColor.b= 0.;
glClearColor(0.0,0.0,0.0,1.0);
}
else{
ClearColor.r= ClearColor.g= ClearColor.b= 0.8;
glClearColor(0.8,0.8,0.8,1.0);
LineColor= 0.0;
}
}


/*:53*/
#line 75 "MathView3D.w"

/*858:*/
#line 13984 "MathView3D.w"

/*860:*/
#line 13994 "MathView3D.w"

void RenderString(char*label)
{
while(*label)
glutStrokeCharacter(GLUT_STROKE_ROMAN,*(label++));
}

/*:860*/
#line 13985 "MathView3D.w"

/*862:*/
#line 14004 "MathView3D.w"

double expt(double a,register int n)
{
double x;

x= 1.;
if(n>0)for(;n>0;n--)x*= a;
else for(;n<0;n++)x/= a;
return x;
}


/*:862*/
#line 13986 "MathView3D.w"

/*864:*/
#line 14021 "MathView3D.w"

double nicenum(double x,int round)
{
int expv;
double f;
double nf;

expv= floor(log10(x));
f= x/expt(10.,expv);
if(round)
if(f<1.5)nf= 1.;
else if(f<3.)nf= 2.;
else if(f<7.)nf= 5.;
else nf= 10.;
else
if(f<=1.)nf= 1.;
else if(f<=2.)nf= 2.;
else if(f<=5.)nf= 5.;
else nf= 10.;
return nf*expt(10.,expv);
}

/*:864*/
#line 13987 "MathView3D.w"

/*866:*/
#line 14051 "MathView3D.w"

void loose_label(double min,double max,
Point3D*p1,Point3D*p2,double fntscale,
double alpha,Point3D*axis,
Point3D*offset)
{double t;
Point3D p;
char str[6],temp[20];
int nfrac;
double d;
double graphmin,graphmax;
double range,x,boxSize;

if(min==max)return;
boxSize= MAX3(offset->x,offset->y,offset->z);

range= nicenum(max-min,0);
d= nicenum(range/(NTICK-1),1);
graphmin= floor(min/d)*d;
graphmax= ceil(max/d)*d;
nfrac= MAX(-floor(log10(d)),0);
sprintf(str,"%%.%df",nfrac);

glPushAttrib(GL_ALL_ATTRIB_BITS);
for(x= graphmin;x<graphmax+.5*d;x+= d){
sprintf(temp,str,x);
t= (x-min)/(max-min);
if(t<0.0||t>1.0)continue;
p.x= p1->x*(1.-t)+t*p2->x;
p.y= p1->y*(1.-t)+t*p2->y;
p.z= p1->z*(1.-t)+t*p2->z;
glPushMatrix();
glTranslated(p.x,p.y,p.z);
glEnable(GL_LIGHTING);
glutSolidCube(boxSize);
glDisable(GL_LIGHTING);
glPopMatrix();
glPushMatrix();
glTranslated(p.x+offset->x,p.y+offset->y,p.z+offset->z);
glRotated(alpha,axis->x,axis->y,axis->z);
glScaled(fntscale,fntscale,fntscale);
RenderString(temp);
glPopMatrix();
}
glPopAttrib();
}

/*:866*/
#line 13988 "MathView3D.w"

/*868:*/
#line 14103 "MathView3D.w"

void DrawAxis(int type,Point3D*origin,
Point3D*p1,Point3D*p2,
char*xlabel,char*ylabel,char*zlabel,double fntScale,int numlabel)
{double f;
double scale;

glPushAttrib(GL_ALL_ATTRIB_BITS);

scale= 0.5*MAX3(p2->x-p1->x,p2->y-p1->y,p2->z-p1->z);
glColor3d(1.0-ClearColor.r,1.0-ClearColor.g,1.0-ClearColor.b);
if(NoAxis==type)return;
glDisable(GL_LIGHTING);
/*869:*/
#line 14132 "MathView3D.w"

glColor3f(0.0f,1.0f,0.0f);
glBegin(GL_LINES);
glVertex3d(p1->x,origin->y,origin->z);
glVertex3d(p2->x,origin->y,origin->z);
glEnd();
glBegin(GL_LINES);
glVertex3d(origin->x,p1->y,origin->z);
glVertex3d(origin->x,p2->y,origin->z);
glEnd();
glBegin(GL_LINES);
glVertex3d(origin->x,origin->y,p1->z);
glVertex3d(origin->x,origin->y,p2->z);
glEnd();

/*:869*/
#line 14116 "MathView3D.w"

glEnable(GL_LIGHTING);
if(CrossAxis==type)
/*870:*/
#line 14149 "MathView3D.w"

{
glPushMatrix();
glEnable(GL_LIGHTING);
glTranslated(p2->x,origin->y,origin->z);
glRotated(0.5*PI*ToDegree,0.,1.,0.);
glColor3d(0.0,1.0,0.0);
glutSolidCone(AxisArrowBase,2.*AxisArrowBase,18,2);
glPopMatrix();
glPushMatrix();
glTranslated(origin->x,p2->y,origin->z);
glRotated(-0.5*PI*ToDegree,1.,0.,0.);
glutSolidCone(AxisArrowBase,2.*AxisArrowBase,18,2);
glPopMatrix();
glPushMatrix();
glTranslated(origin->x,origin->y,p2->z);
glutSolidCone(AxisArrowBase,2.*AxisArrowBase,18,2);
glDisable(GL_LIGHTING);
glPopMatrix();
}

/*:870*/
#line 14119 "MathView3D.w"

if(type>CrossAxis){
if(BoxAxis==(type&BoxAxis)){
/*871:*/
#line 14170 "MathView3D.w"

glDisable(GL_LIGHTING);
glColor3d(1.0-ClearColor.r,1.0-ClearColor.g,1.0-ClearColor.b);

glBegin(GL_LINE_STRIP);
glVertex3d(p1->x,p1->y,p1->z);
glVertex3d(p2->x,p1->y,p1->z);
glVertex3d(p2->x,p2->y,p1->z);
glVertex3d(p1->x,p2->y,p1->z);
glVertex3d(p1->x,p1->y,p1->z);

glVertex3d(p1->x,p1->y,p2->z);
glVertex3d(p2->x,p1->y,p2->z);
glVertex3d(p2->x,p2->y,p2->z);
glVertex3d(p1->x,p2->y,p2->z);
glVertex3d(p1->x,p1->y,p2->z);
glEnd();
glBegin(GL_LINES);
glVertex3d(p2->x,p1->y,p1->z);
glVertex3d(p2->x,p1->y,p2->z);
glEnd();
glBegin(GL_LINES);
glVertex3d(p2->x,p2->y,p1->z);
glVertex3d(p2->x,p2->y,p2->z);
glEnd();
glBegin(GL_LINES);
glVertex3d(p1->x,p2->y,p1->z);
glVertex3d(p1->x,p2->y,p2->z);
glEnd();
glBegin(GL_LINE_STRIP);
glVertex3d(origin->x,p1->y,p1->z);
glVertex3d(origin->x,p2->y,p1->z);
glVertex3d(origin->x,p2->y,p2->z);
glVertex3d(origin->x,p1->y,p2->z);
glVertex3d(origin->x,p1->y,p1->z);
glEnd();
glBegin(GL_LINE_STRIP);
glVertex3d(p1->x,origin->y,p1->z);
glVertex3d(p2->x,origin->y,p1->z);
glVertex3d(p2->x,origin->y,p2->z);
glVertex3d(p1->x,origin->y,p2->z);
glVertex3d(p1->x,origin->y,p1->z);
glEnd();
glBegin(GL_LINE_STRIP);
glVertex3d(p1->x,p1->y,origin->z);
glVertex3d(p2->x,p1->y,origin->z);
glVertex3d(p2->x,p2->y,origin->z);
glVertex3d(p1->x,p2->y,origin->z);
glVertex3d(p1->x,p1->y,origin->z);
glEnd();
glEnable(GL_LIGHTING);

/*:871*/
#line 14122 "MathView3D.w"

}
else{
/*872:*/
#line 14222 "MathView3D.w"

glDisable(GL_LIGHTING);
glColor3d(1.0-ClearColor.r,1.0-ClearColor.g,1.0-ClearColor.b);

glBegin(GL_LINE_STRIP);
glVertex3d(p1->x,p1->y,origin->z);
glVertex3d(p2->x,p1->y,origin->z);
glVertex3d(p2->x,p2->y,origin->z);
glVertex3d(p1->x,p2->y,origin->z);
glVertex3d(p1->x,p1->y,origin->z);
glEnd();
glBegin(GL_LINE_STRIP);
glVertex3d(p1->x,origin->y,p1->z);
glVertex3d(p2->x,origin->y,p1->z);
glVertex3d(p2->x,origin->y,p2->z);
glVertex3d(p1->x,origin->y,p2->z);
glVertex3d(p1->x,origin->y,p1->z);
glEnd();
glBegin(GL_LINE_STRIP);
glVertex3d(origin->x,p1->y,p1->z);
glVertex3d(origin->x,p2->y,p1->z);
glVertex3d(origin->x,p2->y,p2->z);
glVertex3d(origin->x,p1->y,p2->z);
glVertex3d(origin->x,p1->y,p1->z);
glEnd();

glEnable(GL_LIGHTING);


/*:872*/
#line 14125 "MathView3D.w"

}
}
/*873:*/
#line 14251 "MathView3D.w"

{Point3D tp1,tp2,axis,offset;

if(BlackBackground)
glColor3d(1.0,1.0,1.0);
else
glColor3d(0.0,0.0,0.0);

f= fntScale*scale/119.5;
/*874:*/
#line 14283 "MathView3D.w"

if(*xlabel){
glPushMatrix();
glTranslated(p2->x+AxisArrowBase,origin->y+AxisArrowBase,origin->z+AxisArrowBase);
glRotated(0.5*PI*ToDegree,1.0,0.0,0.0);
glScaled(f,f,f);
RenderString(xlabel);
glPopMatrix();
}

/*:874*/
#line 14260 "MathView3D.w"

/*875:*/
#line 14293 "MathView3D.w"

if(*ylabel){
glPushMatrix();
glTranslated(origin->x+AxisArrowBase,p2->y+AxisArrowBase,origin->z+AxisArrowBase);
glRotated(0.5*PI*ToDegree,1.0,0.0,0.0);
glScaled(f,f,f);
RenderString(ylabel);
glPopMatrix();
}

/*:875*/
#line 14261 "MathView3D.w"

/*876:*/
#line 14303 "MathView3D.w"

if(*zlabel){
glPushMatrix();
glTranslated(origin->x+AxisArrowBase,origin->y+AxisArrowBase,p2->z+2.*AxisArrowBase);
glRotated(0.5*PI*ToDegree,1.0,0.0,0.0);
glScaled(f,f,f);
RenderString(zlabel);
glPopMatrix();
}

/*:876*/
#line 14262 "MathView3D.w"

if(numlabel){
f*= 0.5;
axis.x= 1.0;axis.y= 0.0;axis.z= 0.0;

tp1= tp2= AxisWindow->Origin;
tp1.x= BoundP1.x;tp2.x= BoundP2.x;
offset.x= 0.0;offset.y= AxisArrowBase;offset.z= 0.0;
loose_label(BoundP1.x,BoundP2.x,&tp1,&tp2,f,0.0,&axis,&offset);

tp1= tp2= AxisWindow->Origin;
tp1.y= BoundP1.y;tp2.y= BoundP2.y;
offset.x= AxisArrowBase;offset.y= 0;offset.z= 0.0;
loose_label(BoundP1.y,BoundP2.y,&tp1,&tp2,f,0.0,&axis,&offset);

tp1= tp2= AxisWindow->Origin;
tp1.z= BoundP1.z;tp2.z= BoundP2.z;
loose_label(BoundP1.z,BoundP2.z,&tp1,&tp2,f,0.5*PI*ToDegree,&axis,&offset);
}
}

/*:873*/
#line 14128 "MathView3D.w"

glPopAttrib();
}

/*:868*/
#line 13989 "MathView3D.w"


/*:858*/
#line 76 "MathView3D.w"

/*239:*/
#line 4391 "MathView3D.w"

/*247:*/
#line 4583 "MathView3D.w"

void DefineMaterial(long what,RGBColor scolor)
{GLfloat matAmb[4]= {0.35,0.35,0.35,1.0};
GLfloat matAmbBack[4]= {0.3,0.3,0.3,1.0};
GLfloat matDiff[4]= {0.60,0.60,0.60,1.0};
GLfloat matSpec[4]= {0.25,0.25,0.25,1.0};
GLfloat matSpecBack[4]= {0.1,0.1,0.1,1.0};
GLfloat matShine= 60.0;

if(INHERITE_COLOR==(what&INHERITE_COLOR)){
matAmb[0]*= scolor.r;
matAmb[1]*= scolor.g;
matAmb[2]*= scolor.b;
matDiff[0]*= scolor.r;
matDiff[1]*= scolor.g;
matDiff[2]*= scolor.b;
}
glMaterialfv(GL_FRONT,GL_AMBIENT,matAmb);
glMaterialfv(GL_BACK,GL_AMBIENT,matAmbBack);
glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,matDiff);
glMaterialfv(GL_FRONT,GL_SPECULAR,matSpec);
glMaterialfv(GL_BACK,GL_SPECULAR,matSpecBack);
glMaterialf(GL_FRONT_AND_BACK,GL_SHININESS,matShine);

}


/*:247*/
#line 4392 "MathView3D.w"

/*172:*/
#line 3205 "MathView3D.w"

void MakeCone(Point3D*p1,Point3D*p2,double r,int n)
{Point3D v1,v2,v3,vn,p[4],testn;
double norm;
double dphi;
int i,interchange;

v1.x= p2->x-p1->x;
v1.y= p2->y-p1->y;
v1.z= p2->z-p1->z;
norm= Norm(&v1);
v1.x/= norm;v1.y/= norm;v1.z/= norm;
OrthogonalSet(&v1,&v2,&v3);

v1.x= p2->x-p1->x;
v1.y= p2->y-p1->y;
v1.z= p2->z-p1->z;
dphi= TWO_PI/n;
/*173:*/
#line 3248 "MathView3D.w"


ConePoint(r,0.0,&v2,&v3,p);
IncPoint3D(p[0],(*p1));
p[1]= p[0];
IncPoint3D(p[1],v1);

ConePoint(r,dphi,&v2,&v3,p+2);
IncPoint3D(p[2],(*p1));

SurfaceNormal(&testn,p,p+1,p+2);
interchange= Dot(&testn,&v2)<0.0;
if(interchange){
SwapPoint3D(&v2,&v3);
}
if(MVFaceFlip){
SwapPoint3D(&v2,&v3);
}

/*:173*/
#line 3223 "MathView3D.w"

ConePoint(r,0.0,&v2,&v3,p);
vn= *p;
IncPoint3D(p[0],(*p1))
AddPoint3D(p[1],p[0],v1)
glBegin(GL_TRIANGLE_STRIP);
glNormal3dv((GLdouble*)(&vn));
glVertex3dv((GLdouble*)p);
glVertex3dv((GLdouble*)(p+1));
for(i= 1;i<=n;i++){
ConePoint(r,i*dphi,&v2,&v3,p+2);
vn= p[2];
IncPoint3D(p[2],(*p1))
AddPoint3D(p[3],p[2],v1)
glNormal3dv((GLdouble*)(&vn));
glVertex3dv((GLdouble*)(p+2));
glVertex3dv((GLdouble*)(p+3));

}
glEnd();

}

/*:172*/
#line 4393 "MathView3D.w"

/*174:*/
#line 3267 "MathView3D.w"

void MakeConeContinued(Point3D*p1,Point3D*p2,Point3D*p3,
double r,int n)
{Point3D v1,v2,v3,u1,u2,u3,vn1,vn2,p[4];
Point3D testn;
double norm;
double dphi;
int i,interchange;

/*175:*/
#line 3310 "MathView3D.w"

v1.x= p2->x-p1->x;v1.y= p2->y-p1->y;v1.z= p2->z-p1->z;
norm= Norm(&v1);
v1.x/= norm;v1.y/= norm;v1.z/= norm;
u1.x= p3->x-p2->x;u1.y= p3->y-p2->y;u1.z= p3->z-p2->z;
norm= Norm(&u1);
u1.x/= norm;u1.y/= norm;u1.z/= norm;
Cross(&v2,&v1,&u1);
if(fabs(v2.x)+fabs(v2.y)+fabs(v2.z)<TINY){
OrthogonalSet(&v1,&v2,&v3);
}
v3.x= v3.y= v3.z= 1.0;
GramSchmidt(&v1,&v2,&v3);

/*:175*/
#line 3276 "MathView3D.w"

/*176:*/
#line 3324 "MathView3D.w"

if(Dot(&u1,&v1)<CosSkipCont){
MakeCone(p1,p2,r,n);
return;
}

/*:176*/
#line 3277 "MathView3D.w"

u2= v2;u3= v3;
GramSchmidt(&u1,&u2,&u3);
dphi= TWO_PI/n;
/*177:*/
#line 3337 "MathView3D.w"


ConePoint(r,0.0,&v2,&v3,p);
IncPoint3D(p[0],(*p1));
ConePoint(r,0.0,&u2,&u3,p+1);
IncPoint3D(p[1],(*p2));
ConePoint(r,dphi,&v2,&v3,p+2);
IncPoint3D(p[2],(*p1));

SurfaceNormal(&testn,p,p+1,p+2);
interchange= Dot(&testn,&v2)<0.0;
if(interchange){
SwapPoint3D(&v2,&v3);
SwapPoint3D(&u2,&u3);
}
if(MVFaceFlip){
SwapPoint3D(&v2,&v3);
SwapPoint3D(&u2,&u3);
}

/*:177*/
#line 3281 "MathView3D.w"

ConePoint(r,0.0,&v2,&v3,p);
vn1= *p;
IncPoint3D(p[0],(*p1))
ConePoint(r,0.0,&u2,&u3,p+1);
vn2= p[1];
IncPoint3D(p[1],(*p2))
glBegin(GL_TRIANGLE_STRIP);
glNormal3dv((GLdouble*)(&vn1));
glVertex3dv((GLdouble*)p);
glNormal3dv((GLdouble*)(&vn2));
glVertex3dv((GLdouble*)(p+1));
for(i= 1;i<=n;i++){
ConePoint(r,i*dphi,&v2,&v3,p+2);
vn1= p[2];
IncPoint3D(p[2],(*p1))
ConePoint(r,i*dphi,&u2,&u3,p+3);
vn2= p[3];
IncPoint3D(p[3],(*p2));
glNormal3dv((GLdouble*)(&vn1));
glVertex3dv((GLdouble*)(p+2));

glNormal3dv((GLdouble*)(&vn2));
glVertex3dv((GLdouble*)(p+3));
}
glEnd();

}

/*:174*/
#line 4394 "MathView3D.w"

/*269:*/
#line 5006 "MathView3D.w"

void GLSetTextures(TextureList*lst)
{GLuint tname;

if(!lst)return;
glPixelStorei(GL_UNPACK_ALIGNMENT,1);
while(lst){
if(!glIsTexture(TextureGLName(lst))){
glGenTextures(1,&tname);
TextureGLName(lst)= tname;
glBindTexture(GL_TEXTURE_2D,tname);
glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
glTexImage2D(GL_TEXTURE_2D,
0,
TextureType(lst),
TextureWidth(lst),TextureHeight(lst),
0,
TextureType(lst),
GL_UNSIGNED_BYTE,
TextureData(lst));
}
lst= lst->next;
}
}

/*:269*/
#line 4395 "MathView3D.w"

/*248:*/
#line 4614 "MathView3D.w"

#pragma argsused
void PointSphere(Point3D*cnt,double r,int nphi,int ntheta,
float colorRed,float colorGreen,float colorBlue)
{int i,j,count;
double phi,theta;
double dphi,dtheta;
register double cp,sp,ct,st;
register double cp1,sp1,ct1,st1;
Point3D p0,p;


dphi= TWO_PI/nphi;
dtheta= PI/ntheta;

for(j= 0;j<ntheta;j++){
ct= cos(theta= j*dtheta);
st= sin(theta);
ct1= cos(theta= (j+1)*dtheta);
st1= sin(theta);
for(i= 0,count= 0;i<nphi;i++){
cp= cos(phi= i*dphi);
sp= sin(phi);
cp1= cos(phi= (i+1)*dphi);
sp1= sin(phi);
if((count++&1)){
p.x= cp*st;p.y= sp*st;p.z= ct;
p0= p;
/*249:*/
#line 4655 "MathView3D.w"

glBegin(GL_POLYGON);
/*253:*/
#line 4691 "MathView3D.w"

glNormal3d(p.x,p.y,p.z);
glVertex3d(r*p.x+cnt->x,r*p.y+cnt->y,r*p.z+cnt->z);

/*:253*/
#line 4657 "MathView3D.w"

p.x= cp*st1;p.y= sp*st1;p.z= ct1;
/*253:*/
#line 4691 "MathView3D.w"

glNormal3d(p.x,p.y,p.z);
glVertex3d(r*p.x+cnt->x,r*p.y+cnt->y,r*p.z+cnt->z);

/*:253*/
#line 4659 "MathView3D.w"

p.x= cp1*st1;p.y= sp1*st1;
/*253:*/
#line 4691 "MathView3D.w"

glNormal3d(p.x,p.y,p.z);
glVertex3d(r*p.x+cnt->x,r*p.y+cnt->y,r*p.z+cnt->z);

/*:253*/
#line 4661 "MathView3D.w"

glEnd();

/*:249*/
#line 4642 "MathView3D.w"

/*250:*/
#line 4664 "MathView3D.w"

glBegin(GL_POLYGON);
/*253:*/
#line 4691 "MathView3D.w"

glNormal3d(p.x,p.y,p.z);
glVertex3d(r*p.x+cnt->x,r*p.y+cnt->y,r*p.z+cnt->z);

/*:253*/
#line 4666 "MathView3D.w"

p.x= cp1*st;p.y= sp1*st;p.z= ct;
/*253:*/
#line 4691 "MathView3D.w"

glNormal3d(p.x,p.y,p.z);
glVertex3d(r*p.x+cnt->x,r*p.y+cnt->y,r*p.z+cnt->z);

/*:253*/
#line 4668 "MathView3D.w"

p= p0;
/*253:*/
#line 4691 "MathView3D.w"

glNormal3d(p.x,p.y,p.z);
glVertex3d(r*p.x+cnt->x,r*p.y+cnt->y,r*p.z+cnt->z);

/*:253*/
#line 4670 "MathView3D.w"

glEnd();

/*:250*/
#line 4643 "MathView3D.w"

}
else{
p.x= cp*st1;p.y= sp*st1;p.z= ct1;
p0= p;
/*251:*/
#line 4673 "MathView3D.w"

glBegin(GL_POLYGON);
/*253:*/
#line 4691 "MathView3D.w"

glNormal3d(p.x,p.y,p.z);
glVertex3d(r*p.x+cnt->x,r*p.y+cnt->y,r*p.z+cnt->z);

/*:253*/
#line 4675 "MathView3D.w"

p.x= cp1*st1;p.y= sp1*st1;p.z= ct1;
/*253:*/
#line 4691 "MathView3D.w"

glNormal3d(p.x,p.y,p.z);
glVertex3d(r*p.x+cnt->x,r*p.y+cnt->y,r*p.z+cnt->z);

/*:253*/
#line 4677 "MathView3D.w"

p.x= cp1*st;p.y= sp1*st;p.z= ct;
/*253:*/
#line 4691 "MathView3D.w"

glNormal3d(p.x,p.y,p.z);
glVertex3d(r*p.x+cnt->x,r*p.y+cnt->y,r*p.z+cnt->z);

/*:253*/
#line 4679 "MathView3D.w"

glEnd();

/*:251*/
#line 4648 "MathView3D.w"

/*252:*/
#line 4682 "MathView3D.w"

glBegin(GL_POLYGON);
/*253:*/
#line 4691 "MathView3D.w"

glNormal3d(p.x,p.y,p.z);
glVertex3d(r*p.x+cnt->x,r*p.y+cnt->y,r*p.z+cnt->z);

/*:253*/
#line 4684 "MathView3D.w"

p.x= cp*st;p.y= sp*st;p.z= ct;
/*253:*/
#line 4691 "MathView3D.w"

glNormal3d(p.x,p.y,p.z);
glVertex3d(r*p.x+cnt->x,r*p.y+cnt->y,r*p.z+cnt->z);

/*:253*/
#line 4686 "MathView3D.w"

p= p0;
/*253:*/
#line 4691 "MathView3D.w"

glNormal3d(p.x,p.y,p.z);
glVertex3d(r*p.x+cnt->x,r*p.y+cnt->y,r*p.z+cnt->z);

/*:253*/
#line 4688 "MathView3D.w"

glEnd();

/*:252*/
#line 4649 "MathView3D.w"

}
}
}
}

/*:248*/
#line 4396 "MathView3D.w"

/*254:*/
#line 4695 "MathView3D.w"

void GLPrintPointList(PointList*lst)
{
Point3D*p;

if(0.0>=PointSphereSize){
glDisable(GL_LIGHTING);
while(lst){
p= PointPool+lst->point;
glBegin(GL_POINTS);
glColor3dv((double*)&(lst->color));
glVertex3dv((double*)p);
glEnd();
lst= lst->next;
}
glEnable(GL_LIGHTING);
}
else{
while(lst){
p= PointPool+lst->point;
DefineMaterial(INHERITE_COLOR,lst->color);
PointSphere(p,lst->size*PointSphereSize,
PointSphereSegmentsPhi,
PointSphereSegmentsTheta,
lst->color.r,
lst->color.g,
lst->color.b
);
lst= lst->next;
}
}
}

/*:254*/
#line 4397 "MathView3D.w"

/*255:*/
#line 4728 "MathView3D.w"

void GLPrintLinePolyList(LinePolyList*lst)
{int i,thisI;
int j1,j2,j3;
Point3D p,normal;
RGBColor lc;

while(lst){
switch(lst->what){
case Polygon3D:
if(!WireFrame)
DefineMaterial(INHERITE_COLOR,lst->color);
else{
lc.r= lc.g= lc.b= LineColor;
DefineMaterial(INHERITE_COLOR,lst->color);
}
glPolygonMode(GL_FRONT_AND_BACK,(WireFrame?GL_LINE:GL_FILL));
glBegin(GL_POLYGON);
for(i= 0;i<lst->pno;i++){
thisI= lst->points[i];
/*256:*/
#line 4765 "MathView3D.w"

if(NormalPool[thisI]&&SmoothShading)
glNormal3dv((GLdouble*)*(NormalPool+thisI));
else{
/*92:*/
#line 1870 "MathView3D.w"

j1= lst->points[0==i?lst->pno-1:i-1];
j2= lst->points[i];
j3= lst->points[lst->pno-1==i?0:i+1];

/*:92*/
#line 4769 "MathView3D.w"

if(SurfaceNormal(&normal,PointPool+j1,PointPool+j2,PointPool+j3))
glNormal3dv((GLdouble*)&normal);
}
glVertex3dv((GLdouble*)(PointPool+thisI));

/*:256*/
#line 4748 "MathView3D.w"

}
glEnd();
break;
case Line3D:
if(fabs(LineTubeSize)>MachineEpsilon){
glPolygonMode(GL_FRONT_AND_BACK,WireFrame?GL_LINE:GL_FILL);
/*257:*/
#line 4775 "MathView3D.w"

{
DefineMaterial(INHERITE_COLOR,lst->color);
if(2==lst->pno)
MakeCone(PointPool+lst->points[0],
PointPool+lst->points[1],
LineTubeSize,TubeSegments);
else{
for(i= 0;i<lst->pno-2;i++)
MakeConeContinued(
PointPool+lst->points[i],
PointPool+lst->points[i+1],
PointPool+lst->points[i+2],
LineTubeSize,TubeSegments);
i= lst->pno-2;
if(lst->points[0]==lst->points[i+1]){
MakeConeContinued(
PointPool+lst->points[i],
PointPool+lst->points[0],
PointPool+lst->points[1],
LineTubeSize,TubeSegments);
}
else
MakeCone(PointPool+lst->points[i],
PointPool+lst->points[i+1],
LineTubeSize,TubeSegments);
}
}

/*:257*/
#line 4755 "MathView3D.w"

}
else
/*258:*/
#line 4804 "MathView3D.w"

{
glBegin(lst->pno>2?GL_LINE_STRIP:GL_LINES);
for(i= 0;i<lst->pno;i++){
p= PointPool[lst->points[i]];
glVertex3d(p.x,p.y,p.z);
}
glEnd();
}


/*:258*/
#line 4758 "MathView3D.w"

break;
}
lst= lst->next;
}
}

/*:255*/
#line 4398 "MathView3D.w"

/*259:*/
#line 4816 "MathView3D.w"

void GLPrintTexPolyList(int texMapType,LinePolyList*lst)
{int i,thisI;
int j1,j2,j3;
Point3D normal,*tp;
register double scaleS,scaleT;

glEnable(GL_TEXTURE_2D);
scaleS= SceneTextureMap->scaleS;
scaleT= SceneTextureMap->scaleT;
if((GL_RGBA==TextureType(SceneTextureMap->tex))||
(GL_LUMINANCE_ALPHA==TextureType(SceneTextureMap->tex)))
glEnable(GL_BLEND);
glBindTexture(GL_TEXTURE_2D,SceneTextureID);
glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
while(lst){
if(TEXTURE_MAP_PLANAR==texMapType)
/*260:*/
#line 4844 "MathView3D.w"

{
glBegin(GL_POLYGON);
for(i= 0;i<lst->pno;i++){
thisI= lst->points[i];
tp= lst->pTexturePoints+i;
glTexCoord2d(scaleS*tp->x,scaleT*tp->y);
/*256:*/
#line 4765 "MathView3D.w"

if(NormalPool[thisI]&&SmoothShading)
glNormal3dv((GLdouble*)*(NormalPool+thisI));
else{
/*92:*/
#line 1870 "MathView3D.w"

j1= lst->points[0==i?lst->pno-1:i-1];
j2= lst->points[i];
j3= lst->points[lst->pno-1==i?0:i+1];

/*:92*/
#line 4769 "MathView3D.w"

if(SurfaceNormal(&normal,PointPool+j1,PointPool+j2,PointPool+j3))
glNormal3dv((GLdouble*)&normal);
}
glVertex3dv((GLdouble*)(PointPool+thisI));

/*:256*/
#line 4851 "MathView3D.w"

}
glEnd();
}

/*:260*/
#line 4833 "MathView3D.w"

else
/*261:*/
#line 4870 "MathView3D.w"

{
if(CrossXPlusAxis(lst)){
glBegin(GL_POLYGON);
for(i= 0;i<lst->pno;i++){
thisI= lst->points[i];
tp= lst->pTexturePoints+i;
glTexCoord2d(scaleS*(fabs(tp->x)<0.75?1.0+tp->x:tp->x),scaleT*tp->y);
/*256:*/
#line 4765 "MathView3D.w"

if(NormalPool[thisI]&&SmoothShading)
glNormal3dv((GLdouble*)*(NormalPool+thisI));
else{
/*92:*/
#line 1870 "MathView3D.w"

j1= lst->points[0==i?lst->pno-1:i-1];
j2= lst->points[i];
j3= lst->points[lst->pno-1==i?0:i+1];

/*:92*/
#line 4769 "MathView3D.w"

if(SurfaceNormal(&normal,PointPool+j1,PointPool+j2,PointPool+j3))
glNormal3dv((GLdouble*)&normal);
}
glVertex3dv((GLdouble*)(PointPool+thisI));

/*:256*/
#line 4878 "MathView3D.w"

}
glEnd();
}
else
/*260:*/
#line 4844 "MathView3D.w"

{
glBegin(GL_POLYGON);
for(i= 0;i<lst->pno;i++){
thisI= lst->points[i];
tp= lst->pTexturePoints+i;
glTexCoord2d(scaleS*tp->x,scaleT*tp->y);
/*256:*/
#line 4765 "MathView3D.w"

if(NormalPool[thisI]&&SmoothShading)
glNormal3dv((GLdouble*)*(NormalPool+thisI));
else{
/*92:*/
#line 1870 "MathView3D.w"

j1= lst->points[0==i?lst->pno-1:i-1];
j2= lst->points[i];
j3= lst->points[lst->pno-1==i?0:i+1];

/*:92*/
#line 4769 "MathView3D.w"

if(SurfaceNormal(&normal,PointPool+j1,PointPool+j2,PointPool+j3))
glNormal3dv((GLdouble*)&normal);
}
glVertex3dv((GLdouble*)(PointPool+thisI));

/*:256*/
#line 4851 "MathView3D.w"

}
glEnd();
}

/*:260*/
#line 4883 "MathView3D.w"


}

/*:261*/
#line 4835 "MathView3D.w"

lst= lst->next;
}
if((GL_RGBA==TextureType(SceneTextureMap->tex))||
(GL_LUMINANCE_ALPHA==TextureType(SceneTextureMap->tex)))
glDisable(GL_BLEND);
glDisable(GL_TEXTURE_2D);
}

/*:259*/
#line 4399 "MathView3D.w"

/*263:*/
#line 4911 "MathView3D.w"

void GLPrintMesh3DList(Mesh3DList*lst)
{
DefineMaterial(INHERITE_COLOR,lst->mColor);
if(!SmoothShading)glShadeModel(GL_FLAT);
glPolygonMode(GL_FRONT_AND_BACK,WireFrame?GL_LINE:GL_FILL);
while(lst){
glPushAttrib(GL_ALL_ATTRIB_BITS);
if(lst->mTexture&&lst->mTexture->tex)
/*268:*/
#line 4983 "MathView3D.w"

{register int i,j;
register double scaleS,scaleT;
int n= lst->dimU-1,m= lst->dimV-1;
Point3D*tp;

scaleS= lst->mTexture->scaleS;
scaleT= lst->mTexture->scaleT;
glEnable(GL_TEXTURE_2D);
if((GL_RGBA==MeshTextureType(lst))||(GL_LUMINANCE_ALPHA==MeshTextureType(lst)))
glEnable(GL_BLEND);
glBindTexture(GL_TEXTURE_2D,MeshTextureID(lst));
for(i= 0;i<n;i++){
for(j= 0;j<m;j++){
/*265:*/
#line 4949 "MathView3D.w"

glBegin(GL_QUADS);
MeshSmoothTexPoint(i,j)
MeshSmoothTexPoint(i+1,j)
MeshSmoothTexPoint(i+1,j+1)
MeshSmoothTexPoint(i,j+1)
glEnd();


/*:265*/
#line 4997 "MathView3D.w"

}
}
if((GL_RGBA==MeshTextureType(lst))||(GL_LUMINANCE_ALPHA==MeshTextureType(lst)))
glDisable(GL_BLEND);
glDisable(GL_TEXTURE_2D);
}


/*:268*/
#line 4920 "MathView3D.w"

else if(lst->mpolyColors)
/*267:*/
#line 4970 "MathView3D.w"

{register int i,j;
int n= lst->dimU-1,m= lst->dimV-1;

for(i= 0;i<n;i++){
for(j= 0;j<m;j++){
glColorMaterial(GL_FRONT_AND_BACK,GL_DIFFUSE);
glColor3dv((double*)MeshColorPtr(lst,i,j));
/*264:*/
#line 4941 "MathView3D.w"

glBegin(GL_QUADS);
MeshSmoothPoint(i,j)
MeshSmoothPoint(i+1,j)
MeshSmoothPoint(i+1,j+1)
MeshSmoothPoint(i,j+1)
glEnd();

/*:264*/
#line 4978 "MathView3D.w"

}
}
}

/*:267*/
#line 4922 "MathView3D.w"

else
/*266:*/
#line 4958 "MathView3D.w"

{register int i,j;
int n= lst->dimU-1,m= lst->dimV-1;

glColor3d(1.,1.,1.);
for(i= 0;i<n;i++){
for(j= 0;j<m;j++){
/*264:*/
#line 4941 "MathView3D.w"

glBegin(GL_QUADS);
MeshSmoothPoint(i,j)
MeshSmoothPoint(i+1,j)
MeshSmoothPoint(i+1,j+1)
MeshSmoothPoint(i,j+1)
glEnd();

/*:264*/
#line 4965 "MathView3D.w"

}
}
}

/*:266*/
#line 4924 "MathView3D.w"

glPopAttrib();
lst= lst->next;
}
if(!SmoothShading)glShadeModel(GL_SMOOTH);

}

/*:263*/
#line 4400 "MathView3D.w"

/*262:*/
#line 4887 "MathView3D.w"

void GLPrintOutlinePolyList(LinePolyList*lst,RGBColor c)
{int i;
Point3D p;

glDisable(GL_LIGHTING);

glLineWidth(2.0f);
glColor3d(c.r,c.g,c.b);
glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
while(lst){
glBegin(GL_POLYGON);
for(i= 0;i<lst->pno;i++){
p= PointPool[lst->points[i]];
glVertex3d(p.x,p.y,p.z);
}
glEnd();
lst= lst->next;
}
glLineWidth(1.0f);
glEnable(GL_LIGHTING);
}

/*:262*/
#line 4401 "MathView3D.w"

/*244:*/
#line 4518 "MathView3D.w"

double CalcScaling(Point3D*p1,Point3D*p2)
{double is;

ScaleOfScene.x= (p2->x-p1->x);
ScaleOfScene.y= (p2->y-p1->y);
ScaleOfScene.z= (p2->z-p1->z);
MaxScaleOfScene= MAX3(ScaleOfScene.x,ScaleOfScene.y,ScaleOfScene.z);
if(MaxScaleOfScene!=0.0)
is= 2.0/MaxScaleOfScene;
else
is= 1.0;
return is;
}

/*:244*/
#line 4402 "MathView3D.w"

/*245:*/
#line 4533 "MathView3D.w"

void ScalingToUnitCube(Point3D*p1,Point3D*p2)
{double is;

is= CalcScaling(p1,p2);
glScaled(BoxRatios.x*is,BoxRatios.y*is,BoxRatios.z*is);
}

/*:245*/
#line 4403 "MathView3D.w"

/*246:*/
#line 4546 "MathView3D.w"

void DefineLightSources(
Point3D*lightpos,
RGBColor*lightc,
int lno,
RGBColor ambient)
{float flt[4];
int i;

glEnable(GL_LIGHTING);
SET4VEC(flt,ambient);
glLightModelfv(GL_LIGHT_MODEL_AMBIENT,flt);
glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
glFrontFace(GL_CCW);
for(i= 0;i<lno;i++){
glEnable(GL_LIGHT0+i);
flt[0]= lightpos[i].x;flt[1]= lightpos[i].y;flt[2]= lightpos[i].z;
flt[3]= 1.0;
glLightfv(GL_LIGHT0+i,GL_POSITION,flt);
SET4VEC(flt,lightc[i]);
glLightfv(GL_LIGHT0+i,GL_AMBIENT,flt);
glLightfv(GL_LIGHT0+i,GL_DIFFUSE,flt);
glLightfv(GL_LIGHT0+i,GL_SPECULAR,flt);
glLightf(GL_LIGHT0+i,GL_CONSTANT_ATTENUATION,1.0f);
glLightf(GL_LIGHT0+i,GL_LINEAR_ATTENUATION,0.25f);
glLightf(GL_LIGHT0+i,GL_QUADRATIC_ATTENUATION,0.0f);
}
for(i= lno;i<MAX_LIGHT;i++)
glDisable(GL_LIGHT0+i);
glEnable(GL_COLOR_MATERIAL);
glColorMaterial(GL_FRONT,GL_DIFFUSE);
}

/*:246*/
#line 4404 "MathView3D.w"

/*242:*/
#line 4475 "MathView3D.w"

void CreateObject(void)
{
if(PointPool||Meshes){
glPushMatrix();

if(ObjID<0)
ObjID= glGenLists(1);
else{
glDeleteLists(ObjID,1);
ObjID= glGenLists(1);
}
glNewList(ObjID,GL_COMPILE);
if(Points)
GLPrintPointList(Points);
if(Lines)
GLPrintLinePolyList(Lines);
if(Polygons){
if(Textures&&SceneTextureMap&&!WireFrame)
GLPrintTexPolyList(SceneTextureMap->type,Polygons);
else
GLPrintLinePolyList(Polygons);
}
if(PolyOutline)
GLPrintOutlinePolyList(Polygons,BlackBackground?White:Black);
if(Meshes)
GLPrintMesh3DList(Meshes);
if(Axis.AxisType){
DrawAxis(Axis.AxisType,
&Axis.Origin,&BoundP1,&BoundP2,
Axis.XLabel,Axis.YLabel,Axis.ZLabel,
Axis.FontScale,Axis.NumericLabels);
}
glEndList();
glPopMatrix();
}
}

/*:242*/
#line 4405 "MathView3D.w"

/*241:*/
#line 4421 "MathView3D.w"

void RecalcView(double fac,int autoplanes)
{double vr,vphi,vtheta;
Point3D vp;
double nearPlane,farPlane;

glMatrixMode(GL_PROJECTION);
glLoadIdentity();
if(autoplanes){
Point3D p1,p2;

p2.x= p2.y= p2.z= 0.5*sqrt(3.);
p1.x= p1.y= p1.z= -p2.x;
EstimateViewDepth(&nearPlane,&farPlane,
&ViewPoint,&p1,&p2);
nearPlane*= 0.25;
}
else{
nearPlane= MVNearPlane;
farPlane= MVFarPlane;
}
gluPerspective(MVViewAngle,
1.0,
nearPlane,farPlane);
glMatrixMode(GL_MODELVIEW);
glLoadIdentity();
SphericalFromCartesian(&ViewPoint,&vr,&vtheta,&vphi);
vp.x= fac*vr*cos(vphi)*sin(vtheta);
vp.y= fac*vr*sin(vphi)*sin(vtheta);
vp.z= fac*vr*cos(vtheta);
CenterOfScene.x= 0.5*(BoundP1.x+BoundP2.x);
CenterOfScene.y= 0.5*(BoundP1.y+BoundP2.y);
CenterOfScene.z= 0.5*(BoundP1.z+BoundP2.z);
gluLookAt(vp.x,vp.y,vp.z,
0.0,0.0,0.0,
ViewVertical.x,ViewVertical.y,ViewVertical.z);

DefineLightSources(LightPosition,LightColor,LightNo,AmbientColor);
glEnable(GL_NORMALIZE);
glBlendFunc(
GL_SRC_ALPHA,
GL_ONE_MINUS_SRC_ALPHA);
ScalingToUnitCube(&BoundP1,&BoundP2);
glTranslated(-CenterOfScene.x,-CenterOfScene.y,-CenterOfScene.z);

glPushMatrix();

}


/*:241*/
#line 4406 "MathView3D.w"

/*240:*/
#line 4411 "MathView3D.w"

void InitView(int autoplanes)
{
glEnable(GL_DEPTH_TEST);
glShadeModel(GL_SMOOTH);
if(Textures)GLSetTextures(Textures);
CreateObject();
RecalcView(1.0,autoplanes);
}

/*:240*/
#line 4407 "MathView3D.w"



/*:239*/
#line 77 "MathView3D.w"

/*67:*/
#line 1546 "MathView3D.w"

/*66:*/
#line 1538 "MathView3D.w"

void MathPrintPosition(FILE*f,int index)
{Point3D*p;

p= PointPool+index;
fprintf(f,"{ "DBL_FMT", "DBL_FMT", "DBL_FMT"}",p->x,p->y,p->z);
}

/*:66*/
#line 1547 "MathView3D.w"


/*:67*//*75:*/
#line 1621 "MathView3D.w"

/*74:*/
#line 1602 "MathView3D.w"

void MathPrintPointList(FILE*f,PointList*lst)
{
fprintf(f,"{");
while(lst){
fprintf(f,"\t{PointSize["DBL_FMT"], RGBColor["DBL_FMT", "DBL_FMT", "
DBL_FMT"], Point[",
lst->size,
lst->color.r,
lst->color.g,
lst->color.b);
MathPrintPosition(f,lst->point);
fprintf(f,"]}");
if(lst->next)fprintf(f,",\n");
lst= lst->next;
}
fprintf(f,"}");
}

/*:74*/
#line 1622 "MathView3D.w"


/*:75*//*91:*/
#line 1867 "MathView3D.w"

/*90:*/
#line 1848 "MathView3D.w"

void MathPrintLinePolyList(FILE*f,LinePolyList*lst)
{int i;

fprintf(f,"{");
while(lst){
fprintf(f,"%s[{",Line3D==lst->what?"Line":"Polygon");
for(i= 0;i<lst->pno;i++){
MathPrintPosition(f,lst->points[i]);
if(i!=lst->pno-1)
fprintf(f,",");
}
fprintf(f,lst->next?"}]},\n":"}]}");
lst= lst->next;
}
fprintf(f,"}\n");
}


/*:90*/
#line 1868 "MathView3D.w"


/*:91*/
#line 78 "MathView3D.w"

/*9:*/
#line 242 "MathView3D.w"

void MVTrackball(double*angle,
Point3D*axis,
int x1,int y1,
int x2,int y2)
{double phi1,theta1,phi2,theta2,n,arg;
Point3D v1,v2;

phi1= PI*(2.0*x1-ScreenWidth)/((double)ScreenWidth);
phi2= PI*(2.0*x2-ScreenWidth)/((double)ScreenWidth);
theta1= PI*(ScreenHeight-y1)/((double)ScreenHeight);
theta2= PI*(ScreenHeight-y2)/((double)ScreenHeight);
CartesianFromSpherical(1.0,theta1,phi1,&v1);
CartesianFromSpherical(1.0,theta2,phi2,&v2);
Cross(axis,&v2,&v1);
n= Norm(axis);
if(n>0.0){
axis->x/= n;
axis->y/= n;
axis->z/= n;
}
else{
axis->x= axis->y= 0.0;
axis->z= 1.0;
}
if(ROTATE_Z==(DoRotate&ROTATE_Z)){
*angle= phi2-phi1;
return;
}
else if((ROTATE_X==(DoRotate&ROTATE_X))||
(ROTATE_Y==(DoRotate&ROTATE_Y))){
*angle= TWO_PI*(y2-y1)/((double)ScreenHeight);
return;
}
n= Norm(&v1);
if(n>0.0){
v1.x/= n;
v1.y/= n;
v1.z/= n;
}
n= Norm(&v2);
if(n>0.0){
v2.x/= n;
v2.y/= n;
v2.z/= n;
}
arg= Dot(&v1,&v2);
if(arg<-1.)arg= -1.0;
else if(arg>1.0)arg= 1.0;
*angle= acos(arg);
}

/*:9*/
#line 79 "MathView3D.w"

/*287:*/
#line 5507 "MathView3D.w"

/*289:*/
#line 5538 "MathView3D.w"

void WritePOVVector(FILE*f,char*pre,Point3D*p,char*post)
{
fprintf(f,"%s<"DBL_FMT", "DBL_FMT", "DBL_FMT">%s",pre,p->x,p->y,p->z,post);
}

/*:289*/
#line 5508 "MathView3D.w"

/*295:*/
#line 5625 "MathView3D.w"

int InColorArray(RGBColor*c,RGBColor**colors,int cno)
{int i;

for(i= 0;i<cno;i++){
if(fabs(c->r-colors[i]->r)<TINY&&
fabs(c->g-colors[i]->g)<TINY&&
fabs(c->b-colors[i]->b)<TINY)return i;
}
return-1;
}

/*:295*/
#line 5509 "MathView3D.w"

/*294:*/
#line 5600 "MathView3D.w"

RGBColor**PolyColorTable(int*cno,LinePolyList*lst)
{RGBColor**cptr;
int i,polyno,cindex;

polyno= LengthLinePolyList(lst);
cptr= (RGBColor**)calloc(polyno,sizeof(RGBColor*));
for(i= 0;i<polyno;i++)
cptr[i]= NULL;
*cno= 1;
cptr[0]= &(lst->color);
lst->aux1= 0;
lst= lst->next;
while(lst){
cindex= InColorArray(&(lst->color),cptr,*cno);
if(cindex<0){
cptr[cindex= (*cno)]= &(lst->color);
(*cno)++;
}
lst->aux1= cindex;
lst= lst->next;
}
return cptr;
}

/*:294*/
#line 5510 "MathView3D.w"

/*297:*/
#line 5648 "MathView3D.w"

int WhiteSurface(RGBColor*c)
{
return(fabs(1.0-c->r)<TINY)&&(fabs(1.0-c->g)<TINY)&&(fabs(1.0-c->b)<TINY);
}

/*:297*/
#line 5511 "MathView3D.w"

/*321:*/
#line 6144 "MathView3D.w"

void WritePOVColor(FILE*f,
RGBColor*c)
{fprintf(f,"texture {"
"pigment { color <"DBL_FMT", "DBL_FMT", "DBL_FMT">}"
"finish { DefaultFinish }"
"}\n",c->r,c->g,c->b);
}

/*:321*/
#line 5512 "MathView3D.w"

/*296:*/
#line 5637 "MathView3D.w"

void WritePOVColorTab(FILE*f,RGBColor**colors,int cno)
{int i;

fprintf(f,"\n/* Color table */\n");
for(i= 0;i<cno;i++)
fprintf(f,"#declare  MVColor%d = rgb <"DBL_FMT", "DBL_FMT", "DBL_FMT">;\n",
i,colors[i]->r,colors[i]->g,colors[i]->b);
fprintf(f,"/* End color table */\n\n");
}

/*:296*/
#line 5513 "MathView3D.w"

/*293:*/
#line 5588 "MathView3D.w"

int CheckNormals(int*points,int no)
{int res,i;

res= NULL!=NormalPool[points[0]];
for(i= 1;i<no&&res;i++)
res= res&&(NULL!=NormalPool[points[i]]);
return res;
}

/*:293*/
#line 5514 "MathView3D.w"

/*307:*/
#line 5837 "MathView3D.w"

void WritePOVFlatTriangle(FILE*f,
Point3D*p1,Point3D*p2,Point3D*p3,int*mcount,int*c)
{
fprintf(f,"\ttriangle {\n");
WritePOVVector(f,"\t",p1,",\n");
WritePOVVector(f,"\t",p2,",\n");
WritePOVVector(f,"\t",p3,"\n");
/*308:*/
#line 5849 "MathView3D.w"

if(mcount&&c)
fprintf(f,"\ttexture { Mesh%dColor%d}\n",*mcount,*c);

/*:308*/
#line 5845 "MathView3D.w"

fprintf(f,"\t}\n");
}

/*:307*/
#line 5515 "MathView3D.w"

/*306:*/
#line 5820 "MathView3D.w"

void WritePOVSmoothTriangle(FILE*f,
Point3D*p1,Point3D*p2,Point3D*p3,
Point3D*n1,Point3D*n2,Point3D*n3,int*mcount,int*c)
{
if(n1&&n2&&n3){
fprintf(f,"\tsmooth_triangle {\n");
WritePOVVector(f,"\t",p1,", ");WritePOVVector(f,"",n1,",\n");
WritePOVVector(f,"\t",p2,", ");WritePOVVector(f,"",n2,",\n");
WritePOVVector(f,"\t",p3,", ");WritePOVVector(f,"",n3,"\n");
/*308:*/
#line 5849 "MathView3D.w"

if(mcount&&c)
fprintf(f,"\ttexture { Mesh%dColor%d}\n",*mcount,*c);

/*:308*/
#line 5830 "MathView3D.w"

fprintf(f,"\t}\n");
}
else
WritePOVFlatTriangle(f,p1,p2,p3,mcount,c);
}

/*:306*/
#line 5516 "MathView3D.w"

/*291:*/
#line 5558 "MathView3D.w"

void POVTrackballPosition(FILE*f)
{char mstr[256];
Point3D p1;

p1.x= -CenterOfScene.x;
p1.y= -CenterOfScene.y;
p1.z= -CenterOfScene.z;
WritePOVVector(f,"\ttranslate ",&p1,"\n");
sprintf(mstr,"\trotate "DBL_FMT"*",ToDegree*TrackAngle);
WritePOVVector(f,mstr,&TrackAxis,"\n");
WritePOVVector(f,"\ttranslate ",&CenterOfScene,"\n");
}

/*:291*/
#line 5517 "MathView3D.w"

/*316:*/
#line 6008 "MathView3D.w"

void MakePOVCone(FILE*f,Point3D*p1,Point3D*p2,double r,int n)
{Point3D v1,v2,v3,vn[4],p[4];
double norm;
double dphi;
int i;

v1.x= p2->x-p1->x;
v1.y= p2->y-p1->y;
v1.z= p2->z-p1->z;
norm= Norm(&v1);
v1.x/= norm;v1.y/= norm;v1.z/= norm;
OrthogonalSet(&v1,&v2,&v3);

v1.x= p2->x-p1->x;
v1.y= p2->y-p1->y;
v1.z= p2->z-p1->z;

dphi= TWO_PI/n;
ConePoint(r,0.0,&v2,&v3,p);
vn[0]= vn[1]= *p;
IncPoint3D(p[0],(*p1))
AddPoint3D(p[1],p[0],v1)
for(i= 1;i<=n;i++){
ConePoint(r,i*dphi,&v2,&v3,p+2);
vn[2]= p[2];
IncPoint3D(p[2],(*p1))
WritePOVSmoothTriangle(f,p,p+1,p+2,vn,vn+1,vn+2,NULL,NULL);
AddPoint3D(p[3],p[2],v1)
vn[3]= vn[2];
WritePOVSmoothTriangle(f,p+2,p+1,p+3,vn+2,vn+1,vn+3,NULL,NULL);
p[0]= p[2];vn[0]= vn[2];
p[1]= p[3];vn[1]= vn[3];
}
}

/*:316*/
#line 5518 "MathView3D.w"

/*317:*/
#line 6044 "MathView3D.w"

void MakePOVConeContinued(FILE*f,Point3D*p1,Point3D*p2,Point3D*p3,
double r,int n)
{Point3D v1,v2,v3,u1,u2,u3,vn[4],p[4];
double norm;
double dphi;
int i;

v1.x= p2->x-p1->x;v1.y= p2->y-p1->y;v1.z= p2->z-p1->z;
norm= Norm(&v1);
v1.x/= norm;v1.y/= norm;v1.z/= norm;
OrthogonalSet(&v1,&v2,&v3);
u1.x= p3->x-p2->x;u1.y= p3->y-p2->y;u1.z= p3->z-p2->z;
norm= Norm(&u1);
u1.x/= norm;u1.y/= norm;u1.z/= norm;
/*318:*/
#line 6084 "MathView3D.w"

if(Dot(&u1,&v1)<CosSkipCont){
MakePOVCone(f,p1,p2,r,n);
MakePOVCone(f,p2,p3,r,n);
return;
}

/*:318*/
#line 6059 "MathView3D.w"

u2= v2;u3= v3;
GramSchmidt(&u1,&u2,&u3);
dphi= TWO_PI/n;
ConePoint(r,0.0,&v2,&v3,p);
vn[0]= *p;
IncPoint3D(p[0],(*p1))
ConePoint(r,0.0,&u2,&u3,p+1);
vn[1]= p[1];
IncPoint3D(p[1],(*p2))
for(i= 1;i<=n;i++){
ConePoint(r,i*dphi,&v2,&v3,p+2);
vn[2]= p[2];
IncPoint3D(p[2],(*p1))
WritePOVSmoothTriangle(f,p,p+1,p+2,vn,vn+1,vn+2,NULL,NULL);
ConePoint(r,i*dphi,&u2,&u3,p+3);
vn[3]= p[3];
IncPoint3D(p[3],(*p2))
WritePOVSmoothTriangle(f,p+2,p+1,p+3,vn+2,vn+1,vn+3,NULL,NULL);
p[0]= p[2];vn[0]= vn[2];
p[1]= p[3];vn[1]= vn[3];

}
}

/*:317*/
#line 5519 "MathView3D.w"

/*292:*/
#line 5572 "MathView3D.w"

void WritePOVPoints(FILE*f,PointList*lst)
{
fprintf(f,"/* Begin Points */\n\n union {\n");
while(lst){
fprintf(f,"\tsphere {\n");
WritePOVVector(f,"\t",PointPool+lst->point," ");
fprintf(f,DBL_FMT" \n\t}\n",PointSphereSize*lst->size);
lst= lst->next;
}
if(TrackAngle!=0.0)
POVTrackballPosition(f);
fprintf(f,"}\n/* End of Points */\n");

}

/*:292*/
#line 5520 "MathView3D.w"

/*319:*/
#line 6091 "MathView3D.w"

void WritePOVLines(FILE*f,LinePolyList*lst)
{int i,cnt;

fprintf(f,"/* Begin lines */\n");
if(fabs(LineTubeSize)>MachineEpsilon){
cnt= 1;
while(lst){
fprintf(f,"/* Line %d */\n",cnt);
fprintf(f,"mesh {\n");
/*320:*/
#line 6117 "MathView3D.w"

{if(2==lst->pno)
MakePOVCone(f,PointPool+lst->points[0],
PointPool+lst->points[1],
LineTubeSize,TubeSegments);
else{
for(i= 0;i<lst->pno-2;i++)
MakePOVConeContinued(f,
PointPool+lst->points[i],
PointPool+lst->points[i+1],
PointPool+lst->points[i+2],
LineTubeSize,TubeSegments);
i= lst->pno-2;
if(lst->points[0]==lst->points[i+1]){
MakePOVConeContinued(f,
PointPool+lst->points[i],
PointPool+lst->points[0],
PointPool+lst->points[1],
LineTubeSize,TubeSegments);
}
else
MakePOVCone(f,PointPool+lst->points[i],
PointPool+lst->points[i+1],
LineTubeSize,TubeSegments);
}
}

/*:320*/
#line 6101 "MathView3D.w"

fprintf(f,"texture {\n"
"\t pigment { color rgb <"DBL_FMT", "DBL_FMT", "DBL_FMT">}\n"
"\t finish { DefaultFinish }\n"
"\t}\n",lst->color.r,lst->color.g,lst->color.b);
if(TrackAngle!=0.0)
POVTrackballPosition(f);
fprintf(f,"}\n/* End of Line %d*/\n",cnt++);
lst= lst->next;
}
}
else
fprintf(f,"// Line output skiped due to zero LineTubeSize\n");
fprintf(f,"/* End lines */");
}

/*:319*/
#line 5521 "MathView3D.w"

/*298:*/
#line 5654 "MathView3D.w"

void WritePOVTriangles(FILE*f,LinePolyList*polys,int smooth,int colorNo)
{int i,j,thisI;
Point3D*p;
int isSmooth;
LinePolyList*lst;

fprintf(f,"/* Begin polygones */\n\n");
for(j= 0;j<colorNo;j++){
lst= polys;
fprintf(f,"object {\n");
fprintf(f,"mesh {\n");
while(lst){
if(j!=lst->aux1){
lst= lst->next;
continue;
}
/*305:*/
#line 5800 "MathView3D.w"

isSmooth= smooth&&CheckNormals(lst->points,lst->pno);
if(lst->pno>=3){
fprintf(f,"%s {\n",isSmooth?"smooth_triangle":"triangle");
for(i= 0;i<lst->pno;i++){
p= PointPool+(thisI= lst->points[i]);
WritePOVVector(f,"\t",p,"");
if(NormalPool[thisI]&&isSmooth)
WritePOVVector(f,", ",NormalPool[thisI],"");
fprintf(f,"%s",i==lst->pno-1?"\n":",\n");
}
fprintf(f,"}\n");
}

/*:305*/
#line 5671 "MathView3D.w"

lst= lst->next;
}
fprintf(f,"}\n");
fprintf(f,"texture {\n"
"\t  pigment { MVColor%d }\n"
"\t  finish { DefaultFinish }"
"\t}\n",j);
if(TrackAngle!=0.0)
POVTrackballPosition(f);
fprintf(f,"}\n");
}
fprintf(f,"\n/* End polygones  */\n");
}

/*:298*/
#line 5522 "MathView3D.w"

/*299:*/
#line 5686 "MathView3D.w"

void WriteTexPOVTriangles(FILE*f,LinePolyList*polys,int smooth,
const char*pTexPrefix)
{int i,thisI;
Point3D*p,*t,*axis,*scale;
double angle,scS,scT;
int isSmooth;
LinePolyList*lst;
int tmapType;

static int texFileCount= 1;

if(!*TextureFile(SceneTextureMap->tex))
/*378:*/
#line 7166 "MathView3D.w"

{
sprintf(TextureFile(SceneTextureMap->tex),
"%stex%d.png",pTexPrefix,texFileCount++);
PNGSaveBitmap(
TextureFile(SceneTextureMap->tex),
TextureWidth(SceneTextureMap->tex),
TextureHeight(SceneTextureMap->tex),
TextureType(SceneTextureMap->tex),
TextureData(SceneTextureMap->tex),GL_FALSE,1.0);
}

/*:378*/
#line 5699 "MathView3D.w"

tmapType= SceneTextureMap->type;
t= SceneTextureMap->translate;
axis= SceneTextureMap->axis;
angle= SceneTextureMap->angle;
scale= SceneTextureMap->scale;
scS= SceneTextureMap->scaleS;
scT= SceneTextureMap->scaleT;
fprintf(f,"/* Begin polygones */\n\n");
lst= polys;
fprintf(f,"object {\n");
fprintf(f,"mesh {\n");
while(lst){
/*305:*/
#line 5800 "MathView3D.w"

isSmooth= smooth&&CheckNormals(lst->points,lst->pno);
if(lst->pno>=3){
fprintf(f,"%s {\n",isSmooth?"smooth_triangle":"triangle");
for(i= 0;i<lst->pno;i++){
p= PointPool+(thisI= lst->points[i]);
WritePOVVector(f,"\t",p,"");
if(NormalPool[thisI]&&isSmooth)
WritePOVVector(f,", ",NormalPool[thisI],"");
fprintf(f,"%s",i==lst->pno-1?"\n":",\n");
}
fprintf(f,"}\n");
}

/*:305*/
#line 5712 "MathView3D.w"

lst= lst->next;
}
fprintf(f,"}\n");
/*300:*/
#line 5735 "MathView3D.w"

if(t)
fprintf(f,"\t translate <"DBL_FMT", "DBL_FMT", "DBL_FMT">\n",t->x,t->y,t->z);
if(axis&&angle!=0.0)
fprintf(f,"\t rotate "DBL_FMT"*<"DBL_FMT", "DBL_FMT", "DBL_FMT">\n",
ToDegree*angle,
axis->x,axis->y,axis->z);
if(scale)
fprintf(f,"\t scale <"DBL_FMT","DBL_FMT","DBL_FMT">\n",scale->x,scale->y,scale->z);

/*:300*/
#line 5716 "MathView3D.w"

fprintf(f,"texture {\n"
"\t  pigment {\n"
"\t    image_map {\n"
"\t       png \"%s\"\n",TextureFile(SceneTextureMap->tex));
/*302:*/
#line 5756 "MathView3D.w"

switch(tmapType){
case TEXTURE_MAP_PLANAR:
fprintf(f,"\t       map_type 0\n");
break;
case TEXTURE_MAP_CYLINDRICAL:
fprintf(f,"\t       map_type 2\n");
break;
case TEXTURE_MAP_SPHERICAL:
fprintf(f,"\t       map_type 1\n");
break;
case TEXTURE_MAP_TORUS:
fprintf(f,"\t       map_type 5\n");
break;
default:
fprintf(f,"#warning \"MathGL3d::Unsupported mapping.\\n\"\n");
fprintf(f,"\t       map_type 0\n");
break;
}
fprintf(f,"\t       interpolate 2\n");

/*:302*/
#line 5721 "MathView3D.w"

fprintf(f,"\t      }\n");
/*304:*/
#line 5782 "MathView3D.w"

switch(tmapType){
case TEXTURE_MAP_PLANAR:
/*303:*/
#line 5777 "MathView3D.w"

fprintf(f,"\t     scale <"DBL_FMT", "DBL_FMT", 1>\n",
scS>0.0?1.0/scS:1.0,
scT>0.0?1.0/scT:1.0);

/*:303*/
#line 5785 "MathView3D.w"

break;
case TEXTURE_MAP_CYLINDRICAL:
case TEXTURE_MAP_SPHERICAL:
case TEXTURE_MAP_TORUS:
if(scS!=1.0||scT!=1.0)
fprintf(f,
"#warning \"MathGL3d::Texture scaling for spherical, "
"cylindrical and torus mapping not possible.\\n\"\n");
fprintf(f,"\t    scale < -1, 1, 1>\n");
fprintf(f,"\t    rotate 90*x\n");
fprintf(f,"\t    rotate 180*z\n");
break;
}

/*:304*/
#line 5723 "MathView3D.w"

fprintf(f,
"\t   }\n"
"\t  finish { DefaultFinish }"
"\t}\n");
/*301:*/
#line 5745 "MathView3D.w"

if(scale)
fprintf(f,"\t scale <"DBL_FMT","DBL_FMT","DBL_FMT">\n",1.0/scale->x,1.0/scale->y,1.0/scale->z);
if(axis&&angle!=0.0)
fprintf(f,"\t rotate "DBL_FMT"*<"DBL_FMT", "DBL_FMT", "DBL_FMT">\n",
-ToDegree*angle,
axis->x,axis->y,axis->z);
if(t)
fprintf(f,"\t translate <"DBL_FMT", "DBL_FMT", "DBL_FMT">\n",-t->x,-t->y,-t->z);


/*:301*/
#line 5728 "MathView3D.w"

if(TrackAngle!=0.0)
POVTrackballPosition(f);
fprintf(f,"}\n");
fprintf(f,"\n/* End polygones  */\n");
}

/*:299*/
#line 5523 "MathView3D.w"

/*309:*/
#line 5853 "MathView3D.w"

void WritePOVMeshes(FILE*f,Mesh3DList*lst,int smooth,
const char*mTexPrefix)
{int ord= 1;
Mesh3DList*head= lst;

fprintf(f,"/* Begin Meshes */\n\n");
while(lst){
/*312:*/
#line 5927 "MathView3D.w"

{
int n,m,i,j,k,haveColors= GL_FALSE;
RGBColor*c;

if(MeshTextureQ(lst)&&lst->mTexture->type>=TEXTURE_MAP_MESH_ST)
fprintf(f,"#warning \"MathGL3d::Mesh texture map not supported.\\n\"\n");
n= lst->dimU-1;
m= lst->dimV-1;
if(lst->mpolyColors&&!MeshTextureQ(lst)){
/*313:*/
#line 5950 "MathView3D.w"

for(k= i= 0;i<n;i++)
for(j= 0;j<m;j++){
c= MeshColorPtr(lst,i,j);
fprintf(f,"#declare Mesh%dColor%d=texture {\n"
"\t pigment { color rgb <"DBL_FMT", "DBL_FMT", "DBL_FMT">}\n"
"\t finish { DefaultFinish }\n"
"\t}\n",ord,k++,c->r,c->g,c->b);
}

/*:313*/
#line 5937 "MathView3D.w"

haveColors= GL_TRUE;
}
fprintf(f,"#declare  GeneralMesh%d= mesh {\n",ord);
if(smooth){
/*314:*/
#line 5970 "MathView3D.w"

for(k= i= 0;i<n;i++)
for(j= 0;j<m;j++){
if(TrueMeshTriangle1(lst,i,j))
WritePOVSmoothTriangle(f,
MeshCoordPtr(lst,i+1,j),
MeshCoordPtr(lst,i,j),
MeshCoordPtr(lst,i,j+1),
MeshNormalPtr(lst,i+1,j),
MeshNormalPtr(lst,i,j),
MeshNormalPtr(lst,i,j+1),POVMeshCount,POVMeshColor);
if(TrueMeshTriangle2(lst,i,j))
WritePOVSmoothTriangle(f,
MeshCoordPtr(lst,i+1,j),
MeshCoordPtr(lst,i,j+1),
MeshCoordPtr(lst,i+1,j+1),
MeshNormalPtr(lst,i+1,j),
MeshNormalPtr(lst,i,j+1),
MeshNormalPtr(lst,i+1,j+1),POVMeshCount,POVMeshColor);
k++;
}

/*:314*/
#line 5942 "MathView3D.w"

}
else{
/*315:*/
#line 5992 "MathView3D.w"

for(k= i= 0;i<n;i++)
for(j= 0;j<m;j++){
if(TrueMeshTriangle1(lst,i,j))
WritePOVFlatTriangle(f,
MeshCoordPtr(lst,i+1,j),
MeshCoordPtr(lst,i,j),
MeshCoordPtr(lst,i,j+1),POVMeshCount,POVMeshColor);
if(TrueMeshTriangle2(lst,i,j))
WritePOVFlatTriangle(f,
MeshCoordPtr(lst,i+1,j),
MeshCoordPtr(lst,i,j+1),
MeshCoordPtr(lst,i+1,j+1),POVMeshCount,POVMeshColor);
k++;
}

/*:315*/
#line 5945 "MathView3D.w"

}
fprintf(f,"}\n");
}

/*:312*/
#line 5861 "MathView3D.w"

ord++;
lst= lst->next;
}
ord= 1;
lst= head;
while(lst){
if((!MeshTextureQ(lst)||lst->mTexture->type>=TEXTURE_MAP_MESH_ST))
/*310:*/
#line 5878 "MathView3D.w"

{
fprintf(f,"\n"
"object {\n"
"\tGeneralMesh%d\n"
"\ttexture{\n"
"\t\tpigment {rgb <"DBL_FMT","DBL_FMT","DBL_FMT"> }\n"
"\t\tfinish { DefaultFinish }\n"
"\t}\n}\n",ord,lst->mColor.r,lst->mColor.g,lst->mColor.b);
}

/*:310*/
#line 5869 "MathView3D.w"

else
/*311:*/
#line 5889 "MathView3D.w"

{
Point3D*t,*axis,*scale;
double angle,scS,scT;
int tmapType;
Mesh3DList*mlst= lst;
static int texFileCount= 1;

if(!*MeshTextureFile(mlst))
/*385:*/
#line 7301 "MathView3D.w"

{
sprintf(MeshTextureFile(mlst),"%stex%d.png",mTexPrefix,texFileCount++);
PNGSaveBitmap(
MeshTextureFile(mlst),
TextureWidth(mlst->mTexture->tex),
TextureHeight(mlst->mTexture->tex),
TextureType(mlst->mTexture->tex),
TextureData(mlst->mTexture->tex),GL_FALSE,1.0);
}

/*:385*/
#line 5898 "MathView3D.w"

tmapType= lst->mTexture->type;
t= lst->mTexture->translate;
axis= lst->mTexture->axis;
angle= lst->mTexture->angle;
scale= lst->mTexture->scale;
scS= lst->mTexture->scaleS;
scT= lst->mTexture->scaleT;

fprintf(f,"\nobject {\n\tGeneralMesh%d\n",ord);
/*300:*/
#line 5735 "MathView3D.w"

if(t)
fprintf(f,"\t translate <"DBL_FMT", "DBL_FMT", "DBL_FMT">\n",t->x,t->y,t->z);
if(axis&&angle!=0.0)
fprintf(f,"\t rotate "DBL_FMT"*<"DBL_FMT", "DBL_FMT", "DBL_FMT">\n",
ToDegree*angle,
axis->x,axis->y,axis->z);
if(scale)
fprintf(f,"\t scale <"DBL_FMT","DBL_FMT","DBL_FMT">\n",scale->x,scale->y,scale->z);

/*:300*/
#line 5908 "MathView3D.w"

fprintf(f,"\ttexture{\n\t\tpigment {\n");
fprintf(f,"\t    image_map {\n"
"\t       png \"%s\"\n",TextureFile(lst->mTexture->tex));
/*302:*/
#line 5756 "MathView3D.w"

switch(tmapType){
case TEXTURE_MAP_PLANAR:
fprintf(f,"\t       map_type 0\n");
break;
case TEXTURE_MAP_CYLINDRICAL:
fprintf(f,"\t       map_type 2\n");
break;
case TEXTURE_MAP_SPHERICAL:
fprintf(f,"\t       map_type 1\n");
break;
case TEXTURE_MAP_TORUS:
fprintf(f,"\t       map_type 5\n");
break;
default:
fprintf(f,"#warning \"MathGL3d::Unsupported mapping.\\n\"\n");
fprintf(f,"\t       map_type 0\n");
break;
}
fprintf(f,"\t       interpolate 2\n");

/*:302*/
#line 5912 "MathView3D.w"

fprintf(f,"\t      }\n");
/*304:*/
#line 5782 "MathView3D.w"

switch(tmapType){
case TEXTURE_MAP_PLANAR:
/*303:*/
#line 5777 "MathView3D.w"

fprintf(f,"\t     scale <"DBL_FMT", "DBL_FMT", 1>\n",
scS>0.0?1.0/scS:1.0,
scT>0.0?1.0/scT:1.0);

/*:303*/
#line 5785 "MathView3D.w"

break;
case TEXTURE_MAP_CYLINDRICAL:
case TEXTURE_MAP_SPHERICAL:
case TEXTURE_MAP_TORUS:
if(scS!=1.0||scT!=1.0)
fprintf(f,
"#warning \"MathGL3d::Texture scaling for spherical, "
"cylindrical and torus mapping not possible.\\n\"\n");
fprintf(f,"\t    scale < -1, 1, 1>\n");
fprintf(f,"\t    rotate 90*x\n");
fprintf(f,"\t    rotate 180*z\n");
break;
}

/*:304*/
#line 5914 "MathView3D.w"

fprintf(f,"}\n\t\tfinish { DefaultFinish }\n\t}\n");
/*301:*/
#line 5745 "MathView3D.w"

if(scale)
fprintf(f,"\t scale <"DBL_FMT","DBL_FMT","DBL_FMT">\n",1.0/scale->x,1.0/scale->y,1.0/scale->z);
if(axis&&angle!=0.0)
fprintf(f,"\t rotate "DBL_FMT"*<"DBL_FMT", "DBL_FMT", "DBL_FMT">\n",
-ToDegree*angle,
axis->x,axis->y,axis->z);
if(t)
fprintf(f,"\t translate <"DBL_FMT", "DBL_FMT", "DBL_FMT">\n",-t->x,-t->y,-t->z);


/*:301*/
#line 5916 "MathView3D.w"

fprintf(f,"}\n");
}

/*:311*/
#line 5871 "MathView3D.w"

ord++;
lst= lst->next;
}
fprintf(f,"/* End Meshes */\n");
}

/*:309*/
#line 5524 "MathView3D.w"

/*288:*/
#line 5530 "MathView3D.w"

void ScaleToWorld(Point3D*pw,Point3D*p)
{
pw->x= CenterOfScene.x+MaxScaleOfScene*p->x;
pw->y= CenterOfScene.y+MaxScaleOfScene*p->y;
pw->z= CenterOfScene.z+MaxScaleOfScene*p->z;
}

/*:288*/
#line 5525 "MathView3D.w"

/*290:*/
#line 5544 "MathView3D.w"

void WriteLights(FILE*f,Point3D*lightp,RGBColor*lcolor,int n,int shaddow)
{Point3D ls;
int i;

for(i= 0;i<n;i++){
ScaleToWorld(&ls,lightp+i);
WritePOVVector(f,"light_source {\n\t",&ls,"\n");
WritePOVVector(f,"\tcolor rgb ",(Point3D*)(lcolor+i),"\n");
if(!shaddow)fprintf(f,"\tshadowless\n");
fprintf(f,"}\n\n");
}
}

/*:290*/
#line 5526 "MathView3D.w"

/*322:*/
#line 6154 "MathView3D.w"

void WritePOVProlog(FILE*f,const char*povname)
{Point3D vp,p;
#if 0
Point3D addLight[8]= {{2,2,2},{-2,2,2},{-2,-2,2},{2,-2,2},
{2,2,-2},{-2,2,-2},{-2,-2,-2},{2,-2,-2}};
RGBColor addLColor[8]= {{0.25,0.25,0.25},
{0.25,0.25,0.25},{0.25,0.25,0.25},{0.25,0.25,0.25},
{0.1,0.1,0.1},{0.1,0.1,0.1},{0.1,0.1,0.1},{0.1,0.1,0.1}};
#endif
fprintf(f,"//**********************************************\n"
"//  File was created by "BANNER"\n// Version "VERSION"\n"
"// (c) by Jens-Peer Kuska 1997,98,99\n"
"//**********************************************\n");
fprintf(f,"// POVray 3.x input file: %s     \n",povname);
if(*ThreeScriptName){
fprintf(f,"// 3Script source       : %s     \n",ThreeScriptName);
}
fprintf(f,"// Bounding box:\n"
"// x: "DBL_FMT"  "DBL_FMT
"\n// y: "DBL_FMT" "DBL_FMT
"\n// z: "DBL_FMT" "DBL_FMT" \n//\n",
BoundP1.x,BoundP2.x,BoundP1.y,BoundP2.y,BoundP1.z,BoundP2.z);
fprintf(f,"// Unique Points: %d\n",PoolSize);
fprintf(f,"// Line         : %ld\n",LengthLinePolyList(Lines));
fprintf(f,"// Polygons     : %ld\n",LengthLinePolyList(Polygons));
fprintf(f,"// Meshes       : %ld\n",LengthMesh3DList(Meshes));

fprintf(f,"// end of header *******************************\n\n");
fprintf(f,"#declare DefaultFinish=finish { phong 0.8 ambient 0.4}\n");
fprintf(f,"#default {"
"texture {"
"pigment { color <1, 1, 1>}"
"finish { DefaultFinish }"
"}}\n");
if(AmbientColor.r>0.0||AmbientColor.g>0.0||AmbientColor.r>0.0)
fprintf(f,"global_settings { ambient_light color rgb <"
DBL_FMT", "DBL_FMT", "DBL_FMT">}\n\n",
AmbientColor.r,AmbientColor.g,AmbientColor.b);
ScaleToWorld(&vp,&ViewPoint);
p.x= p.z= 0.0;p.y= Norm(&vp);
fprintf(f,"camera {\n");
WritePOVVector(f,"\tlocation  ",&vp,"\n");
WritePOVVector(f,"\tdirection ",&p,"\n");
p.x= p.y= 0.0;p.z= 0.5*Camera*MaxScaleOfScene;
WritePOVVector(f,"\tup        ",&p,"\n");
p.x= 0.5*Camera*MaxScaleOfScene;p.y= p.z= 0.0;

WritePOVVector(f,"\tright     ",&p,"\n");
WritePOVVector(f,"\tsky       ",&ZAxis,"\n");
WritePOVVector(f,"\tlook_at   ",&CenterOfScene,"\n }\n\n");
WriteLights(f,LightPosition,LightColor,LightNo,GL_FALSE);
#if 0
WriteLights(f,addLight,addLColor,8,GL_TRUE);
#endif
fprintf(f,"/* Graphics data begin */\n");
}


/*:322*/
#line 5527 "MathView3D.w"

/*323:*/
#line 6213 "MathView3D.w"

int WritePOVray(const char*povout,int smooth,
const char*pTexPrefix,const char*mTexPrefix)
{FILE*f;
RGBColor**ctab;
int cno;
#if MACINTOSH_MATHLINK
f= MacCreateFile(povout,'TEXT','POV3');
#else
f= fopen(povout,"w");
#endif
if(!f)return 1;
WritePOVProlog(f,povout);
if(Polygons){
ctab= PolyColorTable(&cno,Polygons);
WritePOVColorTab(f,ctab,cno);
free(ctab);
}
if(Points&&(0.0<PointSphereSize))
WritePOVPoints(f,Points);
if(Lines)
WritePOVLines(f,Lines);
if(Polygons){
if(!SceneTextureMap)
WritePOVTriangles(f,Polygons,smooth,cno);
else
WriteTexPOVTriangles(f,Polygons,smooth,pTexPrefix);
}
if(Meshes)
WritePOVMeshes(f,Meshes,smooth,mTexPrefix);
fclose(f);
return 0;
}

/*:323*/
#line 5528 "MathView3D.w"


/*:287*/
#line 80 "MathView3D.w"

/*324:*/
#line 6248 "MathView3D.w"

/*325:*/
#line 6262 "MathView3D.w"

void QDNormal(FILE*f,Point3D*p)
{
fprintf(f,"Container (\n"
"\t AttributeSet()\n"
"\t Normal ( "DBL_FMT" "DBL_FMT" "DBL_FMT" )\n)\n",p->x,p->y,p->z);
}

/*:325*/
#line 6249 "MathView3D.w"

/*326:*/
#line 6270 "MathView3D.w"

void QDNegNormal(FILE*f,Point3D*p)
{
fprintf(f,"Container (\n"
"\t AttributeSet()\n"
"\t Normal ( "DBL_FMT" "DBL_FMT" "DBL_FMT" )\n)\n",-(p->x),-(p->y),-(p->z));
}

/*:326*/
#line 6250 "MathView3D.w"

/*327:*/
#line 6278 "MathView3D.w"

void QDNormalColor(FILE*f,Point3D*p,RGBColor*c)
{
fprintf(f,
"Container (\n"
"\t AttributeSet()\n"
"\t Normal ( "DBL_FMT" "DBL_FMT" "DBL_FMT" )\n"
"\t DiffuseColor( "DBL_FMT" "DBL_FMT" "DBL_FMT")\n"
")\n",p->x,p->y,p->z,c->r,c->g,c->b);
}

/*:327*/
#line 6251 "MathView3D.w"

/*335:*/
#line 6330 "MathView3D.w"

void MakeQDCone(FILE*f,Point3D*p1,Point3D*p2,RGBColor*c,double r,int n)
{Point3D v1,v2,v3,vn,p[4],testn;
double norm;
double dphi;
int i,interchange;
Point3D*normals;

v1.x= p2->x-p1->x;
v1.y= p2->y-p1->y;
v1.z= p2->z-p1->z;
norm= Norm(&v1);
v1.x/= norm;v1.y/= norm;v1.z/= norm;
OrthogonalSet(&v1,&v2,&v3);

v1.x= p2->x-p1->x;
v1.y= p2->y-p1->y;
v1.z= p2->z-p1->z;
dphi= TWO_PI/n;
/*173:*/
#line 3248 "MathView3D.w"


ConePoint(r,0.0,&v2,&v3,p);
IncPoint3D(p[0],(*p1));
p[1]= p[0];
IncPoint3D(p[1],v1);

ConePoint(r,dphi,&v2,&v3,p+2);
IncPoint3D(p[2],(*p1));

SurfaceNormal(&testn,p,p+1,p+2);
interchange= Dot(&testn,&v2)<0.0;
if(interchange){
SwapPoint3D(&v2,&v3);
}
if(MVFaceFlip){
SwapPoint3D(&v2,&v3);
}

/*:173*/
#line 6349 "MathView3D.w"

ConePoint(r,0.0,&v2,&v3,p);
normals= (Point3D*)calloc(n+1,sizeof(Point3D));
vn= *p;
normals[0].x= -p->x;normals[0].y= -p->y;normals[0].z= -p->z;
IncPoint3D(p[0],(*p1))
AddPoint3D(p[1],p[0],v1)
/*329:*/
#line 6295 "MathView3D.w"

fprintf(f,"Container (\n  TriGrid (\n    %d  2\n",n+1);

/*:329*/
#line 6356 "MathView3D.w"

/*330:*/
#line 6298 "MathView3D.w"

fprintf(f,"  "DBL_FMT" "DBL_FMT" "DBL_FMT"  "DBL_FMT" "DBL_FMT" "DBL_FMT"\n",
p->x,p->y,p->z,
p[1].x,p[1].y,p[1].z);

/*:330*/
#line 6357 "MathView3D.w"

for(i= 1;i<=n;i++){
ConePoint(r,i*dphi,&v2,&v3,p+2);
vn= p[2];
IncPoint3D(p[2],(*p1))
AddPoint3D(p[3],p[2],v1)
normals[i].x= -vn.x;
normals[i].y= -vn.y;
normals[i].z= -vn.z;
/*331:*/
#line 6303 "MathView3D.w"

fprintf(f,"  "DBL_FMT" "DBL_FMT" "DBL_FMT"  "DBL_FMT" "DBL_FMT" "DBL_FMT"\n",
p[2].x,p[2].y,p[2].z,
p[3].x,p[3].y,p[3].z);

/*:331*/
#line 6366 "MathView3D.w"

}
/*332:*/
#line 6308 "MathView3D.w"

fprintf(f,")\n");

/*:332*/
#line 6368 "MathView3D.w"

/*333:*/
#line 6311 "MathView3D.w"

fprintf(f,"Container (\n  VertexAttributeSetList ( %d Exclude 0)\n",2*(n+1));
for(i= 0;i<=n;i++){
QDNormalColor(f,normals+i,c);
QDNormalColor(f,normals+i,c);
}
fprintf(f,")\n");
fprintf(f,")\n");

/*:333*/
#line 6369 "MathView3D.w"

free(normals);
}


/*:335*/
#line 6252 "MathView3D.w"

/*336:*/
#line 6374 "MathView3D.w"

void MakeQDConeContinued(FILE*f,Point3D*p1,Point3D*p2,Point3D*p3,
RGBColor*c,
double r,int n)
{Point3D v1,v2,v3,u1,u2,u3,p[4];
Point3D testn;
double norm;
double dphi;
int i,interchange;
Point3D*normals1,*normals2;

/*175:*/
#line 3310 "MathView3D.w"

v1.x= p2->x-p1->x;v1.y= p2->y-p1->y;v1.z= p2->z-p1->z;
norm= Norm(&v1);
v1.x/= norm;v1.y/= norm;v1.z/= norm;
u1.x= p3->x-p2->x;u1.y= p3->y-p2->y;u1.z= p3->z-p2->z;
norm= Norm(&u1);
u1.x/= norm;u1.y/= norm;u1.z/= norm;
Cross(&v2,&v1,&u1);
if(fabs(v2.x)+fabs(v2.y)+fabs(v2.z)<TINY){
OrthogonalSet(&v1,&v2,&v3);
}
v3.x= v3.y= v3.z= 1.0;
GramSchmidt(&v1,&v2,&v3);

/*:175*/
#line 6385 "MathView3D.w"

/*337:*/
#line 6421 "MathView3D.w"

if(Dot(&u1,&v1)<CosSkipCont){
MakeQDCone(f,p1,p2,c,r,n);
return;
}


/*:337*/
#line 6386 "MathView3D.w"

normals1= (Point3D*)calloc(n+1,sizeof(Point3D));
normals2= (Point3D*)calloc(n+1,sizeof(Point3D));

u2= v2;u3= v3;
GramSchmidt(&u1,&u2,&u3);
dphi= TWO_PI/n;
/*177:*/
#line 3337 "MathView3D.w"


ConePoint(r,0.0,&v2,&v3,p);
IncPoint3D(p[0],(*p1));
ConePoint(r,0.0,&u2,&u3,p+1);
IncPoint3D(p[1],(*p2));
ConePoint(r,dphi,&v2,&v3,p+2);
IncPoint3D(p[2],(*p1));

SurfaceNormal(&testn,p,p+1,p+2);
interchange= Dot(&testn,&v2)<0.0;
if(interchange){
SwapPoint3D(&v2,&v3);
SwapPoint3D(&u2,&u3);
}
if(MVFaceFlip){
SwapPoint3D(&v2,&v3);
SwapPoint3D(&u2,&u3);
}

/*:177*/
#line 6393 "MathView3D.w"

ConePoint(r,0.0,&v2,&v3,p);
normals1[0].x= -p->x;normals1[0].y= -p->y;normals1[0].z= -p->z;
IncPoint3D(p[0],(*p1))
ConePoint(r,0.0,&u2,&u3,p+1);
normals2[0].x= -p[1].x;normals2[0].y= -p[1].y;normals2[0].z= -p[1].z;
IncPoint3D(p[1],(*p2))
/*329:*/
#line 6295 "MathView3D.w"

fprintf(f,"Container (\n  TriGrid (\n    %d  2\n",n+1);

/*:329*/
#line 6400 "MathView3D.w"

/*330:*/
#line 6298 "MathView3D.w"

fprintf(f,"  "DBL_FMT" "DBL_FMT" "DBL_FMT"  "DBL_FMT" "DBL_FMT" "DBL_FMT"\n",
p->x,p->y,p->z,
p[1].x,p[1].y,p[1].z);

/*:330*/
#line 6401 "MathView3D.w"

for(i= 1;i<=n;i++){
ConePoint(r,i*dphi,&v2,&v3,p+2);
normals1[i].x= -p[2].x;
normals1[i].y= -p[2].y;
normals1[i].z= -p[2].z;
IncPoint3D(p[2],(*p1))
ConePoint(r,i*dphi,&u2,&u3,p+3);
normals2[i].x= -p[3].x;
normals2[i].y= -p[3].y;
normals2[i].z= -p[3].z;
IncPoint3D(p[3],(*p2));
/*331:*/
#line 6303 "MathView3D.w"

fprintf(f,"  "DBL_FMT" "DBL_FMT" "DBL_FMT"  "DBL_FMT" "DBL_FMT" "DBL_FMT"\n",
p[2].x,p[2].y,p[2].z,
p[3].x,p[3].y,p[3].z);

/*:331*/
#line 6413 "MathView3D.w"

}
/*332:*/
#line 6308 "MathView3D.w"

fprintf(f,")\n");

/*:332*/
#line 6415 "MathView3D.w"

/*334:*/
#line 6320 "MathView3D.w"

fprintf(f,"Container (\n  VertexAttributeSetList ( %d Exclude 0)\n",2*(n+1));
for(i= 0;i<=n;i++){
QDNormalColor(f,normals1+i,c);
QDNormalColor(f,normals2+i,c);
}
fprintf(f,")\n");
fprintf(f,")\n");


/*:334*/
#line 6416 "MathView3D.w"

free(normals1);
free(normals2);
}

/*:336*/
#line 6253 "MathView3D.w"

/*348:*/
#line 6625 "MathView3D.w"

void WriteQDLights(FILE*f,Point3D*lightp,RGBColor*lcolor,int n)
{Point3D ls;
double nls;
int i;

fprintf(f,"Container (\n"
"  ViewHints()\n"
"  Container (\n"
"    InteractiveRenderer()\n"
"   )\n");
for(i= 0;i<n;i++){
ScaleToWorld(&ls,lightp+i);
ls.x-= CenterOfScene.x;
ls.y-= CenterOfScene.y;
ls.z-= CenterOfScene.z;
nls= Norm(&ls);
if(nls>0.0){
ls.x/= nls;
ls.y/= nls;
ls.z/= nls;
}
fprintf(f,"Container (\n"
"  DirectionalLight (\n"
"    "DBL_FMT" "DBL_FMT" "DBL_FMT"\n"
"    False\n"
"   )\n",
ls.x,ls.y,ls.z);
fprintf(f,"  Light (\n"
"    True\n    1.0\n"
"    "DBL_FMT" "DBL_FMT" "DBL_FMT"\n   )\n )\n",
lcolor[i].r,lcolor[i].g,lcolor[i].b);
}
fprintf(f,"Container (\n"
"  AmbientLight()\n"
"  Light(\n"
"    True\n"
"    1.0\n"
"    "DBL_FMT" "DBL_FMT" "DBL_FMT"\n"
"   )\n"
" )\n",AmbientColor.r,AmbientColor.g,AmbientColor.b);
fprintf(f,"ImageClearColor ( 1\n"
"  %s\n )\n",BlackBackground?"0 0 0":"0.8 0.8 0.8");
fprintf(f,")\n");
}


/*:348*/
#line 6254 "MathView3D.w"

/*347:*/
#line 6591 "MathView3D.w"

void WriteQDCamera(FILE*f,Point3D*viewpoint)
{
Point3D vp;
Point3D u1,u2,u3;
double nrm;

ScaleToWorld(&vp,viewpoint);
vp.x= CenterOfScene.x-vp.x;
vp.y= CenterOfScene.y-vp.y;
vp.z= CenterOfScene.z-vp.z;
u3= vp;
nrm= Norm(&u3);
u3.x/= nrm;u3.y/= nrm;u3.z/= nrm;
u1.x= 1.0;u1.y= u1.z= 0.0;
u2.x= 0.0;u2.y= 1.0;u2.z= 0.0;
GramSchmidt(&u3,&u1,&u2);
fprintf(f,"Container(\n  ViewHints()\n  Container (\n");
fprintf(f,"  ViewAngleAspectCamera (\n    "DBL_FMT" 1\n   )\n",MVViewAngle);
fprintf(f,"  CameraPlacement (\n"
"    "DBL_FMT" "DBL_FMT" "DBL_FMT"\n"
"    "DBL_FMT" "DBL_FMT" "DBL_FMT"\n"
"    0 0 1)\n",-vp.x,-vp.y,-vp.z,
CenterOfScene.x,
CenterOfScene.y,
CenterOfScene.z);
fprintf(f,"  CameraViewPort(\n"
"    -0.5 0.5\n"
"    1.0\n"
"    1.0\n"
"   )\n");
fprintf(f,")\n)\n");
}

/*:347*/
#line 6255 "MathView3D.w"

/*345:*/
#line 6566 "MathView3D.w"

PointList*WriteQDPoint(FILE*f,PointList*lst)
{Point3D*p;
double sz;
p= PointPool+lst->point;
if(0==PointSphereSize)
fprintf(f,"Point ("DBL_FMT" "DBL_FMT" "DBL_FMT")\n",p->x,p->y,p->z);
else{
sz= lst->size*PointSphereSize;
fprintf(f,"Ellipsoid (\n"
""DBL_FMT"  0   0\n"
" 0  "DBL_FMT"  0\n"
" 0   0  "DBL_FMT"\n"
" "DBL_FMT" "DBL_FMT" "DBL_FMT"\n0\n1\n0\n1\n)\n",
sz,sz,sz,
p->x,p->y,p->z);
}
return lst->next;
}

/*:345*/
#line 6256 "MathView3D.w"

/*339:*/
#line 6435 "MathView3D.w"

LinePolyList*WriteQDLine(FILE*f,LinePolyList*lst)
{Point3D*p;
int i;

if(fabs(LineTubeSize)<MachineEpsilon){
fprintf(f,"Container (\n");
if(lst->pno>2){
fprintf(f,"PolyLine(\n  %d\n",lst->pno);
/*338:*/
#line 6428 "MathView3D.w"

for(i= 0;i<lst->pno;i++){
p= PointPool+lst->points[i];
fprintf(f,"\t "DBL_FMT" "DBL_FMT" "DBL_FMT"\n",p->x,p->y,p->z);
}


/*:338*/
#line 6444 "MathView3D.w"

fprintf(f,")\n");
}
else{
fprintf(f,"Line (\n");
/*338:*/
#line 6428 "MathView3D.w"

for(i= 0;i<lst->pno;i++){
p= PointPool+lst->points[i];
fprintf(f,"\t "DBL_FMT" "DBL_FMT" "DBL_FMT"\n",p->x,p->y,p->z);
}


/*:338*/
#line 6449 "MathView3D.w"

fprintf(f,")\n");
}
/*328:*/
#line 6289 "MathView3D.w"

fprintf(f,"Container (\n"
"\tAttributeSet()\n"
"\tDiffuseColor( "DBL_FMT" "DBL_FMT" "DBL_FMT")\n)\n)",
lst->color.r,lst->color.g,lst->color.b);

/*:328*/
#line 6452 "MathView3D.w"

}
else{
/*340:*/
#line 6460 "MathView3D.w"

{
if(2==lst->pno)
MakeQDCone(f,PointPool+lst->points[0],
PointPool+lst->points[1],
&lst->color,LineTubeSize,TubeSegments);
else{
for(i= 0;i<lst->pno-2;i++)
MakeQDConeContinued(f,
PointPool+lst->points[i],
PointPool+lst->points[i+1],
PointPool+lst->points[i+2],
&lst->color,
LineTubeSize,TubeSegments);
i= lst->pno-2;
if(lst->points[0]==lst->points[i+1]){
MakeQDConeContinued(f,
PointPool+lst->points[i],
PointPool+lst->points[0],
PointPool+lst->points[1],
&lst->color,
LineTubeSize,TubeSegments);
}
else
MakeQDCone(f,PointPool+lst->points[i],
PointPool+lst->points[i+1],
&lst->color,
LineTubeSize,TubeSegments);

}
}

/*:340*/
#line 6455 "MathView3D.w"

}
return lst->next;
}

/*:339*/
#line 6257 "MathView3D.w"

/*341:*/
#line 6492 "MathView3D.w"

LinePolyList*WriteQDPolygon(FILE*f,LinePolyList*lst)
{int i;
Point3D*p;

fprintf(f,"Container (\n"
"  Polygon (\n"
"\t %d\n",lst->pno);
/*338:*/
#line 6428 "MathView3D.w"

for(i= 0;i<lst->pno;i++){
p= PointPool+lst->points[i];
fprintf(f,"\t "DBL_FMT" "DBL_FMT" "DBL_FMT"\n",p->x,p->y,p->z);
}


/*:338*/
#line 6500 "MathView3D.w"

fprintf(f,"  )\n");
if(SmoothShading){
fprintf(f,"Container (\n  VertexAttributeSetList ( %d Exclude 0)\n",lst->pno);
for(i= 0;i<lst->pno;i++){
p= NormalPool[lst->points[i]];
QDNegNormal(f,p);
}
fprintf(f,")\n");
}
/*328:*/
#line 6289 "MathView3D.w"

fprintf(f,"Container (\n"
"\tAttributeSet()\n"
"\tDiffuseColor( "DBL_FMT" "DBL_FMT" "DBL_FMT")\n)\n)",
lst->color.r,lst->color.g,lst->color.b);

/*:328*/
#line 6510 "MathView3D.w"

return lst->next;
}

/*:341*/
#line 6258 "MathView3D.w"

/*342:*/
#line 6514 "MathView3D.w"

Mesh3DList*WriteQDMesh(FILE*f,Mesh3DList*lst)
{int i,j,mu;
Point3D*p;
RGBColor*c;

mu= lst->dimU*lst->dimV;
fprintf(f,"Container (\n");
fprintf(f,"TriGrid (\n %d\n %d\n",lst->dimU,lst->dimV);
for(i= 0;i<lst->dimU;i++)
for(j= 0;j<lst->dimV;j++){
p= MeshCoordPtr(lst,i,j);
fprintf(f,"\t"DBL_FMT" "DBL_FMT" "DBL_FMT"\n",p->x,p->y,p->z);
}
fprintf(f,")\n");
/*343:*/
#line 6535 "MathView3D.w"

if(SmoothShading){
fprintf(f,"Container ( VertexAttributeSetList ( %d Exclude 0)\n",mu);
for(i= 0;i<lst->dimU;i++)
for(j= 0;j<lst->dimV;j++){
p= MeshNormalPtr(lst,i,j);
QDNegNormal(f,p);
}
fprintf(f,")\n");
}

/*:343*/
#line 6529 "MathView3D.w"

/*344:*/
#line 6546 "MathView3D.w"

if(MESH_FACE_COLOR==(lst->attrib&MESH_FACE_COLOR)){
mu= (lst->dimU-1)*(lst->dimV-1);
fprintf(f,"Container ( FaceAttributeSetList ( %d Exclude 0)\n",2*mu);
for(i= 0;i<lst->dimU-1;i++)
for(j= 0;j<lst->dimV-1;j++){
c= MeshColorPtr(lst,i,j);
fprintf(f,"Container (\n"
"AttributeSet()\n"
"DiffuseColor( "DBL_FMT" "DBL_FMT" "DBL_FMT" )\n"
")\n",c->r,c->g,c->b);
fprintf(f,"Container (\n"
"AttributeSet()\n"
"DiffuseColor( "DBL_FMT" "DBL_FMT" "DBL_FMT" )\n"
")\n",c->r,c->g,c->b);
}
fprintf(f,")\n");
}


/*:344*/
#line 6530 "MathView3D.w"

fprintf(f,")\n");
return lst->next;
}

/*:342*/
#line 6259 "MathView3D.w"

/*349:*/
#line 6672 "MathView3D.w"

int Write3DMF(const char*fname,int WriteLight,int WriteView)
{FILE*qdfile;
PointList*pplst= Points;
LinePolyList*llst= Lines;
LinePolyList*plst= Polygons;
Mesh3DList*mlst= Meshes;
#if MACINTOSH_MATHLINK
qdfile= MacCreateFile(fname,'3DMF','ttxt');
#else
qdfile= fopen(fname,"w");
#endif
if(!qdfile)return 1;
/*346:*/
#line 6586 "MathView3D.w"

fprintf(qdfile,
"3DMetafile ( 1 5 Stream Label0>)\n"
"CString (\"Created by MathGL3d (c) Jens-Peer Kuska 1997,98,99\")\n");

/*:346*/
#line 6685 "MathView3D.w"

if(WriteLight)WriteQDLights(qdfile,LightPosition,LightColor,LightNo);
fprintf(qdfile,"BeginGroup( DisplayGroup() )\n");
fprintf(qdfile,"# Points in scene\n");
while(pplst)pplst= WriteQDPoint(qdfile,pplst);
fprintf(qdfile,"# Polygons in scene\n");
while(plst)plst= WriteQDPolygon(qdfile,plst);
fprintf(qdfile,"# Lines in scene\n");
while(llst)llst= WriteQDLine(qdfile,llst);
fprintf(qdfile,"# Meshes in scene\n");
while(mlst)mlst= WriteQDMesh(qdfile,mlst);
fprintf(qdfile,"EndGroup()\n");
if(WriteView)WriteQDCamera(qdfile,&ViewPoint);
fclose(qdfile);
return 0;
}

/*:349*/
#line 6260 "MathView3D.w"


/*:324*/
#line 81 "MathView3D.w"

/*350:*/
#line 6703 "MathView3D.w"

/*392:*/
#line 7450 "MathView3D.w"

void VROrientation(Point3D*orient,double*angle,
Point3D*viewpoint,Point3D*lookat,Point3D*up)
{Point3D vp,vert,zero;
double d;
Quaternion qRotMajor,qRotMinor;
Quaternion qtmp;

/*393:*/
#line 7482 "MathView3D.w"

vp.x= -(viewpoint->x-lookat->x);
vp.y= -(viewpoint->y-lookat->y);
vp.z= -(viewpoint->z-lookat->z);
d= Norm(&vp);
if(d!=0.0){
vp.x/= d;
vp.y/= d;
vp.z/= d;
}

/*:393*/
#line 7458 "MathView3D.w"

/*394:*/
#line 7493 "MathView3D.w"

d= Dot(&vp,up);
vert.x= up->x-d*vp.x;
vert.y= up->y-d*vp.y;
vert.z= up->z-d*vp.z;
d= Norm(&vert);
if(d!=0.0){
vert.x/= d;
vert.y/= d;
vert.z/= d;
}

/*:394*/
#line 7459 "MathView3D.w"

Cross(orient,&vp,&ZAxis);
d= Norm(orient);
if(d<TINY){
if(orient->z>0.0)
MakeQuaternion(&qRotMajor,0.5*PI,&YAxis);
else{
zero.x= zero.y= zero.z= 0.0;
MakeQuaternion(&qRotMajor,0.5*PI,&zero);
}
}
else{
*angle= asin(Dot(orient,orient));
orient->x/= d;
orient->y/= d;
orient->z/= d;
MakeQuaternion(&qRotMajor,acos(-vp.z),orient);
}
/*395:*/
#line 7510 "MathView3D.w"

{
Point3D newY;
double twist;

QuaternionRotate(&newY,&qRotMajor,&YAxis);
twist= acos(Dot(&newY,&vert));
MakeQuaternion(&qRotMinor,twist,&ZAxis);
}

/*:395*/
#line 7477 "MathView3D.w"

QuaternionTimes(&qtmp,&qRotMajor,&qRotMinor);
GetQuaternionRotation(angle,orient,&qtmp);
}

/*:392*/
#line 6704 "MathView3D.w"

/*354:*/
#line 6759 "MathView3D.w"

void WriteVRMaterial(FILE*f,RGBColor*c)
{
fprintf(f,
"  appearance Appearance {\n"
"    material Material{\n"
"      diffuseColor "DBL_FMT" "DBL_FMT" "DBL_FMT"\n"
"      ambientIntensity 0.0\n"
"      specularColor 0.0 0.0 0.0\n"
"      emissiveColor 0.0 0.0 0.0\n"
"      shininess 0.2\n"
"      transparency 0.0\n"
"     }\n"
"   }\n",c->r,c->g,c->b);
}

/*:354*/
#line 6705 "MathView3D.w"

/*351:*/
#line 6715 "MathView3D.w"

void WriteVRPoints(FILE*f,PointList*plst)
{PointList*lst;
Point3D*p;

if(0.0==PointSphereSize)
/*352:*/
#line 6727 "MathView3D.w"

{
lst= plst;
fprintf(f,"DEF MGL_POINTS Shape {\n");
fprintf(f,
" geometry PointSet {\n"
"   coord Coordinate {\n"
"     point [\n");
while(lst){
p= PointPool+lst->point;
fprintf(f,"     "DBL_FMT" "DBL_FMT" "DBL_FMT"%s\n",p->x,p->y,p->z,(lst->next?",":""));
lst= lst->next;
}
lst= plst;
fprintf(f,
"     ]\n"
"   }\n"
"   color Color {\n"
"     color [\n");
while(lst){
fprintf(f,"     "DBL_FMT" "DBL_FMT" "DBL_FMT"\n",lst->color.r,lst->color.g,lst->color.b);
lst= lst->next;
}
fprintf(f,
"      ]\n"
"   }\n"
" }\n");
}

/*:352*/
#line 6721 "MathView3D.w"

else
/*355:*/
#line 6775 "MathView3D.w"

{double sz;

lst= plst;
fprintf(f,
" Group {\n"
"   children [\n");
while(lst){
p= PointPool+lst->point;
sz= lst->size*PointSphereSize;
fprintf(f,
"  Transform {\n");
fprintf(f,
"    children Shape {\n");
/*353:*/
#line 6756 "MathView3D.w"

WriteVRMaterial(f,&lst->color);

/*:353*/
#line 6789 "MathView3D.w"

fprintf(f,
"  geometry Sphere {\n"
"     radius "DBL_FMT"\n"
"    }\n"
"}\n",sz);
fprintf(f,"    translation "DBL_FMT" "DBL_FMT" "DBL_FMT"\n",p->x,p->y,p->z);
fprintf(f,
"   }\n");
lst= lst->next;
}
fprintf(f,"  ]\n");
}

/*:355*/
#line 6723 "MathView3D.w"

fprintf(f," }\n");
}

/*:351*/
#line 6706 "MathView3D.w"

/*356:*/
#line 6803 "MathView3D.w"

void MakeVRCone(FILE*f,Point3D*p1,Point3D*p2,RGBColor*c,double r,int n)
{Point3D v1,v2,v3,p[4],testn;
double norm;
double dphi;
int i,interchange;

v1.x= p2->x-p1->x;
v1.y= p2->y-p1->y;
v1.z= p2->z-p1->z;
norm= Norm(&v1);
v1.x/= norm;v1.y/= norm;v1.z/= norm;
OrthogonalSet(&v1,&v2,&v3);

v1.x= p2->x-p1->x;
v1.y= p2->y-p1->y;
v1.z= p2->z-p1->z;
dphi= TWO_PI/n;
/*173:*/
#line 3248 "MathView3D.w"


ConePoint(r,0.0,&v2,&v3,p);
IncPoint3D(p[0],(*p1));
p[1]= p[0];
IncPoint3D(p[1],v1);

ConePoint(r,dphi,&v2,&v3,p+2);
IncPoint3D(p[2],(*p1));

SurfaceNormal(&testn,p,p+1,p+2);
interchange= Dot(&testn,&v2)<0.0;
if(interchange){
SwapPoint3D(&v2,&v3);
}
if(MVFaceFlip){
SwapPoint3D(&v2,&v3);
}

/*:173*/
#line 6821 "MathView3D.w"

ConePoint(r,0.0,&v2,&v3,p);
IncPoint3D(p[0],(*p1))
AddPoint3D(p[1],p[0],v1)
/*359:*/
#line 6878 "MathView3D.w"

fprintf(f,"Shape {\n");
WriteVRMaterial(f,c);
fprintf(f,
"  geometry IndexedFaceSet {\n"
"    coord Coordinate {\n"
"      point [\n");

/*:359*/
#line 6825 "MathView3D.w"

/*361:*/
#line 6892 "MathView3D.w"

fprintf(f,
"      "DBL_FMT" "DBL_FMT" "DBL_FMT", "DBL_FMT" "DBL_FMT" "DBL_FMT",\n",
p->x,p->y,p->z,p[1].x,p[1].y,p[1].z);

/*:361*/
#line 6826 "MathView3D.w"

for(i= 1;i<=n;i++){
ConePoint(r,i*dphi,&v2,&v3,p+2);
IncPoint3D(p[2],(*p1))
AddPoint3D(p[3],p[2],v1)
/*362:*/
#line 6897 "MathView3D.w"

fprintf(f,
"      "DBL_FMT" "DBL_FMT" "DBL_FMT", "DBL_FMT" "DBL_FMT" "DBL_FMT",\n",
p[2].x,p[2].y,p[2].z,p[3].x,p[3].y,p[3].z);


/*:362*/
#line 6831 "MathView3D.w"

}
/*363:*/
#line 6903 "MathView3D.w"

fprintf(f,
"       ]\n"
"     }\n"
"    coordIndex [\n");
for(i= 0;i<2*n;i+= 2)
fprintf(f,
"     %d %d %d %d -1,\n",i,i+1,i+3,i+2);
fprintf(f,"     ]\n");

/*:363*/
#line 6833 "MathView3D.w"

/*360:*/
#line 6886 "MathView3D.w"

fprintf(f,
"    creaseAngle "DBL_FMT"\n"
"   }\n"
" }\n",2.0*dphi);

/*:360*/
#line 6834 "MathView3D.w"

}


/*:356*/
#line 6707 "MathView3D.w"

/*357:*/
#line 6838 "MathView3D.w"

void MakeVRConeContinued(FILE*f,Point3D*p1,Point3D*p2,Point3D*p3,
RGBColor*c,
double r,int n)
{Point3D v1,v2,v3,u1,u2,u3,p[4];
Point3D testn;
double norm;
double dphi;
int i,interchange;

/*175:*/
#line 3310 "MathView3D.w"

v1.x= p2->x-p1->x;v1.y= p2->y-p1->y;v1.z= p2->z-p1->z;
norm= Norm(&v1);
v1.x/= norm;v1.y/= norm;v1.z/= norm;
u1.x= p3->x-p2->x;u1.y= p3->y-p2->y;u1.z= p3->z-p2->z;
norm= Norm(&u1);
u1.x/= norm;u1.y/= norm;u1.z/= norm;
Cross(&v2,&v1,&u1);
if(fabs(v2.x)+fabs(v2.y)+fabs(v2.z)<TINY){
OrthogonalSet(&v1,&v2,&v3);
}
v3.x= v3.y= v3.z= 1.0;
GramSchmidt(&v1,&v2,&v3);

/*:175*/
#line 6848 "MathView3D.w"

/*358:*/
#line 6872 "MathView3D.w"

if(Dot(&u1,&v1)<CosSkipCont){
MakeVRCone(f,p1,p2,c,r,n);
return;
}

/*:358*/
#line 6849 "MathView3D.w"


u2= v2;u3= v3;
GramSchmidt(&u1,&u2,&u3);
dphi= TWO_PI/n;
/*177:*/
#line 3337 "MathView3D.w"


ConePoint(r,0.0,&v2,&v3,p);
IncPoint3D(p[0],(*p1));
ConePoint(r,0.0,&u2,&u3,p+1);
IncPoint3D(p[1],(*p2));
ConePoint(r,dphi,&v2,&v3,p+2);
IncPoint3D(p[2],(*p1));

SurfaceNormal(&testn,p,p+1,p+2);
interchange= Dot(&testn,&v2)<0.0;
if(interchange){
SwapPoint3D(&v2,&v3);
SwapPoint3D(&u2,&u3);
}
if(MVFaceFlip){
SwapPoint3D(&v2,&v3);
SwapPoint3D(&u2,&u3);
}

/*:177*/
#line 6854 "MathView3D.w"

ConePoint(r,0.0,&v2,&v3,p);
IncPoint3D(p[0],(*p1))
ConePoint(r,0.0,&u2,&u3,p+1);
IncPoint3D(p[1],(*p2))
/*359:*/
#line 6878 "MathView3D.w"

fprintf(f,"Shape {\n");
WriteVRMaterial(f,c);
fprintf(f,
"  geometry IndexedFaceSet {\n"
"    coord Coordinate {\n"
"      point [\n");

/*:359*/
#line 6859 "MathView3D.w"

/*361:*/
#line 6892 "MathView3D.w"

fprintf(f,
"      "DBL_FMT" "DBL_FMT" "DBL_FMT", "DBL_FMT" "DBL_FMT" "DBL_FMT",\n",
p->x,p->y,p->z,p[1].x,p[1].y,p[1].z);

/*:361*/
#line 6860 "MathView3D.w"

for(i= 1;i<=n;i++){
ConePoint(r,i*dphi,&v2,&v3,p+2);
IncPoint3D(p[2],(*p1))
ConePoint(r,i*dphi,&u2,&u3,p+3);
IncPoint3D(p[3],(*p2));
/*362:*/
#line 6897 "MathView3D.w"

fprintf(f,
"      "DBL_FMT" "DBL_FMT" "DBL_FMT", "DBL_FMT" "DBL_FMT" "DBL_FMT",\n",
p[2].x,p[2].y,p[2].z,p[3].x,p[3].y,p[3].z);


/*:362*/
#line 6866 "MathView3D.w"

}
/*363:*/
#line 6903 "MathView3D.w"

fprintf(f,
"       ]\n"
"     }\n"
"    coordIndex [\n");
for(i= 0;i<2*n;i+= 2)
fprintf(f,
"     %d %d %d %d -1,\n",i,i+1,i+3,i+2);
fprintf(f,"     ]\n");

/*:363*/
#line 6868 "MathView3D.w"

/*360:*/
#line 6886 "MathView3D.w"

fprintf(f,
"    creaseAngle "DBL_FMT"\n"
"   }\n"
" }\n",2.0*dphi);

/*:360*/
#line 6869 "MathView3D.w"

}

/*:357*/
#line 6708 "MathView3D.w"

/*364:*/
#line 6913 "MathView3D.w"

LinePolyList*WriteVRLines(FILE*f,int*count,LinePolyList*lst)
{int i;
Point3D*p;
fprintf(f,"DEF MGL_LINE%d Shape {\n",*count);
if(fabs(LineTubeSize)>MachineEpsilon)
/*366:*/
#line 6933 "MathView3D.w"

{int n= lst->pno-1;
fprintf(f,"  geometry IndexedLineSet {\n");
fprintf(f,"    coord Coordinate {\n"
"      point [\n");
/*365:*/
#line 6927 "MathView3D.w"

for(i= 0;i<lst->pno;i++){
p= PointPool+lst->points[i];
fprintf(f,"      "DBL_FMT" "DBL_FMT" "DBL_FMT"%s",p->x,p->y,p->z,n==i?"\n":",\n");
}

/*:365*/
#line 6938 "MathView3D.w"

fprintf(f,"       ]\n"
"     }\n");
fprintf(f,"   coordIndex [\n");
for(i= 0;i<lst->pno;i++)
fprintf(f," %d\n",i);
fprintf(f,"    ]\n");
fprintf(f,"   }\n");
}

/*:366*/
#line 6919 "MathView3D.w"

else
/*367:*/
#line 6948 "MathView3D.w"

{double dphi;
double r;
int n= lst->pno-1;

r= LineTubeSize;
dphi= TWO_PI/TubeSegments;
/*353:*/
#line 6756 "MathView3D.w"

WriteVRMaterial(f,&lst->color);

/*:353*/
#line 6955 "MathView3D.w"

fprintf(f,"  geometry Extrusion {\n");
fprintf(f,"    crossSection [\n");
for(i= 0;i<=TubeSegments;i++)
fprintf(f,"     "DBL_FMT" "DBL_FMT"%s",
r*cos(i*dphi),r*sin(i*dphi),i==TubeSegments?"\n":",\n");
fprintf(f,"     ]\n");
fprintf(f,"    spine [\n");
/*365:*/
#line 6927 "MathView3D.w"

for(i= 0;i<lst->pno;i++){
p= PointPool+lst->points[i];
fprintf(f,"      "DBL_FMT" "DBL_FMT" "DBL_FMT"%s",p->x,p->y,p->z,n==i?"\n":",\n");
}

/*:365*/
#line 6963 "MathView3D.w"

fprintf(f,"     ]\n");
fprintf(f,"    beginCap FALSE\n"
"    endCap FALSE\n"
"    solid FALSE\n");
fprintf(f,"    creaseAngle "DBL_FMT"\n",2.0*dphi);
fprintf(f,"   }\n");
}


/*:367*/
#line 6921 "MathView3D.w"

fprintf(f,"}\n");
(*count)++;
return lst->next;
}

/*:364*/
#line 6709 "MathView3D.w"

/*368:*/
#line 6973 "MathView3D.w"

LinePolyList*WriteNoExtrudedVRLines(FILE*f,int*count,LinePolyList*lst)
{int i;

fprintf(f,"DEF MGL_NOEEXTLN%d Group {\n",*count);
fprintf(f,"  children [\n");
/*369:*/
#line 6985 "MathView3D.w"

{
if(2==lst->pno)
MakeVRCone(f,PointPool+lst->points[0],
PointPool+lst->points[1],
&lst->color,LineTubeSize,TubeSegments);
else{
for(i= 0;i<lst->pno-2;i++)
MakeVRConeContinued(f,
PointPool+lst->points[i],
PointPool+lst->points[i+1],
PointPool+lst->points[i+2],
&lst->color,
LineTubeSize,TubeSegments);
i= lst->pno-2;
if(lst->points[0]==lst->points[i+1]){
MakeVRConeContinued(f,
PointPool+lst->points[i],
PointPool+lst->points[0],
PointPool+lst->points[1],
&lst->color,
LineTubeSize,TubeSegments);
}
else
MakeVRCone(f,PointPool+lst->points[i],
PointPool+lst->points[i+1],
&lst->color,
LineTubeSize,TubeSegments);

}
}


/*:369*/
#line 6979 "MathView3D.w"

fprintf(f,"   ]\n""}\n");
(*count)++;
return lst->next;
}

/*:368*/
#line 6710 "MathView3D.w"

/*370:*/
#line 7018 "MathView3D.w"

void WriteVRPolygons(FILE*f,const char*pTexPrefix,LinePolyList*plst)
{int i;
LinePolyList*lst;
RGBColor**ctab;
int cno;
static int texFileCount= 1;

ctab= PolyColorTable(&cno,plst);
fprintf(f,
"DEF MGL_POLYGONS Shape {\n"
"  appearance Appearance {\n"
"    material Material{\n"
"      diffuseColor 1.0 1.0 1.0\n"
"      ambientIntensity 0.0\n"
"      specularColor 0.0 0.0 0.0\n"
"      emissiveColor 0.0 0.0 0.0\n"
"      shininess 0.2\n"
"      transparency 0.0\n"
"     }\n");
if(SceneTextureMap)
/*377:*/
#line 7150 "MathView3D.w"

{
if(!*TextureFile(SceneTextureMap->tex))
/*378:*/
#line 7166 "MathView3D.w"

{
sprintf(TextureFile(SceneTextureMap->tex),
"%stex%d.png",pTexPrefix,texFileCount++);
PNGSaveBitmap(
TextureFile(SceneTextureMap->tex),
TextureWidth(SceneTextureMap->tex),
TextureHeight(SceneTextureMap->tex),
TextureType(SceneTextureMap->tex),
TextureData(SceneTextureMap->tex),GL_FALSE,1.0);
}

/*:378*/
#line 7153 "MathView3D.w"

fprintf(f,
"texture ImageTexture {\n"
"  url \"%s\"\n"
"  repeatS TRUE\n"
"  repeatT TRUE\n"
" }\n",TextureFile(SceneTextureMap->tex));
fprintf(f,
"textureTransform TextureTransform {\n"
"  scale "DBL_FMT" "DBL_FMT"\n"
" }\n",SceneTextureMap->scaleS,SceneTextureMap->scaleT);
}

/*:377*/
#line 7039 "MathView3D.w"

fprintf(f,"   }\n");
if(!SceneTextureMap)
/*371:*/
#line 7049 "MathView3D.w"

{
fprintf(f,
"  geometry IndexedFaceSet {\n"
"    solid FALSE\n"
"    coord Coordinate {\n");
/*379:*/
#line 7178 "MathView3D.w"

fprintf(f,
"      point [\n");
for(i= 0;i<PoolSize;i++)
fprintf(f,"\t"DBL_FMT" "DBL_FMT" "DBL_FMT",\n",
PointPool[i].x,PointPool[i].y,PointPool[i].z);
fprintf(f,
"       ]\n");

/*:379*/
#line 7055 "MathView3D.w"

fprintf(f,"     }\n");
/*380:*/
#line 7187 "MathView3D.w"

fprintf(f,
" coordIndex [\n");
lst= plst;
while(lst){
fprintf(f,"\t");
for(i= 0;i<lst->pno;i++)
fprintf(f,"%d ",lst->points[i]);
fprintf(f,"-1,\n");
lst= lst->next;
}
fprintf(f,
"  ]\n");


/*:380*/
#line 7057 "MathView3D.w"

/*381:*/
#line 7202 "MathView3D.w"

if(SmoothShading){
Point3D*p;

fprintf(f,
"  normalPerVertex TRUE\n"
"  normal Normal {\n"
"    vector [\n");
for(i= 0;i<PoolSize;i++){
p= NormalPool[i];
if(p)
fprintf(f,"\t "DBL_FMT" "DBL_FMT" "DBL_FMT"\n",p->x,p->y,p->z);
else
fprintf(f,"\t 0 0 0\n");
}
fprintf(f,
"     ]\n"
"   }\n");
}


/*:381*/
#line 7058 "MathView3D.w"

/*382:*/
#line 7223 "MathView3D.w"

{
fprintf(f,
"colorPerVertex FALSE\n"
" color Color {\n"
"  color [\n");
for(i= 0;i<cno;i++)
fprintf(f,"    "DBL_FMT" "DBL_FMT" "DBL_FMT"\n",ctab[i]->r,ctab[i]->g,ctab[i]->b);
fprintf(f,
"   ]\n"
" }\n");
lst= plst;
fprintf(f,
"colorIndex [\n");
while(lst){
fprintf(f," %d,\n",lst->aux1);
lst= lst->next;
}
fprintf(f,
" ]\n");
}

/*:382*/
#line 7059 "MathView3D.w"

fprintf(f,"}\n");
}

/*:371*/
#line 7042 "MathView3D.w"

else
/*372:*/
#line 7063 "MathView3D.w"

{int i,pcount;
LinePolyList*polys;
Point3D*p,*tp;

fprintf(f,
"  geometry IndexedFaceSet {\n"
"    solid FALSE\n"
"    coord Coordinate {\n");
/*373:*/
#line 7080 "MathView3D.w"

polys= plst;
fprintf(f,"  point [");
while(polys){
for(i= 0;i<polys->pno;i++){
p= PointPool+polys->points[i];
fprintf(f," "DBL_FMT" "DBL_FMT" "DBL_FMT" %s\n",p->x,p->y,p->z,
VRLastPoint(i,polys)?"":",");
}
polys= polys->next;
}
fprintf(f,"   ]\n       }\n");

/*:373*/
#line 7072 "MathView3D.w"

/*374:*/
#line 7093 "MathView3D.w"

polys= plst;
fprintf(f,"  coordIndex [\n");
pcount= 0;
while(polys){
for(i= 0;i<polys->pno;i++)
fprintf(f,"%d, ",pcount++);
fprintf(f," -1,\n");
polys= polys->next;
}
fprintf(f,"    ]\n");

/*:374*/
#line 7073 "MathView3D.w"

/*375:*/
#line 7105 "MathView3D.w"

if(SmoothShading){
fprintf(f,
"  normalPerVertex TRUE\n"
"  normal Normal {\n"
"    vector [\n");
polys= plst;
while(polys){
for(i= 0;i<polys->pno;i++){
p= NormalPool[polys->points[i]];
if(p)
fprintf(f,"\t "DBL_FMT" "DBL_FMT" "DBL_FMT"\n",p->x,p->y,p->z);
else
fprintf(f,"\t 0 0 0\n");
}
polys= polys->next;
}
fprintf(f,
"     ]\n"
"   }\n");
}


/*:375*/
#line 7074 "MathView3D.w"

/*376:*/
#line 7128 "MathView3D.w"

fprintf(f," texCoord TextureCoordinate {\n");
fprintf(f,"   point [\n");
polys= plst;
while(polys){
if(CrossXPlusAxis(polys)){
for(i= 0;i<polys->pno;i++){
tp= polys->pTexturePoints+i;
fprintf(f," "DBL_FMT" "DBL_FMT" %s\n",
(fabs(tp->x)<0.5?1.0+tp->x:tp->x),tp->y,
VRLastPoint(i,polys)?"":",");
}
}
else
for(i= 0;i<polys->pno;i++){
tp= polys->pTexturePoints+i;
fprintf(f," "DBL_FMT" "DBL_FMT" %s\n",tp->x,tp->y,VRLastPoint(i,polys)?"":",");
}
polys= polys->next;
}
fprintf(f,"    ]\n   }\n");

/*:376*/
#line 7075 "MathView3D.w"

fprintf(f,"   }\n");
}

/*:372*/
#line 7044 "MathView3D.w"

fprintf(f," }\n");
free(ctab);
}

/*:370*/
#line 6711 "MathView3D.w"

/*383:*/
#line 7245 "MathView3D.w"

Mesh3DList*WriteVRMesh(FILE*f,const char*mTexPrefix,
int*count,Mesh3DList*mlst)
{static int texFileCount= 1;

fprintf(f,
"DEF MGL_MESH%d Shape {\n",*count);
fprintf(f,
"  appearance Appearance {\n"
"    material Material{\n"
"      diffuseColor "DBL_FMT" "DBL_FMT" "DBL_FMT"\n"
"      ambientIntensity 0.0\n"
"      specularColor 0.0 0.0 0.0\n"
"      emissiveColor 0.0 0.0 0.0\n"
"      shininess 0.2\n"
"      transparency 0.0\n"
"     }\n",mlst->mColor.r,mlst->mColor.g,mlst->mColor.b);
/*384:*/
#line 7283 "MathView3D.w"

{
if(MeshTextureQ(mlst)){
if(!*MeshTextureFile(mlst))
/*385:*/
#line 7301 "MathView3D.w"

{
sprintf(MeshTextureFile(mlst),"%stex%d.png",mTexPrefix,texFileCount++);
PNGSaveBitmap(
MeshTextureFile(mlst),
TextureWidth(mlst->mTexture->tex),
TextureHeight(mlst->mTexture->tex),
TextureType(mlst->mTexture->tex),
TextureData(mlst->mTexture->tex),GL_FALSE,1.0);
}

/*:385*/
#line 7287 "MathView3D.w"

fprintf(f,
"texture ImageTexture {\n"
"  url \"%s\"\n"
"  repeatS TRUE\n"
"  repeatT TRUE\n"
" }\n",MeshTextureFile(mlst));
fprintf(f,
"textureTransform TextureTransform {\n"
"  scale "DBL_FMT" "DBL_FMT"\n"
" }\n",mlst->mTexture->scaleS,mlst->mTexture->scaleT);
}
}

/*:384*/
#line 7262 "MathView3D.w"

fprintf(f,
"   }\n");
fprintf(f,
"  geometry IndexedFaceSet {\n"
"    solid FALSE\n"
"    coord Coordinate {\n");
/*386:*/
#line 7312 "MathView3D.w"

{int i,j;
int n,m;
Point3D*p;

n= mlst->dimU;
m= mlst->dimV;
fprintf(f,
"  point [\n");
for(i= 0;i<n;i++)
for(j= 0;j<m;j++){
p= MeshCoordPtr(mlst,i,j);
fprintf(f,DBL_FMT" "DBL_FMT" "DBL_FMT"\n",p->x,p->y,p->z);
}
fprintf(f,
"   ]\n");
}

/*:386*/
#line 7269 "MathView3D.w"

fprintf(f,
"     }\n");
/*387:*/
#line 7330 "MathView3D.w"

{int i,j;
int n,m;

n= mlst->dimU-1;
m= mlst->dimV-1;
fprintf(f,"  coordIndex [\n");
for(i= 0;i<n;i++)
for(j= 0;j<m;j++){
fprintf(f,"    %d %d %d %d -1\n",
MeshIndexExpanded(mlst,i,j),
MeshIndexExpanded(mlst,i+1,j),
MeshIndexExpanded(mlst,i+1,j+1),
MeshIndexExpanded(mlst,i,j+1));
}
fprintf(f,"   ]\n");
}

/*:387*/
#line 7272 "MathView3D.w"

/*388:*/
#line 7348 "MathView3D.w"

if(SmoothShading){
register Point3D*p;
register int i,j;
register int n,m;

n= mlst->dimU;
m= mlst->dimV;
fprintf(f,
"  normalPerVertex TRUE\n"
"  normal Normal {\n"
"    vector [\n");
for(i= 0;i<n;i++)
for(j= 0;j<m;j++){
p= MeshNormalPtr(mlst,i,j);
fprintf(f,"\t "DBL_FMT" "DBL_FMT" "DBL_FMT"\n",p->x,p->y,p->z);
}
fprintf(f,
"     ]\n"
"   }\n");
}

/*:388*/
#line 7273 "MathView3D.w"

/*389:*/
#line 7370 "MathView3D.w"

if(mlst->mpolyColors){
register int i,j;
register int n,m;
RGBColor*c;

n= mlst->dimU;
m= mlst->dimV;
fprintf(f,
"colorPerVertex %s\n"
" color Color {\n"
"  color [\n",
VRBool(MESH_VERTEX_COLOR==(mlst->attrib&MESH_VERTEX_COLOR)));
for(i= 0;i<n;i++)
for(j= 0;j<m;j++){
c= MeshColorPtr(mlst,i,j);
fprintf(f,"    "DBL_FMT" "DBL_FMT" "DBL_FMT"\n",c->r,c->g,c->b);
}
fprintf(f,
"   ]\n"
" }\n");
}

/*:389*/
#line 7274 "MathView3D.w"

/*390:*/
#line 7393 "MathView3D.w"

if(MeshTextureQ(mlst)){
int i,j;
int n,m;
Point3D*p;

n= mlst->dimU;
m= mlst->dimV;
fprintf(f,
"texCoord TextureCoordinate {\n"
"  point [\n");
for(i= 0;i<n;i++)
for(j= 0;j<m;j++){
p= MeshTexCoordPtr(mlst,i,j);
fprintf(f," "DBL_FMT" "DBL_FMT" %s\n",p->x,p->y,i==(n-1)?",":"");
}
fprintf(f,"   ]\n"
" }\n");
}

/*:390*/
#line 7275 "MathView3D.w"

fprintf(f,
"   }\n"
" }\n");
(*count)++;
return mlst->next;
}

/*:383*/
#line 6712 "MathView3D.w"

/*391:*/
#line 7417 "MathView3D.w"

int WriteVRML(char*fname,int headlight,int useextrusion,
const char*pTexPrefix,const char*mTexPrefix)
{FILE*vrfile;
PointList*pplst= Points;
LinePolyList*llst= Lines;
LinePolyList*plst= Polygons;
Mesh3DList*mlst= Meshes;
int count;

vrfile= fopen(fname,"w");
if(!vrfile)return 1;
/*396:*/
#line 7520 "MathView3D.w"

fprintf(vrfile,
"#VRML V2.0 utf8\n"
"# Created with %s %s \n"
"# (c) Jens-Peer Kuska 1997-99\n",
BANNER,VERSION);
fprintf(vrfile,
"##Bounding box:\n"
"## x:  "DBL_FMT" "DBL_FMT"\n"
"## y:  "DBL_FMT" "DBL_FMT"\n"
"## z:  "DBL_FMT" "DBL_FMT"\n",
BoundP1.x,BoundP2.x,
BoundP1.y,BoundP2.y,
BoundP1.z,BoundP2.z);
fprintf(vrfile,
"WorldInfo {\n"
"  title \"%s\"\n"
"  info  \"%s\n%s\n%s\"\n }\n\n",
"MathGL3d Output",
"VRML 2.0 model was automatic created from a Mathematica 3D Graphic",
"and the free OpenGL viewer for Mathematica. MathGL3d is written by Jens-Peer Kuska.",
"see <http://www.mpae.gwdg.de/~kuska/mview3d.html> for more information.");
fprintf(vrfile,
"NavigationInfo {\n"
"  avatarSize [ "DBL_FMT" "DBL_FMT" "DBL_FMT" ]\n",
BoundP2.x-BoundP1.x,
BoundP2.y-BoundP1.y,
BoundP2.z-BoundP1.z);
fprintf(vrfile,"  headlight %s\n",VRBool(headlight));
fprintf(vrfile,"  speed 1.0\n");
fprintf(vrfile,"  type [\"EXAMINE\" \"WALK\" \"FLY\"]\n");
fprintf(vrfile,"  visibilityLimit 0.0\n");
fprintf(vrfile," }\n");

/*:396*/
#line 7429 "MathView3D.w"

/*397:*/
#line 7554 "MathView3D.w"

{Point3D vp,orient;
double r,th,phi,alpha;
double fac= 5.0;


ScaleToWorld(&vp,&ViewPoint);
SphericalFromCartesian(&vp,&r,&th,&phi);
VROrientation(&orient,&alpha,&vp,&CenterOfScene,&ViewVertical);
fprintf(vrfile,
" Viewpoint {\n"
"      fieldOfView "DBL_FMT"\n"
"      position "DBL_FMT" "DBL_FMT" "DBL_FMT"\n"
"      orientation "DBL_FMT" "DBL_FMT" "DBL_FMT" "DBL_FMT"\n"
"      description \"DefaultViewpoint\"\n"
"     }\n",
PI*MVViewAngle/180.,
CenterOfScene.x+vp.x,CenterOfScene.y+vp.y,CenterOfScene.z+vp.z,
orient.x,orient.y,orient.z,alpha);
fprintf(vrfile,
"Viewpoint {\n"
"  position "DBL_FMT" "DBL_FMT" "DBL_FMT"\n"
"  description \"Top View\""
" }\n",
CenterOfScene.x,
CenterOfScene.y,
CenterOfScene.z+fac*(BoundP2.z-BoundP1.z)
);
fprintf(vrfile,
"Viewpoint {\n"
"  position "DBL_FMT" "DBL_FMT" "DBL_FMT"\n"
"  orientation 1 0 0 "DBL_FMT"\n"
"  description \"Bottom View\""
" }\n",
CenterOfScene.x,
CenterOfScene.y,
CenterOfScene.z-fac*(BoundP2.z-BoundP1.z),
PI
);
fprintf(vrfile,
"Viewpoint {\n"
"  position "DBL_FMT" "DBL_FMT" "DBL_FMT"\n"
"  orientation 1 0 0 "DBL_FMT"\n"
"  description \"Front View\""
" }\n",
CenterOfScene.x,
CenterOfScene.y-fac*(BoundP2.y-BoundP1.y),
CenterOfScene.z,
0.5*PI
);
fprintf(vrfile,
"Viewpoint {\n"
"  position "DBL_FMT" "DBL_FMT" "DBL_FMT"\n"
"  orientation 1 0 0 "DBL_FMT"\n"
"  description \"Backface View\""
" }\n",
CenterOfScene.x,
CenterOfScene.y+fac*(BoundP2.y-BoundP1.y),
CenterOfScene.z,
-0.5*PI
);

}

/*:397*/
#line 7430 "MathView3D.w"

/*398:*/
#line 7618 "MathView3D.w"

{int i;
Point3D lp;

for(i= 0;i<LightNo;i++){
ScaleToWorld(&lp,LightPosition+i);

fprintf(vrfile,
"PointLight {\n"
"  on TRUE\n"
"  intensity 1\n"
"  ambientIntensity 0\n");
fprintf(vrfile,
"  color "DBL_FMT" "DBL_FMT" "DBL_FMT"\n"
"  location "DBL_FMT" "DBL_FMT" "DBL_FMT"\n",
LightColor[i].r,LightColor[i].g,LightColor[i].b,
lp.x,lp.y,lp.z);
fprintf(vrfile,
"  radius "DBL_FMT"\n"
"  attenuation 1 0 0\n"
" }\n");
}
}

/*:398*/
#line 7431 "MathView3D.w"

fprintf(vrfile,"Group {\n");
fprintf(vrfile,"  children [\n");
if(pplst)WriteVRPoints(vrfile,pplst);
count= 0;
if(useextrusion||fabs(LineTubeSize)>MachineEpsilon)
while(llst)llst= WriteVRLines(vrfile,&count,llst);
else
while(llst)llst= WriteNoExtrudedVRLines(vrfile,&count,llst);
if(plst)WriteVRPolygons(vrfile,pTexPrefix,plst);
count= 0;
while(mlst)mlst= WriteVRMesh(vrfile,mTexPrefix,&count,mlst);
fprintf(vrfile,"  ]\n");
fprintf(vrfile,"}");
if(!vrfile)return 1;
fclose(vrfile);
return 0;
}

/*:391*/
#line 6713 "MathView3D.w"


/*:350*/
#line 82 "MathView3D.w"

/*399:*/
#line 7644 "MathView3D.w"

/*405:*/
#line 7720 "MathView3D.w"

void WriteDXF3dFace(FILE*f,int layer,
int color,int*points,int pno)
{int i,k;

fprintf(f,"0\n3DFACE\n8\n%d\n",layer);
fprintf(f,"62\n%d\n",color);
for(i= 0;i<pno;i++){
k= 1;
fprintf(f,"%d\n"DBL_FMT"\n",i+10*(k++),PointPool[points[i]].x);
fprintf(f,"%d\n"DBL_FMT"\n",i+10*(k++),PointPool[points[i]].y);
fprintf(f,"%d\n"DBL_FMT"\n",i+10*(k++),PointPool[points[i]].z);
}
}

/*:405*/
#line 7645 "MathView3D.w"

/*407:*/
#line 7756 "MathView3D.w"

int CheckTubeTetra(Point3D*p1,Point3D*p2,Point3D*p3,Point3D*p4,
Point3D*v1,Point3D*v2,Point3D*v3,Point3D*v4)
{Point3D n;
Point3D nn1,nn2;

n.x= 0.25*(v1->x+v2->x+v3->x+v4->x);
n.y= 0.25*(v1->y+v2->y+v3->y+v4->y);
n.z= 0.25*(v1->z+v2->z+v3->z+v4->z);
SurfaceNormal(&nn1,p4,p1,p2);
SurfaceNormal(&nn2,p2,p3,p4);
nn1.x+= nn2.x;nn1.y= nn2.y;nn1.z= nn2.z;
if(Dot(&nn1,&n)<0)return GL_FALSE;
return GL_TRUE;
}

/*:407*/
#line 7646 "MathView3D.w"

/*406:*/
#line 7736 "MathView3D.w"

void WriteDXFTriangle(FILE*f,int layer,
int color,Point3D*p1,Point3D*p2,Point3D*p3)
{
fprintf(f,"0\n3DFACE\n8\n%d\n",layer);
fprintf(f,"62\n%d\n",color);
fprintf(f,"%d\n"DBL_FMT"\n%d\n"DBL_FMT"\n%d\n"DBL_FMT"\n",
10,p1->x,20,p1->y,30,p1->z);
fprintf(f,"%d\n"DBL_FMT"\n%d\n"DBL_FMT"\n%d\n"DBL_FMT"\n",
11,p2->x,21,p2->y,31,p2->z);
fprintf(f,"%d\n"DBL_FMT"\n%d\n"DBL_FMT"\n%d\n"DBL_FMT"\n",
12,p3->x,22,p3->y,32,p3->z);
fprintf(f,"%d\n"DBL_FMT"\n%d\n"DBL_FMT"\n%d\n"DBL_FMT"\n",
13,p1->x,23,p1->y,33,p1->z);
}

/*:406*/
#line 7647 "MathView3D.w"

/*408:*/
#line 7773 "MathView3D.w"

void WriteDXFTetra(FILE*f,int layer,
int color,Point3D*p1,Point3D*p2,Point3D*p3,Point3D*p4)
{
fprintf(f,"0\n3DFACE\n8\n%d\n",layer);
fprintf(f,"62\n%d\n",color);
fprintf(f,"%d\n"DBL_FMT"\n%d\n"DBL_FMT"\n%d\n"DBL_FMT"\n",
10,p1->x,20,p1->y,30,p1->z);
fprintf(f,"%d\n"DBL_FMT"\n%d\n"DBL_FMT"\n%d\n"DBL_FMT"\n",
11,p2->x,21,p2->y,31,p2->z);
fprintf(f,"%d\n"DBL_FMT"\n%d\n"DBL_FMT"\n%d\n"DBL_FMT"\n",
12,p3->x,22,p3->y,32,p3->z);
fprintf(f,"%d\n"DBL_FMT"\n%d\n"DBL_FMT"\n%d\n"DBL_FMT"\n",
13,p4->x,23,p4->y,33,p4->z);
}

/*:408*/
#line 7648 "MathView3D.w"

/*419:*/
#line 7968 "MathView3D.w"

void MakeDXFSphere(FILE*dxffile,
int layer,int color,Point3D*pnt,
double r,int nphi,int nth)
{double dth,dp;
int i,j;
double cp,cp1,st,st1,sp,sp1,ct,ct1,phi,th;
Point3D p,p1,p2,p3;

dp= TWO_PI/nphi;
dth= PI/nth;
for(i= 0;i<nphi;i++){
cp= cos(phi= i*dp);
sp= sin(phi);
cp1= cos(phi= (i+1)*dp);
sp1= sin(phi);
for(j= 0;j<nth;j++){
ct= r*cos(th= (j+1)*dth);
st= r*sin(th);
ct1= r*cos(th= j*dth);
st1= r*sin(th);
p.x= p1.x= p2.x= p3.x= pnt->x;
p.y= p1.y= p2.y= p3.y= pnt->y;
p.z= p1.z= p2.z= p3.z= pnt->z;
p.x+= cp*st;p.y+= sp*st;p.z+= ct;
p1.x+= cp1*st;p1.y+= sp1*st;p1.z+= ct;
p2.x+= cp1*st1;p2.y+= sp1*st1;p2.z+= ct1;
p3.x+= cp*st1;p3.y+= sp*st1;p3.z+= ct1;
WriteDXFTetra(dxffile,layer,color,&p,&p1,&p2,&p3);
}
}
}

/*:419*/
#line 7649 "MathView3D.w"

/*418:*/
#line 7954 "MathView3D.w"

void WriteDXFPoints(FILE*dxffile,int layer,int color,PointList*lst)
{
fprintf(dxffile,"999\nPoints\n");
while(lst){
MakeDXFSphere(dxffile,layer,color,PointPool+(lst->point),
lst->size*PointSphereSize,
PointSphereSegmentsPhi,
PointSphereSegmentsTheta);
lst= lst->next;
}

}

/*:418*/
#line 7650 "MathView3D.w"

/*412:*/
#line 7829 "MathView3D.w"

void MakeDXFCone(FILE*f,int layer,int color,
Point3D*p1,Point3D*p2,double r,int n)
{Point3D v1,v2,v3,vn[4],p[4];
double norm;
double dphi;
int i;

v1.x= p2->x-p1->x;
v1.y= p2->y-p1->y;
v1.z= p2->z-p1->z;
norm= Norm(&v1);
v1.x/= norm;v1.y/= norm;v1.z/= norm;
OrthogonalSet(&v1,&v2,&v3);

v1.x= p2->x-p1->x;
v1.y= p2->y-p1->y;
v1.z= p2->z-p1->z;

dphi= TWO_PI/n;
ConePoint(r,0.0,&v2,&v3,p);
vn[0]= vn[1]= *p;
IncPoint3D(p[0],(*p1))
AddPoint3D(p[1],p[0],v1)
for(i= 1;i<=n;i++){
ConePoint(r,i*dphi,&v2,&v3,p+2);
vn[2]= p[2];
IncPoint3D(p[2],(*p1))
AddPoint3D(p[3],p[2],v1)
vn[3]= vn[2];
/*413:*/
#line 7865 "MathView3D.w"

if(CheckTubeTetra(p,p+1,p+3,p+2,vn,vn+1,vn+3,vn+2))
WriteDXFTetra(f,layer,color,p,p+1,p+3,p+2);
else
WriteDXFTetra(f,layer,color,p+2,p+3,p+1,p);

/*:413*/
#line 7859 "MathView3D.w"

p[0]= p[2];vn[0]= vn[2];
p[1]= p[3];vn[1]= vn[3];
}
}

/*:412*/
#line 7651 "MathView3D.w"

/*414:*/
#line 7871 "MathView3D.w"

void MakeDXFConeContinued(FILE*f,int layer,int color,
Point3D*p1,Point3D*p2,Point3D*p3,
double r,int n)
{Point3D v1,v2,v3,u1,u2,u3,vn[4],p[4];
double norm;
double dphi;
int i;

v1.x= p2->x-p1->x;v1.y= p2->y-p1->y;v1.z= p2->z-p1->z;
norm= Norm(&v1);
v1.x/= norm;v1.y/= norm;v1.z/= norm;
OrthogonalSet(&v1,&v2,&v3);
u1.x= p3->x-p2->x;u1.y= p3->y-p2->y;u1.z= p3->z-p2->z;
norm= Norm(&u1);
u1.x/= norm;u1.y/= norm;u1.z/= norm;
/*415:*/
#line 7910 "MathView3D.w"

if(Dot(&u1,&v1)<CosSkipCont){
MakeDXFCone(f,layer,color,p1,p2,r,n);
MakeDXFCone(f,layer,color,p2,p3,r,n);
return;
}

/*:415*/
#line 7887 "MathView3D.w"

u2= v2;u3= v3;
GramSchmidt(&u1,&u2,&u3);
dphi= TWO_PI/n;
ConePoint(r,0.0,&v2,&v3,p);
vn[0]= *p;
IncPoint3D(p[0],(*p1))
ConePoint(r,0.0,&u2,&u3,p+1);
vn[1]= p[1];
IncPoint3D(p[1],(*p2))
for(i= 1;i<=n;i++){
ConePoint(r,i*dphi,&v2,&v3,p+2);
vn[2]= p[2];
IncPoint3D(p[2],(*p1))
ConePoint(r,i*dphi,&u2,&u3,p+3);
vn[3]= p[3];
IncPoint3D(p[3],(*p2))
/*413:*/
#line 7865 "MathView3D.w"

if(CheckTubeTetra(p,p+1,p+3,p+2,vn,vn+1,vn+3,vn+2))
WriteDXFTetra(f,layer,color,p,p+1,p+3,p+2);
else
WriteDXFTetra(f,layer,color,p+2,p+3,p+1,p);

/*:413*/
#line 7904 "MathView3D.w"

p[0]= p[2];vn[0]= vn[2];
p[1]= p[3];vn[1]= vn[3];
}
}

/*:414*/
#line 7652 "MathView3D.w"

/*409:*/
#line 7789 "MathView3D.w"

void WriteDXFMeshs(FILE*f,int layer,int color,Mesh3DList*lst)
{
while(lst){
/*410:*/
#line 7798 "MathView3D.w"

{int i,j,n,m;

n= lst->dimU-1;
m= lst->dimV-1;
for(i= 0;i<n;i++)
for(j= 0;j<m;j++)
WriteDXFTetra(f,layer,color,
MeshCoordPtr(lst,i,j),
MeshCoordPtr(lst,i+1,j),
MeshCoordPtr(lst,i+1,j+1),
MeshCoordPtr(lst,i,j+1));
}


/*:410*/
#line 7793 "MathView3D.w"

lst= lst->next;
}
}

/*:409*/
#line 7653 "MathView3D.w"

/*411:*/
#line 7813 "MathView3D.w"

void WriteDXFPolys(FILE*f,int layer,int color,LinePolyList*lst)
{
fprintf(f,"999\nPolygons\n");
while(lst){
if(3==lst->pno)
WriteDXFTriangle(f,layer,color,
PointPool+lst->points[0],
PointPool+lst->points[1],
PointPool+lst->points[2]);
else
WriteDXF3dFace(f,layer,color,lst->points,lst->pno);
lst= lst->next;
}
}

/*:411*/
#line 7654 "MathView3D.w"

/*416:*/
#line 7917 "MathView3D.w"

void WriteDXFLines(FILE*f,int layer,int color,LinePolyList*lst)
{int i;
fprintf(f,"999\nCovered lines\n");
while(lst){
/*417:*/
#line 7927 "MathView3D.w"

{if(2==lst->pno)
MakeDXFCone(f,layer,color,PointPool+lst->points[0],
PointPool+lst->points[1],
LineTubeSize,TubeSegments);
else{
for(i= 0;i<lst->pno-2;i++)
MakeDXFConeContinued(f,layer,color,
PointPool+lst->points[i],
PointPool+lst->points[i+1],
PointPool+lst->points[i+2],
LineTubeSize,TubeSegments);
i= lst->pno-2;
if(lst->points[0]==lst->points[i+1]){
MakeDXFConeContinued(f,layer,color,
PointPool+lst->points[i],
PointPool+lst->points[0],
PointPool+lst->points[1],
LineTubeSize,TubeSegments);
}
else
MakeDXFCone(f,layer,color,PointPool+lst->points[i],
PointPool+lst->points[i+1],
LineTubeSize,TubeSegments);
}
}

/*:417*/
#line 7922 "MathView3D.w"

lst= lst->next;
}
}

/*:416*/
#line 7655 "MathView3D.w"

/*400:*/
#line 7659 "MathView3D.w"

int WriteDXF(char*dxfname)
{FILE*dxffile;

dxffile= fopen(dxfname,"w");
if(dxffile){
/*401:*/
#line 7695 "MathView3D.w"

fprintf(dxffile,"999\n"
"DXF created from MathGL3d "VERSION
"\n0\n"
"SECTION\n2\n"
"HEADER\n0\nENDSEC\n");

fprintf(dxffile,"0\nSECTION\n2\nTABLES\n0\n"
"TABLE\n3\n"
"LAYER\n0\n"
"LAYER\n2\n1\n0\n"
"LAYER\n2\n2\n0\n"
"LAYER\n2\n3\n0\n"
"ENDTAB\n0\nENDSEC\n");

fprintf(dxffile,"0\nSECTION\n2\nBLOCKS\n0\nENDSEC\n");

/*:401*/
#line 7665 "MathView3D.w"

/*403:*/
#line 7714 "MathView3D.w"

fprintf(dxffile,"0\nSECTION\n2\nENTITIES\n");

/*:403*/
#line 7666 "MathView3D.w"


if(Meshes)
WriteDXFMeshs(dxffile,MESH_LAYER,4,Meshes);
if(Polygons)
WriteDXFPolys(dxffile,POLY_LAYER,1,Polygons);
if(Lines&&fabs(LineTubeSize)>MachineEpsilon)
WriteDXFLines(dxffile,LINE_LAYER,2,Lines);
if(Points&&PointSphereSize>0.0)
WriteDXFPoints(dxffile,POINT_LAYER,3,Points);
/*404:*/
#line 7717 "MathView3D.w"

fprintf(dxffile,"0\nENDSEC\n");

/*:404*/
#line 7676 "MathView3D.w"

/*402:*/
#line 7712 "MathView3D.w"
fprintf(dxffile,"0\nEOF\n");

/*:402*/
#line 7677 "MathView3D.w"

fclose(dxffile);
return 0;
}
return 1;
}

/*:400*/
#line 7656 "MathView3D.w"



/*:399*/
#line 83 "MathView3D.w"

/*420:*/
#line 8004 "MathView3D.w"

/*421:*/
#line 8010 "MathView3D.w"

char*RemoveSpaces(char*line)
{int i,n;

n= strlen(line);
for(i= 0;i<=n;i++){
if(10==line[i])line[i]= '\0';
else if(13==line[i])line[i]= '\0';
}
/*422:*/
#line 8023 "MathView3D.w"

while(*line&&(isspace(*line)||*line=='\10'||*line=='\13'))
memmove(line,line+1,strlen(line));
while(*line&&isspace(line[strlen(line)]))
line[strlen(line)]= '\0';

/*:422*/
#line 8019 "MathView3D.w"

return line;
}

/*:421*/
#line 8005 "MathView3D.w"

/*423:*/
#line 8070 "MathView3D.w"

int ReadDXF(char*dxfname)
{FILE*dxf;
static char line[DXF_LINESIZE];
static char dxftoken[DXF_LINESIZE];
static char layerName[DXF_LINESIZE];
int inEntities= GL_FALSE;
double xCoord[10],yCoord[10],zCoord[10];
int lbuffer[10];
double xVertex,yVertex,zVertex;
double xMin,yMin,zMin,xMax,yMax,zMax;
int what,i;
int groupCode,dxfColor,dxfLayer,pointNo,pointNo1;
double red,green,blue;
double psize;
dxf= fopen(dxfname,"r");
if(!dxf)return GL_TRUE;
dxfLayer= dxfColor= 1;
xMin= yMin= zMin= FLT_MAX;
xMax= yMax= zMax= -FLT_MAX;
psize= 0.001;
red= green= blue= 1.0;
while(!feof(dxf)){
DXFGetLine(line)
if(DXFEndQ(line))break;
if(DXFEntitiesQ(line)){
inEntities= GL_TRUE;
continue;
}
if(!inEntities)continue;
what= DXFNone;
if(!*line)continue;
if(isalpha(*line))
strcpy(dxftoken,line);
else
strcpy(dxftoken,"");
if(DXFPointQ(line))/*427:*/
#line 8186 "MathView3D.w"

{
what= DXFPoint;
/*426:*/
#line 8166 "MathView3D.w"

do{
DXFGetInteger(groupCode,line)
if(DXFVertexQ(line))break;
if(8==groupCode){
DXFGetLine(line)
if(isdigit(*line)){
dxfLayer= atoi(line);
strcpy(layerName,"");
}
else
strcpy(layerName,line);
DXFGetInteger(groupCode,line)
}
if(62==groupCode){
DXFGetInteger(dxfColor,line)
}
if(feof(dxf))break;
}while(groupCode<10||groupCode>=39||0==groupCode);

/*:426*/
#line 8189 "MathView3D.w"

pointNo= -1;
/*424:*/
#line 8122 "MathView3D.w"

while(groupCode!=0){
if(groupCode>=10&&groupCode<19)
DXFGetDouble(xCoord[groupCode-10],line)
if(groupCode>=20&&groupCode<29)
DXFGetDouble(yCoord[groupCode-20],line)
if(groupCode>=30&&groupCode<39)
DXFGetDouble(zCoord[groupCode-30],line)
if(groupCode>=10&&groupCode<39)
pointNo++;
DXFGetLine(line)
if(isdigit(*line))
groupCode= atoi(line);
else{
break;
}
}
pointNo= (pointNo+1)/3;

/*:424*/
#line 8191 "MathView3D.w"

}

/*:427*/
#line 8106 "MathView3D.w"

if(DXFPolyLineQ(line))/*428:*/
#line 8194 "MathView3D.w"

{
what= DXFLine3D;
/*426:*/
#line 8166 "MathView3D.w"

do{
DXFGetInteger(groupCode,line)
if(DXFVertexQ(line))break;
if(8==groupCode){
DXFGetLine(line)
if(isdigit(*line)){
dxfLayer= atoi(line);
strcpy(layerName,"");
}
else
strcpy(layerName,line);
DXFGetInteger(groupCode,line)
}
if(62==groupCode){
DXFGetInteger(dxfColor,line)
}
if(feof(dxf))break;
}while(groupCode<10||groupCode>=39||0==groupCode);

/*:426*/
#line 8197 "MathView3D.w"

pointNo= 0;
do{
if(DXFVertexQ(line)){
/*425:*/
#line 8144 "MathView3D.w"

{int coordFound= GL_FALSE;

DXFGetInteger(groupCode,line)
while(groupCode!=0){
if(groupCode>=10&&groupCode<19)
DXFGetDouble(xVertex,line)
if(groupCode>=20&&groupCode<29)
DXFGetDouble(yVertex,line)
if(groupCode>=30&&groupCode<39)
DXFGetDouble(zVertex,line)
if(groupCode>=10&&groupCode<39){
coordFound= GL_TRUE;
}
else
DXFGetLine(line)
DXFGetInteger(groupCode,line)
}
}

/*:425*/
#line 8201 "MathView3D.w"

xCoord[pointNo]= xVertex;
yCoord[pointNo]= yVertex;
zCoord[pointNo]= zVertex;
pointNo++;
}
DXFGetLine(line);
}while(!feof(dxf)&&!DXFEndSeqQ(line));
}

/*:428*/
#line 8107 "MathView3D.w"

if(DXF3DFaceQ(line))/*429:*/
#line 8211 "MathView3D.w"

{
what= DXFPoly3D;
/*426:*/
#line 8166 "MathView3D.w"

do{
DXFGetInteger(groupCode,line)
if(DXFVertexQ(line))break;
if(8==groupCode){
DXFGetLine(line)
if(isdigit(*line)){
dxfLayer= atoi(line);
strcpy(layerName,"");
}
else
strcpy(layerName,line);
DXFGetInteger(groupCode,line)
}
if(62==groupCode){
DXFGetInteger(dxfColor,line)
}
if(feof(dxf))break;
}while(groupCode<10||groupCode>=39||0==groupCode);

/*:426*/
#line 8214 "MathView3D.w"

pointNo= -1;
/*424:*/
#line 8122 "MathView3D.w"

while(groupCode!=0){
if(groupCode>=10&&groupCode<19)
DXFGetDouble(xCoord[groupCode-10],line)
if(groupCode>=20&&groupCode<29)
DXFGetDouble(yCoord[groupCode-20],line)
if(groupCode>=30&&groupCode<39)
DXFGetDouble(zCoord[groupCode-30],line)
if(groupCode>=10&&groupCode<39)
pointNo++;
DXFGetLine(line)
if(isdigit(*line))
groupCode= atoi(line);
else{
break;
}
}
pointNo= (pointNo+1)/3;

/*:424*/
#line 8216 "MathView3D.w"

}


/*:429*/
#line 8108 "MathView3D.w"

/*430:*/
#line 8220 "MathView3D.w"

switch(what){
case DXFNone:break;
case DXFPoint:
DXFUpdateBBox(xCoord[0],yCoord[0],zCoord[0]);
i= AddToPointPool(xCoord[0],yCoord[0],zCoord[0]);
Points= PrependPointList(Points,i,psize,red,green,blue);
break;
case DXFLine3D:
if(pointNo>0){
for(i= 0;i<pointNo;i++){
DXFUpdateBBox(xCoord[i],yCoord[i],zCoord[i]);
lbuffer[i]= AddToPointPool(xCoord[i],yCoord[i],zCoord[i]);
}
Lines= PrependLinePolyList(Lines,Line3D,lbuffer,pointNo,red,green,blue);
}
break;
case DXFPoly3D:
if(pointNo>0){
pointNo1= pointNo-1;

if(xCoord[0]==xCoord[pointNo1]&&
yCoord[0]==yCoord[pointNo1]&&
zCoord[0]==zCoord[pointNo1])
pointNo= pointNo1;
else if(xCoord[pointNo-2]==xCoord[pointNo1]&&
yCoord[pointNo-2]==yCoord[pointNo1]&&
zCoord[pointNo-2]==zCoord[pointNo1])
pointNo= pointNo1;
for(i= 0;i<pointNo;i++){
DXFUpdateBBox(xCoord[i],yCoord[i],zCoord[i]);
lbuffer[i]= AddToPointPool(xCoord[i],yCoord[i],zCoord[i]);
}
Polygons= PrependLinePolyList(Polygons,Polygon3D,lbuffer,pointNo,red,green,blue);

}
break;
default:fprintf(stderr,"ReadDXF::Unkonwn primitive %s found.",dxftoken);
}

/*:430*/
#line 8109 "MathView3D.w"

}
BoundP1.x= xMin;BoundP1.y= yMin;BoundP1.z= zMin;
BoundP2.x= xMax;BoundP2.y= yMax;BoundP2.z= zMax;
/*431:*/
#line 8260 "MathView3D.w"

{
LightNo= 0;
i= 0;

LightPosition[i].x= 1.85198;
LightPosition[i].y= -1.79677;
LightPosition[i].z= 1.44622;
LightColor[i].r= 1.0;
LightColor[i].g= 0.0;
LightColor[i].b= 0.0;
i++;

LightPosition[i].x= 1.73331;
LightPosition[i].y= -1.36661;
LightPosition[i].z= 2.34114;
LightColor[i].r= 0.0;
LightColor[i].g= 1.0;
LightColor[i].b= 0.0;
i++;

LightPosition[i].x= 0.769322;
LightPosition[i].y= -1.63254;
LightPosition[i].z= 2.34114;
LightColor[i].r= 0.0;
LightColor[i].g= 0.0;
LightColor[i].b= 1.0;
i++;
LightNo= i;
}

/*:431*/
#line 8113 "MathView3D.w"

/*432:*/
#line 8291 "MathView3D.w"

{
ViewPoint.x= 1.3;
ViewPoint.y= -2.4;
ViewPoint.z= 2.0;
}

/*:432*/
#line 8114 "MathView3D.w"

fclose(dxf);
return GL_FALSE;
}

/*:423*/
#line 8006 "MathView3D.w"


/*:420*/
#line 84 "MathView3D.w"

/*433:*/
#line 8299 "MathView3D.w"

/*434:*/
#line 8308 "MathView3D.w"

int WriteOFF(char*offname)
{FILE*off;
int nface,i;
LinePolyList*lst;

if(!Polygons)return 1;
off= fopen(offname,"w");
if(!off)return 1;
fprintf(off,"OFF\n");
fprintf(off,"# Off file generated by MathGL3D (c) Jens-Peer Kuska\n");
nface= LengthLinePolyList(Polygons);
fprintf(off,"%d %d %d\n",PoolSize,nface,3*nface);
for(i= 0;i<PoolSize;i++)
fprintf(off,DBL_FMT" "DBL_FMT" "DBL_FMT"\n",PointPool[i].x,PointPool[i].y,PointPool[i].z);
fprintf(off,"# Faces\n");
lst= Polygons;
while(lst){
fprintf(off,"%d ",lst->pno);
for(i= 0;i<lst->pno;i++)
fprintf(off,"%d ",lst->points[i]);
fprintf(off,"\n");
lst= lst->next;
}
fclose(off);
return 0;
}

/*:434*/
#line 8300 "MathView3D.w"

/*435:*/
#line 8336 "MathView3D.w"

char*StripOFFComment(char*line)
{unsigned int i;

for(i= 0;i<strlen(line);i++){
if('#'==line[i]||'\n'==line[i]){
line[i]= '\0';
break;
}
}
/*422:*/
#line 8023 "MathView3D.w"

while(*line&&(isspace(*line)||*line=='\10'||*line=='\13'))
memmove(line,line+1,strlen(line));
while(*line&&isspace(line[strlen(line)]))
line[strlen(line)]= '\0';

/*:422*/
#line 8346 "MathView3D.w"

return line;
}

/*:435*/
#line 8301 "MathView3D.w"

/*436:*/
#line 8350 "MathView3D.w"

char*NextDataLine(FILE*f,char*line)
{
strcpy(line,"");
do{
fgets(line,OFF_LINESIZE,f);
StripOFFComment(line);
}while(!*line&&!feof(f));
return line;
}


/*:436*/
#line 8302 "MathView3D.w"

/*437:*/
#line 8371 "MathView3D.w"

int ReadOFF(char*offname)
{char line[OFF_LINESIZE],*rest;
FILE*off;
int i,j,nvertex,nface,npoly;
double xv,yv,zv;
double xMin,xMax,yMin,yMax,zMin,zMax;
int lbuffer[64],fail;
int*pointmap;

off= fopen(offname,"r");
if(!off)return 1;
NextDataLine(off,line);
if(!strstr(line,"OFF")){
fclose(off);
return 1;
}
xMin= yMin= zMin= FLT_MAX;
xMax= yMax= zMax= -FLT_MAX;

while(!feof(off)){
NextDataLine(off,line);
rest= line;
nvertex= ScanInt(&rest,0,&fail);
nface= ScanInt(&rest,0,&fail);
pointmap= (int*)calloc(nvertex,sizeof(int));
for(i= 0;i<nvertex;i++){
NextDataLine(off,line);
rest= line;
xv= ScanDouble(&rest,0.0,&fail);
yv= ScanDouble(&rest,0.0,&fail);
zv= ScanDouble(&rest,0.0,&fail);
DXFUpdateBBox(xv,yv,zv);
pointmap[i]= AddToPointPool(xv,yv,zv);
}
for(i= 0;i<nface;i++){
NextDataLine(off,line);
rest= line;
npoly= ScanInt(&rest,0,&fail);
for(j= 0;j<npoly;j++){
lbuffer[j]= ScanInt(&rest,0,&fail);
lbuffer[j]= pointmap[lbuffer[j]];
}
if(npoly)
Polygons= PrependLinePolyList(Polygons,Polygon3D,lbuffer,npoly,1.0,1.0,1.0);
}
}
BoundP1.x= xMin;BoundP1.y= yMin;BoundP1.z= zMin;
BoundP2.x= xMax;BoundP2.y= yMax;BoundP2.z= zMax;
free(pointmap);
/*431:*/
#line 8260 "MathView3D.w"

{
LightNo= 0;
i= 0;

LightPosition[i].x= 1.85198;
LightPosition[i].y= -1.79677;
LightPosition[i].z= 1.44622;
LightColor[i].r= 1.0;
LightColor[i].g= 0.0;
LightColor[i].b= 0.0;
i++;

LightPosition[i].x= 1.73331;
LightPosition[i].y= -1.36661;
LightPosition[i].z= 2.34114;
LightColor[i].r= 0.0;
LightColor[i].g= 1.0;
LightColor[i].b= 0.0;
i++;

LightPosition[i].x= 0.769322;
LightPosition[i].y= -1.63254;
LightPosition[i].z= 2.34114;
LightColor[i].r= 0.0;
LightColor[i].g= 0.0;
LightColor[i].b= 1.0;
i++;
LightNo= i;
}

/*:431*/
#line 8421 "MathView3D.w"

/*432:*/
#line 8291 "MathView3D.w"

{
ViewPoint.x= 1.3;
ViewPoint.y= -2.4;
ViewPoint.z= 2.0;
}

/*:432*/
#line 8422 "MathView3D.w"

fclose(off);
return 0;
}



/*:437*/
#line 8303 "MathView3D.w"


/*:433*/
#line 85 "MathView3D.w"

/*270:*/
#line 5036 "MathView3D.w"

/*280:*/
#line 5257 "MathView3D.w"

void WritePSHeader(FILE*file)
{
GLfloat clearColor[4],viewport[4];
GLfloat lineWidth;
int i;




glGetFloatv(GL_VIEWPORT,viewport);
glGetFloatv(GL_COLOR_CLEAR_VALUE,clearColor);
glGetFloatv(GL_LINE_WIDTH,&lineWidth);
glGetFloatv(GL_POINT_SIZE,&pointSize);


fputs("%!PS-Adobe-2.0 EPSF-2.0\n",file);

fprintf(file,"%%%%Creator: %s %s %s (using OpenGL feedback)\n",
BANNER,VERSION,ARCH);
fprintf(file,"%%%%BoundingBox: %g %g %g %g\n",
viewport[0],viewport[1],viewport[2],viewport[3]);
fputs("%%EndComments\n",file);
fputs("\n",file);
fputs("gsave\n",file);
fputs("\n",file);



fputs("% the gouraudtriangle PostScript fragement below is free\n",file);
fputs("% written by Frederic Delhoume (delhoume@ilog.fr)\n",file);
fprintf(file,"/threshold %g def\n",EPS_GOURAUD_THRESHOLD);
for(i= 0;gouraudtriangleEPS[i];i++){
fprintf(file,"%s\n",gouraudtriangleEPS[i]);
}

fprintf(file,"\n%g setlinewidth\n",lineWidth);


fprintf(file,"%g %g %g setrgbcolor\n",
clearColor[0],clearColor[1],clearColor[2]);
fprintf(file,"%g %g %g %g rectfill\n\n",
viewport[0],viewport[1],viewport[2],viewport[3]);

}
/*:280*/
#line 5037 "MathView3D.w"

/*283:*/
#line 5353 "MathView3D.w"

GLfloat*WritePSPrimitives(FILE*file,GLfloat*loc)
{
int token;
int nvertices,i;
GLfloat red,green,blue;
int smooth;
GLfloat dx,dy,dr,dg,db,absR,absG,absB,colormax;
int steps;
Feedback3DColor*vertex;
GLfloat xstep,ystep,rstep,gstep,bstep;
GLfloat xnext,ynext,rnext,gnext,bnext,distance;

token= *loc;
loc++;
switch(token){
case GL_LINE_RESET_TOKEN:
case GL_LINE_TOKEN:
/*284:*/
#line 5388 "MathView3D.w"

vertex= (Feedback3DColor*)loc;

dr= vertex[1].red-vertex[0].red;
dg= vertex[1].green-vertex[0].green;
db= vertex[1].blue-vertex[0].blue;

if(dr!=0||dg!=0||db!=0){

dx= vertex[1].x-vertex[0].x;
dy= vertex[1].y-vertex[0].y;
distance= sqrt(dx*dx+dy*dy);

absR= fabs(dr);
absG= fabs(dg);
absB= fabs(db);

colormax= MAX(absR,MAX(absG,absB));
steps= MAX(1.0,colormax*distance*EPS_SMOOTH_LINE_FACTOR);

xstep= dx/steps;
ystep= dy/steps;

rstep= dr/steps;
gstep= dg/steps;
bstep= db/steps;

xnext= vertex[0].x;
ynext= vertex[0].y;
rnext= vertex[0].red;
gnext= vertex[0].green;
bnext= vertex[0].blue;



xnext-= xstep/2.0;
ynext-= ystep/2.0;
rnext-= rstep/2.0;
gnext-= gstep/2.0;
bnext-= bstep/2.0;
}else{

steps= 0;
}

fprintf(file,"%g %g %g setrgbcolor\n",
vertex[0].red,vertex[0].green,vertex[0].blue);
fprintf(file,"%g %g moveto\n",vertex[0].x,vertex[0].y);

for(i= 0;i<steps;i++){
xnext+= xstep;
ynext+= ystep;
rnext+= rstep;
gnext+= gstep;
bnext+= bstep;
fprintf(file,"%g %g lineto stroke\n",xnext,ynext);
fprintf(file,"%g %g %g setrgbcolor\n",rnext,gnext,bnext);
fprintf(file,"%g %g moveto\n",xnext,ynext);
}
fprintf(file,"%g %g lineto stroke\n",vertex[1].x,vertex[1].y);

loc+= 14;


/*:284*/
#line 5371 "MathView3D.w"

break;
case GL_POLYGON_TOKEN:
/*285:*/
#line 5452 "MathView3D.w"

nvertices= *loc;
loc++;

vertex= (Feedback3DColor*)loc;

if(nvertices>0){
red= vertex[0].red;
green= vertex[0].green;
blue= vertex[0].blue;
smooth= GL_FALSE;
for(i= 1;i<nvertices;i++){
if(red!=vertex[i].red||green!=vertex[i].green||blue!=vertex[i].blue){
smooth= GL_TRUE;
break;
}
}
if(smooth){


for(i= 0;i<nvertices-2;i++){
fprintf(file,"[%g %g %g %g %g %g]",
vertex[0].x,vertex[i+1].x,vertex[i+2].x,
vertex[0].y,vertex[i+1].y,vertex[i+2].y);
fprintf(file," [%g %g %g] [%g %g %g] [%g %g %g] gouraudtriangle\n",
vertex[0].red,vertex[0].green,vertex[0].blue,
vertex[i+1].red,vertex[i+1].green,vertex[i+1].blue,
vertex[i+2].red,vertex[i+2].green,vertex[i+2].blue);
}
}else{

fprintf(file,"newpath\n");
fprintf(file,"%g %g %g setrgbcolor\n",red,green,blue);


fprintf(file,"%g %g moveto\n",vertex[0].x,vertex[0].y);
for(i= 1;i<nvertices;i++){
fprintf(file,"%g %g lineto\n",vertex[i].x,vertex[i].y);
}
fprintf(file,"closepath fill\n\n");
}
}
loc+= nvertices*7;



/*:285*/
#line 5374 "MathView3D.w"

break;
case GL_POINT_TOKEN:
/*286:*/
#line 5498 "MathView3D.w"

vertex= (Feedback3DColor*)loc;
fprintf(file,"%g %g %g setrgbcolor\n",vertex[0].red,vertex[0].green,vertex[0].blue);
fprintf(file,"%g %g %g 0 360 arc fill\n\n",vertex[0].x,vertex[0].y,pointSize/2.0);
loc+= 7;



/*:286*/
#line 5377 "MathView3D.w"

break;
default:

fprintf(stderr,
"Incomplete implementation.  Unexpected token (%d).\n",token);
break;
}
return loc;
}

/*:283*/
#line 5038 "MathView3D.w"

/*273:*/
#line 5113 "MathView3D.w"

/*275:*/
#line 5126 "MathView3D.w"

static int compfeedbackbuff(const void*a,const void*b)
{DepthIndex*p1= (DepthIndex*)a;
DepthIndex*p2= (DepthIndex*)b;
GLfloat diff= p2->depth-p1->depth;

if(diff>0.0)
return 1;
else if(diff<0.0)
return-1;
else
return 0;

}

/*:275*/
#line 5114 "MathView3D.w"

/*276:*/
#line 5141 "MathView3D.w"

void
SortedPSWrite(FILE*file,GLint size,GLfloat*buffer)
{
int token;
GLfloat*loc,*end;
Feedback3DColor*vertex;
GLfloat depthSum;
int nprimitives,item;
DepthIndex*prims;
int nvertices,i;

end= buffer+size;
/*277:*/
#line 5168 "MathView3D.w"

nprimitives= 0;
loc= buffer;
while(loc<end){
token= *loc;
loc++;
switch(token){
case GL_LINE_TOKEN:
case GL_LINE_RESET_TOKEN:
loc+= 14;
nprimitives++;
break;
case GL_POLYGON_TOKEN:
nvertices= *loc;
loc++;
loc+= (7*nvertices);
nprimitives++;
break;
case GL_POINT_TOKEN:
loc+= 7;
nprimitives++;
break;
default:

printf("Incomplete implementation.  Unexpected token (%d).\n",
token);
}
}

/*:277*/
#line 5154 "MathView3D.w"

/*278:*/
#line 5205 "MathView3D.w"

prims= (DepthIndex*)malloc(sizeof(DepthIndex)*nprimitives);

item= 0;
loc= buffer;
while(loc<end){
prims[item].ptr= loc;
token= *loc;
loc++;
switch(token){
case GL_LINE_TOKEN:
case GL_LINE_RESET_TOKEN:
vertex= (Feedback3DColor*)loc;
depthSum= vertex[0].z+vertex[1].z;
prims[item].depth= depthSum/2.0;
loc+= 14;
break;
case GL_POLYGON_TOKEN:
nvertices= *loc;
loc++;
vertex= (Feedback3DColor*)loc;
depthSum= vertex[0].z;
for(i= 1;i<nvertices;i++){
depthSum+= vertex[i].z;
}
prims[item].depth= depthSum/nvertices;
loc+= (7*nvertices);
break;
case GL_POINT_TOKEN:
vertex= (Feedback3DColor*)loc;
prims[item].depth= vertex[0].z;
loc+= 7;
break;
default:
assert(1);
}
item++;
}

/*:278*/
#line 5155 "MathView3D.w"

assert(item==nprimitives);
/*279:*/
#line 5251 "MathView3D.w"

qsort(prims,nprimitives,sizeof(DepthIndex),compfeedbackbuff);

/*:279*/
#line 5157 "MathView3D.w"



for(item= 0;item<nprimitives;item++){
(void)WritePSPrimitives(file,prims[item].ptr);
}

free(prims);
}


/*:276*/
#line 5115 "MathView3D.w"



/*:273*/
#line 5039 "MathView3D.w"

/*271:*/
#line 5046 "MathView3D.w"

void WritePostScript(char*psfilename)
{GLfloat*buffer;
GLenum errCode;
long int buffsize;
long int truebsize;
long int buffac= 64;
GLint retval;
FILE*psfile;
void MVDisplay(void);

buffsize= buffac*MAX(PoolSize,4096);
buffer= (GLfloat*)calloc(buffsize,sizeof(GLfloat));
do{
if(!buffer)return;
glFeedbackBuffer(buffsize,GL_3D_COLOR,buffer);
(void)glRenderMode(GL_FEEDBACK);
DoPSOutput= GL_TRUE;
MVDisplay();
DoPSOutput= GL_FALSE;
retval= glRenderMode(GL_RENDER);
if(GL_NO_ERROR!=(errCode= glGetError())){
const GLubyte*errStr;
free(buffer);
errStr= gluErrorString(errCode);
fprintf(stderr,"OpenGL Error :: %s\n",errStr);
return;
}
if(retval>0)
truebsize= retval;
else{
free(buffer);
buffac*= 2;
buffsize= buffac*MAX(PoolSize,4096);
buffer= (GLfloat*)calloc(buffsize,sizeof(GLfloat));
}
}
while(retval<0);
if(psfilename){
#if MACINTOSH_MATHLINK
psfile= MacCreateFile(psfilename,'TEXT','gsVR');
#else
psfile= fopen(psfilename,"w");
#endif
if(psfile){
WritePSHeader(psfile);
if(!WireFrame)
SortedPSWrite(psfile,truebsize,buffer);
else
/*272:*/
#line 5103 "MathView3D.w"

{GLfloat*loc,*end;
loc= buffer;
end= buffer+truebsize;
while(loc<end){
loc= WritePSPrimitives(psfile,loc);
}
}

/*:272*/
#line 5095 "MathView3D.w"

fprintf(psfile,"\nshowpage\n");
fclose(psfile);
}
}
free(buffer);
}

/*:271*/
#line 5040 "MathView3D.w"


/*:270*/
#line 86 "MathView3D.w"


/*21:*/
#line 596 "MathView3D.w"

#if MACINTOSH_MATHLINK
/*22:*/
#line 606 "MathView3D.w"

void MacintoshSleep(UInt32);

void MacintoshSleep(UInt32 ms)
{EventRecord theEvent;
EventMask mask= everyEvent;
UInt32 sleep= 1;

Delay(1,&sleep);
}

/*:22*/
#line 598 "MathView3D.w"

/*23:*/
#line 619 "MathView3D.w"

int MacPreviewPlacement(int,int,int);
int MacPreviewPlacement(int parent,int winid,int activeid)
{int x,y,wy,wx,cwy,cwx;

glutSetWindow(MVDrawWindow);
x= glutGet(GLUT_WINDOW_X);
y= glutGet(GLUT_WINDOW_Y);
wx= glutGet(GLUT_WINDOW_WIDTH);
wy= glutGet(GLUT_WINDOW_HEIGHT);
glutSetWindow(winid);
cwx= glutGet(GLUT_WINDOW_WIDTH);
cwy= glutGet(GLUT_WINDOW_HEIGHT);
glutPositionWindow(x,y+wy);
glutPopWindow();
glutShowWindow();
glutPostRedisplay();
glutSetWindow(parent);
glutShowWindow();
muiSetActiveUIList(activeid);
#if 0
fprintf(stderr,"MUI Lists View   -> %d   Light-> %d   Axis->%d\n",
VCWindow?VCWindow->muiID:-1,
LCWindow?LCWindow->muiID:-1,
AxisWindow?AxisWindow->muiID:-1);
fprintf(stderr,"MUI       Active ->%d\n",muiGetActiveUIList());
#endif
return winid;
}

/*:23*/
#line 599 "MathView3D.w"

/*26:*/
#line 654 "MathView3D.w"

FILE*MacCreateFile(const char*filename,unsigned long ftype,unsigned long creator)
{
FSSpec fspec;
OSErr err;
short int fileRef;
int i;
FILE*f;
unsigned char macfname[256];

macfname[0]= strlen(filename);
for(i= 0;i<strlen(filename);i++)
macfname[i+1]= (unsigned char)filename[i];
err= FSMakeFSSpec(0L,0L,(const unsigned char*)macfname,&fspec);
if(err==noErr||err==fnfErr){
if(err==noErr)
FSpDelete(&fspec);
err= FSpCreate(&fspec,creator,ftype,smSystemScript);
if(err==noErr){
err= FSpOpenDF(&fspec,fsCurPerm,&fileRef);
if(err==noErr)FSClose(fileRef);
}
else{
fprintf(stderr,"Error %d create the file %s of the given type ...\n",err,filename);
return NULL;
}
}
else{
fprintf(stderr,"Error creating FSSpec %d...\n",err);
return NULL;
}
f= fopen(filename,"a");
return f;
}


/*:26*/
#line 600 "MathView3D.w"

/*29:*/
#line 697 "MathView3D.w"

void MacGetFullPath(char*fullPath,
long int dirID,
int vRefNum)
{
CInfoPBRec myPB;
unsigned char dirName[255];
char cdir[255];
OSErr myErr;
int i;

myPB.dirInfo.ioNamePtr= dirName;
myPB.dirInfo.ioVRefNum= vRefNum;
myPB.dirInfo.ioDrParID= dirID;
myPB.dirInfo.ioFDirIndex= -1;
do{
myPB.dirInfo.ioDrDirID= myPB.dirInfo.ioDrParID;
myErr= PBGetCatInfo(&myPB,GL_FALSE);
for(i= 0;i<dirName[0];i++)
cdir[i]= dirName[i+1];
cdir[i++]= ':';cdir[i]= '\0';
strncat(cdir,fullPath,255);
strcpy(fullPath,cdir);
}while(myPB.dirInfo.ioDrDirID!=fsRtDirID);
}

/*:29*/
#line 601 "MathView3D.w"

/*32:*/
#line 728 "MathView3D.w"

int MacQueryLoadFilename(char*filename)
{short numTypes= 1;
SFTypeList loadTypes= {'TEXT',0};
StandardFileReply theSFReply;
int n,i;

StandardGetFile(NULL,numTypes,loadTypes,&theSFReply);
if(theSFReply.sfGood){
n= theSFReply.sfFile.name[0];
for(i= 0;i<n;i++)
filename[i]= theSFReply.sfFile.name[i+1];
filename[n]= '\0';
MacGetFullPath(filename,theSFReply.sfFile.parID,theSFReply.sfFile.vRefNum);
return GL_TRUE;
}
else
return GL_FALSE;
}

/*:32*/
#line 602 "MathView3D.w"

/*34:*/
#line 751 "MathView3D.w"

int MacQuerySaveFilename(char*cprompt,char*filename)
{StandardFileReply theSFReply;
int n,i;
unsigned char fsuggest[256];
unsigned char prompt[256];

n= prompt[0]= strlen(cprompt);
for(i= 0;i<n;i++)
prompt[i+1]= cprompt[i];
n= fsuggest[0]= strlen(filename);
for(i= 0;i<n;i++)
fsuggest[i+1]= filename[i];

StandardPutFile(prompt,fsuggest,&theSFReply);
if(theSFReply.sfGood){
n= theSFReply.sfFile.name[0];
for(i= 0;i<n;i++)
filename[i]= theSFReply.sfFile.name[i+1];
filename[n]= '\0';
MacGetFullPath(filename,theSFReply.sfFile.parID,theSFReply.sfFile.vRefNum);
return GL_TRUE;
}
else
return GL_FALSE;
}


/*:34*/
#line 603 "MathView3D.w"

#endif

/*:21*/
#line 88 "MathView3D.w"

/*765:*/
#line 12600 "MathView3D.w"

/*776:*/
#line 12766 "MathView3D.w"

void DisplayPreview(void)
{

/*779:*/
#line 12813 "MathView3D.w"

{double ct,st,cp,sp;

ct= cos(VCWindow->PreviewTheta);st= sin(VCWindow->PreviewTheta);
cp= cos(VCWindow->PreviewPhi);sp= sin(VCWindow->PreviewPhi);
VCWindow->PreViewPoint.x= VCWindow->PreviewR*cp*st;
VCWindow->PreViewPoint.y= VCWindow->PreviewR*sp*st;
VCWindow->PreViewPoint.z= VCWindow->PreviewR*ct;
}

/*:779*/
#line 12770 "MathView3D.w"

glPushMatrix();

glViewport(0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
glClearColor(1.0,1.0,1.0,1.0);
glColor3f(0.0,0.0,0.0);

glMatrixMode(GL_PROJECTION);
glLoadIdentity();
gluPerspective(0.5*MVViewAngle,
1.0,
MVNearPlane,MVFarPlane);
glMatrixMode(GL_MODELVIEW);
gluLookAt(VCWindow->PreViewPoint.x,
VCWindow->PreViewPoint.y,
VCWindow->PreViewPoint.z,
0.0,0.0,0.0,
0.0,0.0,1.0);
glScaled(VCWindow->PreviewBox.x,
VCWindow->PreviewBox.y,
VCWindow->PreviewBox.z);
glPushMatrix();
glScaled(0.5,0.5,0.5);
glutWireCube(2.0);
glutWireIcosahedron();
glPopMatrix();
glutSwapBuffers();
glPopMatrix();
}

/*:776*/
#line 12601 "MathView3D.w"

/*778:*/
#line 12806 "MathView3D.w"

void VCReshape(int w,int h)
{
if(w!=VCWindow->CWinX||h!=VCWindow->CWinY)
glutReshapeWindow(VCWindow->CWinX,VCWindow->CWinY);
}

/*:778*/
#line 12602 "MathView3D.w"

/*786:*/
#line 12932 "MathView3D.w"

/*789:*/
#line 12950 "MathView3D.w"

#pragma argsused
void QuitPressed(muiObject*obj,enum muiReturnValue r)
{
exit(0);
}

/*:789*/
#line 12933 "MathView3D.w"

/*791:*/
#line 12960 "MathView3D.w"

#pragma argsused
void HidePressed(muiObject*obj,enum muiReturnValue r)
{
#if !HAVE_GLUT_SUBWINDOW
if(VCWindow&&obj==VCWindow->HideButton){
glutSetWindow(VCWindow->PreviewWinID);
glutHideWindow();
glutSetWindow(VCWindow->WinID);
}
else if(LCWindow&&obj==LCWindow->HideButton){
glutSetWindow(LCWindow->PreviewWinID);
glutHideWindow();
glutSetWindow(LCWindow->WinID);
}
else if(AxisWindow&&obj==AxisWindow->HideButton){
glutSetWindow(AxisWindow->PreviewWinID);
glutHideWindow();
glutSetWindow(AxisWindow->WinID);
}
#endif
glutHideWindow();
glutSetWindow(MVDrawWindow);
}

/*:791*/
#line 12934 "MathView3D.w"

/*793:*/
#line 12988 "MathView3D.w"

#pragma argsused
void ApplyViewPoint(muiObject*obj,enum muiReturnValue r)
{int res;
void CopyToVCWindow(void);
int CopyFromVCWindow(void);

res= CopyFromVCWindow();
if(res>=0){
AveragedNormals(Polygons);
CopyToVCWindow();
}
glutSetWindow(MVDrawWindow);
if(res>=0){
/*682:*/
#line 11321 "MathView3D.w"

glutSetWindow(MVDrawWindow);
glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
if(ObjID>=0)glDeleteLists(ObjID,1);
ObjID= -1;
InitView(GL_TRUE);
glutPostRedisplay();
MakeNewView= GL_TRUE;

/*:682*/
#line 13002 "MathView3D.w"

}
else{
InitView(GL_TRUE);
glutPostRedisplay();
MakeNewView= GL_TRUE;
}
glutSetWindow(VCWindow->WinID);
glutPostRedisplay();
}

/*:793*/
#line 12935 "MathView3D.w"

/*795:*/
#line 13021 "MathView3D.w"

void RSliderMove(muiObject*obj,enum muiReturnValue ret)
{float rs;
if(MUI_NO_ACTION!=ret){
rs= muiGetHSVal(obj);
VCWindow->PreviewR= 1.0+(MVFarPlane-MVNearPlane)*rs;
/*804:*/
#line 13131 "MathView3D.w"

glutSetWindow(VCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(VCWindow->WinID);
/*805:*/
#line 13139 "MathView3D.w"

{double r,ct,st,cp,sp;

ct= cos(VCWindow->PreviewTheta);st= sin(VCWindow->PreviewTheta);
cp= cos(VCWindow->PreviewPhi);sp= sin(VCWindow->PreviewPhi);
r= VCWindow->PreviewR;
sprintf(VCWindow->ViewPointStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
r*cp*st,r*sp*st,r*ct);
muiSetTBString(VCWindow->ViewPointBox,VCWindow->ViewPointStr);

}

/*:805*/
#line 13135 "MathView3D.w"

/*806:*/
#line 13151 "MathView3D.w"

{
sprintf(VCWindow->BoxRatioStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
VCWindow->PreviewBox.x,
VCWindow->PreviewBox.y,
VCWindow->PreviewBox.z);
muiSetTBString(VCWindow->BoxRatioBox,VCWindow->BoxRatioStr);
}


/*:806*/
#line 13136 "MathView3D.w"



/*:804*/
#line 13027 "MathView3D.w"

}
}

/*:795*/
#line 12936 "MathView3D.w"

/*796:*/
#line 13031 "MathView3D.w"

void PhiSliderMove(muiObject*obj,enum muiReturnValue ret)
{float ps;
if(MUI_NO_ACTION!=ret){
ps= muiGetHSVal(obj);
VCWindow->PreviewPhi= TWO_PI*ps;
/*804:*/
#line 13131 "MathView3D.w"

glutSetWindow(VCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(VCWindow->WinID);
/*805:*/
#line 13139 "MathView3D.w"

{double r,ct,st,cp,sp;

ct= cos(VCWindow->PreviewTheta);st= sin(VCWindow->PreviewTheta);
cp= cos(VCWindow->PreviewPhi);sp= sin(VCWindow->PreviewPhi);
r= VCWindow->PreviewR;
sprintf(VCWindow->ViewPointStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
r*cp*st,r*sp*st,r*ct);
muiSetTBString(VCWindow->ViewPointBox,VCWindow->ViewPointStr);

}

/*:805*/
#line 13135 "MathView3D.w"

/*806:*/
#line 13151 "MathView3D.w"

{
sprintf(VCWindow->BoxRatioStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
VCWindow->PreviewBox.x,
VCWindow->PreviewBox.y,
VCWindow->PreviewBox.z);
muiSetTBString(VCWindow->BoxRatioBox,VCWindow->BoxRatioStr);
}


/*:806*/
#line 13136 "MathView3D.w"



/*:804*/
#line 13037 "MathView3D.w"

}
}

/*:796*/
#line 12937 "MathView3D.w"

/*797:*/
#line 13041 "MathView3D.w"

void ThetaSliderMove(muiObject*obj,enum muiReturnValue ret)
{float ts;
if(MUI_NO_ACTION!=ret){
ts= 1.0-muiGetVSVal(obj);
if(ts==0.0)ts= TINY;
VCWindow->PreviewTheta= PI*ts;
/*804:*/
#line 13131 "MathView3D.w"

glutSetWindow(VCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(VCWindow->WinID);
/*805:*/
#line 13139 "MathView3D.w"

{double r,ct,st,cp,sp;

ct= cos(VCWindow->PreviewTheta);st= sin(VCWindow->PreviewTheta);
cp= cos(VCWindow->PreviewPhi);sp= sin(VCWindow->PreviewPhi);
r= VCWindow->PreviewR;
sprintf(VCWindow->ViewPointStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
r*cp*st,r*sp*st,r*ct);
muiSetTBString(VCWindow->ViewPointBox,VCWindow->ViewPointStr);

}

/*:805*/
#line 13135 "MathView3D.w"

/*806:*/
#line 13151 "MathView3D.w"

{
sprintf(VCWindow->BoxRatioStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
VCWindow->PreviewBox.x,
VCWindow->PreviewBox.y,
VCWindow->PreviewBox.z);
muiSetTBString(VCWindow->BoxRatioBox,VCWindow->BoxRatioStr);
}


/*:806*/
#line 13136 "MathView3D.w"



/*:804*/
#line 13048 "MathView3D.w"

}
}

/*:797*/
#line 12938 "MathView3D.w"

/*798:*/
#line 13052 "MathView3D.w"

void XSliderMove(muiObject*obj,enum muiReturnValue ret)
{float ts;
if(MUI_NO_ACTION!=ret){
ts= muiGetVSVal(obj);
VCWindow->PreviewBox.x= ts;
/*804:*/
#line 13131 "MathView3D.w"

glutSetWindow(VCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(VCWindow->WinID);
/*805:*/
#line 13139 "MathView3D.w"

{double r,ct,st,cp,sp;

ct= cos(VCWindow->PreviewTheta);st= sin(VCWindow->PreviewTheta);
cp= cos(VCWindow->PreviewPhi);sp= sin(VCWindow->PreviewPhi);
r= VCWindow->PreviewR;
sprintf(VCWindow->ViewPointStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
r*cp*st,r*sp*st,r*ct);
muiSetTBString(VCWindow->ViewPointBox,VCWindow->ViewPointStr);

}

/*:805*/
#line 13135 "MathView3D.w"

/*806:*/
#line 13151 "MathView3D.w"

{
sprintf(VCWindow->BoxRatioStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
VCWindow->PreviewBox.x,
VCWindow->PreviewBox.y,
VCWindow->PreviewBox.z);
muiSetTBString(VCWindow->BoxRatioBox,VCWindow->BoxRatioStr);
}


/*:806*/
#line 13136 "MathView3D.w"



/*:804*/
#line 13058 "MathView3D.w"

}
}

/*:798*/
#line 12939 "MathView3D.w"

/*799:*/
#line 13062 "MathView3D.w"

void YSliderMove(muiObject*obj,enum muiReturnValue ret)
{float ts;
if(MUI_NO_ACTION!=ret){
ts= muiGetVSVal(obj);
VCWindow->PreviewBox.y= ts;
/*804:*/
#line 13131 "MathView3D.w"

glutSetWindow(VCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(VCWindow->WinID);
/*805:*/
#line 13139 "MathView3D.w"

{double r,ct,st,cp,sp;

ct= cos(VCWindow->PreviewTheta);st= sin(VCWindow->PreviewTheta);
cp= cos(VCWindow->PreviewPhi);sp= sin(VCWindow->PreviewPhi);
r= VCWindow->PreviewR;
sprintf(VCWindow->ViewPointStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
r*cp*st,r*sp*st,r*ct);
muiSetTBString(VCWindow->ViewPointBox,VCWindow->ViewPointStr);

}

/*:805*/
#line 13135 "MathView3D.w"

/*806:*/
#line 13151 "MathView3D.w"

{
sprintf(VCWindow->BoxRatioStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
VCWindow->PreviewBox.x,
VCWindow->PreviewBox.y,
VCWindow->PreviewBox.z);
muiSetTBString(VCWindow->BoxRatioBox,VCWindow->BoxRatioStr);
}


/*:806*/
#line 13136 "MathView3D.w"



/*:804*/
#line 13068 "MathView3D.w"

}
}

/*:799*/
#line 12940 "MathView3D.w"

/*800:*/
#line 13072 "MathView3D.w"

void ZSliderMove(muiObject*obj,enum muiReturnValue ret)
{float ts;
if(MUI_NO_ACTION!=ret){
ts= muiGetVSVal(obj);
VCWindow->PreviewBox.z= ts;
/*804:*/
#line 13131 "MathView3D.w"

glutSetWindow(VCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(VCWindow->WinID);
/*805:*/
#line 13139 "MathView3D.w"

{double r,ct,st,cp,sp;

ct= cos(VCWindow->PreviewTheta);st= sin(VCWindow->PreviewTheta);
cp= cos(VCWindow->PreviewPhi);sp= sin(VCWindow->PreviewPhi);
r= VCWindow->PreviewR;
sprintf(VCWindow->ViewPointStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
r*cp*st,r*sp*st,r*ct);
muiSetTBString(VCWindow->ViewPointBox,VCWindow->ViewPointStr);

}

/*:805*/
#line 13135 "MathView3D.w"

/*806:*/
#line 13151 "MathView3D.w"

{
sprintf(VCWindow->BoxRatioStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
VCWindow->PreviewBox.x,
VCWindow->PreviewBox.y,
VCWindow->PreviewBox.z);
muiSetTBString(VCWindow->BoxRatioBox,VCWindow->BoxRatioStr);
}


/*:806*/
#line 13136 "MathView3D.w"



/*:804*/
#line 13078 "MathView3D.w"

}
}

/*:800*/
#line 12941 "MathView3D.w"

/*801:*/
#line 13082 "MathView3D.w"

void ManualViewPoint(muiObject*obj,enum muiReturnValue ret)
{double xc,yc,zc;

if(MUI_TEXTBOX_RETURN==ret){
/*803:*/
#line 13116 "MathView3D.w"

{char txtStr[512],*l;
int i,n,fail1,fail2,fail3;

strcpy(txtStr,muiGetTBString(obj));
n= strlen(txtStr);
for(i= 0;i<n;i++)
if('{'==txtStr[i]||'}'==txtStr[i]||'['==txtStr[i]||']'==txtStr[i])txtStr[i]= ' ';
l= txtStr;
xc= ScanDouble(&l,0.0,&fail1);
yc= ScanDouble(&l,0.0,&fail2);
zc= ScanDouble(&l,0.0,&fail3);
if(fail1||fail2||fail3)return;
}

/*:803*/
#line 13087 "MathView3D.w"

VCWindow->PreViewPoint.x= xc;
VCWindow->PreViewPoint.y= yc;
VCWindow->PreViewPoint.z= zc;
SphericalFromCartesian(&(VCWindow->PreViewPoint),
&(VCWindow->PreviewR),
&(VCWindow->PreviewTheta),
&(VCWindow->PreviewPhi));
/*783:*/
#line 12902 "MathView3D.w"

muiSetHSValue(VCWindow->RSlider,(VCWindow->PreviewR-1.0)/(MVFarPlane-MVNearPlane));
muiSetVSValue(VCWindow->ThetaSlider,1.0-VCWindow->PreviewTheta/PI);
muiSetHSValue(VCWindow->PhiSlider,VCWindow->PreviewPhi/TWO_PI);


/*:783*/
#line 13095 "MathView3D.w"

/*804:*/
#line 13131 "MathView3D.w"

glutSetWindow(VCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(VCWindow->WinID);
/*805:*/
#line 13139 "MathView3D.w"

{double r,ct,st,cp,sp;

ct= cos(VCWindow->PreviewTheta);st= sin(VCWindow->PreviewTheta);
cp= cos(VCWindow->PreviewPhi);sp= sin(VCWindow->PreviewPhi);
r= VCWindow->PreviewR;
sprintf(VCWindow->ViewPointStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
r*cp*st,r*sp*st,r*ct);
muiSetTBString(VCWindow->ViewPointBox,VCWindow->ViewPointStr);

}

/*:805*/
#line 13135 "MathView3D.w"

/*806:*/
#line 13151 "MathView3D.w"

{
sprintf(VCWindow->BoxRatioStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
VCWindow->PreviewBox.x,
VCWindow->PreviewBox.y,
VCWindow->PreviewBox.z);
muiSetTBString(VCWindow->BoxRatioBox,VCWindow->BoxRatioStr);
}


/*:806*/
#line 13136 "MathView3D.w"



/*:804*/
#line 13096 "MathView3D.w"

}
}

/*:801*/
#line 12942 "MathView3D.w"

/*802:*/
#line 13100 "MathView3D.w"

void ManualBoundingBox(muiObject*obj,enum muiReturnValue ret)
{double xc,yc,zc;

if(MUI_TEXTBOX_RETURN==ret){
/*803:*/
#line 13116 "MathView3D.w"

{char txtStr[512],*l;
int i,n,fail1,fail2,fail3;

strcpy(txtStr,muiGetTBString(obj));
n= strlen(txtStr);
for(i= 0;i<n;i++)
if('{'==txtStr[i]||'}'==txtStr[i]||'['==txtStr[i]||']'==txtStr[i])txtStr[i]= ' ';
l= txtStr;
xc= ScanDouble(&l,0.0,&fail1);
yc= ScanDouble(&l,0.0,&fail2);
zc= ScanDouble(&l,0.0,&fail3);
if(fail1||fail2||fail3)return;
}

/*:803*/
#line 13105 "MathView3D.w"

if(xc>=0.0&&xc<=1.0)VCWindow->PreviewBox.x= xc;
if(yc>=0.0&&yc<=1.0)VCWindow->PreviewBox.y= yc;
if(zc>=0.0&&zc<=1.0)VCWindow->PreviewBox.z= zc;
/*785:*/
#line 12926 "MathView3D.w"

muiSetVSValue(VCWindow->XScaleSlider,VCWindow->PreviewBox.x);
muiSetVSValue(VCWindow->YScaleSlider,VCWindow->PreviewBox.y);
muiSetVSValue(VCWindow->ZScaleSlider,VCWindow->PreviewBox.z);

/*:785*/
#line 13109 "MathView3D.w"

/*804:*/
#line 13131 "MathView3D.w"

glutSetWindow(VCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(VCWindow->WinID);
/*805:*/
#line 13139 "MathView3D.w"

{double r,ct,st,cp,sp;

ct= cos(VCWindow->PreviewTheta);st= sin(VCWindow->PreviewTheta);
cp= cos(VCWindow->PreviewPhi);sp= sin(VCWindow->PreviewPhi);
r= VCWindow->PreviewR;
sprintf(VCWindow->ViewPointStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
r*cp*st,r*sp*st,r*ct);
muiSetTBString(VCWindow->ViewPointBox,VCWindow->ViewPointStr);

}

/*:805*/
#line 13135 "MathView3D.w"

/*806:*/
#line 13151 "MathView3D.w"

{
sprintf(VCWindow->BoxRatioStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
VCWindow->PreviewBox.x,
VCWindow->PreviewBox.y,
VCWindow->PreviewBox.z);
muiSetTBString(VCWindow->BoxRatioBox,VCWindow->BoxRatioStr);
}


/*:806*/
#line 13136 "MathView3D.w"



/*:804*/
#line 13110 "MathView3D.w"

}
}

/*:802*/
#line 12943 "MathView3D.w"

/*:786*//*883:*/
#line 14439 "MathView3D.w"

/*884:*/
#line 14448 "MathView3D.w"

void AWReshape(int w,int h)
{
if(w!=AxisWindow->CWinX||h!=AxisWindow->CWinY)
glutReshapeWindow(AxisWindow->CWinX,AxisWindow->CWinY);
}


/*:884*/
#line 14440 "MathView3D.w"

/*886:*/
#line 14480 "MathView3D.w"

void CopyToAxisWindow(void)
{

AxisWindow->AxisType= Axis.AxisType;
AxisWindow->Origin= Axis.Origin;
AxisWindow->NumericLabels= Axis.NumericLabels;
AxisWindow->FontScale= Axis.FontScale;
strcpy(AxisWindow->XLabel,Axis.XLabel);
strcpy(AxisWindow->YLabel,Axis.YLabel);
strcpy(AxisWindow->ZLabel,Axis.ZLabel);
/*891:*/
#line 14547 "MathView3D.w"

muiSetTBString(AxisWindow->XLabelBox,AxisWindow->XLabel);
muiSetTBString(AxisWindow->YLabelBox,AxisWindow->YLabel);
muiSetTBString(AxisWindow->ZLabelBox,AxisWindow->ZLabel);

/*:891*/
#line 14491 "MathView3D.w"

/*894:*/
#line 14583 "MathView3D.w"

{
sprintf(AxisWindow->OriginStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
AxisWindow->Origin.x,
AxisWindow->Origin.y,
AxisWindow->Origin.z);
muiSetTBString(AxisWindow->OriginBox,AxisWindow->OriginStr);
}


/*:894*/
#line 14492 "MathView3D.w"

#if !HAVE_GLUT_SUBWINDOW
MacPreviewPlacement(AxisWindow->WinID,AxisWindow->PreviewWinID,AxisWindow->muiID);
#endif
}

/*:886*/
#line 14441 "MathView3D.w"

/*885:*/
#line 14456 "MathView3D.w"

#pragma argsused
void ApplyAxis(muiObject*obj,enum muiReturnValue r)
{
Axis.AxisType= AxisWindow->AxisType;
Axis.Origin= AxisWindow->Origin;
Axis.NumericLabels= AxisWindow->NumericLabels;
Axis.FontScale= AxisWindow->FontScale;
/*890:*/
#line 14542 "MathView3D.w"

strcpy(AxisWindow->XLabel,muiGetTBString(AxisWindow->XLabelBox));
strcpy(AxisWindow->YLabel,muiGetTBString(AxisWindow->YLabelBox));
strcpy(AxisWindow->ZLabel,muiGetTBString(AxisWindow->ZLabelBox));

/*:890*/
#line 14464 "MathView3D.w"

strcpy(Axis.XLabel,AxisWindow->XLabel);
strcpy(Axis.YLabel,AxisWindow->YLabel);
strcpy(Axis.ZLabel,AxisWindow->ZLabel);

glutSetWindow(MVDrawWindow);
InitView(GL_TRUE);
glutPostRedisplay();
MakeNewView= GL_TRUE;
glutSetWindow(AxisWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(AxisWindow->WinID);
glutPostRedisplay();

}

/*:885*/
#line 14442 "MathView3D.w"

/*887:*/
#line 14498 "MathView3D.w"

#pragma argsused
void SwitchAxisSolid(muiObject*obj,enum muiReturnValue res)
{
if(MUI_BUTTON_PRESS==res){
/*893:*/
#line 14573 "MathView3D.w"

AxisWindow->NumericLabels= muiGetActive(AxisWindow->NumLabelRButton);
/*894:*/
#line 14583 "MathView3D.w"

{
sprintf(AxisWindow->OriginStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
AxisWindow->Origin.x,
AxisWindow->Origin.y,
AxisWindow->Origin.z);
muiSetTBString(AxisWindow->OriginBox,AxisWindow->OriginStr);
}


/*:894*/
#line 14575 "MathView3D.w"

glutSetWindow(AxisWindow->PreviewWinID);
SetBackgroundColor(BlackBackground);

glutPostRedisplay();
glutSetWindow(AxisWindow->WinID);


/*:893*/
#line 14503 "MathView3D.w"

}
}

/*:887*/
#line 14443 "MathView3D.w"

/*888:*/
#line 14507 "MathView3D.w"

#pragma argsused
void SwitchAxis(muiObject*obj,enum muiReturnValue res)
{
if(muiGetActive(AxisWindow->CrossRButton))
AxisWindow->AxisType= CrossAxis;
else if(muiGetActive(AxisWindow->BoxRButton))
AxisWindow->AxisType= BoxAxis;
else if(muiGetActive(AxisWindow->PlaneRButton))
AxisWindow->AxisType= PlaneAxis;
else
AxisWindow->AxisType= NoAxis;
if(MUI_BUTTON_PRESS==res){
/*893:*/
#line 14573 "MathView3D.w"

AxisWindow->NumericLabels= muiGetActive(AxisWindow->NumLabelRButton);
/*894:*/
#line 14583 "MathView3D.w"

{
sprintf(AxisWindow->OriginStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
AxisWindow->Origin.x,
AxisWindow->Origin.y,
AxisWindow->Origin.z);
muiSetTBString(AxisWindow->OriginBox,AxisWindow->OriginStr);
}


/*:894*/
#line 14575 "MathView3D.w"

glutSetWindow(AxisWindow->PreviewWinID);
SetBackgroundColor(BlackBackground);

glutPostRedisplay();
glutSetWindow(AxisWindow->WinID);


/*:893*/
#line 14520 "MathView3D.w"

}
}

/*:888*/
#line 14444 "MathView3D.w"

/*889:*/
#line 14524 "MathView3D.w"

void AxesTextChange(muiObject*obj,enum muiReturnValue res)
{double xc,yc,zc;

if(MUI_TEXTBOX_RETURN==res){
if(obj==AxisWindow->XLabelBox||obj==AxisWindow->YLabelBox||obj==AxisWindow->ZLabelBox){
/*890:*/
#line 14542 "MathView3D.w"

strcpy(AxisWindow->XLabel,muiGetTBString(AxisWindow->XLabelBox));
strcpy(AxisWindow->YLabel,muiGetTBString(AxisWindow->YLabelBox));
strcpy(AxisWindow->ZLabel,muiGetTBString(AxisWindow->ZLabelBox));

/*:890*/
#line 14530 "MathView3D.w"

}
else if(obj==AxisWindow->OriginBox){
/*803:*/
#line 13116 "MathView3D.w"

{char txtStr[512],*l;
int i,n,fail1,fail2,fail3;

strcpy(txtStr,muiGetTBString(obj));
n= strlen(txtStr);
for(i= 0;i<n;i++)
if('{'==txtStr[i]||'}'==txtStr[i]||'['==txtStr[i]||']'==txtStr[i])txtStr[i]= ' ';
l= txtStr;
xc= ScanDouble(&l,0.0,&fail1);
yc= ScanDouble(&l,0.0,&fail2);
zc= ScanDouble(&l,0.0,&fail3);
if(fail1||fail2||fail3)return;
}

/*:803*/
#line 14533 "MathView3D.w"

AxisWindow->Origin.x= xc;
AxisWindow->Origin.y= yc;
AxisWindow->Origin.z= zc;
}
/*893:*/
#line 14573 "MathView3D.w"

AxisWindow->NumericLabels= muiGetActive(AxisWindow->NumLabelRButton);
/*894:*/
#line 14583 "MathView3D.w"

{
sprintf(AxisWindow->OriginStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
AxisWindow->Origin.x,
AxisWindow->Origin.y,
AxisWindow->Origin.z);
muiSetTBString(AxisWindow->OriginBox,AxisWindow->OriginStr);
}


/*:894*/
#line 14575 "MathView3D.w"

glutSetWindow(AxisWindow->PreviewWinID);
SetBackgroundColor(BlackBackground);

glutPostRedisplay();
glutSetWindow(AxisWindow->WinID);


/*:893*/
#line 14538 "MathView3D.w"

}
}

/*:889*/
#line 14445 "MathView3D.w"

/*892:*/
#line 14552 "MathView3D.w"

void OriginSliderMove(muiObject*obj,enum muiReturnValue ret)
{float ts;
if(MUI_SLIDER_RETURN==ret){
ts= muiGetVSVal(obj);

if(obj==AxisWindow->XOriginSlider)
AxisWindow->Origin.x= ts*(BoundP2.x-BoundP1.x)+BoundP1.x;
else if(obj==AxisWindow->YOriginSlider)
AxisWindow->Origin.y= ts*(BoundP2.y-BoundP1.y)+BoundP1.y;
else if(obj==AxisWindow->ZOriginSlider)
AxisWindow->Origin.z= ts*(BoundP2.z-BoundP1.z)+BoundP1.z;
else if(obj==AxisWindow->FontScaleSlider){
ts= muiGetHSVal(obj);
AxisWindow->FontScale= ts;
}
/*893:*/
#line 14573 "MathView3D.w"

AxisWindow->NumericLabels= muiGetActive(AxisWindow->NumLabelRButton);
/*894:*/
#line 14583 "MathView3D.w"

{
sprintf(AxisWindow->OriginStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
AxisWindow->Origin.x,
AxisWindow->Origin.y,
AxisWindow->Origin.z);
muiSetTBString(AxisWindow->OriginBox,AxisWindow->OriginStr);
}


/*:894*/
#line 14575 "MathView3D.w"

glutSetWindow(AxisWindow->PreviewWinID);
SetBackgroundColor(BlackBackground);

glutPostRedisplay();
glutSetWindow(AxisWindow->WinID);


/*:893*/
#line 14568 "MathView3D.w"

}
}


/*:892*/
#line 14446 "MathView3D.w"


/*:883*/
#line 12603 "MathView3D.w"

/*770:*/
#line 12680 "MathView3D.w"

void CopyToVCWindow(void)
{
#if 0
fprintf(stderr,"Main      %d  ",MVDrawWindow);
fprintf(stderr,"Viewpoint %d   Preview %d\n",VCWindow->WinID,VCWindow->PreviewWinID);
fprintf(stderr,"Light     %d   Preview %d\n",LCWindow->WinID,LCWindow->PreviewWinID);
fprintf(stderr,"Axes      %d   Preview %d\n",AxisWindow->WinID,AxisWindow->PreviewWinID);
fprintf(stderr,"Active    %d\n",glutGetWindow());
#endif
VCWindow->PreViewPoint= ViewPoint;
SphericalFromCartesian(&ViewPoint,
&(VCWindow->PreviewR),
&(VCWindow->PreviewTheta),
&(VCWindow->PreviewPhi));
while(VCWindow->PreviewPhi<0.0)VCWindow->PreviewPhi+= TWO_PI;
VCWindow->PreviewBox= BoxRatios;;
/*805:*/
#line 13139 "MathView3D.w"

{double r,ct,st,cp,sp;

ct= cos(VCWindow->PreviewTheta);st= sin(VCWindow->PreviewTheta);
cp= cos(VCWindow->PreviewPhi);sp= sin(VCWindow->PreviewPhi);
r= VCWindow->PreviewR;
sprintf(VCWindow->ViewPointStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
r*cp*st,r*sp*st,r*ct);
muiSetTBString(VCWindow->ViewPointBox,VCWindow->ViewPointStr);

}

/*:805*/
#line 12697 "MathView3D.w"

/*806:*/
#line 13151 "MathView3D.w"

{
sprintf(VCWindow->BoxRatioStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
VCWindow->PreviewBox.x,
VCWindow->PreviewBox.y,
VCWindow->PreviewBox.z);
muiSetTBString(VCWindow->BoxRatioBox,VCWindow->BoxRatioStr);
}


/*:806*/
#line 12698 "MathView3D.w"

/*785:*/
#line 12926 "MathView3D.w"

muiSetVSValue(VCWindow->XScaleSlider,VCWindow->PreviewBox.x);
muiSetVSValue(VCWindow->YScaleSlider,VCWindow->PreviewBox.y);
muiSetVSValue(VCWindow->ZScaleSlider,VCWindow->PreviewBox.z);

/*:785*/
#line 12699 "MathView3D.w"

/*783:*/
#line 12902 "MathView3D.w"

muiSetHSValue(VCWindow->RSlider,(VCWindow->PreviewR-1.0)/(MVFarPlane-MVNearPlane));
muiSetVSValue(VCWindow->ThetaSlider,1.0-VCWindow->PreviewTheta/PI);
muiSetHSValue(VCWindow->PhiSlider,VCWindow->PreviewPhi/TWO_PI);


/*:783*/
#line 12700 "MathView3D.w"

strcpy(VCWindow->FilePrefix,WriteFilePrefix);
muiSetTBString(VCWindow->FilePrefixBox,VCWindow->FilePrefix);
strcpy(VCWindow->Load3Script,"None");
muiSetTBString(VCWindow->Load3ScriptBox,VCWindow->Load3Script);
strcpy(VCWindow->LoadDXF,"None");
muiSetTBString(VCWindow->LoadDXFBox,VCWindow->LoadDXF);
#if !HAVE_GLUT_SUBWINDOW
MacPreviewPlacement(VCWindow->WinID,VCWindow->PreviewWinID,VCWindow->muiID);
#endif
}

/*:770*/
#line 12604 "MathView3D.w"

/*772:*/
#line 12715 "MathView3D.w"

int CopyFromVCWindow(void)
{int fail= -1;

ViewPoint= VCWindow->PreViewPoint;
BoxRatios= VCWindow->PreviewBox;
strcpy(VCWindow->FilePrefix,muiGetTBString(VCWindow->FilePrefixBox));
strcpy(VCWindow->Load3Script,muiGetTBString(VCWindow->Load3ScriptBox));
strcpy(VCWindow->LoadDXF,muiGetTBString(VCWindow->LoadDXFBox));
strcpy(WriteFilePrefix,VCWindow->FilePrefix);
if(0!=strcmp(VCWindow->Load3Script,"None")&&
0!=strcmp(VCWindow->Load3Script,"")){
fail= Read3Script(VCWindow->Load3Script);
if(!fail){
AveragedNormals(Polygons);
strcpy(VCWindow->Load3Script,"None");
}
}
if(0!=strcmp(VCWindow->LoadDXF,"None")&&
0!=strcmp(VCWindow->LoadDXF,"")){
fail= ReadDXF(VCWindow->LoadDXF);
if(!fail){
AveragedNormals(Polygons);
strcpy(VCWindow->LoadDXF,"None");
}
}
muiSetTBString(VCWindow->Load3ScriptBox,VCWindow->Load3Script);
muiSetTBString(VCWindow->LoadDXFBox,VCWindow->LoadDXF);
return fail;
}

/*:772*/
#line 12605 "MathView3D.w"

/*768:*/
#line 12645 "MathView3D.w"

ViewCntrlWin*CreateViewControl(void)
{ViewCntrlWin*vc;
int xPos,yPos,xEnd;

VCWindow= vc= (ViewCntrlWin*)malloc(sizeof(ViewCntrlWin));
if(NULL==vc)return NULL;
vc->CWinX= 380;
vc->CWinY= 500;
glutInitWindowSize(vc->CWinX,vc->CWinY);
vc->WinID= glutCreateWindow("View Point Control");
vc->PreViewPoint= ViewPoint;
SphericalFromCartesian(&ViewPoint,
&(vc->PreviewR),
&(vc->PreviewTheta),
&(vc->PreviewPhi));
vc->PreviewBox= BoxRatios;
muiInit();
muiNewUIList(vc->WinID);
/*780:*/
#line 12823 "MathView3D.w"

{int yOffset= 0;

xPos= 10;
yPos= 100;
#if !HAVE_GLUT_SUBWINDOW
xPos+= 30;
yPos+= 30;
yOffset+= 90;
#endif
vc->ApplyButton= muiNewButton(xPos,yPos,yOffset+10,yOffset+35);
muiLoadButton(vc->ApplyButton,"Apply");
muiSetCallback(vc->ApplyButton,ApplyViewPoint);

xPos+= 100;
yPos+= 100;
vc->HideButton= muiNewButton(xPos,yPos,yOffset+10,yOffset+35);
muiLoadButton(vc->HideButton,"Hide");
muiSetCallback(vc->HideButton,HidePressed);
xPos+= 170;
yPos+= 170;
if(!MathLinkMode){
vc->QuitButton= muiNewButton(xPos,yPos,yOffset+10,yOffset+35);
muiLoadButton(vc->QuitButton,"Quit");
muiSetCallback(vc->QuitButton,QuitPressed);
}
}

/*:780*/
#line 12664 "MathView3D.w"

/*781:*/
#line 12851 "MathView3D.w"

xPos= 10;
xEnd= vc->CWinX-10;
yPos= vc->CWinY-25;
(void*)muiNewBoldLabel(xPos,yPos,"Export Prefix:");
vc->FilePrefixBox= muiNewTextbox(xPos+90,xEnd,yPos-15);
muiSetTBString(vc->FilePrefixBox,"mview3d");
yPos-= 30;
(void*)muiNewBoldLabel(xPos,yPos,"Load 3Script: ");
vc->Load3ScriptBox= muiNewTextbox(xPos+90,xEnd,yPos-15);
muiSetTBString(vc->Load3ScriptBox,"None");
yPos-= 30;
(void*)muiNewBoldLabel(xPos,yPos,"Load DXF file: ");
vc->LoadDXFBox= muiNewTextbox(xPos+90,xEnd,yPos-15);
muiSetTBString(vc->LoadDXFBox,"None");
yPos-= 30;
(void*)muiNewBoldLabel(xPos,yPos,"ViewPoint -> ");
vc->ViewPointBox= muiNewTextbox(xPos+90,xEnd,yPos-15);
muiSetCallback(vc->ViewPointBox,ManualViewPoint);

/*805:*/
#line 13139 "MathView3D.w"

{double r,ct,st,cp,sp;

ct= cos(VCWindow->PreviewTheta);st= sin(VCWindow->PreviewTheta);
cp= cos(VCWindow->PreviewPhi);sp= sin(VCWindow->PreviewPhi);
r= VCWindow->PreviewR;
sprintf(VCWindow->ViewPointStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
r*cp*st,r*sp*st,r*ct);
muiSetTBString(VCWindow->ViewPointBox,VCWindow->ViewPointStr);

}

/*:805*/
#line 12871 "MathView3D.w"

yPos-= 30;
(void*)muiNewBoldLabel(xPos,yPos,"BoxRatios -> ");
vc->BoxRatioBox= muiNewTextbox(xPos+90,xEnd,yPos-15);
muiSetCallback(vc->BoxRatioBox,ManualBoundingBox);
/*806:*/
#line 13151 "MathView3D.w"

{
sprintf(VCWindow->BoxRatioStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
VCWindow->PreviewBox.x,
VCWindow->PreviewBox.y,
VCWindow->PreviewBox.z);
muiSetTBString(VCWindow->BoxRatioBox,VCWindow->BoxRatioStr);
}


/*:806*/
#line 12876 "MathView3D.w"



/*:781*/
#line 12665 "MathView3D.w"

/*779:*/
#line 12813 "MathView3D.w"

{double ct,st,cp,sp;

ct= cos(VCWindow->PreviewTheta);st= sin(VCWindow->PreviewTheta);
cp= cos(VCWindow->PreviewPhi);sp= sin(VCWindow->PreviewPhi);
VCWindow->PreViewPoint.x= VCWindow->PreviewR*cp*st;
VCWindow->PreViewPoint.y= VCWindow->PreviewR*sp*st;
VCWindow->PreViewPoint.z= VCWindow->PreviewR*ct;
}

/*:779*/
#line 12666 "MathView3D.w"

/*782:*/
#line 12879 "MathView3D.w"

(void*)muiNewLabel(248,303,"theta");
vc->ThetaSlider= muiNewVSlider(250,95,295,195,5);
muiSetCallback(vc->ThetaSlider,ThetaSliderMove);
muiSetVSValue(vc->ThetaSlider,1.0-vc->PreviewTheta/PI);

#if HAVE_GLUT_SUBWINDOW
xPos= 10;
yPos= 55;
#else
xPos= 10;
yPos= vc->CWinY-245;
#endif
(void*)muiNewLabel(xPos,yPos,"r");
vc->RSlider= muiNewHSlider(xPos+30,yPos-5,240,200,10);
muiSetCallback(vc->RSlider,RSliderMove);
muiSetHSValue(vc->RSlider,(vc->PreviewR-1.0)/8.0);
yPos+= 25;
(void*)muiNewLabel(xPos,yPos,"phi");
vc->PhiSlider= muiNewHSlider(xPos+30,yPos-5,240,200,10);
muiSetCallback(vc->PhiSlider,PhiSliderMove);
muiSetHSValue(vc->PhiSlider,1.0-vc->PreviewPhi/TWO_PI);

/*:782*/
#line 12667 "MathView3D.w"

/*784:*/
#line 12908 "MathView3D.w"

(void*)muiNewBoldLabel(295,vc->CWinY-175,"Box Scaling");

(void*)muiNewLabel(305,303,"x");
vc->XScaleSlider= muiNewVSlider(295,95,295,195,5);
muiSetCallback(vc->XScaleSlider,XSliderMove);
muiSetVSValue(vc->XScaleSlider,vc->PreviewBox.x);

(void*)muiNewLabel(330,303,"y");
vc->YScaleSlider= muiNewVSlider(320,95,295,195,5);
muiSetCallback(vc->YScaleSlider,YSliderMove);
muiSetVSValue(vc->YScaleSlider,vc->PreviewBox.y);

(void*)muiNewLabel(355,303,"z");
vc->ZScaleSlider= muiNewVSlider(345,95,295,195,5);
muiSetCallback(vc->ZScaleSlider,ZSliderMove);
muiSetVSValue(vc->ZScaleSlider,vc->PreviewBox.z);

/*:784*/
#line 12668 "MathView3D.w"

glutReshapeFunc(VCReshape);
muiAttachUIList(vc->WinID);
vc->muiID= muiGetActiveUIList();
/*774:*/
#line 12750 "MathView3D.w"

#if !HAVE_GLUT_SUBWINDOW
glutInitWindowSize(200,200);
vc->PreviewWinID= glutCreateWindow("Preview");
#else
vc->PreviewWinID= glutCreateSubWindow(vc->WinID,40,vc->CWinY-305,200,200);
#endif
glutDisplayFunc(DisplayPreview);
DisplayPreview();
glutHideWindow();
glutSetWindow(vc->WinID);
glutPostRedisplay();

/*:774*/
#line 12672 "MathView3D.w"

return vc;
}

/*:768*/
#line 12606 "MathView3D.w"


/*:765*/
#line 89 "MathView3D.w"

/*807:*/
#line 13164 "MathView3D.w"

/*829:*/
#line 13570 "MathView3D.w"

/*830:*/
#line 13585 "MathView3D.w"

#pragma argsused
void ApplyLightSources(muiObject*obj,enum muiReturnValue r)
{
int i;
LightNo= LCWindow->LNo;
for(i= 0;i<LightNo;i++){
LightPosition[i]= LCWindow->LightPosition[i];
LightColor[i]= LCWindow->LightColor[i];
}
glutSetWindow(MVDrawWindow);
for(i= 0;i<MAX_LIGHT;i++)
glDisable(GL_LIGHT0+i);
InitView(GL_TRUE);
glutPostRedisplay();
MakeNewView= GL_TRUE;
glutSetWindow(LCWindow->WinID);
glutPostRedisplay();
}

/*:830*/
#line 13571 "MathView3D.w"

/*835:*/
#line 13669 "MathView3D.w"

void LRSliderMove(muiObject*obj,enum muiReturnValue ret)
{float rs;

if(MUI_SLIDER_RETURN==ret&&LCWindow->LNo>=0){
rs= muiGetHSVal(obj);
LCWindow->PreviewR= 1.0+8.0*rs;
/*845:*/
#line 13790 "MathView3D.w"

{double r,ct,st,cp,sp;

/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13793 "MathView3D.w"
;
ct= cos(LCWindow->PreviewTheta);st= sin(LCWindow->PreviewTheta);
cp= cos(LCWindow->PreviewPhi);sp= sin(LCWindow->PreviewPhi);
r= LCWindow->PreviewR;
CurrentPos.x= r*cp*st;
CurrentPos.y= r*sp*st;
CurrentPos.z= r*ct;
}

/*:845*/
#line 13676 "MathView3D.w"

/*843:*/
#line 13767 "MathView3D.w"

/*826:*/
#line 13515 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13519 "MathView3D.w"
;
SphericalFromCartesian(
LCWindow->LightPosition+what,
&(LCWindow->PreviewR),
&(LCWindow->PreviewTheta),
&(LCWindow->PreviewPhi));
while(LCWindow->PreviewPhi<0.0)LCWindow->PreviewPhi+= TWO_PI;

muiSetEnable(LCWindow->RSlider,GL_TRUE);
muiSetEnable(LCWindow->ThetaSlider,GL_TRUE);
muiSetEnable(LCWindow->PhiSlider,GL_TRUE);
muiSetHSValue(LCWindow->RSlider,(LCWindow->PreviewR-1.0)/8.0);
muiSetVSValue(LCWindow->ThetaSlider,1.0-LCWindow->PreviewTheta/PI);
muiSetHSValue(LCWindow->PhiSlider,LCWindow->PreviewPhi/TWO_PI);
}
else{
muiSetEnable(LCWindow->RSlider,GL_FALSE);
muiSetEnable(LCWindow->ThetaSlider,GL_FALSE);
muiSetEnable(LCWindow->PhiSlider,GL_FALSE);

}

/*:826*/
#line 13768 "MathView3D.w"

/*844:*/
#line 13776 "MathView3D.w"

{
if(LCWindow->LNo>=0){
/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13779 "MathView3D.w"
;
/*845:*/
#line 13790 "MathView3D.w"

{double r,ct,st,cp,sp;

/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13793 "MathView3D.w"
;
ct= cos(LCWindow->PreviewTheta);st= sin(LCWindow->PreviewTheta);
cp= cos(LCWindow->PreviewPhi);sp= sin(LCWindow->PreviewPhi);
r= LCWindow->PreviewR;
CurrentPos.x= r*cp*st;
CurrentPos.y= r*sp*st;
CurrentPos.z= r*ct;
}

/*:845*/
#line 13780 "MathView3D.w"

sprintf(LCWindow->PositionStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
CurrentPos.x,CurrentPos.y,CurrentPos.z);
}
else
strcpy(LCWindow->PositionStr,"");
muiSetTBString(LCWindow->PositionBox,LCWindow->PositionStr);

}

/*:844*/
#line 13769 "MathView3D.w"

/*825:*/
#line 13493 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13497 "MathView3D.w"
;
muiSetEnable(LCWindow->RedSlider,GL_TRUE);
muiSetEnable(LCWindow->GreenSlider,GL_TRUE);
muiSetEnable(LCWindow->BlueSlider,GL_TRUE);

muiSetVSValue(LCWindow->RedSlider,LCWindow->LightColor[what].r);
muiSetVSValue(LCWindow->GreenSlider,LCWindow->LightColor[what].g);
muiSetVSValue(LCWindow->BlueSlider,LCWindow->LightColor[what].b);
}
else{
muiSetVSValue(LCWindow->RedSlider,0.0);
muiSetVSValue(LCWindow->GreenSlider,0.0);
muiSetVSValue(LCWindow->BlueSlider,0.0);
muiSetEnable(LCWindow->RedSlider,GL_FALSE);
muiSetEnable(LCWindow->GreenSlider,GL_FALSE);
muiSetEnable(LCWindow->BlueSlider,GL_FALSE);
}

/*:825*/
#line 13770 "MathView3D.w"

/*846:*/
#line 13806 "MathView3D.w"

{if(LCWindow->LNo>=0){
sprintf(LCWindow->ColorStr,"["DBL_FMT", "DBL_FMT", "DBL_FMT"]",
CurrentColor.r,
CurrentColor.g,
CurrentColor.b);
}
else
strcpy(LCWindow->ColorStr,"");
muiSetTBString(LCWindow->ColorBox,LCWindow->ColorStr);
}

/*:846*/
#line 13771 "MathView3D.w"

glutSetWindow(LCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(LCWindow->WinID);

/*:843*/
#line 13677 "MathView3D.w"

}
}

/*:835*/
#line 13572 "MathView3D.w"

/*836:*/
#line 13681 "MathView3D.w"

void LPhiSliderMove(muiObject*obj,enum muiReturnValue ret)
{float ps;

if(MUI_SLIDER_RETURN==ret&&LCWindow->LNo>=0){
ps= muiGetHSVal(obj);
LCWindow->PreviewPhi= TWO_PI*ps;
/*845:*/
#line 13790 "MathView3D.w"

{double r,ct,st,cp,sp;

/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13793 "MathView3D.w"
;
ct= cos(LCWindow->PreviewTheta);st= sin(LCWindow->PreviewTheta);
cp= cos(LCWindow->PreviewPhi);sp= sin(LCWindow->PreviewPhi);
r= LCWindow->PreviewR;
CurrentPos.x= r*cp*st;
CurrentPos.y= r*sp*st;
CurrentPos.z= r*ct;
}

/*:845*/
#line 13688 "MathView3D.w"

/*843:*/
#line 13767 "MathView3D.w"

/*826:*/
#line 13515 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13519 "MathView3D.w"
;
SphericalFromCartesian(
LCWindow->LightPosition+what,
&(LCWindow->PreviewR),
&(LCWindow->PreviewTheta),
&(LCWindow->PreviewPhi));
while(LCWindow->PreviewPhi<0.0)LCWindow->PreviewPhi+= TWO_PI;

muiSetEnable(LCWindow->RSlider,GL_TRUE);
muiSetEnable(LCWindow->ThetaSlider,GL_TRUE);
muiSetEnable(LCWindow->PhiSlider,GL_TRUE);
muiSetHSValue(LCWindow->RSlider,(LCWindow->PreviewR-1.0)/8.0);
muiSetVSValue(LCWindow->ThetaSlider,1.0-LCWindow->PreviewTheta/PI);
muiSetHSValue(LCWindow->PhiSlider,LCWindow->PreviewPhi/TWO_PI);
}
else{
muiSetEnable(LCWindow->RSlider,GL_FALSE);
muiSetEnable(LCWindow->ThetaSlider,GL_FALSE);
muiSetEnable(LCWindow->PhiSlider,GL_FALSE);

}

/*:826*/
#line 13768 "MathView3D.w"

/*844:*/
#line 13776 "MathView3D.w"

{
if(LCWindow->LNo>=0){
/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13779 "MathView3D.w"
;
/*845:*/
#line 13790 "MathView3D.w"

{double r,ct,st,cp,sp;

/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13793 "MathView3D.w"
;
ct= cos(LCWindow->PreviewTheta);st= sin(LCWindow->PreviewTheta);
cp= cos(LCWindow->PreviewPhi);sp= sin(LCWindow->PreviewPhi);
r= LCWindow->PreviewR;
CurrentPos.x= r*cp*st;
CurrentPos.y= r*sp*st;
CurrentPos.z= r*ct;
}

/*:845*/
#line 13780 "MathView3D.w"

sprintf(LCWindow->PositionStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
CurrentPos.x,CurrentPos.y,CurrentPos.z);
}
else
strcpy(LCWindow->PositionStr,"");
muiSetTBString(LCWindow->PositionBox,LCWindow->PositionStr);

}

/*:844*/
#line 13769 "MathView3D.w"

/*825:*/
#line 13493 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13497 "MathView3D.w"
;
muiSetEnable(LCWindow->RedSlider,GL_TRUE);
muiSetEnable(LCWindow->GreenSlider,GL_TRUE);
muiSetEnable(LCWindow->BlueSlider,GL_TRUE);

muiSetVSValue(LCWindow->RedSlider,LCWindow->LightColor[what].r);
muiSetVSValue(LCWindow->GreenSlider,LCWindow->LightColor[what].g);
muiSetVSValue(LCWindow->BlueSlider,LCWindow->LightColor[what].b);
}
else{
muiSetVSValue(LCWindow->RedSlider,0.0);
muiSetVSValue(LCWindow->GreenSlider,0.0);
muiSetVSValue(LCWindow->BlueSlider,0.0);
muiSetEnable(LCWindow->RedSlider,GL_FALSE);
muiSetEnable(LCWindow->GreenSlider,GL_FALSE);
muiSetEnable(LCWindow->BlueSlider,GL_FALSE);
}

/*:825*/
#line 13770 "MathView3D.w"

/*846:*/
#line 13806 "MathView3D.w"

{if(LCWindow->LNo>=0){
sprintf(LCWindow->ColorStr,"["DBL_FMT", "DBL_FMT", "DBL_FMT"]",
CurrentColor.r,
CurrentColor.g,
CurrentColor.b);
}
else
strcpy(LCWindow->ColorStr,"");
muiSetTBString(LCWindow->ColorBox,LCWindow->ColorStr);
}

/*:846*/
#line 13771 "MathView3D.w"

glutSetWindow(LCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(LCWindow->WinID);

/*:843*/
#line 13689 "MathView3D.w"

}
}

/*:836*/
#line 13573 "MathView3D.w"

/*837:*/
#line 13693 "MathView3D.w"

void LThetaSliderMove(muiObject*obj,enum muiReturnValue ret)
{float ts;

if(MUI_SLIDER_RETURN==ret&&LCWindow->LNo>=0){
ts= 1.0-muiGetVSVal(obj);
LCWindow->PreviewTheta= PI*ts;
/*845:*/
#line 13790 "MathView3D.w"

{double r,ct,st,cp,sp;

/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13793 "MathView3D.w"
;
ct= cos(LCWindow->PreviewTheta);st= sin(LCWindow->PreviewTheta);
cp= cos(LCWindow->PreviewPhi);sp= sin(LCWindow->PreviewPhi);
r= LCWindow->PreviewR;
CurrentPos.x= r*cp*st;
CurrentPos.y= r*sp*st;
CurrentPos.z= r*ct;
}

/*:845*/
#line 13700 "MathView3D.w"

/*843:*/
#line 13767 "MathView3D.w"

/*826:*/
#line 13515 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13519 "MathView3D.w"
;
SphericalFromCartesian(
LCWindow->LightPosition+what,
&(LCWindow->PreviewR),
&(LCWindow->PreviewTheta),
&(LCWindow->PreviewPhi));
while(LCWindow->PreviewPhi<0.0)LCWindow->PreviewPhi+= TWO_PI;

muiSetEnable(LCWindow->RSlider,GL_TRUE);
muiSetEnable(LCWindow->ThetaSlider,GL_TRUE);
muiSetEnable(LCWindow->PhiSlider,GL_TRUE);
muiSetHSValue(LCWindow->RSlider,(LCWindow->PreviewR-1.0)/8.0);
muiSetVSValue(LCWindow->ThetaSlider,1.0-LCWindow->PreviewTheta/PI);
muiSetHSValue(LCWindow->PhiSlider,LCWindow->PreviewPhi/TWO_PI);
}
else{
muiSetEnable(LCWindow->RSlider,GL_FALSE);
muiSetEnable(LCWindow->ThetaSlider,GL_FALSE);
muiSetEnable(LCWindow->PhiSlider,GL_FALSE);

}

/*:826*/
#line 13768 "MathView3D.w"

/*844:*/
#line 13776 "MathView3D.w"

{
if(LCWindow->LNo>=0){
/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13779 "MathView3D.w"
;
/*845:*/
#line 13790 "MathView3D.w"

{double r,ct,st,cp,sp;

/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13793 "MathView3D.w"
;
ct= cos(LCWindow->PreviewTheta);st= sin(LCWindow->PreviewTheta);
cp= cos(LCWindow->PreviewPhi);sp= sin(LCWindow->PreviewPhi);
r= LCWindow->PreviewR;
CurrentPos.x= r*cp*st;
CurrentPos.y= r*sp*st;
CurrentPos.z= r*ct;
}

/*:845*/
#line 13780 "MathView3D.w"

sprintf(LCWindow->PositionStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
CurrentPos.x,CurrentPos.y,CurrentPos.z);
}
else
strcpy(LCWindow->PositionStr,"");
muiSetTBString(LCWindow->PositionBox,LCWindow->PositionStr);

}

/*:844*/
#line 13769 "MathView3D.w"

/*825:*/
#line 13493 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13497 "MathView3D.w"
;
muiSetEnable(LCWindow->RedSlider,GL_TRUE);
muiSetEnable(LCWindow->GreenSlider,GL_TRUE);
muiSetEnable(LCWindow->BlueSlider,GL_TRUE);

muiSetVSValue(LCWindow->RedSlider,LCWindow->LightColor[what].r);
muiSetVSValue(LCWindow->GreenSlider,LCWindow->LightColor[what].g);
muiSetVSValue(LCWindow->BlueSlider,LCWindow->LightColor[what].b);
}
else{
muiSetVSValue(LCWindow->RedSlider,0.0);
muiSetVSValue(LCWindow->GreenSlider,0.0);
muiSetVSValue(LCWindow->BlueSlider,0.0);
muiSetEnable(LCWindow->RedSlider,GL_FALSE);
muiSetEnable(LCWindow->GreenSlider,GL_FALSE);
muiSetEnable(LCWindow->BlueSlider,GL_FALSE);
}

/*:825*/
#line 13770 "MathView3D.w"

/*846:*/
#line 13806 "MathView3D.w"

{if(LCWindow->LNo>=0){
sprintf(LCWindow->ColorStr,"["DBL_FMT", "DBL_FMT", "DBL_FMT"]",
CurrentColor.r,
CurrentColor.g,
CurrentColor.b);
}
else
strcpy(LCWindow->ColorStr,"");
muiSetTBString(LCWindow->ColorBox,LCWindow->ColorStr);
}

/*:846*/
#line 13771 "MathView3D.w"

glutSetWindow(LCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(LCWindow->WinID);

/*:843*/
#line 13701 "MathView3D.w"

}
}

/*:837*/
#line 13574 "MathView3D.w"

/*838:*/
#line 13705 "MathView3D.w"

void RedSliderMove(muiObject*obj,enum muiReturnValue ret)
{float ts;
if(MUI_SLIDER_RETURN==ret&&LCWindow->sLight>=0){
ts= muiGetVSVal(obj);
/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13710 "MathView3D.w"
;
CurrentColor.r= ts;
/*843:*/
#line 13767 "MathView3D.w"

/*826:*/
#line 13515 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13519 "MathView3D.w"
;
SphericalFromCartesian(
LCWindow->LightPosition+what,
&(LCWindow->PreviewR),
&(LCWindow->PreviewTheta),
&(LCWindow->PreviewPhi));
while(LCWindow->PreviewPhi<0.0)LCWindow->PreviewPhi+= TWO_PI;

muiSetEnable(LCWindow->RSlider,GL_TRUE);
muiSetEnable(LCWindow->ThetaSlider,GL_TRUE);
muiSetEnable(LCWindow->PhiSlider,GL_TRUE);
muiSetHSValue(LCWindow->RSlider,(LCWindow->PreviewR-1.0)/8.0);
muiSetVSValue(LCWindow->ThetaSlider,1.0-LCWindow->PreviewTheta/PI);
muiSetHSValue(LCWindow->PhiSlider,LCWindow->PreviewPhi/TWO_PI);
}
else{
muiSetEnable(LCWindow->RSlider,GL_FALSE);
muiSetEnable(LCWindow->ThetaSlider,GL_FALSE);
muiSetEnable(LCWindow->PhiSlider,GL_FALSE);

}

/*:826*/
#line 13768 "MathView3D.w"

/*844:*/
#line 13776 "MathView3D.w"

{
if(LCWindow->LNo>=0){
/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13779 "MathView3D.w"
;
/*845:*/
#line 13790 "MathView3D.w"

{double r,ct,st,cp,sp;

/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13793 "MathView3D.w"
;
ct= cos(LCWindow->PreviewTheta);st= sin(LCWindow->PreviewTheta);
cp= cos(LCWindow->PreviewPhi);sp= sin(LCWindow->PreviewPhi);
r= LCWindow->PreviewR;
CurrentPos.x= r*cp*st;
CurrentPos.y= r*sp*st;
CurrentPos.z= r*ct;
}

/*:845*/
#line 13780 "MathView3D.w"

sprintf(LCWindow->PositionStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
CurrentPos.x,CurrentPos.y,CurrentPos.z);
}
else
strcpy(LCWindow->PositionStr,"");
muiSetTBString(LCWindow->PositionBox,LCWindow->PositionStr);

}

/*:844*/
#line 13769 "MathView3D.w"

/*825:*/
#line 13493 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13497 "MathView3D.w"
;
muiSetEnable(LCWindow->RedSlider,GL_TRUE);
muiSetEnable(LCWindow->GreenSlider,GL_TRUE);
muiSetEnable(LCWindow->BlueSlider,GL_TRUE);

muiSetVSValue(LCWindow->RedSlider,LCWindow->LightColor[what].r);
muiSetVSValue(LCWindow->GreenSlider,LCWindow->LightColor[what].g);
muiSetVSValue(LCWindow->BlueSlider,LCWindow->LightColor[what].b);
}
else{
muiSetVSValue(LCWindow->RedSlider,0.0);
muiSetVSValue(LCWindow->GreenSlider,0.0);
muiSetVSValue(LCWindow->BlueSlider,0.0);
muiSetEnable(LCWindow->RedSlider,GL_FALSE);
muiSetEnable(LCWindow->GreenSlider,GL_FALSE);
muiSetEnable(LCWindow->BlueSlider,GL_FALSE);
}

/*:825*/
#line 13770 "MathView3D.w"

/*846:*/
#line 13806 "MathView3D.w"

{if(LCWindow->LNo>=0){
sprintf(LCWindow->ColorStr,"["DBL_FMT", "DBL_FMT", "DBL_FMT"]",
CurrentColor.r,
CurrentColor.g,
CurrentColor.b);
}
else
strcpy(LCWindow->ColorStr,"");
muiSetTBString(LCWindow->ColorBox,LCWindow->ColorStr);
}

/*:846*/
#line 13771 "MathView3D.w"

glutSetWindow(LCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(LCWindow->WinID);

/*:843*/
#line 13712 "MathView3D.w"

}
}

/*:838*/
#line 13575 "MathView3D.w"

/*839:*/
#line 13716 "MathView3D.w"

void GreenSliderMove(muiObject*obj,enum muiReturnValue ret)
{float ts;
if(MUI_SLIDER_RETURN==ret&&LCWindow->sLight>=0){
ts= muiGetVSVal(obj);
/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13721 "MathView3D.w"
;
CurrentColor.g= ts;
/*843:*/
#line 13767 "MathView3D.w"

/*826:*/
#line 13515 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13519 "MathView3D.w"
;
SphericalFromCartesian(
LCWindow->LightPosition+what,
&(LCWindow->PreviewR),
&(LCWindow->PreviewTheta),
&(LCWindow->PreviewPhi));
while(LCWindow->PreviewPhi<0.0)LCWindow->PreviewPhi+= TWO_PI;

muiSetEnable(LCWindow->RSlider,GL_TRUE);
muiSetEnable(LCWindow->ThetaSlider,GL_TRUE);
muiSetEnable(LCWindow->PhiSlider,GL_TRUE);
muiSetHSValue(LCWindow->RSlider,(LCWindow->PreviewR-1.0)/8.0);
muiSetVSValue(LCWindow->ThetaSlider,1.0-LCWindow->PreviewTheta/PI);
muiSetHSValue(LCWindow->PhiSlider,LCWindow->PreviewPhi/TWO_PI);
}
else{
muiSetEnable(LCWindow->RSlider,GL_FALSE);
muiSetEnable(LCWindow->ThetaSlider,GL_FALSE);
muiSetEnable(LCWindow->PhiSlider,GL_FALSE);

}

/*:826*/
#line 13768 "MathView3D.w"

/*844:*/
#line 13776 "MathView3D.w"

{
if(LCWindow->LNo>=0){
/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13779 "MathView3D.w"
;
/*845:*/
#line 13790 "MathView3D.w"

{double r,ct,st,cp,sp;

/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13793 "MathView3D.w"
;
ct= cos(LCWindow->PreviewTheta);st= sin(LCWindow->PreviewTheta);
cp= cos(LCWindow->PreviewPhi);sp= sin(LCWindow->PreviewPhi);
r= LCWindow->PreviewR;
CurrentPos.x= r*cp*st;
CurrentPos.y= r*sp*st;
CurrentPos.z= r*ct;
}

/*:845*/
#line 13780 "MathView3D.w"

sprintf(LCWindow->PositionStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
CurrentPos.x,CurrentPos.y,CurrentPos.z);
}
else
strcpy(LCWindow->PositionStr,"");
muiSetTBString(LCWindow->PositionBox,LCWindow->PositionStr);

}

/*:844*/
#line 13769 "MathView3D.w"

/*825:*/
#line 13493 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13497 "MathView3D.w"
;
muiSetEnable(LCWindow->RedSlider,GL_TRUE);
muiSetEnable(LCWindow->GreenSlider,GL_TRUE);
muiSetEnable(LCWindow->BlueSlider,GL_TRUE);

muiSetVSValue(LCWindow->RedSlider,LCWindow->LightColor[what].r);
muiSetVSValue(LCWindow->GreenSlider,LCWindow->LightColor[what].g);
muiSetVSValue(LCWindow->BlueSlider,LCWindow->LightColor[what].b);
}
else{
muiSetVSValue(LCWindow->RedSlider,0.0);
muiSetVSValue(LCWindow->GreenSlider,0.0);
muiSetVSValue(LCWindow->BlueSlider,0.0);
muiSetEnable(LCWindow->RedSlider,GL_FALSE);
muiSetEnable(LCWindow->GreenSlider,GL_FALSE);
muiSetEnable(LCWindow->BlueSlider,GL_FALSE);
}

/*:825*/
#line 13770 "MathView3D.w"

/*846:*/
#line 13806 "MathView3D.w"

{if(LCWindow->LNo>=0){
sprintf(LCWindow->ColorStr,"["DBL_FMT", "DBL_FMT", "DBL_FMT"]",
CurrentColor.r,
CurrentColor.g,
CurrentColor.b);
}
else
strcpy(LCWindow->ColorStr,"");
muiSetTBString(LCWindow->ColorBox,LCWindow->ColorStr);
}

/*:846*/
#line 13771 "MathView3D.w"

glutSetWindow(LCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(LCWindow->WinID);

/*:843*/
#line 13723 "MathView3D.w"

}
}

/*:839*/
#line 13576 "MathView3D.w"

/*840:*/
#line 13727 "MathView3D.w"

void BlueSliderMove(muiObject*obj,enum muiReturnValue ret)
{float ts;
if(MUI_SLIDER_RETURN==ret&&LCWindow->sLight>=0){
ts= muiGetVSVal(obj);
/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13732 "MathView3D.w"
;
CurrentColor.b= ts;
/*843:*/
#line 13767 "MathView3D.w"

/*826:*/
#line 13515 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13519 "MathView3D.w"
;
SphericalFromCartesian(
LCWindow->LightPosition+what,
&(LCWindow->PreviewR),
&(LCWindow->PreviewTheta),
&(LCWindow->PreviewPhi));
while(LCWindow->PreviewPhi<0.0)LCWindow->PreviewPhi+= TWO_PI;

muiSetEnable(LCWindow->RSlider,GL_TRUE);
muiSetEnable(LCWindow->ThetaSlider,GL_TRUE);
muiSetEnable(LCWindow->PhiSlider,GL_TRUE);
muiSetHSValue(LCWindow->RSlider,(LCWindow->PreviewR-1.0)/8.0);
muiSetVSValue(LCWindow->ThetaSlider,1.0-LCWindow->PreviewTheta/PI);
muiSetHSValue(LCWindow->PhiSlider,LCWindow->PreviewPhi/TWO_PI);
}
else{
muiSetEnable(LCWindow->RSlider,GL_FALSE);
muiSetEnable(LCWindow->ThetaSlider,GL_FALSE);
muiSetEnable(LCWindow->PhiSlider,GL_FALSE);

}

/*:826*/
#line 13768 "MathView3D.w"

/*844:*/
#line 13776 "MathView3D.w"

{
if(LCWindow->LNo>=0){
/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13779 "MathView3D.w"
;
/*845:*/
#line 13790 "MathView3D.w"

{double r,ct,st,cp,sp;

/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13793 "MathView3D.w"
;
ct= cos(LCWindow->PreviewTheta);st= sin(LCWindow->PreviewTheta);
cp= cos(LCWindow->PreviewPhi);sp= sin(LCWindow->PreviewPhi);
r= LCWindow->PreviewR;
CurrentPos.x= r*cp*st;
CurrentPos.y= r*sp*st;
CurrentPos.z= r*ct;
}

/*:845*/
#line 13780 "MathView3D.w"

sprintf(LCWindow->PositionStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
CurrentPos.x,CurrentPos.y,CurrentPos.z);
}
else
strcpy(LCWindow->PositionStr,"");
muiSetTBString(LCWindow->PositionBox,LCWindow->PositionStr);

}

/*:844*/
#line 13769 "MathView3D.w"

/*825:*/
#line 13493 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13497 "MathView3D.w"
;
muiSetEnable(LCWindow->RedSlider,GL_TRUE);
muiSetEnable(LCWindow->GreenSlider,GL_TRUE);
muiSetEnable(LCWindow->BlueSlider,GL_TRUE);

muiSetVSValue(LCWindow->RedSlider,LCWindow->LightColor[what].r);
muiSetVSValue(LCWindow->GreenSlider,LCWindow->LightColor[what].g);
muiSetVSValue(LCWindow->BlueSlider,LCWindow->LightColor[what].b);
}
else{
muiSetVSValue(LCWindow->RedSlider,0.0);
muiSetVSValue(LCWindow->GreenSlider,0.0);
muiSetVSValue(LCWindow->BlueSlider,0.0);
muiSetEnable(LCWindow->RedSlider,GL_FALSE);
muiSetEnable(LCWindow->GreenSlider,GL_FALSE);
muiSetEnable(LCWindow->BlueSlider,GL_FALSE);
}

/*:825*/
#line 13770 "MathView3D.w"

/*846:*/
#line 13806 "MathView3D.w"

{if(LCWindow->LNo>=0){
sprintf(LCWindow->ColorStr,"["DBL_FMT", "DBL_FMT", "DBL_FMT"]",
CurrentColor.r,
CurrentColor.g,
CurrentColor.b);
}
else
strcpy(LCWindow->ColorStr,"");
muiSetTBString(LCWindow->ColorBox,LCWindow->ColorStr);
}

/*:846*/
#line 13771 "MathView3D.w"

glutSetWindow(LCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(LCWindow->WinID);

/*:843*/
#line 13734 "MathView3D.w"

}
}

/*:840*/
#line 13577 "MathView3D.w"

/*831:*/
#line 13605 "MathView3D.w"

#pragma argsused
void NewLightPressed(muiObject*obj,enum muiReturnValue res)
{int save;

if(LCWindow->LNo<MAX_LIGHT){
(LCWindow->LNo)++;
LCWindow->LNames[LCWindow->LNo-1]= LCWindow->LNameN[LCWindow->LNo-1];
save= LCWindow->sLight;
LCWindow->sLight= LCWindow->LNo-1;
CurrentPos.x= 0.0;
CurrentPos.y= 0.0;
CurrentPos.z= 1.5;
LCWindow->PreviewR= 1.5;
LCWindow->PreviewPhi= 0.0;
LCWindow->PreviewTheta= 0.0;
CurrentColor.r= CurrentColor.g= CurrentColor.b= 1.0;
LCWindow->LNames[LCWindow->LNo]= NULL;
LCWindow->sLight= save;
muiSetTLStrings(LCWindow->LTextList,(char**)(LCWindow->LNames));
/*843:*/
#line 13767 "MathView3D.w"

/*826:*/
#line 13515 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13519 "MathView3D.w"
;
SphericalFromCartesian(
LCWindow->LightPosition+what,
&(LCWindow->PreviewR),
&(LCWindow->PreviewTheta),
&(LCWindow->PreviewPhi));
while(LCWindow->PreviewPhi<0.0)LCWindow->PreviewPhi+= TWO_PI;

muiSetEnable(LCWindow->RSlider,GL_TRUE);
muiSetEnable(LCWindow->ThetaSlider,GL_TRUE);
muiSetEnable(LCWindow->PhiSlider,GL_TRUE);
muiSetHSValue(LCWindow->RSlider,(LCWindow->PreviewR-1.0)/8.0);
muiSetVSValue(LCWindow->ThetaSlider,1.0-LCWindow->PreviewTheta/PI);
muiSetHSValue(LCWindow->PhiSlider,LCWindow->PreviewPhi/TWO_PI);
}
else{
muiSetEnable(LCWindow->RSlider,GL_FALSE);
muiSetEnable(LCWindow->ThetaSlider,GL_FALSE);
muiSetEnable(LCWindow->PhiSlider,GL_FALSE);

}

/*:826*/
#line 13768 "MathView3D.w"

/*844:*/
#line 13776 "MathView3D.w"

{
if(LCWindow->LNo>=0){
/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13779 "MathView3D.w"
;
/*845:*/
#line 13790 "MathView3D.w"

{double r,ct,st,cp,sp;

/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13793 "MathView3D.w"
;
ct= cos(LCWindow->PreviewTheta);st= sin(LCWindow->PreviewTheta);
cp= cos(LCWindow->PreviewPhi);sp= sin(LCWindow->PreviewPhi);
r= LCWindow->PreviewR;
CurrentPos.x= r*cp*st;
CurrentPos.y= r*sp*st;
CurrentPos.z= r*ct;
}

/*:845*/
#line 13780 "MathView3D.w"

sprintf(LCWindow->PositionStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
CurrentPos.x,CurrentPos.y,CurrentPos.z);
}
else
strcpy(LCWindow->PositionStr,"");
muiSetTBString(LCWindow->PositionBox,LCWindow->PositionStr);

}

/*:844*/
#line 13769 "MathView3D.w"

/*825:*/
#line 13493 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13497 "MathView3D.w"
;
muiSetEnable(LCWindow->RedSlider,GL_TRUE);
muiSetEnable(LCWindow->GreenSlider,GL_TRUE);
muiSetEnable(LCWindow->BlueSlider,GL_TRUE);

muiSetVSValue(LCWindow->RedSlider,LCWindow->LightColor[what].r);
muiSetVSValue(LCWindow->GreenSlider,LCWindow->LightColor[what].g);
muiSetVSValue(LCWindow->BlueSlider,LCWindow->LightColor[what].b);
}
else{
muiSetVSValue(LCWindow->RedSlider,0.0);
muiSetVSValue(LCWindow->GreenSlider,0.0);
muiSetVSValue(LCWindow->BlueSlider,0.0);
muiSetEnable(LCWindow->RedSlider,GL_FALSE);
muiSetEnable(LCWindow->GreenSlider,GL_FALSE);
muiSetEnable(LCWindow->BlueSlider,GL_FALSE);
}

/*:825*/
#line 13770 "MathView3D.w"

/*846:*/
#line 13806 "MathView3D.w"

{if(LCWindow->LNo>=0){
sprintf(LCWindow->ColorStr,"["DBL_FMT", "DBL_FMT", "DBL_FMT"]",
CurrentColor.r,
CurrentColor.g,
CurrentColor.b);
}
else
strcpy(LCWindow->ColorStr,"");
muiSetTBString(LCWindow->ColorBox,LCWindow->ColorStr);
}

/*:846*/
#line 13771 "MathView3D.w"

glutSetWindow(LCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(LCWindow->WinID);

/*:843*/
#line 13625 "MathView3D.w"

}

}

/*:831*/
#line 13578 "MathView3D.w"

/*832:*/
#line 13630 "MathView3D.w"

#pragma argsused
void DeleteLightPressed(muiObject*obj,enum muiReturnValue res)
{
if(LCWindow->LNo>0){
(LCWindow->LNo)--;
LCWindow->LNames[LCWindow->LNo]= NULL;
muiSetTLStrings(LCWindow->LTextList,(char**)(LCWindow->LNames));
/*843:*/
#line 13767 "MathView3D.w"

/*826:*/
#line 13515 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13519 "MathView3D.w"
;
SphericalFromCartesian(
LCWindow->LightPosition+what,
&(LCWindow->PreviewR),
&(LCWindow->PreviewTheta),
&(LCWindow->PreviewPhi));
while(LCWindow->PreviewPhi<0.0)LCWindow->PreviewPhi+= TWO_PI;

muiSetEnable(LCWindow->RSlider,GL_TRUE);
muiSetEnable(LCWindow->ThetaSlider,GL_TRUE);
muiSetEnable(LCWindow->PhiSlider,GL_TRUE);
muiSetHSValue(LCWindow->RSlider,(LCWindow->PreviewR-1.0)/8.0);
muiSetVSValue(LCWindow->ThetaSlider,1.0-LCWindow->PreviewTheta/PI);
muiSetHSValue(LCWindow->PhiSlider,LCWindow->PreviewPhi/TWO_PI);
}
else{
muiSetEnable(LCWindow->RSlider,GL_FALSE);
muiSetEnable(LCWindow->ThetaSlider,GL_FALSE);
muiSetEnable(LCWindow->PhiSlider,GL_FALSE);

}

/*:826*/
#line 13768 "MathView3D.w"

/*844:*/
#line 13776 "MathView3D.w"

{
if(LCWindow->LNo>=0){
/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13779 "MathView3D.w"
;
/*845:*/
#line 13790 "MathView3D.w"

{double r,ct,st,cp,sp;

/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13793 "MathView3D.w"
;
ct= cos(LCWindow->PreviewTheta);st= sin(LCWindow->PreviewTheta);
cp= cos(LCWindow->PreviewPhi);sp= sin(LCWindow->PreviewPhi);
r= LCWindow->PreviewR;
CurrentPos.x= r*cp*st;
CurrentPos.y= r*sp*st;
CurrentPos.z= r*ct;
}

/*:845*/
#line 13780 "MathView3D.w"

sprintf(LCWindow->PositionStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
CurrentPos.x,CurrentPos.y,CurrentPos.z);
}
else
strcpy(LCWindow->PositionStr,"");
muiSetTBString(LCWindow->PositionBox,LCWindow->PositionStr);

}

/*:844*/
#line 13769 "MathView3D.w"

/*825:*/
#line 13493 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13497 "MathView3D.w"
;
muiSetEnable(LCWindow->RedSlider,GL_TRUE);
muiSetEnable(LCWindow->GreenSlider,GL_TRUE);
muiSetEnable(LCWindow->BlueSlider,GL_TRUE);

muiSetVSValue(LCWindow->RedSlider,LCWindow->LightColor[what].r);
muiSetVSValue(LCWindow->GreenSlider,LCWindow->LightColor[what].g);
muiSetVSValue(LCWindow->BlueSlider,LCWindow->LightColor[what].b);
}
else{
muiSetVSValue(LCWindow->RedSlider,0.0);
muiSetVSValue(LCWindow->GreenSlider,0.0);
muiSetVSValue(LCWindow->BlueSlider,0.0);
muiSetEnable(LCWindow->RedSlider,GL_FALSE);
muiSetEnable(LCWindow->GreenSlider,GL_FALSE);
muiSetEnable(LCWindow->BlueSlider,GL_FALSE);
}

/*:825*/
#line 13770 "MathView3D.w"

/*846:*/
#line 13806 "MathView3D.w"

{if(LCWindow->LNo>=0){
sprintf(LCWindow->ColorStr,"["DBL_FMT", "DBL_FMT", "DBL_FMT"]",
CurrentColor.r,
CurrentColor.g,
CurrentColor.b);
}
else
strcpy(LCWindow->ColorStr,"");
muiSetTBString(LCWindow->ColorBox,LCWindow->ColorStr);
}

/*:846*/
#line 13771 "MathView3D.w"

glutSetWindow(LCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(LCWindow->WinID);

/*:843*/
#line 13638 "MathView3D.w"

}
}

/*:832*/
#line 13579 "MathView3D.w"

/*834:*/
#line 13651 "MathView3D.w"

#pragma argsused
void TLLightSelected(muiObject*obj,enum muiReturnValue res)
{int save,what;

if(MUI_TEXTLIST_RETURN==res){
save= LCWindow->sLight;
what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13658 "MathView3D.w"
;
if(what>=LCWindow->LNo){
/*827:*/
#line 13541 "MathView3D.w"

muiSetEnable(LCWindow->RSlider,GL_FALSE);
muiSetEnable(LCWindow->ThetaSlider,GL_FALSE);
muiSetEnable(LCWindow->PhiSlider,GL_FALSE);
muiSetEnable(LCWindow->RedSlider,GL_FALSE);
muiSetEnable(LCWindow->GreenSlider,GL_FALSE);
muiSetEnable(LCWindow->BlueSlider,GL_FALSE);
muiSetEnable(LCWindow->PositionBox,GL_FALSE);
muiSetEnable(LCWindow->ColorBox,GL_FALSE);

/*:827*/
#line 13660 "MathView3D.w"

LCWindow->sLight= save;
return;
}
/*828:*/
#line 13551 "MathView3D.w"

if(!muiGetEnable(LCWindow->RSlider))
muiSetEnable(LCWindow->RSlider,GL_TRUE);
if(!muiGetEnable(LCWindow->ThetaSlider))
muiSetEnable(LCWindow->ThetaSlider,GL_TRUE);
if(!muiGetEnable(LCWindow->PhiSlider))
muiSetEnable(LCWindow->PhiSlider,GL_TRUE);
if(!muiGetEnable(LCWindow->RedSlider))
muiSetEnable(LCWindow->RedSlider,GL_TRUE);
if(!muiGetEnable(LCWindow->GreenSlider))
muiSetEnable(LCWindow->GreenSlider,GL_TRUE);
if(!muiGetEnable(LCWindow->BlueSlider))
muiSetEnable(LCWindow->BlueSlider,GL_TRUE);
if(!muiGetEnable(LCWindow->PositionBox))
muiSetEnable(LCWindow->PositionBox,GL_TRUE);
if(!muiGetEnable(LCWindow->ColorBox))
muiSetEnable(LCWindow->ColorBox,GL_TRUE);

/*:828*/
#line 13664 "MathView3D.w"

/*843:*/
#line 13767 "MathView3D.w"

/*826:*/
#line 13515 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13519 "MathView3D.w"
;
SphericalFromCartesian(
LCWindow->LightPosition+what,
&(LCWindow->PreviewR),
&(LCWindow->PreviewTheta),
&(LCWindow->PreviewPhi));
while(LCWindow->PreviewPhi<0.0)LCWindow->PreviewPhi+= TWO_PI;

muiSetEnable(LCWindow->RSlider,GL_TRUE);
muiSetEnable(LCWindow->ThetaSlider,GL_TRUE);
muiSetEnable(LCWindow->PhiSlider,GL_TRUE);
muiSetHSValue(LCWindow->RSlider,(LCWindow->PreviewR-1.0)/8.0);
muiSetVSValue(LCWindow->ThetaSlider,1.0-LCWindow->PreviewTheta/PI);
muiSetHSValue(LCWindow->PhiSlider,LCWindow->PreviewPhi/TWO_PI);
}
else{
muiSetEnable(LCWindow->RSlider,GL_FALSE);
muiSetEnable(LCWindow->ThetaSlider,GL_FALSE);
muiSetEnable(LCWindow->PhiSlider,GL_FALSE);

}

/*:826*/
#line 13768 "MathView3D.w"

/*844:*/
#line 13776 "MathView3D.w"

{
if(LCWindow->LNo>=0){
/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13779 "MathView3D.w"
;
/*845:*/
#line 13790 "MathView3D.w"

{double r,ct,st,cp,sp;

/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13793 "MathView3D.w"
;
ct= cos(LCWindow->PreviewTheta);st= sin(LCWindow->PreviewTheta);
cp= cos(LCWindow->PreviewPhi);sp= sin(LCWindow->PreviewPhi);
r= LCWindow->PreviewR;
CurrentPos.x= r*cp*st;
CurrentPos.y= r*sp*st;
CurrentPos.z= r*ct;
}

/*:845*/
#line 13780 "MathView3D.w"

sprintf(LCWindow->PositionStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
CurrentPos.x,CurrentPos.y,CurrentPos.z);
}
else
strcpy(LCWindow->PositionStr,"");
muiSetTBString(LCWindow->PositionBox,LCWindow->PositionStr);

}

/*:844*/
#line 13769 "MathView3D.w"

/*825:*/
#line 13493 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13497 "MathView3D.w"
;
muiSetEnable(LCWindow->RedSlider,GL_TRUE);
muiSetEnable(LCWindow->GreenSlider,GL_TRUE);
muiSetEnable(LCWindow->BlueSlider,GL_TRUE);

muiSetVSValue(LCWindow->RedSlider,LCWindow->LightColor[what].r);
muiSetVSValue(LCWindow->GreenSlider,LCWindow->LightColor[what].g);
muiSetVSValue(LCWindow->BlueSlider,LCWindow->LightColor[what].b);
}
else{
muiSetVSValue(LCWindow->RedSlider,0.0);
muiSetVSValue(LCWindow->GreenSlider,0.0);
muiSetVSValue(LCWindow->BlueSlider,0.0);
muiSetEnable(LCWindow->RedSlider,GL_FALSE);
muiSetEnable(LCWindow->GreenSlider,GL_FALSE);
muiSetEnable(LCWindow->BlueSlider,GL_FALSE);
}

/*:825*/
#line 13770 "MathView3D.w"

/*846:*/
#line 13806 "MathView3D.w"

{if(LCWindow->LNo>=0){
sprintf(LCWindow->ColorStr,"["DBL_FMT", "DBL_FMT", "DBL_FMT"]",
CurrentColor.r,
CurrentColor.g,
CurrentColor.b);
}
else
strcpy(LCWindow->ColorStr,"");
muiSetTBString(LCWindow->ColorBox,LCWindow->ColorStr);
}

/*:846*/
#line 13771 "MathView3D.w"

glutSetWindow(LCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(LCWindow->WinID);

/*:843*/
#line 13665 "MathView3D.w"

}
}

/*:834*/
#line 13580 "MathView3D.w"

/*833:*/
#line 13642 "MathView3D.w"

#pragma argsused
void SwitchSolid(muiObject*obj,enum muiReturnValue res)
{
if(MUI_BUTTON_PRESS==res){
/*843:*/
#line 13767 "MathView3D.w"

/*826:*/
#line 13515 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13519 "MathView3D.w"
;
SphericalFromCartesian(
LCWindow->LightPosition+what,
&(LCWindow->PreviewR),
&(LCWindow->PreviewTheta),
&(LCWindow->PreviewPhi));
while(LCWindow->PreviewPhi<0.0)LCWindow->PreviewPhi+= TWO_PI;

muiSetEnable(LCWindow->RSlider,GL_TRUE);
muiSetEnable(LCWindow->ThetaSlider,GL_TRUE);
muiSetEnable(LCWindow->PhiSlider,GL_TRUE);
muiSetHSValue(LCWindow->RSlider,(LCWindow->PreviewR-1.0)/8.0);
muiSetVSValue(LCWindow->ThetaSlider,1.0-LCWindow->PreviewTheta/PI);
muiSetHSValue(LCWindow->PhiSlider,LCWindow->PreviewPhi/TWO_PI);
}
else{
muiSetEnable(LCWindow->RSlider,GL_FALSE);
muiSetEnable(LCWindow->ThetaSlider,GL_FALSE);
muiSetEnable(LCWindow->PhiSlider,GL_FALSE);

}

/*:826*/
#line 13768 "MathView3D.w"

/*844:*/
#line 13776 "MathView3D.w"

{
if(LCWindow->LNo>=0){
/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13779 "MathView3D.w"
;
/*845:*/
#line 13790 "MathView3D.w"

{double r,ct,st,cp,sp;

/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13793 "MathView3D.w"
;
ct= cos(LCWindow->PreviewTheta);st= sin(LCWindow->PreviewTheta);
cp= cos(LCWindow->PreviewPhi);sp= sin(LCWindow->PreviewPhi);
r= LCWindow->PreviewR;
CurrentPos.x= r*cp*st;
CurrentPos.y= r*sp*st;
CurrentPos.z= r*ct;
}

/*:845*/
#line 13780 "MathView3D.w"

sprintf(LCWindow->PositionStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
CurrentPos.x,CurrentPos.y,CurrentPos.z);
}
else
strcpy(LCWindow->PositionStr,"");
muiSetTBString(LCWindow->PositionBox,LCWindow->PositionStr);

}

/*:844*/
#line 13769 "MathView3D.w"

/*825:*/
#line 13493 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13497 "MathView3D.w"
;
muiSetEnable(LCWindow->RedSlider,GL_TRUE);
muiSetEnable(LCWindow->GreenSlider,GL_TRUE);
muiSetEnable(LCWindow->BlueSlider,GL_TRUE);

muiSetVSValue(LCWindow->RedSlider,LCWindow->LightColor[what].r);
muiSetVSValue(LCWindow->GreenSlider,LCWindow->LightColor[what].g);
muiSetVSValue(LCWindow->BlueSlider,LCWindow->LightColor[what].b);
}
else{
muiSetVSValue(LCWindow->RedSlider,0.0);
muiSetVSValue(LCWindow->GreenSlider,0.0);
muiSetVSValue(LCWindow->BlueSlider,0.0);
muiSetEnable(LCWindow->RedSlider,GL_FALSE);
muiSetEnable(LCWindow->GreenSlider,GL_FALSE);
muiSetEnable(LCWindow->BlueSlider,GL_FALSE);
}

/*:825*/
#line 13770 "MathView3D.w"

/*846:*/
#line 13806 "MathView3D.w"

{if(LCWindow->LNo>=0){
sprintf(LCWindow->ColorStr,"["DBL_FMT", "DBL_FMT", "DBL_FMT"]",
CurrentColor.r,
CurrentColor.g,
CurrentColor.b);
}
else
strcpy(LCWindow->ColorStr,"");
muiSetTBString(LCWindow->ColorBox,LCWindow->ColorStr);
}

/*:846*/
#line 13771 "MathView3D.w"

glutSetWindow(LCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(LCWindow->WinID);

/*:843*/
#line 13647 "MathView3D.w"

}
}

/*:833*/
#line 13581 "MathView3D.w"

/*841:*/
#line 13738 "MathView3D.w"

void ManualLightPosBox(muiObject*obj,enum muiReturnValue ret)
{double xc,yc,zc;

if(MUI_TEXTBOX_RETURN==ret){
/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13743 "MathView3D.w"
;
/*803:*/
#line 13116 "MathView3D.w"

{char txtStr[512],*l;
int i,n,fail1,fail2,fail3;

strcpy(txtStr,muiGetTBString(obj));
n= strlen(txtStr);
for(i= 0;i<n;i++)
if('{'==txtStr[i]||'}'==txtStr[i]||'['==txtStr[i]||']'==txtStr[i])txtStr[i]= ' ';
l= txtStr;
xc= ScanDouble(&l,0.0,&fail1);
yc= ScanDouble(&l,0.0,&fail2);
zc= ScanDouble(&l,0.0,&fail3);
if(fail1||fail2||fail3)return;
}

/*:803*/
#line 13744 "MathView3D.w"

CurrentPos.x= xc;
CurrentPos.y= yc;
CurrentPos.z= zc;
/*843:*/
#line 13767 "MathView3D.w"

/*826:*/
#line 13515 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13519 "MathView3D.w"
;
SphericalFromCartesian(
LCWindow->LightPosition+what,
&(LCWindow->PreviewR),
&(LCWindow->PreviewTheta),
&(LCWindow->PreviewPhi));
while(LCWindow->PreviewPhi<0.0)LCWindow->PreviewPhi+= TWO_PI;

muiSetEnable(LCWindow->RSlider,GL_TRUE);
muiSetEnable(LCWindow->ThetaSlider,GL_TRUE);
muiSetEnable(LCWindow->PhiSlider,GL_TRUE);
muiSetHSValue(LCWindow->RSlider,(LCWindow->PreviewR-1.0)/8.0);
muiSetVSValue(LCWindow->ThetaSlider,1.0-LCWindow->PreviewTheta/PI);
muiSetHSValue(LCWindow->PhiSlider,LCWindow->PreviewPhi/TWO_PI);
}
else{
muiSetEnable(LCWindow->RSlider,GL_FALSE);
muiSetEnable(LCWindow->ThetaSlider,GL_FALSE);
muiSetEnable(LCWindow->PhiSlider,GL_FALSE);

}

/*:826*/
#line 13768 "MathView3D.w"

/*844:*/
#line 13776 "MathView3D.w"

{
if(LCWindow->LNo>=0){
/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13779 "MathView3D.w"
;
/*845:*/
#line 13790 "MathView3D.w"

{double r,ct,st,cp,sp;

/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13793 "MathView3D.w"
;
ct= cos(LCWindow->PreviewTheta);st= sin(LCWindow->PreviewTheta);
cp= cos(LCWindow->PreviewPhi);sp= sin(LCWindow->PreviewPhi);
r= LCWindow->PreviewR;
CurrentPos.x= r*cp*st;
CurrentPos.y= r*sp*st;
CurrentPos.z= r*ct;
}

/*:845*/
#line 13780 "MathView3D.w"

sprintf(LCWindow->PositionStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
CurrentPos.x,CurrentPos.y,CurrentPos.z);
}
else
strcpy(LCWindow->PositionStr,"");
muiSetTBString(LCWindow->PositionBox,LCWindow->PositionStr);

}

/*:844*/
#line 13769 "MathView3D.w"

/*825:*/
#line 13493 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13497 "MathView3D.w"
;
muiSetEnable(LCWindow->RedSlider,GL_TRUE);
muiSetEnable(LCWindow->GreenSlider,GL_TRUE);
muiSetEnable(LCWindow->BlueSlider,GL_TRUE);

muiSetVSValue(LCWindow->RedSlider,LCWindow->LightColor[what].r);
muiSetVSValue(LCWindow->GreenSlider,LCWindow->LightColor[what].g);
muiSetVSValue(LCWindow->BlueSlider,LCWindow->LightColor[what].b);
}
else{
muiSetVSValue(LCWindow->RedSlider,0.0);
muiSetVSValue(LCWindow->GreenSlider,0.0);
muiSetVSValue(LCWindow->BlueSlider,0.0);
muiSetEnable(LCWindow->RedSlider,GL_FALSE);
muiSetEnable(LCWindow->GreenSlider,GL_FALSE);
muiSetEnable(LCWindow->BlueSlider,GL_FALSE);
}

/*:825*/
#line 13770 "MathView3D.w"

/*846:*/
#line 13806 "MathView3D.w"

{if(LCWindow->LNo>=0){
sprintf(LCWindow->ColorStr,"["DBL_FMT", "DBL_FMT", "DBL_FMT"]",
CurrentColor.r,
CurrentColor.g,
CurrentColor.b);
}
else
strcpy(LCWindow->ColorStr,"");
muiSetTBString(LCWindow->ColorBox,LCWindow->ColorStr);
}

/*:846*/
#line 13771 "MathView3D.w"

glutSetWindow(LCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(LCWindow->WinID);

/*:843*/
#line 13748 "MathView3D.w"

}
}

/*:841*/
#line 13582 "MathView3D.w"

/*842:*/
#line 13752 "MathView3D.w"

void ManualLightColorBox(muiObject*obj,enum muiReturnValue ret)
{double xc,yc,zc;

if(MUI_TEXTBOX_RETURN==ret){
/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13757 "MathView3D.w"
;
/*803:*/
#line 13116 "MathView3D.w"

{char txtStr[512],*l;
int i,n,fail1,fail2,fail3;

strcpy(txtStr,muiGetTBString(obj));
n= strlen(txtStr);
for(i= 0;i<n;i++)
if('{'==txtStr[i]||'}'==txtStr[i]||'['==txtStr[i]||']'==txtStr[i])txtStr[i]= ' ';
l= txtStr;
xc= ScanDouble(&l,0.0,&fail1);
yc= ScanDouble(&l,0.0,&fail2);
zc= ScanDouble(&l,0.0,&fail3);
if(fail1||fail2||fail3)return;
}

/*:803*/
#line 13758 "MathView3D.w"

CurrentColor.r= xc;
CurrentColor.g= yc;
CurrentColor.b= zc;
/*843:*/
#line 13767 "MathView3D.w"

/*826:*/
#line 13515 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13519 "MathView3D.w"
;
SphericalFromCartesian(
LCWindow->LightPosition+what,
&(LCWindow->PreviewR),
&(LCWindow->PreviewTheta),
&(LCWindow->PreviewPhi));
while(LCWindow->PreviewPhi<0.0)LCWindow->PreviewPhi+= TWO_PI;

muiSetEnable(LCWindow->RSlider,GL_TRUE);
muiSetEnable(LCWindow->ThetaSlider,GL_TRUE);
muiSetEnable(LCWindow->PhiSlider,GL_TRUE);
muiSetHSValue(LCWindow->RSlider,(LCWindow->PreviewR-1.0)/8.0);
muiSetVSValue(LCWindow->ThetaSlider,1.0-LCWindow->PreviewTheta/PI);
muiSetHSValue(LCWindow->PhiSlider,LCWindow->PreviewPhi/TWO_PI);
}
else{
muiSetEnable(LCWindow->RSlider,GL_FALSE);
muiSetEnable(LCWindow->ThetaSlider,GL_FALSE);
muiSetEnable(LCWindow->PhiSlider,GL_FALSE);

}

/*:826*/
#line 13768 "MathView3D.w"

/*844:*/
#line 13776 "MathView3D.w"

{
if(LCWindow->LNo>=0){
/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13779 "MathView3D.w"
;
/*845:*/
#line 13790 "MathView3D.w"

{double r,ct,st,cp,sp;

/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13793 "MathView3D.w"
;
ct= cos(LCWindow->PreviewTheta);st= sin(LCWindow->PreviewTheta);
cp= cos(LCWindow->PreviewPhi);sp= sin(LCWindow->PreviewPhi);
r= LCWindow->PreviewR;
CurrentPos.x= r*cp*st;
CurrentPos.y= r*sp*st;
CurrentPos.z= r*ct;
}

/*:845*/
#line 13780 "MathView3D.w"

sprintf(LCWindow->PositionStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
CurrentPos.x,CurrentPos.y,CurrentPos.z);
}
else
strcpy(LCWindow->PositionStr,"");
muiSetTBString(LCWindow->PositionBox,LCWindow->PositionStr);

}

/*:844*/
#line 13769 "MathView3D.w"

/*825:*/
#line 13493 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13497 "MathView3D.w"
;
muiSetEnable(LCWindow->RedSlider,GL_TRUE);
muiSetEnable(LCWindow->GreenSlider,GL_TRUE);
muiSetEnable(LCWindow->BlueSlider,GL_TRUE);

muiSetVSValue(LCWindow->RedSlider,LCWindow->LightColor[what].r);
muiSetVSValue(LCWindow->GreenSlider,LCWindow->LightColor[what].g);
muiSetVSValue(LCWindow->BlueSlider,LCWindow->LightColor[what].b);
}
else{
muiSetVSValue(LCWindow->RedSlider,0.0);
muiSetVSValue(LCWindow->GreenSlider,0.0);
muiSetVSValue(LCWindow->BlueSlider,0.0);
muiSetEnable(LCWindow->RedSlider,GL_FALSE);
muiSetEnable(LCWindow->GreenSlider,GL_FALSE);
muiSetEnable(LCWindow->BlueSlider,GL_FALSE);
}

/*:825*/
#line 13770 "MathView3D.w"

/*846:*/
#line 13806 "MathView3D.w"

{if(LCWindow->LNo>=0){
sprintf(LCWindow->ColorStr,"["DBL_FMT", "DBL_FMT", "DBL_FMT"]",
CurrentColor.r,
CurrentColor.g,
CurrentColor.b);
}
else
strcpy(LCWindow->ColorStr,"");
muiSetTBString(LCWindow->ColorBox,LCWindow->ColorStr);
}

/*:846*/
#line 13771 "MathView3D.w"

glutSetWindow(LCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(LCWindow->WinID);

/*:843*/
#line 13762 "MathView3D.w"

}
}


/*:842*/
#line 13583 "MathView3D.w"


/*:829*/
#line 13165 "MathView3D.w"

/*811:*/
#line 13257 "MathView3D.w"

void CopyToLCWindow(void)
{int i;

LCWindow->LNo= LightNo;
for(i= 0;i<LightNo;i++){
LCWindow->LightPosition[i]= LightPosition[i];
LCWindow->LightColor[i]= LightColor[i];
LCWindow->LNames[i]= LCWindow->LNameN[i];
}
LCWindow->LNames[LightNo]= NULL;
muiSetTLStrings(LCWindow->LTextList,(char**)(LCWindow->LNames));
/*843:*/
#line 13767 "MathView3D.w"

/*826:*/
#line 13515 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13519 "MathView3D.w"
;
SphericalFromCartesian(
LCWindow->LightPosition+what,
&(LCWindow->PreviewR),
&(LCWindow->PreviewTheta),
&(LCWindow->PreviewPhi));
while(LCWindow->PreviewPhi<0.0)LCWindow->PreviewPhi+= TWO_PI;

muiSetEnable(LCWindow->RSlider,GL_TRUE);
muiSetEnable(LCWindow->ThetaSlider,GL_TRUE);
muiSetEnable(LCWindow->PhiSlider,GL_TRUE);
muiSetHSValue(LCWindow->RSlider,(LCWindow->PreviewR-1.0)/8.0);
muiSetVSValue(LCWindow->ThetaSlider,1.0-LCWindow->PreviewTheta/PI);
muiSetHSValue(LCWindow->PhiSlider,LCWindow->PreviewPhi/TWO_PI);
}
else{
muiSetEnable(LCWindow->RSlider,GL_FALSE);
muiSetEnable(LCWindow->ThetaSlider,GL_FALSE);
muiSetEnable(LCWindow->PhiSlider,GL_FALSE);

}

/*:826*/
#line 13768 "MathView3D.w"

/*844:*/
#line 13776 "MathView3D.w"

{
if(LCWindow->LNo>=0){
/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13779 "MathView3D.w"
;
/*845:*/
#line 13790 "MathView3D.w"

{double r,ct,st,cp,sp;

/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13793 "MathView3D.w"
;
ct= cos(LCWindow->PreviewTheta);st= sin(LCWindow->PreviewTheta);
cp= cos(LCWindow->PreviewPhi);sp= sin(LCWindow->PreviewPhi);
r= LCWindow->PreviewR;
CurrentPos.x= r*cp*st;
CurrentPos.y= r*sp*st;
CurrentPos.z= r*ct;
}

/*:845*/
#line 13780 "MathView3D.w"

sprintf(LCWindow->PositionStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
CurrentPos.x,CurrentPos.y,CurrentPos.z);
}
else
strcpy(LCWindow->PositionStr,"");
muiSetTBString(LCWindow->PositionBox,LCWindow->PositionStr);

}

/*:844*/
#line 13769 "MathView3D.w"

/*825:*/
#line 13493 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13497 "MathView3D.w"
;
muiSetEnable(LCWindow->RedSlider,GL_TRUE);
muiSetEnable(LCWindow->GreenSlider,GL_TRUE);
muiSetEnable(LCWindow->BlueSlider,GL_TRUE);

muiSetVSValue(LCWindow->RedSlider,LCWindow->LightColor[what].r);
muiSetVSValue(LCWindow->GreenSlider,LCWindow->LightColor[what].g);
muiSetVSValue(LCWindow->BlueSlider,LCWindow->LightColor[what].b);
}
else{
muiSetVSValue(LCWindow->RedSlider,0.0);
muiSetVSValue(LCWindow->GreenSlider,0.0);
muiSetVSValue(LCWindow->BlueSlider,0.0);
muiSetEnable(LCWindow->RedSlider,GL_FALSE);
muiSetEnable(LCWindow->GreenSlider,GL_FALSE);
muiSetEnable(LCWindow->BlueSlider,GL_FALSE);
}

/*:825*/
#line 13770 "MathView3D.w"

/*846:*/
#line 13806 "MathView3D.w"

{if(LCWindow->LNo>=0){
sprintf(LCWindow->ColorStr,"["DBL_FMT", "DBL_FMT", "DBL_FMT"]",
CurrentColor.r,
CurrentColor.g,
CurrentColor.b);
}
else
strcpy(LCWindow->ColorStr,"");
muiSetTBString(LCWindow->ColorBox,LCWindow->ColorStr);
}

/*:846*/
#line 13771 "MathView3D.w"

glutSetWindow(LCWindow->PreviewWinID);
glutPostRedisplay();
glutSetWindow(LCWindow->WinID);

/*:843*/
#line 13269 "MathView3D.w"

#if !HAVE_GLUT_SUBWINDOW
MacPreviewPlacement(LCWindow->WinID,LCWindow->PreviewWinID,LCWindow->muiID);
#endif

}

/*:811*/
#line 13166 "MathView3D.w"

/*814:*/
#line 13291 "MathView3D.w"

void DisplayLightPreview(void)
{
Point3D*p;

glPushMatrix();

glViewport(0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
glClearColor(0.0,0.0,0.0,0.0);
glColor3f(1.0,1.0,1.0);

glMatrixMode(GL_PROJECTION);
glLoadIdentity();
gluPerspective(MVViewAngle,
1.0,
MVNearPlane,MVFarPlane);
glMatrixMode(GL_MODELVIEW);
gluLookAt(ViewPoint.x,
ViewPoint.y,
ViewPoint.z,
0.0,0.0,0.0,
0.0,0.0,1.0);
glScaled(BoxRatios.x,
BoxRatios.y,
BoxRatios.z);
glEnable(GL_DEPTH_TEST);
glutWireCube(2.0);
glShadeModel(GL_SMOOTH);
/*815:*/
#line 13331 "MathView3D.w"

if(LCWindow->LNo>0&&
LCWindow->sLight>=0&&
LCWindow->sLight<LCWindow->LNo){
glLineWidth(3.0);
glBegin(GL_LINES);
glVertex3d(0.0,0.0,0.0);
p= LCWindow->LightPosition+LCWindow->sLight;
glVertex3d(p->x,p->y,p->z);
glEnd();
glLineWidth(1.0);
}

/*:815*/
#line 13320 "MathView3D.w"

/*847:*/
#line 13818 "MathView3D.w"

DefineLightSources(
LCWindow->LightPosition,
LCWindow->LightColor,
LCWindow->LNo,
AmbientColor);

/*:847*/
#line 13321 "MathView3D.w"

DefineMaterial(INHERITE_COLOR,White);
glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
/*816:*/
#line 13344 "MathView3D.w"

glPushMatrix();
if(muiGetActive(LCWindow->SphereRButton))
glutSolidSphere(1.0,36,18);
else if(muiGetActive(LCWindow->TorusRButton))
glutSolidTorus(0.25,0.75,18,36);
else if(muiGetActive(LCWindow->DodecahedronRButton)){
glPushMatrix();
glScaled(0.58,0.58,0.58);
glutSolidDodecahedron();
glPopMatrix();
}
else if(muiGetActive(LCWindow->CubeRButton))
glutSolidCube(1.73);
else
glutSolidSphere(1.0,18,9);
glPopMatrix();

/*:816*/
#line 13324 "MathView3D.w"


glDisable(GL_LIGHTING);
glutSwapBuffers();
glPopMatrix();
}

/*:814*/
#line 13167 "MathView3D.w"

/*818:*/
#line 13367 "MathView3D.w"

void LCReshape(int w,int h)
{
if(w!=LCWindow->CWinX||h!=LCWindow->CWinY)
glutReshapeWindow(LCWindow->CWinX,LCWindow->CWinY);
}

/*:818*/
#line 13168 "MathView3D.w"

/*810:*/
#line 13213 "MathView3D.w"

LightCntrlWin*CreateLightControl(void)
{LightCntrlWin*vc;
int xPos,yPos,xEnd,i;
static char*lname[]= {"  Light[[1]]",
"  Light[[2]]",
"  Light[[3]]",
"  Light[[4]]",
"  Light[[5]]",
"  Light[[6]]",
"  Light[[7]]",
"  Light[[8]]",NULL};

LCWindow= vc= (LightCntrlWin*)malloc(sizeof(LightCntrlWin));
if(NULL==vc)return NULL;
vc->CWinX= 400;
vc->CWinY= 550;
glutInitWindowSize(vc->CWinX,vc->CWinY);
vc->WinID= glutCreateWindow("Light Control");
vc->PreviewR= 4.4;
vc->PreviewTheta= 1.13;
vc->PreviewPhi= 1.15;
vc->LNo= 0;
vc->sLight= -1;
for(i= 0;i<=MAX_LIGHT;i++)
vc->LNameN[i]= lname[i];
muiInit();
muiNewUIList(vc->WinID);
/*819:*/
#line 13374 "MathView3D.w"

xPos= 115;
yPos= vc->CWinY-40;
vc->NewLSButton= muiNewButton(xPos,xPos+100,yPos,yPos+30);
muiLoadButton(vc->NewLSButton,"New Light");
muiSetCallback(vc->NewLSButton,NewLightPressed);
yPos-= 35;
vc->DelLSButton= muiNewButton(xPos,xPos+100,yPos,yPos+30);
muiLoadButton(vc->DelLSButton,"Delete Light");
muiSetCallback(vc->DelLSButton,DeleteLightPressed);

{int yOffset= 0;

#if !HAVE_GLUT_SUBWINDOW
yOffset+= 90;
#endif
xPos= 10;
yPos= 100;
vc->ApplyButton= muiNewButton(xPos,yPos,10+yOffset,35+yOffset);
muiLoadButton(vc->ApplyButton,"Apply");
muiSetCallback(vc->ApplyButton,ApplyLightSources);
xPos+= 100;
yPos+= 100;
vc->HideButton= muiNewButton(xPos,yPos,10+yOffset,35+yOffset);
muiLoadButton(vc->HideButton,"Hide");
muiSetCallback(vc->HideButton,HidePressed);
xPos+= 170;
yPos+= 170;
if(!MathLinkMode){
vc->QuitButton= muiNewButton(xPos,yPos,10+yOffset,35+yOffset);
muiLoadButton(vc->QuitButton,"Quit");
muiSetCallback(vc->QuitButton,QuitPressed);
}
}

/*:819*/
#line 13241 "MathView3D.w"

/*820:*/
#line 13409 "MathView3D.w"

xPos= 250;
yPos= vc->CWinY-25;
muiNewBoldLabel(xPos,yPos,"Solid for Preview");
yPos-= 30;
vc->CubeRButton= muiNewRadioButton(xPos,yPos);
muiLoadButton(vc->CubeRButton,"Cube");
yPos-= 30;
vc->DodecahedronRButton= muiNewRadioButton(xPos,yPos);
muiLoadButton(vc->DodecahedronRButton,"Dodecahedron");
yPos-= 30;
vc->TorusRButton= muiNewRadioButton(xPos,yPos);
muiLoadButton(vc->TorusRButton,"Torus");
yPos-= 30;
vc->SphereRButton= muiNewRadioButton(xPos,yPos);
muiLoadButton(vc->SphereRButton,"Sphere");

muiLinkButtons(vc->CubeRButton,vc->DodecahedronRButton);
muiLinkButtons(vc->DodecahedronRButton,vc->TorusRButton);
muiLinkButtons(vc->TorusRButton,vc->SphereRButton);
muiSetActive(vc->SphereRButton,GL_TRUE);
muiSetCallback(vc->CubeRButton,SwitchSolid);
muiSetCallback(vc->DodecahedronRButton,SwitchSolid);
muiSetCallback(vc->TorusRButton,SwitchSolid);
muiSetCallback(vc->SphereRButton,SwitchSolid);

/*:820*/
#line 13242 "MathView3D.w"

/*821:*/
#line 13435 "MathView3D.w"

xPos= 10;
xEnd= vc->CWinX-10;
yPos= vc->CWinY-150;
vc->LTextList= muiNewTextList(xPos,yPos,xPos+100,8);
muiSetCallback(vc->LTextList,TLLightSelected);

yPos-= 20;
(void*)muiNewBoldLabel(xPos,yPos-5,"Position -> ");
vc->PositionBox= muiNewTextbox(xPos+120,xEnd,yPos-15);
muiSetCallback(vc->PositionBox,ManualLightPosBox);
/*844:*/
#line 13776 "MathView3D.w"

{
if(LCWindow->LNo>=0){
/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13779 "MathView3D.w"
;
/*845:*/
#line 13790 "MathView3D.w"

{double r,ct,st,cp,sp;

/*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13793 "MathView3D.w"
;
ct= cos(LCWindow->PreviewTheta);st= sin(LCWindow->PreviewTheta);
cp= cos(LCWindow->PreviewPhi);sp= sin(LCWindow->PreviewPhi);
r= LCWindow->PreviewR;
CurrentPos.x= r*cp*st;
CurrentPos.y= r*sp*st;
CurrentPos.z= r*ct;
}

/*:845*/
#line 13780 "MathView3D.w"

sprintf(LCWindow->PositionStr,"{"DBL_FMT", "DBL_FMT", "DBL_FMT"}",
CurrentPos.x,CurrentPos.y,CurrentPos.z);
}
else
strcpy(LCWindow->PositionStr,"");
muiSetTBString(LCWindow->PositionBox,LCWindow->PositionStr);

}

/*:844*/
#line 13446 "MathView3D.w"

yPos-= 30;
(void*)muiNewBoldLabel(xPos,yPos-5,"Color -> RGBColor");
vc->ColorBox= muiNewTextbox(xPos+120,xEnd,yPos-15);
muiSetCallback(vc->ColorBox,ManualLightColorBox);
/*846:*/
#line 13806 "MathView3D.w"

{if(LCWindow->LNo>=0){
sprintf(LCWindow->ColorStr,"["DBL_FMT", "DBL_FMT", "DBL_FMT"]",
CurrentColor.r,
CurrentColor.g,
CurrentColor.b);
}
else
strcpy(LCWindow->ColorStr,"");
muiSetTBString(LCWindow->ColorBox,LCWindow->ColorStr);
}

/*:846*/
#line 13451 "MathView3D.w"



/*:821*/
#line 13243 "MathView3D.w"

/*822:*/
#line 13454 "MathView3D.w"

(void*)muiNewLabel(248,303,"theta");
vc->ThetaSlider= muiNewVSlider(250,95,295,195,5);
muiSetCallback(vc->ThetaSlider,LThetaSliderMove);
muiSetVSValue(vc->ThetaSlider,1.0-vc->PreviewTheta/PI);
#if HAVE_GLUT_SUBWINDOW
xPos= 10;
yPos= 55;
#else
xPos= 10;
yPos= vc->CWinY-295;
#endif
(void*)muiNewLabel(xPos,yPos,"r");
vc->RSlider= muiNewHSlider(xPos+30,yPos-5,240,200,10);
muiSetCallback(vc->RSlider,LRSliderMove);
muiSetHSValue(vc->RSlider,(vc->PreviewR-1.0)/8.0);
yPos+= 25;
(void*)muiNewLabel(xPos,yPos,"phi");
vc->PhiSlider= muiNewHSlider(xPos+30,yPos-5,240,200,10);
muiSetCallback(vc->PhiSlider,LPhiSliderMove);
muiSetHSValue(vc->PhiSlider,vc->PreviewPhi/TWO_PI);

/*:822*/
#line 13244 "MathView3D.w"

/*823:*/
#line 13476 "MathView3D.w"


(void*)muiNewLabel(290,303,"Red");
vc->RedSlider= muiNewVSlider(295,95,295,195,5);
muiSetCallback(vc->RedSlider,RedSliderMove);

(void*)muiNewLabel(315,303,"Green");
vc->GreenSlider= muiNewVSlider(325,95,295,195,5);
muiSetCallback(vc->GreenSlider,GreenSliderMove);

(void*)muiNewLabel(355,303,"Blue");
vc->BlueSlider= muiNewVSlider(355,95,295,195,5);
muiSetCallback(vc->BlueSlider,BlueSliderMove);

/*:823*/
#line 13245 "MathView3D.w"

muiAttachUIList(vc->WinID);
vc->muiID= muiGetActiveUIList();

/*825:*/
#line 13493 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13497 "MathView3D.w"
;
muiSetEnable(LCWindow->RedSlider,GL_TRUE);
muiSetEnable(LCWindow->GreenSlider,GL_TRUE);
muiSetEnable(LCWindow->BlueSlider,GL_TRUE);

muiSetVSValue(LCWindow->RedSlider,LCWindow->LightColor[what].r);
muiSetVSValue(LCWindow->GreenSlider,LCWindow->LightColor[what].g);
muiSetVSValue(LCWindow->BlueSlider,LCWindow->LightColor[what].b);
}
else{
muiSetVSValue(LCWindow->RedSlider,0.0);
muiSetVSValue(LCWindow->GreenSlider,0.0);
muiSetVSValue(LCWindow->BlueSlider,0.0);
muiSetEnable(LCWindow->RedSlider,GL_FALSE);
muiSetEnable(LCWindow->GreenSlider,GL_FALSE);
muiSetEnable(LCWindow->BlueSlider,GL_FALSE);
}

/*:825*/
#line 13249 "MathView3D.w"

/*826:*/
#line 13515 "MathView3D.w"

if(LCWindow->LNo>0){
int what;

what= /*824:*/
#line 13490 "MathView3D.w"

LCWindow->sLight= muiGetTLSelectedItem(LCWindow->LTextList)

/*:824*/
#line 13519 "MathView3D.w"
;
SphericalFromCartesian(
LCWindow->LightPosition+what,
&(LCWindow->PreviewR),
&(LCWindow->PreviewTheta),
&(LCWindow->PreviewPhi));
while(LCWindow->PreviewPhi<0.0)LCWindow->PreviewPhi+= TWO_PI;

muiSetEnable(LCWindow->RSlider,GL_TRUE);
muiSetEnable(LCWindow->ThetaSlider,GL_TRUE);
muiSetEnable(LCWindow->PhiSlider,GL_TRUE);
muiSetHSValue(LCWindow->RSlider,(LCWindow->PreviewR-1.0)/8.0);
muiSetVSValue(LCWindow->ThetaSlider,1.0-LCWindow->PreviewTheta/PI);
muiSetHSValue(LCWindow->PhiSlider,LCWindow->PreviewPhi/TWO_PI);
}
else{
muiSetEnable(LCWindow->RSlider,GL_FALSE);
muiSetEnable(LCWindow->ThetaSlider,GL_FALSE);
muiSetEnable(LCWindow->PhiSlider,GL_FALSE);

}

/*:826*/
#line 13250 "MathView3D.w"

/*812:*/
#line 13276 "MathView3D.w"

#if HAVE_GLUT_SUBWINDOW
vc->PreviewWinID= glutCreateSubWindow(vc->WinID,40,vc->CWinY-305,200,200);
#else
glutInitWindowSize(200,200);
vc->PreviewWinID= glutCreateWindow("Preview");
#endif
glutDisplayFunc(DisplayLightPreview);
DisplayLightPreview();
glutHideWindow();


/*:812*/
#line 13251 "MathView3D.w"

glutSetWindow(vc->WinID);
glutReshapeFunc(LCReshape);
return vc;
}

/*:810*/
#line 13169 "MathView3D.w"


/*:807*/
#line 90 "MathView3D.w"

/*854:*/
#line 13924 "MathView3D.w"

/*857:*/
#line 13944 "MathView3D.w"

void DisplayAxisPreview(void)
{
glPushMatrix();
glPushAttrib(GL_ALL_ATTRIB_BITS);
glViewport(0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);


glColor3f(1.0,1.0,1.0);

glMatrixMode(GL_PROJECTION);
glLoadIdentity();
gluPerspective(MVViewAngle,
1.0,
MVNearPlane,MVFarPlane);
glMatrixMode(GL_MODELVIEW);
gluLookAt(ViewPoint.x,
ViewPoint.y,
ViewPoint.z,
0.0,0.0,0.0,
0.0,0.0,1.0);
ScalingToUnitCube(&BoundP1,&BoundP2);
glTranslated(-CenterOfScene.x,-CenterOfScene.y,-CenterOfScene.z);

glEnable(GL_DEPTH_TEST);
DefineLightSources(LightPosition,LightColor,LightNo,AmbientColor);
DrawAxis(AxisWindow->AxisType,
&AxisWindow->Origin,&BoundP1,&BoundP2,
AxisWindow->XLabel,AxisWindow->YLabel,AxisWindow->ZLabel,
AxisWindow->FontScale,AxisWindow->NumericLabels);
glShadeModel(GL_SMOOTH);

glutSwapBuffers();
glPopAttrib();
glPopMatrix();
}

/*:857*/
#line 13925 "MathView3D.w"

/*853:*/
#line 13895 "MathView3D.w"

AxisCtrlWin*CreateAxisControl(void)
{AxisCtrlWin*vc;
int xPos,yPos,xEnd;

AxisWindow= vc= (AxisCtrlWin*)malloc(sizeof(AxisCtrlWin));
if(NULL==vc)return NULL;
vc->CWinX= 400;
vc->CWinY= 550;
vc->FontScale= 0.25;
glutInitWindowSize(vc->CWinX,vc->CWinY);
vc->WinID= glutCreateWindow("Axis Editor");
muiInit();
muiNewUIList(vc->WinID);
/*877:*/
#line 14313 "MathView3D.w"

{int yOffset= 0;

xPos= 10;
yPos= 100;
#if !HAVE_GLUT_SUBWINDOW
yOffset+= 90;
#endif
vc->ApplyButton= muiNewButton(xPos,yPos,10+yOffset,35+yOffset);
muiLoadButton(vc->ApplyButton,"Apply");
muiSetCallback(vc->ApplyButton,ApplyAxis);

xPos+= 100;
yPos+= 100;
vc->HideButton= muiNewButton(xPos,yPos,10+yOffset,35+yOffset);
muiLoadButton(vc->HideButton,"Hide");
muiSetCallback(vc->HideButton,HidePressed);
if(!MathLinkMode){
xPos+= 170;
yPos+= 170;
vc->QuitButton= muiNewButton(xPos,yPos,10+yOffset,35+yOffset);
muiLoadButton(vc->QuitButton,"Quit");
muiSetCallback(vc->QuitButton,QuitPressed);
}
}


/*:877*/
#line 13909 "MathView3D.w"

/*878:*/
#line 14340 "MathView3D.w"

xPos= 10;
yPos= vc->CWinY-25;
muiNewBoldLabel(xPos,yPos,"Axis Type");
yPos-= 30;
vc->NoneRButton= muiNewRadioButton(xPos,yPos);
muiLoadButton(vc->NoneRButton,"None");
yPos-= 30;
vc->CrossRButton= muiNewRadioButton(xPos,yPos);
muiLoadButton(vc->CrossRButton,"Cross");
yPos-= 30;
vc->BoxRButton= muiNewRadioButton(xPos,yPos);
muiLoadButton(vc->BoxRButton,"Box");
yPos-= 30;
vc->PlaneRButton= muiNewRadioButton(xPos,yPos);
muiLoadButton(vc->PlaneRButton,"Plane");

muiLinkButtons(vc->NoneRButton,vc->CrossRButton);
muiLinkButtons(vc->CrossRButton,vc->BoxRButton);
muiLinkButtons(vc->BoxRButton,vc->PlaneRButton);
muiSetActive(vc->NoneRButton,GL_TRUE);
muiSetCallback(vc->NoneRButton,SwitchAxis);
muiSetCallback(vc->CrossRButton,SwitchAxis);
muiSetCallback(vc->BoxRButton,SwitchAxis);
muiSetCallback(vc->PlaneRButton,SwitchAxis);

/*:878*/
#line 13910 "MathView3D.w"

/*880:*/
#line 14373 "MathView3D.w"

xPos= 155;
xEnd= vc->CWinX-10;
yPos= vc->CWinY-25;
muiNewBoldLabel(xPos,yPos,"AxesLabel");
yPos-= 30;
vc->XLabelBox= muiNewTextbox(xPos,xEnd-60,yPos-15);
muiNewLabel(xEnd-50,yPos,"x-Axis");
yPos-= 40;
vc->YLabelBox= muiNewTextbox(xPos,xEnd-60,yPos-15);
muiNewLabel(xEnd-50,yPos,"y-Axis");
yPos-= 40;
vc->ZLabelBox= muiNewTextbox(xPos,xEnd-60,yPos-15);
muiNewLabel(xEnd-50,yPos,"z-Axis");

xPos= 10;
yPos= vc->CWinY-200;

yPos-= 20;
(void*)muiNewBoldLabel(xPos,yPos-5,"Origin -> ");
vc->OriginBox= muiNewTextbox(xPos+80,xEnd,yPos-15);

muiSetCallback(vc->XLabelBox,AxesTextChange);
muiSetCallback(vc->YLabelBox,AxesTextChange);
muiSetCallback(vc->ZLabelBox,AxesTextChange);
muiSetCallback(vc->OriginBox,AxesTextChange);


/*:880*/
#line 13911 "MathView3D.w"

/*879:*/
#line 14366 "MathView3D.w"

xPos= 10;
yPos= vc->CWinY-190;
vc->NumLabelRButton= muiNewRadioButton(xPos,yPos);
muiLoadButton(vc->NumLabelRButton,"Numeric Labels");
muiSetCallback(vc->NumLabelRButton,SwitchAxisSolid);

/*:879*/
#line 13912 "MathView3D.w"

/*881:*/
#line 14401 "MathView3D.w"


(void*)muiNewLabel(300,303,"x");
vc->XOriginSlider= muiNewVSlider(295,95,295,195,5);
muiSetCallback(vc->XOriginSlider,OriginSliderMove);

(void*)muiNewLabel(330,303,"y");
vc->YOriginSlider= muiNewVSlider(325,95,295,195,5);
muiSetCallback(vc->YOriginSlider,OriginSliderMove);

(void*)muiNewLabel(360,303,"z");
vc->ZOriginSlider= muiNewVSlider(355,95,295,195,5);
muiSetCallback(vc->ZOriginSlider,OriginSliderMove);

xPos= 155;
xEnd= vc->CWinX-10;
yPos= vc->CWinY-185;

(void*)muiNewBoldLabel(xPos,yPos,"FontSize");
vc->FontScaleSlider= muiNewHSlider(xPos+60,yPos-5,xEnd,xPos+60,5);
muiSetHSValue(vc->FontScaleSlider,AxisWindow->FontScale);
muiSetCallback(vc->FontScaleSlider,OriginSliderMove);


/*:881*/
#line 13913 "MathView3D.w"


muiAttachUIList(vc->WinID);
vc->muiID= muiGetActiveUIList();

/*855:*/
#line 13928 "MathView3D.w"

#if HAVE_GLUT_SUBWINDOW
vc->PreviewWinID= glutCreateSubWindow(vc->WinID,10,vc->CWinY-305,250,250);
#else
glutInitWindowSize(250,250);
vc->PreviewWinID= glutCreateWindow("Preview");
#endif
glutDisplayFunc(DisplayAxisPreview);
DisplayAxisPreview();
glutHideWindow();
glutSetWindow(vc->WinID);


/*:855*/
#line 13918 "MathView3D.w"

glutSetWindow(vc->WinID);
glutReshapeFunc(AWReshape);
return vc;
}

/*:853*/
#line 13926 "MathView3D.w"


/*:854*/
#line 91 "MathView3D.w"

/*2:*/
#line 117 "MathView3D.w"

void MVDisplay(void){

if(MakeNewView){
glPopMatrix();
glPushMatrix();
glTranslated(ZoomMove.x,ZoomMove.y,ZoomMove.z);
MakeNewView= GL_FALSE;
}
if(MakeNewModel){
CreateObject();
MakeNewModel= GL_FALSE;
}
if(DoReset){
glPopMatrix();
glPushMatrix();
glTranslated(ZoomMove.x+CenterOfScene.x,
ZoomMove.y+CenterOfScene.y,
ZoomMove.z+CenterOfScene.z);
if(DoTracking)
glRotated(TrackAngle*ToDegree,TrackAxis.x,TrackAxis.y,TrackAxis.z);
glTranslated(-CenterOfScene.x,-CenterOfScene.y,-CenterOfScene.z);
if(DoReset)DoReset= GL_FALSE;
}
else if(DoTracking){
if(!DoTrackSpin&&GetMouseTrack){
Point3D axis;

MVTrackball(&TrackAngle,&axis,ScreenXStart,ScreenYStart,ScreenX,ScreenY);
if(ROTATE_X==(DoRotate&ROTATE_X))TrackAxis= XUnitVector;
else if(ROTATE_Y==(DoRotate&ROTATE_Y))TrackAxis= YUnitVector;
else if(ROTATE_Z==(DoRotate&ROTATE_Z))TrackAxis= ZUnitVector;
else TrackAxis= axis;
DTrackAngle= MIN(TWO_PI/90.0,TrackAngle);
}
else
TrackAngle+= DTrackAngle;
}
if(TrackAngle){
/*3:*/
#line 164 "MathView3D.w"

glPopMatrix();
glPushMatrix();
glTranslated(ZoomMove.x+CenterOfScene.x,
ZoomMove.y+CenterOfScene.y,
ZoomMove.z+CenterOfScene.z);
glRotated(TrackAngle*ToDegree,TrackAxis.x,TrackAxis.y,TrackAxis.z);
glTranslated(-CenterOfScene.x,-CenterOfScene.y,-CenterOfScene.z);

/*:3*/
#line 156 "MathView3D.w"

}
if(!DoPSOutput)glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
if(ObjID>0)
glCallList(ObjID);
if(!DoPSOutput)glutSwapBuffers();
}

/*:2*/
#line 92 "MathView3D.w"

/*4:*/
#line 173 "MathView3D.w"

void MVAnimate(void)
{
if(MathLinkMode&&MathLinkActive){
if(!ProcessMathLink(&LinkToMath))exit(0);
#ifdef __WIN32__
MLCallYieldFunction(
MLYieldFunction(LinkToMath),
LinkToMath,(MLYieldParameters)0
);
#endif
MVSleep(50);
}

if(DoTracking&&MouseInsideMain)
glutPostRedisplay();
else
MVSleep(50);
}


/*:4*/
#line 93 "MathView3D.w"

/*5:*/
#line 194 "MathView3D.w"

void MVMouseButton(int btn,int state,int x,int y)
{
if(btn!=GLUT_LEFT_BUTTON||!(DoTracking||DoRotate))return;
switch(state){
case GLUT_DOWN:
ScreenXStart= ScreenX= x;
ScreenYStart= ScreenY= y;
DoTrackSpin= GL_FALSE;
break;
case GLUT_UP:
ScreenX= x;
ScreenY= y;
if(!DoRotate)DoTrackSpin= GL_TRUE;
break;
}
}

/*:5*/
#line 94 "MathView3D.w"

/*6:*/
#line 212 "MathView3D.w"

void MVMouseMove(int x,int y)
{
if(!DoTracking)return;
ScreenX= x;
ScreenY= y;
glutPostRedisplay();
}

/*:6*/
#line 95 "MathView3D.w"

/*7:*/
#line 221 "MathView3D.w"

void MVMouseEnterMain(int state)
{
if(GLUT_ENTERED==state&&glutGetWindow()!=MVDrawWindow){
glutSetWindow(MVDrawWindow);
}
if(GLUT_ENTERED==state)
MouseInsideMain= GL_TRUE;
else
MouseInsideMain= GL_FALSE;
}

/*:7*/
#line 96 "MathView3D.w"

/*8:*/
#line 233 "MathView3D.w"

void MVReshape(int w,int h)
{
glViewport(0,0,w,h);
ScreenWidth= w;
ScreenHeight= h;
}


/*:8*/
#line 97 "MathView3D.w"

/*35:*/
#line 779 "MathView3D.w"

#ifdef __WIN32__
DWORD AboutThreadID;
char GlVersion[512]= "";
char GlVendor[512]= "";
char GlRender[512]= "";
char MlLinkName[512]= "None";
#pragma argsused
DWORD WINAPI AboutThread(LPVOID p)
{char msg[2048];

sprintf(msg,
"Version "VERSION"\ncompiled: "__DATE__
"\n(c) Jens-Peer Kuska 1997-99\n"
"e-mail:  kuska@informatik.uni-leipzig.de\n\n"
"OpenGL\n  %s\n  %s\n  %s\n\n"
"MathLink\n  %s\n\n"
"Export prefix  : %s\n"
"Unique Points  : %d of %d\n"
"Polygon number : %ld\n"
"Line number    : %ld\n"
"Point number   : %ld\n"
"Mesh number    : %ld\n"
"Texture number : %ld\n",
GlVendor,
GlVersion,
GlRender,
MlLinkName,
WriteFilePrefix,
PoolSize,MaxPoolSize,
LengthLinePolyList(Polygons),
LengthLinePolyList(Lines),
LengthPointList(Points),
LengthMesh3DList(Meshes),
LengthTextureList(Textures)
);
MessageBox(MLIconWindow,
msg,
"About MathGL3d",MB_OK|MB_ICONINFORMATION|MB_TASKMODAL|MB_SETFOREGROUND);
return 0;
}

#endif

/*:35*//*42:*/
#line 1041 "MathView3D.w"

/*41:*/
#line 1004 "MathView3D.w"

char*MVQuerySaveFilename(char*suggest,
char*filter,
char*ext,
int len,int*fail)
{
#ifdef __WIN32__
OPENFILENAME ofn;

ofn.lStructSize= sizeof(ofn);
ofn.hwndOwner= GetForegroundWindow();
ofn.hInstance= NULL;
ofn.lpstrFilter= filter;
ofn.lpstrCustomFilter= NULL;
ofn.nMaxCustFilter= 0;
ofn.nFilterIndex= 1;
ofn.lpstrFile= suggest;
ofn.nMaxFile= len;
ofn.lpstrTitle= NULL;
ofn.lpstrInitialDir= NULL;
ofn.Flags= OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|OFN_EXPLORER|OFN_LONGNAMES;
ofn.nFileOffset= 0;
ofn.nFileExtension= 0;
ofn.lpstrDefExt= ext;
ofn.lCustData= 0L;
ofn.lpfnHook= 0L;
ofn.lpTemplateName= (LPSTR)NULL;
ofn.lpstrTitle= "MathGL3d - Save";
*fail= !GetSaveFileName(&ofn);
return(*fail?NULL:suggest);
#else
*fail= GL_FALSE;
return suggest;
#endif
}

/*:41*/
#line 1042 "MathView3D.w"



/*:42*/
#line 98 "MathView3D.w"

/*43:*/
#line 1045 "MathView3D.w"

void MVWinFileMenu(int what)
{char fname[256];
int res= GL_FALSE;

strcpy(fname,WriteFilePrefix);
switch(what){
#if MACINTOSH_MATHLINK
case MENU_ID_LOAD_XXX:
if(MacQueryLoadFilename(fname)){
glutSetCursor(GLUT_CURSOR_WAIT);
if(strstr(fname,".ts")||strstr(fname,".TS")){
res= Read3Script(fname);
}
else if(strstr(fname,".dxf")||strstr(fname,".DXF")){
res= ReadDXF(fname);
}
glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
if(!res){
AveragedNormals(Polygons);
glutSetWindow(MVDrawWindow);
glutPopWindow();
glutShowWindow();
/*681:*/
#line 11317 "MathView3D.w"

BoxRatios.x= BoxRatios.y= BoxRatios.z= 1.0;
/*682:*/
#line 11321 "MathView3D.w"

glutSetWindow(MVDrawWindow);
glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
if(ObjID>=0)glDeleteLists(ObjID,1);
ObjID= -1;
InitView(GL_TRUE);
glutPostRedisplay();
MakeNewView= GL_TRUE;

/*:682*/
#line 11319 "MathView3D.w"


/*:681*/
#line 1068 "MathView3D.w"

glFinish();
/*37:*/
#line 958 "MathView3D.w"

DoReset= GL_TRUE;
DoTracking= DoTrackSpin= GL_FALSE;
DoRotate= GL_FALSE;
ZoomMove.x= ZoomMove.y= ZoomMove.z= 0.0;
RadialZoom= 1.0;
TrackAngle= 0.0;
TrackAxis.x= TrackAxis.y= 0.0;
TrackAxis.z= 1.0;
ScreenXStart= ScreenX= 0;
ScreenYStart= ScreenY= 0;
BlackBackground= GL_TRUE;
SetBackgroundColor(BlackBackground);
glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
Axis.AxisType= NoAxis;

/*:37*/
#line 1070 "MathView3D.w"

}
aglUpdateCurrent();
}
break;
#endif
case MENU_ID_SAVEPOVRAY:
glutSetCursor(GLUT_CURSOR_WAIT);
strcat(fname,SmoothShading?"s.pov":"f.pov");
#if MACINTOSH_MATHLINK
if(MacQuerySaveFilename("Save POV-Ray file as:",fname)){
WritePOVray(fname,SmoothShading,"mvs","mvm");
}
#else
WritePOVray(fname,SmoothShading,"mvs","mvm");
#endif
glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
break;
case MENU_ID_SAVE_3DMF:
glutSetCursor(GLUT_CURSOR_WAIT);
strcat(fname,".3dmf");
#if MACINTOSH_MATHLINK
if(MacQuerySaveFilename("Save QuickDraw3D file as:",fname)){
Write3DMF(fname,GL_TRUE,GL_TRUE);
}
#else
Write3DMF(fname,GL_TRUE,GL_TRUE);
#endif
glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
break;
case MENU_ID_SAVE_VRML:
glutSetCursor(GLUT_CURSOR_WAIT);
strcat(fname,".wrl");
#if MACINTOSH_MATHLINK
if(MacQuerySaveFilename("Save VRML 2.0 file as:",fname)){
WriteVRML(fname,GL_FALSE,UseVRMLExtrusion,"mvs","mvm");
}
#else
WriteVRML(fname,GL_FALSE,UseVRMLExtrusion,"mvs","mvm");
#endif
glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
break;
case MENU_ID_SAVEPOSTSCRIPT:
glutSetCursor(GLUT_CURSOR_WAIT);
strcat(fname,".eps");
#if MACINTOSH_MATHLINK
if(MacQuerySaveFilename("Save PostScrip file as:",fname)){
WritePostScript(fname);
}
#else
WritePostScript(fname);
#endif
glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
break;
case MENU_ID_SAVE_DXF:
glutSetCursor(GLUT_CURSOR_WAIT);
strcat(fname,".dxf");
#if MACINTOSH_MATHLINK
if(MacQuerySaveFilename("Save DXF file as:",fname)){
WriteDXF(fname);
}
#else
WriteDXF(fname);
#endif
glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
break;
case MENU_ID_SAVE_OFF:
glutSetCursor(GLUT_CURSOR_WAIT);
strcat(fname,".off");
#if MACINTOSH_MATHLINK
if(MacQuerySaveFilename("Save Geomview file as:",fname)){
WriteOFF(fname);
}
#else
WriteOFF(fname);
#endif
glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
break;
case MENU_ID_SAVE_PNG:
strcat(fname,".png");
#if MACINTOSH_MATHLINK
if(MacQuerySaveFilename("Save PNG file as:",fname)){
/*44:*/
#line 1180 "MathView3D.w"

glutSetWindow(MVDrawWindow);
glutPopWindow();
glutShowWindow();

/*:44*/
#line 1152 "MathView3D.w"

PNGSaveFrameBuffer(fname,
0,0,glutGet(GLUT_WINDOW_WIDTH),
glutGet(GLUT_WINDOW_HEIGHT),GL_FALSE,1.0);
}
#else
PNGSaveFrameBuffer(fname,
0,0,glutGet(GLUT_WINDOW_WIDTH),
glutGet(GLUT_WINDOW_HEIGHT),GL_FALSE,1.0);
#endif
break;
case MENU_ID_SAVE_BITMAP_EPS:
strcat(fname,"bm.eps");
#if MACINTOSH_MATHLINK
if(MacQuerySaveFilename("Save PostScrip file as:",fname)){
/*44:*/
#line 1180 "MathView3D.w"

glutSetWindow(MVDrawWindow);
glutPopWindow();
glutShowWindow();

/*:44*/
#line 1167 "MathView3D.w"

EPSSaveFrameBuffer(fname,
0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
}
#else
EPSSaveFrameBuffer(fname,
0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
#endif
break;
}
glutPostRedisplay();
}

/*:43*/
#line 99 "MathView3D.w"

/*46:*/
#line 1195 "MathView3D.w"

void MVWinSpinMenu(int what)
{
/*687:*/
#line 11392 "MathView3D.w"

/*683:*/
#line 11330 "MathView3D.w"

{ViewCntrlWin*vc;

if(VCWindow){
vc= VCWindow;
if(glutGetWindow()==VCWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11336 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11340 "MathView3D.w"

glutSetWindow(VCWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:683*/
#line 11393 "MathView3D.w"

/*684:*/
#line 11348 "MathView3D.w"

{LightCntrlWin*vc;

if(LCWindow){
vc= LCWindow;
if(glutGetWindow()==LCWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11354 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11358 "MathView3D.w"

glutSetWindow(LCWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:684*/
#line 11394 "MathView3D.w"

/*685:*/
#line 11366 "MathView3D.w"

{AxisCtrlWin*vc;

if(AxisWindow){
vc= AxisWindow;
if(glutGetWindow()==AxisWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11372 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11376 "MathView3D.w"

glutSetWindow(AxisWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:685*/
#line 11395 "MathView3D.w"


/*:687*/
#line 1198 "MathView3D.w"

DoRotate= GL_FALSE;
switch(what){
case MENU_ID_SPIN_TRACKBALL:
GetMouseTrack= GL_TRUE;
/*48:*/
#line 1237 "MathView3D.w"

DoTracking= GL_TRUE;
DoTrackSpin= GL_FALSE;
glutSetCursor(GLUT_CURSOR_INFO);

/*:48*/
#line 1203 "MathView3D.w"

break;
case MENU_ID_SPIN_X:
TrackAxis.x= 1.0;TrackAxis.y= TrackAxis.z= 0.0;
/*49:*/
#line 1242 "MathView3D.w"

DoTrackSpin= DoTracking= GL_TRUE;
GetMouseTrack= GL_FALSE;
TrackAngle= 0.0;
DTrackAngle= TWO_PI/80.;
glutSetCursor(GLUT_CURSOR_LEFT_ARROW);

/*:49*/
#line 1207 "MathView3D.w"

break;
case MENU_ID_SPIN_Y:
TrackAxis.y= 1.0;TrackAxis.x= TrackAxis.z= 0.0;
/*49:*/
#line 1242 "MathView3D.w"

DoTrackSpin= DoTracking= GL_TRUE;
GetMouseTrack= GL_FALSE;
TrackAngle= 0.0;
DTrackAngle= TWO_PI/80.;
glutSetCursor(GLUT_CURSOR_LEFT_ARROW);

/*:49*/
#line 1211 "MathView3D.w"

break;
case MENU_ID_SPIN_Z:
TrackAxis.z= 1.0;TrackAxis.x= TrackAxis.y= 0.0;
/*49:*/
#line 1242 "MathView3D.w"

DoTrackSpin= DoTracking= GL_TRUE;
GetMouseTrack= GL_FALSE;
TrackAngle= 0.0;
DTrackAngle= TWO_PI/80.;
glutSetCursor(GLUT_CURSOR_LEFT_ARROW);

/*:49*/
#line 1215 "MathView3D.w"

break;
case MENU_ID_SPIN_SLOW:
DTrackAngle*= 0.5;
break;
case MENU_ID_SPIN_FAST:
DTrackAngle*= 2.0;
break;
case MENU_ID_STOP:
/*47:*/
#line 1230 "MathView3D.w"

DoTrackSpin= GL_FALSE;
TrackAngle= 0.0;
DoTracking= GL_FALSE;
glutSetCursor(GLUT_CURSOR_LEFT_ARROW);


/*:47*/
#line 1224 "MathView3D.w"

break;
}
glutPostRedisplay();
}

/*:46*/
#line 100 "MathView3D.w"

/*51:*/
#line 1260 "MathView3D.w"

void MVWinRotateMenu(int what)
{
/*687:*/
#line 11392 "MathView3D.w"

/*683:*/
#line 11330 "MathView3D.w"

{ViewCntrlWin*vc;

if(VCWindow){
vc= VCWindow;
if(glutGetWindow()==VCWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11336 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11340 "MathView3D.w"

glutSetWindow(VCWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:683*/
#line 11393 "MathView3D.w"

/*684:*/
#line 11348 "MathView3D.w"

{LightCntrlWin*vc;

if(LCWindow){
vc= LCWindow;
if(glutGetWindow()==LCWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11354 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11358 "MathView3D.w"

glutSetWindow(LCWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:684*/
#line 11394 "MathView3D.w"

/*685:*/
#line 11366 "MathView3D.w"

{AxisCtrlWin*vc;

if(AxisWindow){
vc= AxisWindow;
if(glutGetWindow()==AxisWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11372 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11376 "MathView3D.w"

glutSetWindow(AxisWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:685*/
#line 11395 "MathView3D.w"


/*:687*/
#line 1263 "MathView3D.w"

DoTrackSpin= GL_FALSE;
DoRotate= GL_FALSE;
TrackAngle= 0.0;
TrackAxis.x= TrackAxis.y= 0.0;
TrackAxis.z= 1.0;
ScreenXStart= ScreenX;
ScreenYStart= ScreenY;
switch(what){
case MENU_ID_ROTATE:
GetMouseTrack= GL_TRUE;
DoRotate= GL_TRUE;
/*48:*/
#line 1237 "MathView3D.w"

DoTracking= GL_TRUE;
DoTrackSpin= GL_FALSE;
glutSetCursor(GLUT_CURSOR_INFO);

/*:48*/
#line 1275 "MathView3D.w"

break;
case MENU_ID_ROTATE_X:
GetMouseTrack= GL_TRUE;
DoRotate= GL_TRUE|ROTATE_X;
TrackAxis.x= 1.0;TrackAxis.y= TrackAxis.z= 0.0;
/*48:*/
#line 1237 "MathView3D.w"

DoTracking= GL_TRUE;
DoTrackSpin= GL_FALSE;
glutSetCursor(GLUT_CURSOR_INFO);

/*:48*/
#line 1281 "MathView3D.w"

glutSetCursor(GLUT_CURSOR_UP_DOWN);
break;
case MENU_ID_ROTATE_Y:
GetMouseTrack= GL_TRUE;
DoRotate= GL_TRUE|ROTATE_Y;
TrackAxis.x= TrackAxis.z= 0.0;TrackAxis.y= 1.0;
/*48:*/
#line 1237 "MathView3D.w"

DoTracking= GL_TRUE;
DoTrackSpin= GL_FALSE;
glutSetCursor(GLUT_CURSOR_INFO);

/*:48*/
#line 1288 "MathView3D.w"

glutSetCursor(GLUT_CURSOR_UP_DOWN);
break;
case MENU_ID_ROTATE_Z:
GetMouseTrack= GL_TRUE;
DoRotate= GL_TRUE|ROTATE_Z;
TrackAxis.x= TrackAxis.y= 0.0;TrackAxis.z= 1.0;
/*48:*/
#line 1237 "MathView3D.w"

DoTracking= GL_TRUE;
DoTrackSpin= GL_FALSE;
glutSetCursor(GLUT_CURSOR_INFO);

/*:48*/
#line 1295 "MathView3D.w"

glutSetCursor(GLUT_CURSOR_LEFT_RIGHT);
break;
}
glutPostRedisplay();
}

/*:51*/
#line 101 "MathView3D.w"

/*54:*/
#line 1327 "MathView3D.w"

void MVWinAppearMenu(int what);
void MVWinAppearMenu(int what)
{
switch(what){
case MENU_ID_SHOW_SMOOTH:
SmoothShading= GL_TRUE;
WireFrame= GL_FALSE;
break;
case MENU_ID_SHOW_FLAT:
SmoothShading= GL_FALSE;
WireFrame= GL_FALSE;
break;
case MENU_ID_SHOW_WIRE:
WireFrame= GL_TRUE;
PolyOutline= GL_FALSE;
break;
case MENU_ID_TOGGLE_BACK:
BlackBackground= !BlackBackground;
SetBackgroundColor(BlackBackground);
break;
case MENU_ID_SHOW_OUTLINE:
if(!WireFrame)PolyOutline= !PolyOutline;
break;
}
MakeNewModel= GL_TRUE;
glutPostRedisplay();
}


/*:54*/
#line 102 "MathView3D.w"

/*36:*/
#line 824 "MathView3D.w"

void MVWinMainMenu(int what)
{double vr,vtheta,vphi;

switch(what){
case MENU_ID_VIEWCONTROL:
/*37:*/
#line 958 "MathView3D.w"

DoReset= GL_TRUE;
DoTracking= DoTrackSpin= GL_FALSE;
DoRotate= GL_FALSE;
ZoomMove.x= ZoomMove.y= ZoomMove.z= 0.0;
RadialZoom= 1.0;
TrackAngle= 0.0;
TrackAxis.x= TrackAxis.y= 0.0;
TrackAxis.z= 1.0;
ScreenXStart= ScreenX= 0;
ScreenYStart= ScreenY= 0;
BlackBackground= GL_TRUE;
SetBackgroundColor(BlackBackground);
glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
Axis.AxisType= NoAxis;

/*:37*/
#line 830 "MathView3D.w"

/*684:*/
#line 11348 "MathView3D.w"

{LightCntrlWin*vc;

if(LCWindow){
vc= LCWindow;
if(glutGetWindow()==LCWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11354 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11358 "MathView3D.w"

glutSetWindow(LCWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:684*/
#line 831 "MathView3D.w"

/*685:*/
#line 11366 "MathView3D.w"

{AxisCtrlWin*vc;

if(AxisWindow){
vc= AxisWindow;
if(glutGetWindow()==AxisWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11372 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11376 "MathView3D.w"

glutSetWindow(AxisWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:685*/
#line 832 "MathView3D.w"

if(!VCWindow)
VCWindow= CreateViewControl();
CopyToVCWindow();
glutSetWindow(VCWindow->PreviewWinID);
glutShowWindow();
glutPostRedisplay();
glutSetWindow(VCWindow->WinID);
glutShowWindow();
glutPostRedisplay();
break;
case MENU_ID_LIGHTCONTROL:
/*37:*/
#line 958 "MathView3D.w"

DoReset= GL_TRUE;
DoTracking= DoTrackSpin= GL_FALSE;
DoRotate= GL_FALSE;
ZoomMove.x= ZoomMove.y= ZoomMove.z= 0.0;
RadialZoom= 1.0;
TrackAngle= 0.0;
TrackAxis.x= TrackAxis.y= 0.0;
TrackAxis.z= 1.0;
ScreenXStart= ScreenX= 0;
ScreenYStart= ScreenY= 0;
BlackBackground= GL_TRUE;
SetBackgroundColor(BlackBackground);
glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
Axis.AxisType= NoAxis;

/*:37*/
#line 844 "MathView3D.w"

/*683:*/
#line 11330 "MathView3D.w"

{ViewCntrlWin*vc;

if(VCWindow){
vc= VCWindow;
if(glutGetWindow()==VCWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11336 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11340 "MathView3D.w"

glutSetWindow(VCWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:683*/
#line 845 "MathView3D.w"

/*685:*/
#line 11366 "MathView3D.w"

{AxisCtrlWin*vc;

if(AxisWindow){
vc= AxisWindow;
if(glutGetWindow()==AxisWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11372 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11376 "MathView3D.w"

glutSetWindow(AxisWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:685*/
#line 846 "MathView3D.w"

if(!LCWindow)
LCWindow= CreateLightControl();
CopyToLCWindow();
glutSetWindow(LCWindow->PreviewWinID);
glutShowWindow();
glutPostRedisplay();
glutSetWindow(LCWindow->WinID);
glutShowWindow();
glutPostRedisplay();
break;
case MENU_ID_AXISCONTROL:
/*37:*/
#line 958 "MathView3D.w"

DoReset= GL_TRUE;
DoTracking= DoTrackSpin= GL_FALSE;
DoRotate= GL_FALSE;
ZoomMove.x= ZoomMove.y= ZoomMove.z= 0.0;
RadialZoom= 1.0;
TrackAngle= 0.0;
TrackAxis.x= TrackAxis.y= 0.0;
TrackAxis.z= 1.0;
ScreenXStart= ScreenX= 0;
ScreenYStart= ScreenY= 0;
BlackBackground= GL_TRUE;
SetBackgroundColor(BlackBackground);
glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
Axis.AxisType= NoAxis;

/*:37*/
#line 858 "MathView3D.w"

/*683:*/
#line 11330 "MathView3D.w"

{ViewCntrlWin*vc;

if(VCWindow){
vc= VCWindow;
if(glutGetWindow()==VCWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11336 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11340 "MathView3D.w"

glutSetWindow(VCWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:683*/
#line 859 "MathView3D.w"

/*684:*/
#line 11348 "MathView3D.w"

{LightCntrlWin*vc;

if(LCWindow){
vc= LCWindow;
if(glutGetWindow()==LCWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11354 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11358 "MathView3D.w"

glutSetWindow(LCWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:684*/
#line 860 "MathView3D.w"

if(!AxisWindow)
AxisWindow= CreateAxisControl();
CopyToAxisWindow();
glutSetWindow(AxisWindow->PreviewWinID);
glutShowWindow();
SetBackgroundColor(BlackBackground);
glutPostRedisplay();
glutSetWindow(AxisWindow->WinID);
glutShowWindow();
glutPostRedisplay();
break;
case MENU_ID_RESET:
/*37:*/
#line 958 "MathView3D.w"

DoReset= GL_TRUE;
DoTracking= DoTrackSpin= GL_FALSE;
DoRotate= GL_FALSE;
ZoomMove.x= ZoomMove.y= ZoomMove.z= 0.0;
RadialZoom= 1.0;
TrackAngle= 0.0;
TrackAxis.x= TrackAxis.y= 0.0;
TrackAxis.z= 1.0;
ScreenXStart= ScreenX= 0;
ScreenYStart= ScreenY= 0;
BlackBackground= GL_TRUE;
SetBackgroundColor(BlackBackground);
glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
Axis.AxisType= NoAxis;

/*:37*/
#line 873 "MathView3D.w"

break;
case MENU_ID_FLIP_FACE:
ReversePolygons(Polygons);
AveragedNormals(Polygons);
Meshes= FlipMeshNormals(Meshes);
MVFaceFlip= !MVFaceFlip;
MakeNewModel= GL_TRUE;
break;
case MENU_ID_ZOOM_OUT:
RadialZoom*= 1.5;
/*38:*/
#line 974 "MathView3D.w"

SphericalFromCartesian(&ViewPoint,&vr,&vtheta,&vphi);
ZoomMove.x= ViewPoint.x-RadialZoom*vr*cos(vphi)*sin(vtheta);
ZoomMove.y= ViewPoint.y-RadialZoom*vr*sin(vphi)*sin(vtheta);
ZoomMove.z= ViewPoint.z-RadialZoom*vr*cos(vtheta);

MakeNewView= GL_TRUE;


/*:38*/
#line 884 "MathView3D.w"

break;
case MENU_ID_ZOOM_IN:
RadialZoom/= 1.5;
/*38:*/
#line 974 "MathView3D.w"

SphericalFromCartesian(&ViewPoint,&vr,&vtheta,&vphi);
ZoomMove.x= ViewPoint.x-RadialZoom*vr*cos(vphi)*sin(vtheta);
ZoomMove.y= ViewPoint.y-RadialZoom*vr*sin(vphi)*sin(vtheta);
ZoomMove.z= ViewPoint.z-RadialZoom*vr*cos(vtheta);

MakeNewView= GL_TRUE;


/*:38*/
#line 888 "MathView3D.w"

break;
case MENU_ID_ABOUT:
#ifdef __WIN32__
strcpy(GlVendor,(char*)glGetString(GL_VENDOR));
strcpy(GlVersion,(char*)glGetString(GL_VERSION));
strcpy(GlRender,(char*)glGetString(GL_RENDERER));
if(MathLinkMode)
strcpy(MlLinkName,MLName(LinkToMath));
CreateThread(NULL,8096,(LPTHREAD_START_ROUTINE)AboutThread,
NULL,0,&AboutThreadID);
#else
{int i;

printf(BANNER
" (c) Jens-Peer Kuska  Version "VERSION
"\nfor "ARCH"\ncompiled at "__DATE__"\n");
printf("OpenGL:\n\t%s\n\t%s\n\t%s\n",
glGetString(GL_VENDOR),glGetString(GL_VERSION),
glGetString(GL_RENDERER));
if(*ThreeScriptName)
printf("3Script file   : %s\n",ThreeScriptName);
printf("File write     : %s*.*\n",WriteFilePrefix);
if(*DXFInputName)
printf("DXF file       : %s\n",DXFInputName);
printf("Screen         : %d x %d\n",ScreenWidth,ScreenHeight);
printf("Center         : {"DBL_FMT", "DBL_FMT", "DBL_FMT"}\n",
CenterOfScene.x,CenterOfScene.y,CenterOfScene.z);
printf("Scaling        : {"DBL_FMT", "DBL_FMT", "DBL_FMT"}\n",
ScaleOfScene.x,ScaleOfScene.y,ScaleOfScene.z);
printf("View point     : {"DBL_FMT", "DBL_FMT", "DBL_FMT"}\n",
ViewPoint.x,ViewPoint.y,ViewPoint.z);
if(ObjID>0)
printf("Compiled obj   : %d\n",ObjID);
if(PoolSize)
printf("Unique Points  : %d of %d\n",PoolSize,MaxPoolSize);
printf("Polygon number : %ld\n",LengthLinePolyList(Polygons));
printf("Line number    : %ld\n",LengthLinePolyList(Lines));
printf("Point number   : %ld\n",LengthPointList(Points));
printf("Mesh number    : %ld\n",LengthMesh3DList(Meshes));
printf("Texture number : %ld\n",LengthTextureList(Textures));
printf("Polygon shading: %s\n",SmoothShading?"Smooth":"Flat");
printf("Wire frame     : %s\n",WireFrame?"On":"Off");
printf("Poly Outlines  : %s\n",PolyOutline?"On":"Off");
printf("Tracking mode  : %s\n",DoTracking?"On":"Off");
printf("Light sources  :\n");
for(i= 0;i<LightNo;i++)
printf("\t Light[[%d]] -> {{"DBL_FMT
", "DBL_FMT", "DBL_FMT
"},RGBColor["DBL_FMT", "DBL_FMT", "DBL_FMT"]}\n",
i,
LightPosition[i].x,
LightPosition[i].y,
LightPosition[i].z,
LightColor[i].r,
LightColor[i].g,
LightColor[i].b);
}
#endif
break;
case MENU_ID_CLEAR:
mvClearContents();
break;
case MENU_ID_QUIT:
exit(0);
break;
}
glutPostRedisplay();
}

/*:36*/
#line 103 "MathView3D.w"

/*20:*/
#line 564 "MathView3D.w"

void MVCreateMainMenu(void){

/*39:*/
#line 984 "MathView3D.w"

menuIDFiles= glutCreateMenu(MVWinFileMenu);
#if MACINTOSH_MATHLINK
glutAddMenuEntry("Load 3Script, DXF ..",MENU_ID_LOAD_XXX);
#endif
glutAddMenuEntry("Save POVRay file",MENU_ID_SAVEPOVRAY);
glutAddMenuEntry("Save QuickDraw 3D",MENU_ID_SAVE_3DMF);
glutAddMenuEntry("Save VRML 2.0",MENU_ID_SAVE_VRML);
glutAddMenuEntry("Save PostScript file",MENU_ID_SAVEPOSTSCRIPT);
glutAddMenuEntry("Save DXF file",MENU_ID_SAVE_DXF);
glutAddMenuEntry("Save OFF file",MENU_ID_SAVE_OFF);
glutAddMenuEntry("Save PNG bitmap",MENU_ID_SAVE_PNG);
glutAddMenuEntry("Save EPS bitmap",MENU_ID_SAVE_BITMAP_EPS);


/*:39*/
#line 567 "MathView3D.w"

/*45:*/
#line 1185 "MathView3D.w"

menuIDSpin= glutCreateMenu(MVWinSpinMenu);
glutAddMenuEntry("Trackball",MENU_ID_SPIN_TRACKBALL);
glutAddMenuEntry("X axes spin",MENU_ID_SPIN_X);
glutAddMenuEntry("Y axes spin",MENU_ID_SPIN_Y);
glutAddMenuEntry("Z axes spin",MENU_ID_SPIN_Z);
glutAddMenuEntry("Spin slower",MENU_ID_SPIN_SLOW);
glutAddMenuEntry("Spin faster",MENU_ID_SPIN_FAST);
glutAddMenuEntry("Halt",MENU_ID_STOP);

/*:45*/
#line 568 "MathView3D.w"

/*50:*/
#line 1249 "MathView3D.w"

menuIDRotate= glutCreateMenu(MVWinRotateMenu);
glutAddMenuEntry("Free rotation",MENU_ID_ROTATE);
glutAddMenuEntry("X axes rotation",MENU_ID_ROTATE_X);
glutAddMenuEntry("Y axes rotation",MENU_ID_ROTATE_Y);
glutAddMenuEntry("Z axes rotation",MENU_ID_ROTATE_Z);

/*:50*/
#line 569 "MathView3D.w"

/*52:*/
#line 1302 "MathView3D.w"

menuIDAppear= glutCreateMenu(MVWinAppearMenu);
glutAddMenuEntry("Smooth surface",MENU_ID_SHOW_SMOOTH);
glutAddMenuEntry("Flat surface",MENU_ID_SHOW_FLAT);
glutAddMenuEntry("Wire frame",MENU_ID_SHOW_WIRE);
glutAddMenuEntry("Toggle b/gray backround",MENU_ID_TOGGLE_BACK);
glutAddMenuEntry("Toggle polyon outlines",MENU_ID_SHOW_OUTLINE);

/*:52*/
#line 570 "MathView3D.w"

menuIDMain= glutCreateMenu(MVWinMainMenu);
glutAddSubMenu("Files...",menuIDFiles);
glutAddSubMenu("Spin...",menuIDSpin);
glutAddSubMenu("Rotate ...",menuIDRotate);
glutAddSubMenu("Appearence ...",menuIDAppear);
glutAddMenuEntry("Change view point",MENU_ID_VIEWCONTROL);
glutAddMenuEntry("Change lights",MENU_ID_LIGHTCONTROL);
glutAddMenuEntry("Axis Editor",MENU_ID_AXISCONTROL);
glutAddMenuEntry("Reset view",MENU_ID_RESET);
glutAddMenuEntry("Zoom out",MENU_ID_ZOOM_OUT);
glutAddMenuEntry("Zoom in",MENU_ID_ZOOM_IN);
glutAddMenuEntry("Flip faces",MENU_ID_FLIP_FACE);
glutAddMenuEntry("Clear window",MENU_ID_CLEAR);
#ifdef __WIN32__
if(MathLinkMode)
glutAddMenuEntry("About",MENU_ID_ABOUT);
#endif
if(!MathLinkMode){
glutAddMenuEntry("About",MENU_ID_ABOUT);
glutAddMenuEntry("Quit",MENU_ID_QUIT);
}

glutAttachMenu(GLUT_RIGHT_BUTTON);
}

/*:20*/
#line 104 "MathView3D.w"

/*55:*/
#line 1358 "MathView3D.w"

void MathViewExit(void);
void MathViewExit(void)
{int i;
if(PointPool)free(PointPool);
if(NormalPool)
for(i= 0;i<PoolSize;i++)
if(NormalPool[i])free(NormalPool[i]);
free(NormalPool);
Points= RemovePointList(Points);
OctalPointTree= RemovePointOctTree(OctalPointTree);
Lines= RemoveLinePolyList(Lines);
Polygons= RemoveLinePolyList(Polygons);
Meshes= RemoveMesh3DList(Meshes);
Textures= RemoveTextureList(Textures);
SceneTextureMap= DeleteTextureMapping(SceneTextureMap);
if(VCWindow)free(VCWindow);
if(LCWindow)free(LCWindow);
if(AxisWindow)free(AxisWindow);
if(MathLinkMode){
/*754:*/
#line 12443 "MathView3D.w"

if(LinkToMath)
MLClose(LinkToMath);
MLDeinitialize(stdenv);
stdenv= (MLEnvironment)0;
exit(!MLDone);


/*:754*/
#line 1378 "MathView3D.w"

}
}

/*:55*/
#line 105 "MathView3D.w"

/*645:*/
#line 10746 "MathView3D.w"

/*757:*/
#line 12484 "MathView3D.w"

int IsMathLinkCall(int argc,char*argv[])
{int i;
#if MACINTOSH_MATHLINK
return GL_TRUE;
#else
for(i= 0;i<argc;i++){
if(strcmp(argv[i],"-mathlink")==0)return GL_TRUE;
if(strcmp(argv[i],"-linkname")==0)return GL_TRUE;
if(strcmp(argv[i],"-linkmode")==0)return GL_TRUE;
if(strcmp(argv[i],"-linkcreate")==0)return GL_TRUE;
if(strcmp(argv[i],"-linkprotocol")==0)return GL_TRUE;
}
return GL_FALSE;
#endif
}


/*:757*/
#line 10747 "MathView3D.w"

/*756:*/
#line 12454 "MathView3D.w"

static int refuse_to_be_a_frontend(MLINK);

int ProcessMathLink(MLINK*mlp)
{int pkt;

if(MLDone&&MLError(*mlp)){
return 0;
}
while(MLReady(*mlp)&&!MLDone){
DoMathematicaCall= GL_TRUE;
pkt= MLNextPacket(*mlp);
MLAbort= 0;
if(RESUMEPKT==pkt){
if(!refuse_to_be_a_frontend(*mlp))return 0;
/*754:*/
#line 12443 "MathView3D.w"

if(LinkToMath)
MLClose(LinkToMath);
MLDeinitialize(stdenv);
stdenv= (MLEnvironment)0;
exit(!MLDone);


/*:754*/
#line 12469 "MathView3D.w"

}
glutSetWindow(MVDrawWindow);
glutPopWindow();
glutPostRedisplay();
glFinish();
if(pkt!=CALLPKT||!MLDoCallPacket(*mlp))return 0;
MLFlush(*mlp);
MLAbort= 0;
DoMathematicaCall= GL_FALSE;
}
return 1;
}


/*:756*/
#line 10748 "MathView3D.w"

/*742:*/
#line 12197 "MathView3D.w"

int GetGraphics3D(MLINK link,const char*head,long int count)
{long ecount,n,coord,coord1,i;
int k,j;
#if !MACINTOSH_MATHLINK
int plBuffer[8192];
#else
int plBuffer[4096];
#endif
double*v,vec[3],vec1[3];
int cubep[8];
const char*fhead;
double red,green,blue,sred,sgreen,sblue,dummy;
double psize;
int isMesh= GL_FALSE;

fhead= head;
ecount= count;
n= ecount;
red= green= blue= sred= sgreen= sblue= 1.0;
psize= 0.001;
for(i= 0;i<ecount;i++){
/*744:*/
#line 12303 "MathView3D.w"

while(!MLReady(link)){
MVSleep(100);
}

/*:744*/
#line 12219 "MathView3D.w"

MLGetFunction(link,&fhead,&n);
if(!fhead){
MLDisownSymbol(link,fhead);
continue;
}
#if 0
fprintf(stderr,"Get %s with %d arguments\n",fhead,n);
#endif
if(0==strcmp(fhead,"Polygon")){
/*745:*/
#line 12308 "MathView3D.w"

MLCheckFunction(link,"List",&n);
for(k= j= 0;j<n;j++){
/*744:*/
#line 12303 "MathView3D.w"

while(!MLReady(link)){
MVSleep(100);
}

/*:744*/
#line 12311 "MathView3D.w"

MLGetRealList(link,&v,&coord);
if(3==coord){
plBuffer[k++]= AddToPointPool(v[0],v[1],v[2]);
}
MLDisownRealList(link,v,coord);
}
Polygons= PrependLinePolyList(Polygons,Polygon3D,plBuffer,k,sred,sgreen,sblue);

/*:745*/
#line 12229 "MathView3D.w"

}
else if(0==strcmp(fhead,"Line")){
/*746:*/
#line 12320 "MathView3D.w"

{int*plBuffer;
MLCheckFunction(link,"List",&n);
plBuffer= (int*)calloc(n+1,sizeof(int));
for(k= j= 0;j<n;j++){
/*744:*/
#line 12303 "MathView3D.w"

while(!MLReady(link)){
MVSleep(100);
}

/*:744*/
#line 12325 "MathView3D.w"

MLGetRealList(link,&v,&coord);
if(3==coord){
plBuffer[k++]= AddToPointPool(v[0],v[1],v[2]);
}
MLDisownRealList(link,v,coord);
}
Lines= PrependLinePolyList(Lines,Line3D,plBuffer,k,red,green,blue);
free(plBuffer);
}

/*:746*/
#line 12232 "MathView3D.w"

}
else if(0==strcmp(fhead,"Point")){
/*747:*/
#line 12336 "MathView3D.w"

MLGetRealList(link,&v,&coord);
if(3==coord){
k= AddToPointPool(v[0],v[1],v[2]);
Points= PrependPointList(Points,k,psize,red,green,blue);
}
MLDisownRealList(link,v,coord);

/*:747*/
#line 12235 "MathView3D.w"

}
else if(0==strcmp(fhead,"RGBColor")){
/*749:*/
#line 12384 "MathView3D.w"

MLGetReal(link,&red);
MLGetReal(link,&green);
MLGetReal(link,&blue);

/*:749*/
#line 12238 "MathView3D.w"

}
else if(0==strcmp(fhead,"SurfaceColor")){
/*750:*/
#line 12389 "MathView3D.w"

if(MLCheckFunction(link,"RGBColor",&n)){
MLGetReal(link,&sred);
MLGetReal(link,&sgreen);
MLGetReal(link,&sblue);
}


/*:750*/
#line 12241 "MathView3D.w"

}
else if(0==strcmp(fhead,"PointSize")){
MLGetReal(link,&psize);
}
else if(0==strcmp(fhead,"Cuboid")){
/*748:*/
#line 12344 "MathView3D.w"

MLGetRealList(link,&v,&coord);
vec[0]= v[0];vec[1]= v[1];vec[2]= v[2];
MLDisownRealList(link,v,coord);
if(2==n){
MLGetRealList(link,&v,&coord1);
vec1[0]= v[0];vec1[1]= v[1];vec1[2]= v[2];
MLDisownRealList(link,v,coord1);
}
else{
vec1[0]= vec[0]+1.0;vec1[1]= vec[1]+1.0;vec1[2]= vec[2]+1.0;
coord1= 3;
}
if(3==coord&&3==coord1){
plBuffer[3]= cubep[0]= AddToPointPool(vec[0],vec[1],vec[2]);
plBuffer[2]= cubep[1]= AddToPointPool(vec1[0],vec[1],vec[2]);
plBuffer[1]= cubep[2]= AddToPointPool(vec1[0],vec1[1],vec[2]);
plBuffer[0]= cubep[3]= AddToPointPool(vec[0],vec1[1],vec[2]);
plBuffer[4]= cubep[4]= AddToPointPool(vec[0],vec[1],vec1[2]);
plBuffer[5]= cubep[5]= AddToPointPool(vec1[0],vec[1],vec1[2]);
plBuffer[6]= cubep[6]= AddToPointPool(vec1[0],vec1[1],vec1[2]);
plBuffer[7]= cubep[7]= AddToPointPool(vec[0],vec1[1],vec1[2]);
Polygons= PrependLinePolyList(Polygons,Polygon3D,plBuffer,4,sred,sgreen,sblue);
Polygons= PrependLinePolyList(Polygons,Polygon3D,plBuffer+4,4,sred,sgreen,sblue);
plBuffer[0]= cubep[0];plBuffer[1]= cubep[1];
plBuffer[2]= cubep[5];plBuffer[3]= cubep[4];
Polygons= PrependLinePolyList(Polygons,Polygon3D,plBuffer,4,sred,sgreen,sblue);
plBuffer[0]= cubep[1];plBuffer[1]= cubep[2];
plBuffer[2]= cubep[6];plBuffer[3]= cubep[5];
Polygons= PrependLinePolyList(Polygons,Polygon3D,plBuffer,4,sred,sgreen,sblue);
plBuffer[0]= cubep[2];plBuffer[1]= cubep[3];
plBuffer[2]= cubep[7];plBuffer[3]= cubep[6];
Polygons= PrependLinePolyList(Polygons,Polygon3D,plBuffer,4,sred,sgreen,sblue);
plBuffer[0]= cubep[3];plBuffer[1]= cubep[0];
plBuffer[2]= cubep[4];plBuffer[3]= cubep[7];
Polygons= PrependLinePolyList(Polygons,Polygon3D,plBuffer,4,sred,sgreen,sblue);

}


/*:748*/
#line 12247 "MathView3D.w"

}
else if(0==strcmp(fhead,"EdgeForm")){
if(n)
MLGetReal(link,&dummy);
}
else if(0==strcmp(fhead,"Thickness")){
MLGetReal(link,&dummy);
}
else if(0==strcmp(fhead,"List")){
GetGraphics3D(link,fhead,n);
}
else if(0==strcmp(fhead,"MeshGraphics3D")){
/*743:*/
#line 12280 "MathView3D.w"

{int n,m,cn,cm;
double*data;
double*colors;
long dim,cdim;
RGBColor c= {1.0,1.0,1.0};

MLGetInteger(link,&n);
MLGetInteger(link,&m);
MLGetRealList(link,&data,&dim);
MLGetInteger(link,&cn);
MLGetInteger(link,&cm);
MLGetRealList(link,&colors,&cdim);
if(n*m*3==dim&&3*cn*cm==cdim)
Meshes= PrependMesh3DList(Meshes,n,m,(Point3D*)data,&c,cn,cm,colors);
MLDisownRealList(link,data,dim);
MLDisownRealList(link,colors,cdim);
}

/*:743*/
#line 12260 "MathView3D.w"

isMesh= GL_TRUE;
}
else{
MLDisownSymbol(link,fhead);
MLGetNext(link);
continue;
}
MLDisownSymbol(link,fhead);
}
return isMesh;
}


/*:742*/
#line 10749 "MathView3D.w"

/*669:*/
#line 11168 "MathView3D.w"

int MLPolygonReduce(double tol,int iter);
int MLPolygonReduce(double tol,int iter)
{int dpoly;

dpoly= 0;
if(Polygons)
dpoly= SimplifyPolygones(iter,tol,OctalPointTree);
if(dpoly){
/*681:*/
#line 11317 "MathView3D.w"

BoxRatios.x= BoxRatios.y= BoxRatios.z= 1.0;
/*682:*/
#line 11321 "MathView3D.w"

glutSetWindow(MVDrawWindow);
glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
if(ObjID>=0)glDeleteLists(ObjID,1);
ObjID= -1;
InitView(GL_TRUE);
glutPostRedisplay();
MakeNewView= GL_TRUE;

/*:682*/
#line 11319 "MathView3D.w"


/*:681*/
#line 11177 "MathView3D.w"

}
return dpoly;
}


/*:669*/
#line 10750 "MathView3D.w"

/*732:*/
#line 12057 "MathView3D.w"

void mvPutGraphics3D(void)
{long int arglen,n;
int shade,newScene= GL_FALSE;
double*vec;
int isMesh;
long int count;

/*687:*/
#line 11392 "MathView3D.w"

/*683:*/
#line 11330 "MathView3D.w"

{ViewCntrlWin*vc;

if(VCWindow){
vc= VCWindow;
if(glutGetWindow()==VCWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11336 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11340 "MathView3D.w"

glutSetWindow(VCWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:683*/
#line 11393 "MathView3D.w"

/*684:*/
#line 11348 "MathView3D.w"

{LightCntrlWin*vc;

if(LCWindow){
vc= LCWindow;
if(glutGetWindow()==LCWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11354 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11358 "MathView3D.w"

glutSetWindow(LCWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:684*/
#line 11394 "MathView3D.w"

/*685:*/
#line 11366 "MathView3D.w"

{AxisCtrlWin*vc;

if(AxisWindow){
vc= AxisWindow;
if(glutGetWindow()==AxisWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11372 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11376 "MathView3D.w"

glutSetWindow(AxisWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:685*/
#line 11395 "MathView3D.w"


/*:687*/
#line 12065 "MathView3D.w"

if(MLCheckFunction(stdlink,"List",&arglen)){
MLGetInteger(stdlink,&newScene);
/*736:*/
#line 12153 "MathView3D.w"

if(!MLCheckFunction(stdlink,"List",&n)){
MLPutSymbol(stdlink,"$Failed");
return;
}
else{
MLGetRealList(stdlink,&vec,&n);
memcpy((void*)&BoundP1,vec,n*sizeof(double));
MLDisownRealList(stdlink,vec,n);
MLGetRealList(stdlink,&vec,&n);
memcpy((void*)&BoundP2,vec,n*sizeof(double));
MLDisownRealList(stdlink,vec,n);
}

/*:736*/
#line 12068 "MathView3D.w"

/*737:*/
#line 12167 "MathView3D.w"

MLGetRealList(stdlink,&vec,&n);
memcpy((void*)&ViewPoint,vec,n*sizeof(double));
MLDisownRealList(stdlink,vec,n);

/*:737*/
#line 12069 "MathView3D.w"

/*739:*/
#line 12178 "MathView3D.w"

MLGetRealList(stdlink,&vec,&n);
memcpy((void*)&ViewVertical,vec,n*sizeof(double));
MLDisownRealList(stdlink,vec,n);

/*:739*/
#line 12070 "MathView3D.w"

/*740:*/
#line 12183 "MathView3D.w"

MLGetRealList(stdlink,&vec,&n);
LightNo= n/3;
memcpy((void*)LightPosition,vec,n*sizeof(double));
MLDisownRealList(stdlink,vec,n);
MLGetRealList(stdlink,&vec,&n);
memcpy((void*)LightColor,vec,n*sizeof(double));
MLDisownRealList(stdlink,vec,n);

/*:740*/
#line 12071 "MathView3D.w"

/*741:*/
#line 12192 "MathView3D.w"

MLGetRealList(stdlink,&vec,&n);
memcpy((void*)&AmbientColor,vec,n*sizeof(double));
MLDisownRealList(stdlink,vec,n);

/*:741*/
#line 12072 "MathView3D.w"

/*738:*/
#line 12172 "MathView3D.w"

MLGetRealList(stdlink,&vec,&n);
memcpy((void*)&BoxRatios,vec,n*sizeof(double));
MLDisownRealList(stdlink,vec,n);


/*:738*/
#line 12073 "MathView3D.w"

MLGetDouble(stdlink,&PointSphereSize);
MLGetDouble(stdlink,&LineTubeSize);
MLGetDouble(stdlink,&LineSegmentAngle);
/*14:*/
#line 441 "MathView3D.w"

CosSkipCont= cos(ToRadian*(fabs(LineSegmentAngle)<MaxLSAngle?LineSegmentAngle:MaxLSAngle));

/*:14*/
#line 12077 "MathView3D.w"

MLGetInteger(stdlink,&shade);
MLGetInteger(stdlink,&TubeSegments);
MLGetInteger(stdlink,&BlackBackground);


BlackBackground= !BlackBackground;
switch(shade){
case 1:
SmoothShading= GL_TRUE;
WireFrame= GL_FALSE;
break;
case 2:
SmoothShading= GL_FALSE;
WireFrame= GL_FALSE;
break;
case 3:
SmoothShading= GL_FALSE;
WireFrame= GL_TRUE;
break;
}
SetBackgroundColor(BlackBackground);
if(newScene){
int i;
/*649:*/
#line 10821 "MathView3D.w"

if(PointPool)free(PointPool);
PointPool= NULL;
if(NormalPool)
for(i= 0;i<PoolSize;i++)
if(NormalPool[i])free(NormalPool[i]);
if(NormalPool)free(NormalPool);
NormalPool= NULL;
MaxPoolSize= PoolSize= 0;
Points= RemovePointList(Points);
OctalPointTree= RemovePointOctTree(OctalPointTree);
Lines= RemoveLinePolyList(Lines);
Polygons= RemoveLinePolyList(Polygons);
Meshes= RemoveMesh3DList(Meshes);
Textures= RemoveTextureList(Textures);
if(SceneTextureMap)
SceneTextureMap= DeleteTextureMapping(SceneTextureMap);
glDeleteLists(ObjID,1);
ObjID= -1;

/*:649*/
#line 12101 "MathView3D.w"

}
if(MLCheckFunction(stdlink,"List",&count)){
isMesh= GetGraphics3D(stdlink,"List",count);
}
else{
GetGraphics3D(stdlink,"MeshGraphics3D",count);
isMesh= GL_TRUE;
}
}
else{
MLPutSymbol(stdlink,"$Failed");
return;
}
if(Polygons)AveragedNormals(Polygons);
/*682:*/
#line 11321 "MathView3D.w"

glutSetWindow(MVDrawWindow);
glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
if(ObjID>=0)glDeleteLists(ObjID,1);
ObjID= -1;
InitView(GL_TRUE);
glutPostRedisplay();
MakeNewView= GL_TRUE;

/*:682*/
#line 12116 "MathView3D.w"

MLPutFunction(stdlink,"List",4L);
if(isMesh){
MLPutFunction(stdlink,"MVMesh3D",1);
MLPutInteger(stdlink,MeshID(Meshes));
}
else MLPutFunction(stdlink,"MVScene3D",0);
/*733:*/
#line 12129 "MathView3D.w"

MLPutFunction(stdlink,"Rule",2L);
MLPutSymbol(stdlink,"PlotRange");
MLPutList(3L);
MLPutList(2L);
MLPutReal(stdlink,BoundP1.x);
MLPutReal(stdlink,BoundP2.x);
MLPutList(2L);
MLPutReal(stdlink,BoundP1.y);
MLPutReal(stdlink,BoundP2.y);
MLPutList(2L);
MLPutReal(stdlink,BoundP1.z);
MLPutReal(stdlink,BoundP2.z);

/*:733*/
#line 12123 "MathView3D.w"

/*734:*/
#line 12143 "MathView3D.w"

MLPutFunction(stdlink,"Rule",2L);
MLPutSymbol(stdlink,"ViewPoint");
MLPutRealList(stdlink,&ViewPoint.x,3L);

/*:734*/
#line 12124 "MathView3D.w"

/*735:*/
#line 12148 "MathView3D.w"

MLPutFunction(stdlink,"Rule",2L);
MLPutSymbol(stdlink,"AmbientLight");
MLPutRealList(stdlink,&AmbientColor.r,3L);

/*:735*/
#line 12125 "MathView3D.w"

MLFlush(stdlink);
}

/*:732*/
#line 10751 "MathView3D.w"

/*648:*/
#line 10798 "MathView3D.w"

void mvClearContents(void)
{int i;

/*649:*/
#line 10821 "MathView3D.w"

if(PointPool)free(PointPool);
PointPool= NULL;
if(NormalPool)
for(i= 0;i<PoolSize;i++)
if(NormalPool[i])free(NormalPool[i]);
if(NormalPool)free(NormalPool);
NormalPool= NULL;
MaxPoolSize= PoolSize= 0;
Points= RemovePointList(Points);
OctalPointTree= RemovePointOctTree(OctalPointTree);
Lines= RemoveLinePolyList(Lines);
Polygons= RemoveLinePolyList(Polygons);
Meshes= RemoveMesh3DList(Meshes);
Textures= RemoveTextureList(Textures);
if(SceneTextureMap)
SceneTextureMap= DeleteTextureMapping(SceneTextureMap);
glDeleteLists(ObjID,1);
ObjID= -1;

/*:649*/
#line 10802 "MathView3D.w"

DoTracking= DoTrackSpin= GL_FALSE;
ZoomMove.x= ZoomMove.y= ZoomMove.z= 0.0;
RadialZoom= 1.0;
TrackAngle= 0.0;
TrackAxis.x= TrackAxis.y= 0.0;
TrackAxis.z= 1.0;
ScreenXStart= ScreenX= 0;
ScreenYStart= ScreenY= 0;
BoxRatios.x= BoxRatios.y= BoxRatios.z= 1.0;
BlackBackground= GL_TRUE;
MVFaceFlip= GL_FALSE;
Axis.AxisType= NoAxis;
LightNo= 0;
glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
/*681:*/
#line 11317 "MathView3D.w"

BoxRatios.x= BoxRatios.y= BoxRatios.z= 1.0;
/*682:*/
#line 11321 "MathView3D.w"

glutSetWindow(MVDrawWindow);
glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
if(ObjID>=0)glDeleteLists(ObjID,1);
ObjID= -1;
InitView(GL_TRUE);
glutPostRedisplay();
MakeNewView= GL_TRUE;

/*:682*/
#line 11319 "MathView3D.w"


/*:681*/
#line 10817 "MathView3D.w"

/*657:*/
#line 10902 "MathView3D.w"

glutPostRedisplay();
if(DoMathematicaCall){
MLPutSymbol(stdlink,"True");
MLFlush(stdlink);
}

/*:657*/
#line 10818 "MathView3D.w"

}

/*:648*/
#line 10752 "MathView3D.w"

/*659:*/
#line 10923 "MathView3D.w"

void mvSpin(double angle,double*axis,long int n);
void mvSpin(double angle,double*axis,long int n)
{double atrack;

if(3!=n){
MLPutSymbol(stdlink,"$Failed");
return;
}
TrackAxis.x= axis[0];TrackAxis.y= axis[1];TrackAxis.z= axis[2];
atrack= Norm(&TrackAxis);
if(atrack!=0.0){
TrackAxis.x/= atrack;
TrackAxis.y/= atrack;
TrackAxis.z/= atrack;
}
else{
MLPutSymbol(stdlink,"$Failed");
return;
}
TrackAngle= angle;
if(TrackAngle>0.0)
DTrackAngle= MIN(TWO_PI/90.0,TrackAngle);
else
DTrackAngle= MAX(-TWO_PI/90.0,TrackAngle);
DoTrackSpin= GL_TRUE;
DoTracking= GL_TRUE;
GetMouseTrack= GL_FALSE;
MLPutSymbol(stdlink,"Null");
}

/*:659*/
#line 10753 "MathView3D.w"

/*660:*/
#line 10954 "MathView3D.w"

void mvStopSpin(void);
void mvStopSpin(void)
{
/*47:*/
#line 1230 "MathView3D.w"

DoTrackSpin= GL_FALSE;
TrackAngle= 0.0;
DoTracking= GL_FALSE;
glutSetCursor(GLUT_CURSOR_LEFT_ARROW);


/*:47*/
#line 10958 "MathView3D.w"

/*657:*/
#line 10902 "MathView3D.w"

glutPostRedisplay();
if(DoMathematicaCall){
MLPutSymbol(stdlink,"True");
MLFlush(stdlink);
}

/*:657*/
#line 10959 "MathView3D.w"

}


/*:660*/
#line 10754 "MathView3D.w"

/*646:*/
#line 10789 "MathView3D.w"

int mvGetPoolSize(void)
{
return PoolSize;
}

/*:646*/
#line 10755 "MathView3D.w"

/*671:*/
#line 11186 "MathView3D.w"

void mlWritePOVray(const char*fname,const char*ptexpre,const char*mtexpre)
{int res;
char filename[512];
int i,j,n;

/*674:*/
#line 11217 "MathView3D.w"

n= strlen(fname);
for(j= i= 0;i<n;i++)
if(fname[i]!='\\')
filename[j++]= fname[i];
else if(fname[++i]=='\\')
filename[j++]= '\\';
filename[j++]= fname[n];filename[j+1]= '\0';
#ifdef __WIN32__
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= '\\';
#elif MACINTOSH_MATHLINK
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= ':';
#endif

/*:674*/
#line 11192 "MathView3D.w"

res= WritePOVray(filename,SmoothShading,ptexpre,mtexpre);
if(res)
MLPutSymbol(stdlink,"$Failed");
else
MLPutString(stdlink,fname);
}

/*:671*/
#line 10756 "MathView3D.w"

/*673:*/
#line 11203 "MathView3D.w"

void mlWrite3DMF(const char*fname,int savelight,int saveview)
{int res;
char filename[512];
int i,j,n;

/*674:*/
#line 11217 "MathView3D.w"

n= strlen(fname);
for(j= i= 0;i<n;i++)
if(fname[i]!='\\')
filename[j++]= fname[i];
else if(fname[++i]=='\\')
filename[j++]= '\\';
filename[j++]= fname[n];filename[j+1]= '\0';
#ifdef __WIN32__
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= '\\';
#elif MACINTOSH_MATHLINK
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= ':';
#endif

/*:674*/
#line 11209 "MathView3D.w"

res= Write3DMF(filename,savelight,saveview);
if(res)
MLPutSymbol(stdlink,"$Failed");
else
MLPutString(stdlink,fname);
}

/*:673*/
#line 10757 "MathView3D.w"

/*695:*/
#line 11455 "MathView3D.w"

void mlRawWriteVRML(const char*fname,int headlight,int useextrude,
const char*ptexpre,const char*mtexpre)
{char filename[512];
int i,j,n;
int res;

/*674:*/
#line 11217 "MathView3D.w"

n= strlen(fname);
for(j= i= 0;i<n;i++)
if(fname[i]!='\\')
filename[j++]= fname[i];
else if(fname[++i]=='\\')
filename[j++]= '\\';
filename[j++]= fname[n];filename[j+1]= '\0';
#ifdef __WIN32__
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= '\\';
#elif MACINTOSH_MATHLINK
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= ':';
#endif

/*:674*/
#line 11462 "MathView3D.w"

res= WriteVRML(filename,headlight,useextrude,ptexpre,mtexpre);
if(!res)
MLPutString(stdlink,fname);
else
MLPutSymbol(stdlink,"$Failed");
}

/*:695*/
#line 10758 "MathView3D.w"

/*676:*/
#line 11238 "MathView3D.w"

void MVLoad3Script(const char*fname)
{int res;
char filename[512];
int i,j,n;
Point3D b1,b2;

/*687:*/
#line 11392 "MathView3D.w"

/*683:*/
#line 11330 "MathView3D.w"

{ViewCntrlWin*vc;

if(VCWindow){
vc= VCWindow;
if(glutGetWindow()==VCWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11336 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11340 "MathView3D.w"

glutSetWindow(VCWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:683*/
#line 11393 "MathView3D.w"

/*684:*/
#line 11348 "MathView3D.w"

{LightCntrlWin*vc;

if(LCWindow){
vc= LCWindow;
if(glutGetWindow()==LCWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11354 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11358 "MathView3D.w"

glutSetWindow(LCWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:684*/
#line 11394 "MathView3D.w"

/*685:*/
#line 11366 "MathView3D.w"

{AxisCtrlWin*vc;

if(AxisWindow){
vc= AxisWindow;
if(glutGetWindow()==AxisWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11372 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11376 "MathView3D.w"

glutSetWindow(AxisWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:685*/
#line 11395 "MathView3D.w"


/*:687*/
#line 11245 "MathView3D.w"

/*674:*/
#line 11217 "MathView3D.w"

n= strlen(fname);
for(j= i= 0;i<n;i++)
if(fname[i]!='\\')
filename[j++]= fname[i];
else if(fname[++i]=='\\')
filename[j++]= '\\';
filename[j++]= fname[n];filename[j+1]= '\0';
#ifdef __WIN32__
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= '\\';
#elif MACINTOSH_MATHLINK
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= ':';
#endif

/*:674*/
#line 11246 "MathView3D.w"

b1= BoundP1;b2= BoundP2;
res= Read3Script(filename);
if(res)
MLPutSymbol(stdlink,"$Failed");
else{
MLPutString(stdlink,fname);
BoundP1.x= MIN(b1.x,BoundP1.x);
BoundP1.y= MIN(b1.y,BoundP1.y);
BoundP1.z= MIN(b1.z,BoundP1.z);
BoundP2.x= MAX(b2.x,BoundP2.x);
BoundP2.y= MAX(b2.y,BoundP2.y);
BoundP2.z= MAX(b2.z,BoundP2.z);
}
AveragedNormals(Polygons);
/*681:*/
#line 11317 "MathView3D.w"

BoxRatios.x= BoxRatios.y= BoxRatios.z= 1.0;
/*682:*/
#line 11321 "MathView3D.w"

glutSetWindow(MVDrawWindow);
glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
if(ObjID>=0)glDeleteLists(ObjID,1);
ObjID= -1;
InitView(GL_TRUE);
glutPostRedisplay();
MakeNewView= GL_TRUE;

/*:682*/
#line 11319 "MathView3D.w"


/*:681*/
#line 11261 "MathView3D.w"

}


/*:676*/
#line 10759 "MathView3D.w"

/*678:*/
#line 11268 "MathView3D.w"

void mlWriteDXF(const char*fname,
double ts,int tseg,double pssize,double tangle)
{int res;
double tsSave,psSave,cosaSave;
int tsegSave;
char filename[512];
int i,j,n;

tsSave= LineTubeSize;
LineTubeSize= ts;
tsegSave= TubeSegments;
TubeSegments= tseg;
psSave= PointSphereSize;
PointSphereSize= pssize;
cosaSave= CosSkipCont;
CosSkipCont= cos(ToRadian*(fabs(tangle)<MaxLSAngle?tangle:MaxLSAngle));
/*674:*/
#line 11217 "MathView3D.w"

n= strlen(fname);
for(j= i= 0;i<n;i++)
if(fname[i]!='\\')
filename[j++]= fname[i];
else if(fname[++i]=='\\')
filename[j++]= '\\';
filename[j++]= fname[n];filename[j+1]= '\0';
#ifdef __WIN32__
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= '\\';
#elif MACINTOSH_MATHLINK
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= ':';
#endif

/*:674*/
#line 11285 "MathView3D.w"

res= WriteDXF(filename);
LineTubeSize= tsSave;
TubeSegments= tsegSave;
PointSphereSize= psSave;
CosSkipCont= cosaSave;
if(res)
MLPutSymbol(stdlink,"$Failed");
else
MLPutString(stdlink,fname);
}

/*:678*/
#line 10760 "MathView3D.w"

/*680:*/
#line 11300 "MathView3D.w"

void MVLoadDXF(const char*fname)
{int res;
char filename[512];
int i,j,n;

/*687:*/
#line 11392 "MathView3D.w"

/*683:*/
#line 11330 "MathView3D.w"

{ViewCntrlWin*vc;

if(VCWindow){
vc= VCWindow;
if(glutGetWindow()==VCWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11336 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11340 "MathView3D.w"

glutSetWindow(VCWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:683*/
#line 11393 "MathView3D.w"

/*684:*/
#line 11348 "MathView3D.w"

{LightCntrlWin*vc;

if(LCWindow){
vc= LCWindow;
if(glutGetWindow()==LCWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11354 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11358 "MathView3D.w"

glutSetWindow(LCWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:684*/
#line 11394 "MathView3D.w"

/*685:*/
#line 11366 "MathView3D.w"

{AxisCtrlWin*vc;

if(AxisWindow){
vc= AxisWindow;
if(glutGetWindow()==AxisWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11372 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11376 "MathView3D.w"

glutSetWindow(AxisWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:685*/
#line 11395 "MathView3D.w"


/*:687*/
#line 11306 "MathView3D.w"

/*674:*/
#line 11217 "MathView3D.w"

n= strlen(fname);
for(j= i= 0;i<n;i++)
if(fname[i]!='\\')
filename[j++]= fname[i];
else if(fname[++i]=='\\')
filename[j++]= '\\';
filename[j++]= fname[n];filename[j+1]= '\0';
#ifdef __WIN32__
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= '\\';
#elif MACINTOSH_MATHLINK
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= ':';
#endif

/*:674*/
#line 11307 "MathView3D.w"

res= ReadDXF(filename);
if(!res)
MLPutSymbol(stdlink,"$Failed");
else
MLPutString(stdlink,fname);
AveragedNormals(Polygons);
/*681:*/
#line 11317 "MathView3D.w"

BoxRatios.x= BoxRatios.y= BoxRatios.z= 1.0;
/*682:*/
#line 11321 "MathView3D.w"

glutSetWindow(MVDrawWindow);
glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
if(ObjID>=0)glDeleteLists(ObjID,1);
ObjID= -1;
InitView(GL_TRUE);
glutPostRedisplay();
MakeNewView= GL_TRUE;

/*:682*/
#line 11319 "MathView3D.w"


/*:681*/
#line 11314 "MathView3D.w"

}

/*:680*/
#line 10761 "MathView3D.w"

/*689:*/
#line 11400 "MathView3D.w"

void mvWriteOFF(const char*fname)
{int res;
char filename[512];
int i,j,n;

/*674:*/
#line 11217 "MathView3D.w"

n= strlen(fname);
for(j= i= 0;i<n;i++)
if(fname[i]!='\\')
filename[j++]= fname[i];
else if(fname[++i]=='\\')
filename[j++]= '\\';
filename[j++]= fname[n];filename[j+1]= '\0';
#ifdef __WIN32__
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= '\\';
#elif MACINTOSH_MATHLINK
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= ':';
#endif

/*:674*/
#line 11406 "MathView3D.w"

res= WriteOFF(filename);
if(res)
MLPutSymbol(stdlink,"$Failed");
else
MLPutString(stdlink,fname);
}

/*:689*/
#line 10762 "MathView3D.w"

/*691:*/
#line 11417 "MathView3D.w"

void MVLoadOFF(const char*fname)
{int res;
char filename[512];
int i,j,n;

/*687:*/
#line 11392 "MathView3D.w"

/*683:*/
#line 11330 "MathView3D.w"

{ViewCntrlWin*vc;

if(VCWindow){
vc= VCWindow;
if(glutGetWindow()==VCWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11336 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11340 "MathView3D.w"

glutSetWindow(VCWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:683*/
#line 11393 "MathView3D.w"

/*684:*/
#line 11348 "MathView3D.w"

{LightCntrlWin*vc;

if(LCWindow){
vc= LCWindow;
if(glutGetWindow()==LCWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11354 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11358 "MathView3D.w"

glutSetWindow(LCWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:684*/
#line 11394 "MathView3D.w"

/*685:*/
#line 11366 "MathView3D.w"

{AxisCtrlWin*vc;

if(AxisWindow){
vc= AxisWindow;
if(glutGetWindow()==AxisWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11372 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11376 "MathView3D.w"

glutSetWindow(AxisWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:685*/
#line 11395 "MathView3D.w"


/*:687*/
#line 11423 "MathView3D.w"

/*674:*/
#line 11217 "MathView3D.w"

n= strlen(fname);
for(j= i= 0;i<n;i++)
if(fname[i]!='\\')
filename[j++]= fname[i];
else if(fname[++i]=='\\')
filename[j++]= '\\';
filename[j++]= fname[n];filename[j+1]= '\0';
#ifdef __WIN32__
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= '\\';
#elif MACINTOSH_MATHLINK
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= ':';
#endif

/*:674*/
#line 11424 "MathView3D.w"

res= ReadOFF(filename);
if(res)
MLPutSymbol(stdlink,"$Failed");
else
MLPutString(stdlink,fname);
AveragedNormals(Polygons);
/*681:*/
#line 11317 "MathView3D.w"

BoxRatios.x= BoxRatios.y= BoxRatios.z= 1.0;
/*682:*/
#line 11321 "MathView3D.w"

glutSetWindow(MVDrawWindow);
glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
if(ObjID>=0)glDeleteLists(ObjID,1);
ObjID= -1;
InitView(GL_TRUE);
glutPostRedisplay();
MakeNewView= GL_TRUE;

/*:682*/
#line 11319 "MathView3D.w"


/*:681*/
#line 11431 "MathView3D.w"

}

/*:691*/
#line 10763 "MathView3D.w"

/*693:*/
#line 11437 "MathView3D.w"

void mlWritePostScript(const char*fname,int bitmapImage)
{char filename[512];
int i,j,n;

/*674:*/
#line 11217 "MathView3D.w"

n= strlen(fname);
for(j= i= 0;i<n;i++)
if(fname[i]!='\\')
filename[j++]= fname[i];
else if(fname[++i]=='\\')
filename[j++]= '\\';
filename[j++]= fname[n];filename[j+1]= '\0';
#ifdef __WIN32__
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= '\\';
#elif MACINTOSH_MATHLINK
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= ':';
#endif

/*:674*/
#line 11442 "MathView3D.w"

if(bitmapImage)
EPSSaveFrameBuffer(filename,
0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
else
WritePostScript(filename);
MLPutString(stdlink,fname);
}

/*:693*/
#line 10764 "MathView3D.w"

/*697:*/
#line 11473 "MathView3D.w"

void mlRawWritePNG(const char*fname,double gamma,int interlace)
{char filename[512];
int i,j,n,res;

/*687:*/
#line 11392 "MathView3D.w"

/*683:*/
#line 11330 "MathView3D.w"

{ViewCntrlWin*vc;

if(VCWindow){
vc= VCWindow;
if(glutGetWindow()==VCWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11336 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11340 "MathView3D.w"

glutSetWindow(VCWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:683*/
#line 11393 "MathView3D.w"

/*684:*/
#line 11348 "MathView3D.w"

{LightCntrlWin*vc;

if(LCWindow){
vc= LCWindow;
if(glutGetWindow()==LCWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11354 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11358 "MathView3D.w"

glutSetWindow(LCWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:684*/
#line 11394 "MathView3D.w"

/*685:*/
#line 11366 "MathView3D.w"

{AxisCtrlWin*vc;

if(AxisWindow){
vc= AxisWindow;
if(glutGetWindow()==AxisWindow->WinID){
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11372 "MathView3D.w"

glutHideWindow();
}
else{
/*686:*/
#line 11385 "MathView3D.w"

#if MACINTOSH_MATHLINK
glutSetWindow(vc->PreviewWinID);
glutHideWindow();
glutSetWindow(vc->WinID);
#endif

/*:686*/
#line 11376 "MathView3D.w"

glutSetWindow(AxisWindow->WinID);
glutHideWindow();
}
}
glutSetWindow(MVDrawWindow);
}

/*:685*/
#line 11395 "MathView3D.w"


/*:687*/
#line 11478 "MathView3D.w"

/*674:*/
#line 11217 "MathView3D.w"

n= strlen(fname);
for(j= i= 0;i<n;i++)
if(fname[i]!='\\')
filename[j++]= fname[i];
else if(fname[++i]=='\\')
filename[j++]= '\\';
filename[j++]= fname[n];filename[j+1]= '\0';
#ifdef __WIN32__
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= '\\';
#elif MACINTOSH_MATHLINK
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= ':';
#endif

/*:674*/
#line 11479 "MathView3D.w"

glutPostRedisplay();
res= PNGSaveFrameBuffer(filename,
0,0,glutGet(GLUT_WINDOW_WIDTH),
glutGet(GLUT_WINDOW_HEIGHT),interlace,gamma);
if(res)
MLPutString(stdlink,fname);
else
MLPutSymbol(stdlink,"$Failed");
}

/*:697*/
#line 10765 "MathView3D.w"

/*730:*/
#line 12024 "MathView3D.w"

void mlRawWriteDepthBuffer(const char*fname,int fmt,double gamma,
int interlace,double scale)
{char filename[512];
int i,j,n;

/*674:*/
#line 11217 "MathView3D.w"

n= strlen(fname);
for(j= i= 0;i<n;i++)
if(fname[i]!='\\')
filename[j++]= fname[i];
else if(fname[++i]=='\\')
filename[j++]= '\\';
filename[j++]= fname[n];filename[j+1]= '\0';
#ifdef __WIN32__
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= '\\';
#elif MACINTOSH_MATHLINK
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= ':';
#endif

/*:674*/
#line 12030 "MathView3D.w"

PNGSaveDepthBuffer(
filename,
0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT),
fmt,
interlace,gamma,scale
);
MLPutString(stdlink,fname);
}

/*:730*/
#line 10766 "MathView3D.w"

/*731:*/
#line 12042 "MathView3D.w"

void mlRawWriteSIRDS(const char*fname,
int dpiOut,double trueWidth,double eyeSeparation,
double mu,int osample,double scale)
{char filename[512];
int i,j,n;

/*674:*/
#line 11217 "MathView3D.w"

n= strlen(fname);
for(j= i= 0;i<n;i++)
if(fname[i]!='\\')
filename[j++]= fname[i];
else if(fname[++i]=='\\')
filename[j++]= '\\';
filename[j++]= fname[n];filename[j+1]= '\0';
#ifdef __WIN32__
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= '\\';
#elif MACINTOSH_MATHLINK
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= ':';
#endif

/*:674*/
#line 12049 "MathView3D.w"

PNGSaveSIRDS(filename,0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT),
dpiOut,trueWidth,eyeSeparation,mu,osample,scale);
MLPutString(stdlink,fname);
}


/*:731*/
#line 10767 "MathView3D.w"

/*661:*/
#line 10963 "MathView3D.w"

void mvSendBoundingBox(void);
void mvSendBoundingBox(void)
{int i;
if(Points||Lines||Polygons||Meshes){
MLPutFunction(stdlink,"List",2);
MLPutRealList(stdlink,(double*)&BoundP1,3L);
MLPutRealList(stdlink,(double*)&BoundP2,3L);
}
else{
MLPutFunction(stdlink,"List",2);
/*662:*/
#line 10979 "MathView3D.w"

MLPutFunction(stdlink,"List",3);
for(i= 0;i<3;i++)
MLPutSymbol(stdlink,"Indeterminate");


/*:662*/
#line 10974 "MathView3D.w"

/*662:*/
#line 10979 "MathView3D.w"

MLPutFunction(stdlink,"List",3);
for(i= 0;i<3;i++)
MLPutSymbol(stdlink,"Indeterminate");


/*:662*/
#line 10975 "MathView3D.w"

}
}

/*:661*/
#line 10768 "MathView3D.w"

/*663:*/
#line 10985 "MathView3D.w"

void mvSendViewPoint(void)
{
MLPutRealList(stdlink,(double*)&ViewPoint,3L);
}

/*:663*/
#line 10769 "MathView3D.w"

/*664:*/
#line 10991 "MathView3D.w"

void mvSendLightSources(void);
void mvSendLightSources(void)
{int i;

MLPutFunction(stdlink,"List",LightNo);
for(i= 0;i<LightNo;i++){
MLPutFunction(stdlink,"List",2L);
MLPutRealList(stdlink,(double*)(LightPosition+i),3L);
MLPutFunction(stdlink,"RGBColor",3L);
MLPutDouble(stdlink,LightColor[i].r);
MLPutDouble(stdlink,LightColor[i].g);
MLPutDouble(stdlink,LightColor[i].b);
}
}

/*:664*/
#line 10770 "MathView3D.w"

/*718:*/
#line 11806 "MathView3D.w"

/*719:*/
#line 11814 "MathView3D.w"

void MLSendRawPoint(MLINK lnk,int i)
{Point3D*p= PointPool+i;

MLPutFunction(lnk,"List",3);
MLPutDouble(lnk,p->x);
MLPutDouble(lnk,p->y);
MLPutDouble(lnk,p->z);
}

/*:719*/
#line 11807 "MathView3D.w"

/*720:*/
#line 11824 "MathView3D.w"

void MLSendRawMeshPoint(MLINK lnk,Mesh3DList*lst,int i,int j)
{Point3D*p;

p= MeshCoordPtr(lst,i,j);
MLPutFunction(lnk,"List",3);
MLPutDouble(lnk,p->x);
MLPutDouble(lnk,p->y);
MLPutDouble(lnk,p->z);
}

/*:720*/
#line 11808 "MathView3D.w"

/*721:*/
#line 11835 "MathView3D.w"

PointList*MLSendPoint(MLINK lnk,PointList*lst)
{
MLPutFunction(lnk,"Point",1);
MLSendRawPoint(lnk,lst->point);
return lst->next;
}

/*:721*/
#line 11809 "MathView3D.w"

/*722:*/
#line 11843 "MathView3D.w"

LinePolyList*MLSendLinePolygon(MLINK lnk,LinePolyList*lst)
{int i;

MLPutFunction(lnk,lst->what==Polygon3D?"Polygon":"Line",1);
MLPutFunction(lnk,"List",lst->pno);
for(i= 0;i<lst->pno;i++)
MLSendRawPoint(lnk,lst->points[i]);
return lst->next;
}

/*:722*/
#line 11810 "MathView3D.w"

/*723:*/
#line 11856 "MathView3D.w"

Mesh3DList*MLSendMesh3D(MLINK lnk,Mesh3DList*lst)
{int polyNo;
int i,j,n,m;

polyNo= (n= lst->dimU-1)*(m= lst->dimV-1);
MLPutFunction(lnk,"List",polyNo);
if(lst->mpolyColors)
/*725:*/
#line 11876 "MathView3D.w"

{RGBColor*c;

for(i= 0;i<n;i++)
for(j= 0;j<m;j++){
c= MeshColorPtr(lst,i,j);
MLPutFunction(lnk,"List",2);
MLPutFunction(lnk,"SurfaceColor",1);
MLPutFunction(lnk,"RGBColor",3);
MLPutDouble(lnk,c->r);
MLPutDouble(lnk,c->g);
MLPutDouble(lnk,c->b);
/*726:*/
#line 11892 "MathView3D.w"

MLPutFunction(lnk,"Polygon",1);
MLPutFunction(lnk,"List",4);
MLSendRawMeshPoint(lnk,lst,i,j);
MLSendRawMeshPoint(lnk,lst,i+1,j);
MLSendRawMeshPoint(lnk,lst,i+1,j+1);
MLSendRawMeshPoint(lnk,lst,i,j+1);


/*:726*/
#line 11888 "MathView3D.w"

}
}

/*:725*/
#line 11864 "MathView3D.w"

else
/*724:*/
#line 11870 "MathView3D.w"

for(i= 0;i<n;i++)
for(j= 0;j<m;j++){
/*726:*/
#line 11892 "MathView3D.w"

MLPutFunction(lnk,"Polygon",1);
MLPutFunction(lnk,"List",4);
MLSendRawMeshPoint(lnk,lst,i,j);
MLSendRawMeshPoint(lnk,lst,i+1,j);
MLSendRawMeshPoint(lnk,lst,i+1,j+1);
MLSendRawMeshPoint(lnk,lst,i,j+1);


/*:726*/
#line 11873 "MathView3D.w"

}

/*:724*/
#line 11866 "MathView3D.w"

return lst->next;
}

/*:723*/
#line 11811 "MathView3D.w"

/*727:*/
#line 11901 "MathView3D.w"

void MLSendGraphics(void)
{int objNo= 0;
int pointNo,lineNo,polyNo,meshNo;
PointList*plst;
LinePolyList*llst;
Mesh3DList*mlst;

pointNo= lineNo= polyNo= 0;
if(Points){
objNo++;
pointNo= LengthPointList(Points);
}
if(Lines){
objNo++;
lineNo= LengthLinePolyList(Lines);
}
if(Polygons){
objNo++;
polyNo= LengthLinePolyList(Polygons);
}
if(Meshes){
objNo++;
meshNo= LengthMesh3DList(Meshes);
}
MLPutFunction(stdlink,"Graphics3D",1);
MLPutFunction(stdlink,"List",objNo);
if(Points){
plst= Points;
MLPutFunction(stdlink,"List",pointNo);
while(plst)plst= MLSendPoint(stdlink,plst);
}
if(Lines){
llst= Lines;
MLPutFunction(stdlink,"List",lineNo);
while(llst)llst= MLSendLinePolygon(stdlink,llst);
}
if(Polygons){
llst= Polygons;
MLPutFunction(stdlink,"List",polyNo);
while(llst)llst= MLSendLinePolygon(stdlink,llst);
}
if(Meshes){
mlst= Meshes;
MLPutFunction(stdlink,"List",meshNo);
while(mlst)mlst= MLSendMesh3D(stdlink,mlst);
}
}

/*:727*/
#line 11812 "MathView3D.w"


/*:718*/
#line 10771 "MathView3D.w"

/*728:*/
#line 11952 "MathView3D.w"

void MLSendFrameBuffer(void)
{int w,h;
long i,size;
unsigned char*fbuff;
double*dfbuffer;

glutSetWindow(MVDrawWindow);
w= glutGet(GLUT_WINDOW_WIDTH);
h= glutGet(GLUT_WINDOW_HEIGHT);
size= 3*w*h;
if(!size){
MLPutSymbol(stdlink,"$Failed");
return;
}
fbuff= (unsigned char*)calloc(size,sizeof(unsigned char));
dfbuffer= (double*)calloc(size,sizeof(double));
DoPSOutput= GL_TRUE;

glFlush();
glFinish();
glReadBuffer(GL_BACK);







glPixelStorei(GL_PACK_ALIGNMENT,1);
glReadPixels(0,0,w,h,GL_RGB,GL_UNSIGNED_BYTE,fbuff);
for(i= 0;i<size;i++)
dfbuffer[i]= (double)fbuff[i]/(255.0);
free(fbuff);
/*729:*/
#line 12006 "MathView3D.w"

MLPutFunction(stdlink,"RasterArray",1);
MLPutFunction(stdlink,"Map",3);
MLPutFunction(stdlink,"Function",1);
MLPutFunction(stdlink,"Apply",2);
MLPutSymbol(stdlink,"RGBColor");
MLPutFunction(stdlink,"Slot",1);
MLPutInteger(stdlink,1);
MLPutFunction(stdlink,"Partition",2);
MLPutFunction(stdlink,"Partition",2);
MLPutRealList(stdlink,dfbuffer,size);
MLPutInteger(stdlink,3);
MLPutInteger(stdlink,w);
MLPutFunction(stdlink,"List",1);
MLPutInteger(stdlink,2);

/*:729*/
#line 11986 "MathView3D.w"

free(dfbuffer);
DoPSOutput= GL_FALSE;
glDrawBuffer(GL_BACK);
glutPostRedisplay();
}

/*:728*/
#line 10772 "MathView3D.w"

/*699:*/
#line 11512 "MathView3D.w"

void mlLoadPNGTexture(const char*fname,double gamma)
{char filename[512];
int i,j,n;
TextureList*tlst;

/*674:*/
#line 11217 "MathView3D.w"

n= strlen(fname);
for(j= i= 0;i<n;i++)
if(fname[i]!='\\')
filename[j++]= fname[i];
else if(fname[++i]=='\\')
filename[j++]= '\\';
filename[j++]= fname[n];filename[j+1]= '\0';
#ifdef __WIN32__
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= '\\';
#elif MACINTOSH_MATHLINK
n= strlen(filename);
for(i= 0;i<n;i++)
if('/'==filename[i])filename[i]= ':';
#endif

/*:674*/
#line 11518 "MathView3D.w"

tlst= ReadPNGTexture(Textures,filename,gamma,GL_TRUE);
if(!tlst){
MLPutSymbol(stdlink,"$Failed");
return;
}
Textures= tlst;
/*700:*/
#line 11528 "MathView3D.w"

MLPutFunction(stdlink,"List",4);
MLPutFunction(stdlink,"MVTexture2D",1);
MLPutInteger(stdlink,TextureID(tlst));
/*701:*/
#line 11535 "MathView3D.w"

MLPutInteger(stdlink,tlst->texWidth);
MLPutInteger(stdlink,tlst->texHeight);

/*:701*/
#line 11532 "MathView3D.w"

/*702:*/
#line 11539 "MathView3D.w"

switch(TextureType(tlst)){
case GL_LUMINANCE:
MLPutSymbol(stdlink,"MVLuminantTexture");
break;
case GL_LUMINANCE_ALPHA:
MLPutSymbol(stdlink,"MVLuminantAlphaTexture");
break;
case GL_RGB:
MLPutSymbol(stdlink,"MVRGBTexture");
break;
case GL_RGBA:
MLPutSymbol(stdlink,"MVRGBAlphaTexture");
break;
default:
MLPutSymbol(stdlink,"Null");
break;
}

/*:702*/
#line 11533 "MathView3D.w"


/*:700*/
#line 11525 "MathView3D.w"

}

/*:699*/
#line 10773 "MathView3D.w"

/*703:*/
#line 11561 "MathView3D.w"

void mlGraphicsToTexture(int clipborder)
{int width,height,bitdepth,i,j,wclip,hclip;
char*unistr,*endstr,*pbm,*sp,*imagedata,*pbmclip;
int len;
TextureList*tlst;

/*704:*/
#line 11595 "MathView3D.w"

MLGetString(stdlink,(const char**)&unistr);
len= strlen(unistr);
endstr= unistr+len;
imagedata= pbm= (char*)calloc(len,sizeof(char));
for(j= i= 0,sp= unistr;i<len;i++){
pbm[j++]= (char)MLNextCharacter((const char**)&sp,(const char*)endstr);
if(!pbm)break;
}
MLDisownString(stdlink,unistr);

/*:704*/
#line 11568 "MathView3D.w"

/*705:*/
#line 11606 "MathView3D.w"

{
while(*pbm!='\012')pbm++;
sscanf(++pbm,"%d %d\n",&width,&height);
while(*pbm!='\012')pbm++;
sscanf(++pbm,"%d",&bitdepth);
bitdepth= 8;
while(*pbm!='\012')pbm++;
pbm++;
}

/*:705*/
#line 11569 "MathView3D.w"

/*707:*/
#line 11620 "MathView3D.w"

ReverseImageRows(width,height,3,(GLubyte*)pbm);

/*:707*/
#line 11570 "MathView3D.w"

if(clipborder){
ClipRGBImageBorder(width,height,(GLubyte*)pbm,
&wclip,&hclip,(GLubyte**)&pbmclip);
tlst= CreateTexture(Textures,"Automatic",
GL_RGB,wclip,hclip,(GLubyte*)pbmclip,3L*wclip*hclip,
GL_TRUE);
}
else
tlst= CreateTexture(Textures,"Automatic",
GL_RGB,width,height,(GLubyte*)pbm,3L*width*height,
GL_TRUE);
if(!tlst){
MLPutSymbol(stdlink,"$Failed");
free(imagedata);
return;
}
Textures= tlst;
/*700:*/
#line 11528 "MathView3D.w"

MLPutFunction(stdlink,"List",4);
MLPutFunction(stdlink,"MVTexture2D",1);
MLPutInteger(stdlink,TextureID(tlst));
/*701:*/
#line 11535 "MathView3D.w"

MLPutInteger(stdlink,tlst->texWidth);
MLPutInteger(stdlink,tlst->texHeight);

/*:701*/
#line 11532 "MathView3D.w"

/*702:*/
#line 11539 "MathView3D.w"

switch(TextureType(tlst)){
case GL_LUMINANCE:
MLPutSymbol(stdlink,"MVLuminantTexture");
break;
case GL_LUMINANCE_ALPHA:
MLPutSymbol(stdlink,"MVLuminantAlphaTexture");
break;
case GL_RGB:
MLPutSymbol(stdlink,"MVRGBTexture");
break;
case GL_RGBA:
MLPutSymbol(stdlink,"MVRGBAlphaTexture");
break;
default:
MLPutSymbol(stdlink,"Null");
break;
}

/*:702*/
#line 11533 "MathView3D.w"


/*:700*/
#line 11588 "MathView3D.w"

free(imagedata);
}

/*:703*/
#line 10774 "MathView3D.w"

/*708:*/
#line 11630 "MathView3D.w"

void mlRasterToTexture(int bytec,int w,int h,double*dpix,long len)
{GLubyte*pix;
int ttype,i;
TextureList*tlst;

if(len!=bytec*w*h||2==bytec||bytec>3||bytec<1){
MLPutSymbol(stdlink,"$Failed");
return;
}
pix= calloc(len,sizeof(GLubyte));
ttype= (1==bytec?GL_LUMINANCE:GL_RGB);
for(i= 0;i<len;i++)
pix[i]= (GLubyte)(255*fabs(dpix[i]));
tlst= CreateTexture(Textures,"Automatic",ttype,w,h,pix,bytec*w*h,GL_TRUE);
if(!tlst){
MLPutSymbol(stdlink,"$Failed");
free(pix);
return;
}
Textures= tlst;
/*700:*/
#line 11528 "MathView3D.w"

MLPutFunction(stdlink,"List",4);
MLPutFunction(stdlink,"MVTexture2D",1);
MLPutInteger(stdlink,TextureID(tlst));
/*701:*/
#line 11535 "MathView3D.w"

MLPutInteger(stdlink,tlst->texWidth);
MLPutInteger(stdlink,tlst->texHeight);

/*:701*/
#line 11532 "MathView3D.w"

/*702:*/
#line 11539 "MathView3D.w"

switch(TextureType(tlst)){
case GL_LUMINANCE:
MLPutSymbol(stdlink,"MVLuminantTexture");
break;
case GL_LUMINANCE_ALPHA:
MLPutSymbol(stdlink,"MVLuminantAlphaTexture");
break;
case GL_RGB:
MLPutSymbol(stdlink,"MVRGBTexture");
break;
case GL_RGBA:
MLPutSymbol(stdlink,"MVRGBAlphaTexture");
break;
default:
MLPutSymbol(stdlink,"Null");
break;
}

/*:702*/
#line 11533 "MathView3D.w"


/*:700*/
#line 11651 "MathView3D.w"

free(pix);
}

/*:708*/
#line 10775 "MathView3D.w"

/*709:*/
#line 11656 "MathView3D.w"

void mvSendTextureInfo(void)
{TextureList*tlst;
long int tno;

tlst= Textures;
tno= LengthTextureList(tlst);
MLPutFunction(stdlink,"List",tno);
while(tlst){
/*700:*/
#line 11528 "MathView3D.w"

MLPutFunction(stdlink,"List",4);
MLPutFunction(stdlink,"MVTexture2D",1);
MLPutInteger(stdlink,TextureID(tlst));
/*701:*/
#line 11535 "MathView3D.w"

MLPutInteger(stdlink,tlst->texWidth);
MLPutInteger(stdlink,tlst->texHeight);

/*:701*/
#line 11532 "MathView3D.w"

/*702:*/
#line 11539 "MathView3D.w"

switch(TextureType(tlst)){
case GL_LUMINANCE:
MLPutSymbol(stdlink,"MVLuminantTexture");
break;
case GL_LUMINANCE_ALPHA:
MLPutSymbol(stdlink,"MVLuminantAlphaTexture");
break;
case GL_RGB:
MLPutSymbol(stdlink,"MVRGBTexture");
break;
case GL_RGBA:
MLPutSymbol(stdlink,"MVRGBAlphaTexture");
break;
default:
MLPutSymbol(stdlink,"Null");
break;
}

/*:702*/
#line 11533 "MathView3D.w"


/*:700*/
#line 11665 "MathView3D.w"

tlst= tlst->next;
}
}

/*:709*/
#line 10776 "MathView3D.w"

/*713:*/
#line 11690 "MathView3D.w"

#pragma argsused
void mlMapTexture(int geomId,int textureId,
int texMapType,
double*move,long int nm,
double*axis,long int na,
double angle,
double*scale,long int ns,
double scaleU,
double scaleV)
{TextureMapping*tm,*tnew;
Mesh3DList*mlst;
TextureList*tlst;

if(geomId<0||textureId<0)/*714:*/
#line 11744 "MathView3D.w"

{
MLPutSymbol(stdlink,"$Failed");
return;
}

/*:714*/
#line 11704 "MathView3D.w"

/*715:*/
#line 11753 "MathView3D.w"

{
int found= GL_TRUE;

if(0==geomId)
tm= SceneTextureMap;
else{
mlst= Meshes;
found= GL_FALSE;
while(mlst){
if(MeshID(mlst)==geomId){
tm= MeshTexMap(mlst);
found= GL_TRUE;
break;
}
mlst= mlst->next;
}
}
if(!found)/*714:*/
#line 11744 "MathView3D.w"

{
MLPutSymbol(stdlink,"$Failed");
return;
}

/*:714*/
#line 11771 "MathView3D.w"

}

/*:715*/
#line 11705 "MathView3D.w"

/*716:*/
#line 11774 "MathView3D.w"

tlst= Textures;
while(tlst){
if(textureId==TextureID(tlst))break;
tlst= tlst->next;
}
if(!tlst)/*714:*/
#line 11744 "MathView3D.w"

{
MLPutSymbol(stdlink,"$Failed");
return;
}

/*:714*/
#line 11780 "MathView3D.w"



/*:716*/
#line 11706 "MathView3D.w"

tnew= CreateTextureMapping(texMapType,
TILE_TEXTURE_S||TILE_TEXTURE_T,
(Point3D*)move,
(Point3D*)axis,angle,
(Point3D*)scale,
scaleU,scaleV,tlst);
if(!tnew)/*714:*/
#line 11744 "MathView3D.w"

{
MLPutSymbol(stdlink,"$Failed");
return;
}

/*:714*/
#line 11713 "MathView3D.w"


if(tm)DeleteTextureMapping(tm);
tm= tnew;
if(0==geomId)
SceneTextureMap= tm;
else
MeshTexMap(mlst)= tm;
if(SceneTextureMap)
UpdatePolygonTextureCoord(SceneTextureMap,Polygons);
else
UpdateMeshTextureCoordinates(mlst);
/*681:*/
#line 11317 "MathView3D.w"

BoxRatios.x= BoxRatios.y= BoxRatios.z= 1.0;
/*682:*/
#line 11321 "MathView3D.w"

glutSetWindow(MVDrawWindow);
glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
if(ObjID>=0)glDeleteLists(ObjID,1);
ObjID= -1;
InitView(GL_TRUE);
glutPostRedisplay();
MakeNewView= GL_TRUE;

/*:682*/
#line 11319 "MathView3D.w"


/*:681*/
#line 11725 "MathView3D.w"

MLPutFunction(stdlink,"List",3);
MLPutFunction(stdlink,"DoubleLeftArrow",2);
if(0==geomId)
MLPutFunction(stdlink,"MVScene3D",0);
else{
MLPutFunction(stdlink,"MVMesh3D",1);
MLPutInteger(stdlink,MeshID(mlst));
}
MLPutFunction(stdlink,"MVTexture2D",1);
MLPutInteger(stdlink,TextureID(tm->tex));
/*717:*/
#line 11783 "MathView3D.w"

switch(texMapType){
case TEXTURE_MAP_PLANAR:
MLPutSymbol(stdlink,"MVPlaneMapping");
break;
case TEXTURE_MAP_CYLINDRICAL:
MLPutSymbol(stdlink,"MVCylinderMapping");
break;
case TEXTURE_MAP_SPHERICAL:
MLPutSymbol(stdlink,"MVSphereMapping");
break;
case TEXTURE_MAP_TORUS:
MLPutSymbol(stdlink,"MVTorusMapping");
break;
case TEXTURE_MAP_MESH_ST:
MLPutSymbol(stdlink,"MVMeshUVMapping");
break;
case TEXTURE_MAP_MESH_TS:
MLPutSymbol(stdlink,"MVMeshVUMapping");
break;
}


/*:717*/
#line 11736 "MathView3D.w"

MLPutFunction(stdlink,"List",3);
/*701:*/
#line 11535 "MathView3D.w"

MLPutInteger(stdlink,tlst->texWidth);
MLPutInteger(stdlink,tlst->texHeight);

/*:701*/
#line 11738 "MathView3D.w"

/*702:*/
#line 11539 "MathView3D.w"

switch(TextureType(tlst)){
case GL_LUMINANCE:
MLPutSymbol(stdlink,"MVLuminantTexture");
break;
case GL_LUMINANCE_ALPHA:
MLPutSymbol(stdlink,"MVLuminantAlphaTexture");
break;
case GL_RGB:
MLPutSymbol(stdlink,"MVRGBTexture");
break;
case GL_RGBA:
MLPutSymbol(stdlink,"MVRGBAlphaTexture");
break;
default:
MLPutSymbol(stdlink,"Null");
break;
}

/*:702*/
#line 11739 "MathView3D.w"

MLFlush(stdlink);
glutPostRedisplay();
}

/*:713*/
#line 10777 "MathView3D.w"

/*650:*/
#line 10841 "MathView3D.w"

void mvReset(void);
void mvReset(void)
{
/*37:*/
#line 958 "MathView3D.w"

DoReset= GL_TRUE;
DoTracking= DoTrackSpin= GL_FALSE;
DoRotate= GL_FALSE;
ZoomMove.x= ZoomMove.y= ZoomMove.z= 0.0;
RadialZoom= 1.0;
TrackAngle= 0.0;
TrackAxis.x= TrackAxis.y= 0.0;
TrackAxis.z= 1.0;
ScreenXStart= ScreenX= 0;
ScreenYStart= ScreenY= 0;
BlackBackground= GL_TRUE;
SetBackgroundColor(BlackBackground);
glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
Axis.AxisType= NoAxis;

/*:37*/
#line 10845 "MathView3D.w"

/*657:*/
#line 10902 "MathView3D.w"

glutPostRedisplay();
if(DoMathematicaCall){
MLPutSymbol(stdlink,"True");
MLFlush(stdlink);
}

/*:657*/
#line 10846 "MathView3D.w"

}

/*:650*/
#line 10778 "MathView3D.w"

/*656:*/
#line 10891 "MathView3D.w"

void mvZoom(double zoom);
void mvZoom(double zoom)
{double vr,vphi,vtheta;

RadialZoom*= zoom;
/*38:*/
#line 974 "MathView3D.w"

SphericalFromCartesian(&ViewPoint,&vr,&vtheta,&vphi);
ZoomMove.x= ViewPoint.x-RadialZoom*vr*cos(vphi)*sin(vtheta);
ZoomMove.y= ViewPoint.y-RadialZoom*vr*sin(vphi)*sin(vtheta);
ZoomMove.z= ViewPoint.z-RadialZoom*vr*cos(vtheta);

MakeNewView= GL_TRUE;


/*:38*/
#line 10897 "MathView3D.w"

glutPostRedisplay();
MLPutDouble(stdlink,RadialZoom);
}

/*:656*/
#line 10779 "MathView3D.w"

/*651:*/
#line 10849 "MathView3D.w"

void mvShowWindow(void);
void mvShowWindow(void)
{
glutShowWindow();
/*657:*/
#line 10902 "MathView3D.w"

glutPostRedisplay();
if(DoMathematicaCall){
MLPutSymbol(stdlink,"True");
MLFlush(stdlink);
}

/*:657*/
#line 10854 "MathView3D.w"

}

/*:651*/
#line 10780 "MathView3D.w"

/*654:*/
#line 10875 "MathView3D.w"

void mvHideWindow(void);
void mvHideWindow(void)
{
glutHideWindow();
/*657:*/
#line 10902 "MathView3D.w"

glutPostRedisplay();
if(DoMathematicaCall){
MLPutSymbol(stdlink,"True");
MLFlush(stdlink);
}

/*:657*/
#line 10880 "MathView3D.w"

}

/*:654*/
#line 10781 "MathView3D.w"

/*652:*/
#line 10857 "MathView3D.w"

void mvPopWindow(void);
void mvPopWindow(void)
{
glutSetWindow(MVDrawWindow);
glutPopWindow();
/*657:*/
#line 10902 "MathView3D.w"

glutPostRedisplay();
if(DoMathematicaCall){
MLPutSymbol(stdlink,"True");
MLFlush(stdlink);
}

/*:657*/
#line 10863 "MathView3D.w"

}

/*:652*/
#line 10782 "MathView3D.w"

/*653:*/
#line 10866 "MathView3D.w"

void mvPushWindow(void);
void mvPushWindow(void)
{
glutSetWindow(MVDrawWindow);
glutPushWindow();
/*657:*/
#line 10902 "MathView3D.w"

glutPostRedisplay();
if(DoMathematicaCall){
MLPutSymbol(stdlink,"True");
MLFlush(stdlink);
}

/*:657*/
#line 10872 "MathView3D.w"

}

/*:653*/
#line 10783 "MathView3D.w"

/*655:*/
#line 10883 "MathView3D.w"

void mvIconifyWindow(void);
void mvIconifyWindow(void)
{
glutIconifyWindow();
/*657:*/
#line 10902 "MathView3D.w"

glutPostRedisplay();
if(DoMathematicaCall){
MLPutSymbol(stdlink,"True");
MLFlush(stdlink);
}

/*:657*/
#line 10888 "MathView3D.w"

}

/*:655*/
#line 10784 "MathView3D.w"

/*658:*/
#line 10909 "MathView3D.w"

void mvReshape(int w,int h);
void mvReshape(int w,int h)
{int w0,h0;

glutSetWindow(MVDrawWindow);
w0= glutGet(GLUT_WINDOW_WIDTH);
h0= glutGet(GLUT_WINDOW_HEIGHT);
glutReshapeWindow(w,h);
MLPutFunction(stdlink,"List",2);
MLPutInteger(stdlink,w0);
MLPutInteger(stdlink,h0);
}

/*:658*/
#line 10785 "MathView3D.w"

/*665:*/
#line 11016 "MathView3D.w"

void mvRotate3D(int,double*,long int,double,double*,long int,int);
void mvRotate3D(int what,
double*axis,long int l1,
double theta,
double*cntr,long int l2,
int update)
{double rmat[3][3];
double norm,ct,st;
double nx,ny,nz;
register double x,y,z,ox,oy,oz;
double xMin,xMax,yMin,yMax,zMin,zMax;
Mesh3DList*mlst= Meshes;
int i;

if(l1!=3L||l2!=3L){
MLPutSymbol(stdlink,"$Failed");
return;
}
norm= sqrt(axis[0]*axis[0]+axis[1]*axis[1]+axis[2]*axis[2]);
nx= axis[0]/norm;
ny= axis[1]/norm;
nz= axis[2]/norm;
if(norm<1.0e-12){
MLPutSymbol(stdlink,"$Failed");
return;
}
else{
MLPutFunction(stdlink,"List",2);
MLPutFunction(stdlink,"List",3);
MLPutDouble(stdlink,nx);
MLPutDouble(stdlink,ny);
MLPutDouble(stdlink,nz);
MLPutDouble(stdlink,theta);
}
ct= cos(theta);st= sin(theta);
xMin= yMin= zMin= FLT_MAX;
xMax= yMax= zMax= -FLT_MAX;
/*667:*/
#line 11101 "MathView3D.w"

{
rmat[0][0]= nx*nx+(1.-nx*nx)*ct;
rmat[0][1]= nx*ny*(1.-ct)+nz*st;
rmat[0][2]= nx*nz*(1.-ct)-ny*st;

rmat[1][0]= nx*ny*(1.-ct)-nz*st;
rmat[1][1]= ny*ny+(1.-ny*ny)*ct;
rmat[1][2]= ny*nz*(1.-ct)+nx*st;

rmat[2][0]= nx*nz*(1.-ct)+ny*st;
rmat[2][1]= ny*nz*(1.-ct)-nx*st;
rmat[2][2]= nz*nz+(1.-nz*nz)*ct;
}

/*:667*/
#line 11054 "MathView3D.w"

if(TransformSceneQ(what)){
for(i= 0;i<PoolSize;i++){
ox= PointPool[i].x-cntr[0];
oy= PointPool[i].y-cntr[1];
oz= PointPool[i].z-cntr[2];
/*666:*/
#line 11094 "MathView3D.w"

{
x= rmat[0][0]*ox+rmat[0][1]*oy+rmat[0][2]*oz;
y= rmat[1][0]*ox+rmat[1][1]*oy+rmat[1][2]*oz;
z= rmat[2][0]*ox+rmat[2][1]*oy+rmat[2][2]*oz;
}

/*:666*/
#line 11060 "MathView3D.w"

if(BoxUpdateQ(update))DXFUpdateBBox(x,y,z);
PointPool[i].x= x+cntr[0];
PointPool[i].y= y+cntr[1];
PointPool[i].z= z+cntr[2];
}
}
while(mlst){
if(TransformMeshQ(what,MeshID(mlst))){
for(i= 0;i<mlst->mPointNo;i++){
ox= mlst->mPoints[i].x-cntr[0];
oy= mlst->mPoints[i].y-cntr[1];
oz= mlst->mPoints[i].z-cntr[2];
/*666:*/
#line 11094 "MathView3D.w"

{
x= rmat[0][0]*ox+rmat[0][1]*oy+rmat[0][2]*oz;
y= rmat[1][0]*ox+rmat[1][1]*oy+rmat[1][2]*oz;
z= rmat[2][0]*ox+rmat[2][1]*oy+rmat[2][2]*oz;
}

/*:666*/
#line 11073 "MathView3D.w"

if(BoxUpdateQ(update))DXFUpdateBBox(x,y,z);
mlst->mPoints[i].x= x+cntr[0];
mlst->mPoints[i].y= y+cntr[1];
mlst->mPoints[i].z= z+cntr[2];
ox= mlst->mNormals[i].x;oy= mlst->mNormals[i].y;oz= mlst->mNormals[i].z;
/*666:*/
#line 11094 "MathView3D.w"

{
x= rmat[0][0]*ox+rmat[0][1]*oy+rmat[0][2]*oz;
y= rmat[1][0]*ox+rmat[1][1]*oy+rmat[1][2]*oz;
z= rmat[2][0]*ox+rmat[2][1]*oy+rmat[2][2]*oz;
}

/*:666*/
#line 11079 "MathView3D.w"

mlst->mNormals[i].x= x;mlst->mNormals[i].y= y;mlst->mNormals[i].z= z;
}
}
mlst= mlst->next;
}
if(BoxUpdateQ(update)){
BoundP1.x= xMin;BoundP1.y= yMin;BoundP1.z= zMin;
BoundP2.x= xMax;BoundP2.y= yMax;BoundP2.z= zMax;
}
if(DisplayUpdateQ(update)){
/*681:*/
#line 11317 "MathView3D.w"

BoxRatios.x= BoxRatios.y= BoxRatios.z= 1.0;
/*682:*/
#line 11321 "MathView3D.w"

glutSetWindow(MVDrawWindow);
glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
if(ObjID>=0)glDeleteLists(ObjID,1);
ObjID= -1;
InitView(GL_TRUE);
glutPostRedisplay();
MakeNewView= GL_TRUE;

/*:682*/
#line 11319 "MathView3D.w"


/*:681*/
#line 11090 "MathView3D.w"

}
}

/*:665*/
#line 10786 "MathView3D.w"

/*668:*/
#line 11118 "MathView3D.w"

void mvTranslate3D(int what,double nx,double ny,double nz,int update);
void mvTranslate3D(int what,double nx,double ny,double nz,int update)
{double norm;
register double x,y,z;
double xMin,xMax,yMin,yMax,zMin,zMax;
register int i;
Mesh3DList*mlst= Meshes;

norm= sqrt(nx*nx+ny*ny+nz*nz);
if(norm<1.0e-12){
MLPutSymbol(stdlink,"$Failed");
return;
}
else{
MLPutFunction(stdlink,"List",3);
MLPutDouble(stdlink,nx);
MLPutDouble(stdlink,ny);
MLPutDouble(stdlink,nz);
}
xMin= yMin= zMin= FLT_MAX;
xMax= yMax= zMax= -FLT_MAX;
if(TransformSceneQ(what)){
for(i= 0;i<PoolSize;i++){
x= PointPool[i].x+nx;y= PointPool[i].y+ny;z= PointPool[i].z+nz;
if(BoxUpdateQ(update))DXFUpdateBBox(x,y,z);
PointPool[i].x= x;
PointPool[i].y= y;
PointPool[i].z= z;
}
}
while(mlst){
if(TransformMeshQ(what,MeshID(mlst))){
for(i= 0;i<mlst->mPointNo;i++){
x= mlst->mPoints[i].x+= nx;y= mlst->mPoints[i].y+= ny;z= mlst->mPoints[i].z+= nz;
if(BoxUpdateQ(update))DXFUpdateBBox(x,y,z);
}
}
mlst= mlst->next;
}
if(BoxUpdateQ(update)){
BoundP1.x= xMin;BoundP1.y= yMin;BoundP1.z= zMin;
BoundP2.x= xMax;BoundP2.y= yMax;BoundP2.z= zMax;
}
if(DisplayUpdateQ(update)){
/*681:*/
#line 11317 "MathView3D.w"

BoxRatios.x= BoxRatios.y= BoxRatios.z= 1.0;
/*682:*/
#line 11321 "MathView3D.w"

glutSetWindow(MVDrawWindow);
glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
if(ObjID>=0)glDeleteLists(ObjID,1);
ObjID= -1;
InitView(GL_TRUE);
glutPostRedisplay();
MakeNewView= GL_TRUE;

/*:682*/
#line 11319 "MathView3D.w"


/*:681*/
#line 11163 "MathView3D.w"

}
}

/*:668*/
#line 10787 "MathView3D.w"


/*:645*/
#line 106 "MathView3D.w"

/*56:*/
#line 1386 "MathView3D.w"

#pragma argsused
int CreateOGLPackage(int argc,char*argv[]);
int CreateOGLPackage(int argc,char*argv[])
{FILE*f;
char Self[1024];
int i,n;
#ifdef __WIN32__
argc= _argc;
argv= _argv;
#endif
strcpy(Self,argv[0]);
n= strlen(Self);
for(i= 0;i<n;i++)
Self[i]= Self[i]=='\\'?'/':tolower(Self[i]);
f= fopen(MVPACKAGE_NAME,"w");
if(!f)return GL_FALSE;
fprintf(f,"(* File : %s, eventual edit the $MVTexturePath value.  *)\n"
"(* (c) Jens-Peer Kuska 1997,98,99     *)\n"
"(* The file is automatical generated by the +mvcp switch   *)\n",
MVPACKAGE_NAME);
fprintf(f,"(* Mathematica startup file of \"%s\" *)\n"
"(* Version %s *)\n",BANNER,VERSION);
fprintf(f,"(* For %s compiled at %s *)\n\n",ARCH,__DATE__);
fprintf(f,"Begin[\"System`\"]\n"
"  If[!MemberQ[Links[],$MVLink],\n"
"    $MVLink=Install[LinkOpen[\"%s -mathlink\"]],\n"
"    Print[\"The OpenGLViewer is running.\"]\n"
"   ];\n",Self);
#ifdef __WIN32__
fprintf(f,"MVShowWindow[];\n");
#endif
fprintf(f,"Unprotect[MVClose];\n\n"
"Clear[MVClose];\n\n"
"MVClose[]:=Uninstall[$MVLink]\n\n"
"Protect[MVClose];\n\n");
fprintf(f,"End[];\n\n");
fprintf(f,"AppendTo[$MVTexturePath,{\"..\",\"..\",\"Textures\"}];\n");
fclose(f);
#ifdef __WIN32__
MessageBox(MLIconWindow,
"Mathematica package\nOpenGLViewer.m\ngenerated.",
"OpenGLViewer Package",MB_OK|MB_ICONINFORMATION|MB_TASKMODAL|MB_SETFOREGROUND);
#else
printf("Mathematica package %s generated.\n",MVPACKAGE_NAME);
#endif
return GL_TRUE;
}

/*:56*/
#line 107 "MathView3D.w"

/*61:*/
#line 1449 "MathView3D.w"

#ifdef __WIN32__
#pragma argsused
int PASCAL WinMain(HINSTANCE hCurInstance,
HINSTANCE hPrevInstance,
LPSTR lpCmdLine,int nCmdShow){
char*argv[32];
int argc,i;
char**argv_end= argv+32;
char buff[512];
char*buff_start= buff;
#else
int main(int argc,char*argv[]){
#endif
CmdArgLst*cmdln;
long int err;
void MathViewExit(void);

/*60:*/
#line 1445 "MathView3D.w"

Textures= NULL;
TextureCount= 1;

/*:60*/
#line 1467 "MathView3D.w"

#ifdef __WIN32__
MLScanString(argv,&argv_end,&lpCmdLine,&buff_start);
argc= argv_end-argv;
for(i= argc;i>0;i--)
argv[i]= argv[i-1];
argv[0]= "mathview3d";
argc++;
#endif
/*11:*/
#line 296 "MathView3D.w"

cmdln= DeclareCmdArg("mvfile",CMD_STRING,&ThreeScriptName,"       3Script input file.");
cmdln= AppendCmdArg(DeclareCmdArg("mvs",CMD_SWITCH,&SmoothShading,
"shows smooth/flat input data"),cmdln);
cmdln= AppendCmdArg(DeclareCmdArg("mvw",CMD_SWITCH,&WireFrame,
"show data as WireFrame"),cmdln);
cmdln= AppendCmdArg(DeclareCmdArg("mvo",CMD_SWITCH,&PolyOutline,
"show polygon outline"),cmdln);
cmdln= AppendCmdArg(DeclareCmdArg("mvct",CMD_SWITCH,&COMTriangulation,
"triangulate n-gons (n>4) with a new point in the center"),cmdln);

cmdln= AppendCmdArg(DeclareCmdArg("mvts",CMD_DOUBLE,&LineTubeSize,
"tube size of lines -- if given lines appear as tubes"),cmdln);
cmdln= AppendCmdArg(DeclareCmdArg("mvta",CMD_DOUBLE,&LineSegmentAngle,
"maximum angle for smooth line continuation"),cmdln);
cmdln= AppendCmdArg(DeclareCmdArg("mvps",CMD_DOUBLE,&PointSphereSize,
"sphere size of points -- if given points appear as spheres"),cmdln);
cmdln= AppendCmdArg(
DeclareCmdArg("mvpov",CMD_STRING,&POVOutName,
"       POVRay output file."),
cmdln);
cmdln= AppendCmdArg(
DeclareCmdArg("mvdxf",CMD_STRING,&DXFOutName,
"       DXF output file."),
cmdln);
cmdln= AppendCmdArg(
DeclareCmdArg("mv3dm",CMD_STRING,&QD3dOutName,
"       QuickDraw3D output file."),
cmdln);
cmdln= AppendCmdArg(
DeclareCmdArg("mvwrl",CMD_STRING,&VRMLOutName,
"       VMRL 2.0 output file."),
cmdln);
cmdln= AppendCmdArg(DeclareCmdArg("mvue",CMD_SWITCH,&UseVRMLExtrusion,
"use Extrusions in VRML export"),cmdln);
cmdln= AppendCmdArg(
DeclareCmdArg("mvindxf",CMD_STRING,&DXFInputName,"     DXF input file."),
cmdln);
cmdln= AppendCmdArg(
DeclareCmdArg("mvinoff",CMD_STRING,&OFFInputName,"     OFF input file."),
cmdln);
cmdln= AppendCmdArg(
DeclareCmdArg("mvtex",CMD_STRING,&TextureName,
"       PNG file of texture."),
cmdln);
cmdln= AppendCmdArg(
DeclareCmdArg("mvtxm",CMD_STRING,&TexMapType,
"       mapping type of the texture {planar,cylindrical,spherical,torus}."),
cmdln);
cmdln= AppendCmdArg(DeclareCmdArg("mvno",CMD_SWITCH,&NoDisplay,
"run as filter and do not display the 3Script"),cmdln);
cmdln= AppendCmdArg(DeclareCmdArg("mvcp",CMD_SWITCH,&MakePackage,
"Create OpenGLViewer.m package"),cmdln);
cmdln= AppendCmdArg(DeclareCmdArg("mvpr",CMD_DOUBLE,&PolyReduceFactor,
"        value for polygon reduction polygons with distance less are removed."),cmdln);
cmdln= AppendCmdArg(DeclareCmdArg("mvrit",CMD_INTEGER,&PolyReduceIterations,
"        number for polygon reduction iterations."),cmdln);



/*:11*/
#line 1476 "MathView3D.w"

atexit(MathViewExit);
/*59:*/
#line 1440 "MathView3D.w"

#if defined(__WIN32__) && defined(__BORLANDC__)
_control87(MCW_EM,MCW_EM);
#endif

/*:59*/
#line 1478 "MathView3D.w"

#ifdef __WIN32__
glutInit(&_argc,_argv);
#else
glutInit(&argc,argv);
#endif
/*13:*/
#line 386 "MathView3D.w"

#if !MACINTOSH_MATHLINK
if(argc<=1){
/*12:*/
#line 356 "MathView3D.w"

#ifndef __WIN32__
HelpCmdArg(BANNER,cmdln);
#else
MessageBox(MLIconWindow,
"(c) Jens-Peer Kuska 1997-99\n\n"
"usage:  -mvfile 3script -opt val ..\n\n"
"-mvfile <3script>    show 3Script\n"
"-mvindxf <dxf file>  show dxf file\n"
"-mvinoff <off file>  show off file\n\n"
"-mvpov <pov file>    write POVRay file\n"
"-mv3dm <3dmf file>   write QuickDraw3D file\n"
"-mvwrl <wrl file>    write VRML 2 file\n"
"-mvdxf <dxf file>    write dxf file\n\n"
"-mvtex <png file>    use png file as texture\n"
"-mvtxm p|c|s|t       texture map type\n\n"
"+|-mvno            run as filter don't render\n"
"+|-mvs               turn smooth shading on|off\n"
"+|-mvw              wire frame on|off\n"
"+|-mvo               polygon outlines on|off\n"
"-mvts <real>    tube size of lines\n"
"-mvps <real>   point sphere size\n"
"+|-mvcp             create Mathematica package\n\n"
"-mvpr <real>    polygon reduction factor\n"
"-mvrit <int>        iterations for polygon reduction\n",
"MathGL3d "VERSION,
MB_OK|MB_ICONINFORMATION|MB_TASKMODAL|MB_SETFOREGROUND);
#endif
return 1;

/*:12*/
#line 389 "MathView3D.w"

}
#endif
#ifdef __WIN32__
cmdln= ScanCommandline(cmdln,_argc,_argv);
#else
cmdln= ScanCommandline(cmdln,argc,argv);
#endif
if(*ThreeScriptName){
Read3Script(ThreeScriptName);
AveragedNormals(Polygons);
}
if(!(*ThreeScriptName)&&*DXFInputName){
ReadDXF(DXFInputName);
AveragedNormals(Polygons);
}
if(!(*ThreeScriptName)&&!(*DXFInputName)&&*OFFInputName){
ReadOFF(OFFInputName);
AveragedNormals(Polygons);
}
if(MakePackage&&CreateOGLPackage(argc,argv))
return 0;
if(PolyReduceIterations>0&&PolyReduceFactor>0.0&&Polygons){
SimplifyPolygones(PolyReduceIterations,PolyReduceFactor,
OctalPointTree);
}

/*14:*/
#line 441 "MathView3D.w"

CosSkipCont= cos(ToRadian*(fabs(LineSegmentAngle)<MaxLSAngle?LineSegmentAngle:MaxLSAngle));

/*:14*/
#line 416 "MathView3D.w"

#ifdef __WIN32__
MathLinkMode= IsMathLinkCall(_argc,_argv);
#else
MathLinkMode= IsMathLinkCall(argc,argv);
#endif
/*753:*/
#line 12410 "MathView3D.w"

if(MathLinkMode){
#if 0
init_macintosh();
#endif
if(!stdenv)
stdenv= MLInitialize((MLParametersPointer)0);
if(stdenv==(MLEnvironment)0)
return!MLDone;
#if !MACINTOSH_MATHLINK
if(!stdhandler)
stdhandler= MLCreateMessageHandler(stdenv,MLDefaultHandler,0);
#else
if(!stdhandler)
stdhandler= MLCreateMessageHandler(stdenv,
NewMLHandlerProc(MLDefaultHandler),0);
#endif
LinkToMath= MLOpenArgv(stdenv,argv,argv+argc,&err);
if(LinkToMath==(MLINK)0){
MLAlert(stdenv,MLErrorString(stdenv,err));
/*754:*/
#line 12443 "MathView3D.w"

if(LinkToMath)
MLClose(LinkToMath);
MLDeinitialize(stdenv);
stdenv= (MLEnvironment)0;
exit(!MLDone);


/*:754*/
#line 12430 "MathView3D.w"

}
#if MACINTOSH_MATHLINK
if(stdyielder)MLSetYieldFunction(LinkToMath,stdyielder);
#endif
if(stdhandler)MLSetMessageHandler(LinkToMath,stdhandler);
if(!MLInstall(LinkToMath)){
/*754:*/
#line 12443 "MathView3D.w"

if(LinkToMath)
MLClose(LinkToMath);
MLDeinitialize(stdenv);
stdenv= (MLEnvironment)0;
exit(!MLDone);


/*:754*/
#line 12437 "MathView3D.w"

}
MathLinkActive= GL_TRUE;
}


/*:753*/
#line 422 "MathView3D.w"

if(*ThreeScriptName&&*POVOutName)
WritePOVray(POVOutName,SmoothShading,"mvs","mvm");
if(*ThreeScriptName&&*DXFOutName)
WriteDXF(DXFOutName);
if(*ThreeScriptName&&*QD3dOutName)
Write3DMF(QD3dOutName,GL_TRUE,GL_TRUE);
if(*ThreeScriptName&&*VRMLOutName)
WriteVRML(VRMLOutName,GL_FALSE,UseVRMLExtrusion,"mvs","mvm");
if(*TextureName){
Textures= ReadPNGTexture(Textures,TextureName,1.0,GL_FALSE);
if(Textures)
/*138:*/
#line 2687 "MathView3D.w"

{

if(Polygons){
SceneTextureMap= DefaultTextureMapping(Textures);
if(SceneTextureMap)
UpdatePolygonTextureCoord(SceneTextureMap,Polygons);
}
else if(Meshes){
MeshTexMap(Meshes)= DefaultTextureMapping(Textures);
if(MeshTexMap(Meshes))
UpdateMeshTextureCoordinates(Meshes);
}
}

/*:138*/
#line 434 "MathView3D.w"

}
if(NoDisplay){
RemoveCmdArgs(cmdln);
return 0;
}

/*:13*/
#line 1484 "MathView3D.w"

RemoveCmdArgs(cmdln);
glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGBA|GLUT_DEPTH);
#if MACINTOSH_MATHLINK
glutInitWindowSize(300,300);
glutInitWindowPosition(glutGet(GLUT_SCREEN_WIDTH)-310,10);
#endif
MVDrawWindow= glutCreateWindow(BANNER);
#if MACINTOSH_MATHLINK || defined(__WIN32__)
glutReshapeWindow(300,300);
#else
glutReshapeWindow(400,400);
#endif
MVCreateMainMenu();
glutMouseFunc(MVMouseButton);
glutMotionFunc(MVMouseMove);
glutDisplayFunc(MVDisplay);
glutReshapeFunc(MVReshape);
glutEntryFunc(MVMouseEnterMain);
glutIdleFunc(MVAnimate);
InitView(GL_TRUE);
#if MACINTOSH_MATHLINK
glutInitWindowPosition(0,0);
#endif
glutMainLoop();
return 0;
}

/*:61*/
#line 108 "MathView3D.w"

/*439:*/
#line 8432 "MathView3D.w"

:Begin:
:Function: mvGetPoolSize  
:Pattern: MViewPoolSize[] 
:Arguments: {}            
:ArgumentTypes: {}        
:ReturnType: Integer      
:End:

/*:439*//*441:*/
#line 8445 "MathView3D.w"

:Begin: 
:Function: mvPutGraphics3D 
:Pattern: MVPutGraphics3D[l_List] 
:Arguments:{l}                    
:ArgumentTypes: {Manual}          
:ReturnType: Manual               
:End:

/*:441*//*443:*/
#line 8457 "MathView3D.w"

:Begin: 
:Function: mvClearContents 
:Pattern: MVClear[]       
:Arguments: {}            
:ArgumentTypes: {}        
:ReturnType: Manual       
:End:

/*:443*//*445:*/
#line 8469 "MathView3D.w"

:Begin: 
:Function: mvReset 
:Pattern: MVReset[]       
:Arguments: {}            
:ArgumentTypes: {}        
:ReturnType: Manual       
:End:

/*:445*//*447:*/
#line 8481 "MathView3D.w"

:Begin:
:Function: mvShowWindow  
:Pattern: MVShowWindow[] 
:Arguments: {}           
:ArgumentTypes: {}       
:ReturnType: Manual      
:End:

/*:447*//*449:*/
#line 8494 "MathView3D.w"

:Begin:
:Function: mvHideWindow  
:Pattern: MVHideWindow[] 
:Arguments: {}           
:ArgumentTypes: {}       
:ReturnType: Manual      
:End:

/*:449*//*451:*/
#line 8506 "MathView3D.w"

:Begin:
:Function: mvIconifyWindow  
:Pattern: MVIconifyWindow[] 
:Arguments: {}           
:ArgumentTypes: {}       
:ReturnType: Manual      
:End:

/*:451*//*453:*/
#line 8518 "MathView3D.w"

:Begin:
:Function: mvPopWindow  
:Pattern: MVPopWindow[] 
:Arguments: {}           
:ArgumentTypes: {}       
:ReturnType: Manual      
:End:

/*:453*//*455:*/
#line 8531 "MathView3D.w"

:Begin:
:Function: mvPushWindow  
:Pattern: MVPushWindow[] 
:Arguments: {}           
:ArgumentTypes: {}       
:ReturnType: Manual      
:End:

/*:455*//*457:*/
#line 8545 "MathView3D.w"

:Begin: 
:Function: mvStopSpin 
:Pattern: MVStopSpin[]       
:Arguments: {}            
:ArgumentTypes: {}        
:ReturnType: Manual       
:End:

/*:457*//*459:*/
#line 8557 "MathView3D.w"

:Begin: 
:Function: mvZoom             
:Pattern: MVZoom[zoom_Real]   
:Arguments: {zoom}            
:ArgumentTypes: {Real}        
:ReturnType: Manual           
:End:

/*:459*//*461:*/
#line 8569 "MathView3D.w"

:Begin: 
:Function: mvSpin                                        
:Pattern: MVSpin[angle_Real,l_ /;                        
            VectorQ[l,Real==Head[#] &] && 3==Length[l]]  
:Arguments: {angle,l}                                    
:ArgumentTypes:{Real,RealList}                           
:ReturnType: Manual                                      
:End: 

/*:461*//*462:*/
#line 8579 "MathView3D.w"

:Evaluate: MVSpin[angle_?NumericQ,{x_?NumericQ,y_NumericQ,z_NumericQ}]:= 
             MVSpin[N[angle],N[{x,y,z}]] /. 0->0.0          

/*:462*//*464:*/
#line 8587 "MathView3D.w"

:Begin: 
:Function: mvReshape                                     
:Pattern: MVReshapeWindow[width_Integer,height_Integer]  
:Arguments: {width,height}                               
:ArgumentTypes:{Integer,Integer}                         
:ReturnType: Manual                                      
:End: 

/*:464*//*466:*/
#line 8600 "MathView3D.w"

:Begin: 
:Function: mvSendBoundingBox                             
:Pattern: MVGetPlotRange[]                               
:Arguments: {}                                           
:ArgumentTypes:{}                                        
:ReturnType: Manual                                      
:End: 

/*:466*//*468:*/
#line 8612 "MathView3D.w"

:Begin: 
:Function: mvSendViewPoint                               
:Pattern: MVGetViewPoint[]                               
:Arguments: {}                                           
:ArgumentTypes:{}                                        
:ReturnType: Manual                                      
:End: 

/*:468*//*470:*/
#line 8624 "MathView3D.w"

:Begin: 
:Function: mvSendLightSources        
:Pattern:   mvRawGetLightSources[]   
:Arguments: {}             
:ArgumentTypes:{}          
:ReturnType: Manual        
:End: 

/*:470*//*472:*/
#line 8636 "MathView3D.w"

:Begin: 
:Function: mlWritePOVray                        
:Pattern: mvRawWritePOVRay[fname_String,ptexpre_String,mtexpre_String] 
:Arguments: {fname,ptexpre,mtexpre}             
:ArgumentTypes: {String,String,String}          
:ReturnType: Manual                             
:End:                                           

/*:472*//*474:*/
#line 8648 "MathView3D.w"

:Begin: 
:Function: mlWrite3DMF                        
:Pattern: mvWrite3DMF[fname_String,savelight_Integer,   
             saveview_Integer]                
:Arguments: {fname,savelight,saveview}        
:ArgumentTypes: {String,Integer,Integer}      
:ReturnType: Manual                           
:End:                                         

/*:474*//*476:*/
#line 8661 "MathView3D.w"

:Begin:
:Function: mlWriteDXF                            
:Pattern: mvWriteDXF[fname_String,               
            ts_Real,tseg_Integer,psize_Real,tangle_Real]   
:Arguments: {fname,ts,tseg,psize,tangle}         
:ArgumentTypes: {String,Real,Integer,Real,Real}  
:ReturnType: Manual                              
:End:

/*:476*//*478:*/
#line 8674 "MathView3D.w"

:Begin:
:Function: mlWritePostScript                     
:Pattern: mvWritePostScript[fname_String,wbitmap_Integer]  
:Arguments: {fname,wbitmap}                      
:ArgumentTypes: {String,Integer}                 
:ReturnType: Manual                              
:End:

/*:478*//*480:*/
#line 8687 "MathView3D.w"

:Begin:
:Function: mlRawWriteVRML                                
:Pattern: mvRawWriteVRML[fname_String,                   
   headl_Integer,useext_Integer,ptpre_String,mtpre_String]         
:Arguments: {fname,headl,useext,ptpre,mtpre}             
:ArgumentTypes: {String,Integer,Integer,String,String}   
:ReturnType: Manual                              
:End:

/*:480*//*482:*/
#line 8701 "MathView3D.w"

:Begin:
:Function: MLSendGraphics                        
:Pattern: MVGetGraphics3D[]                      
:Arguments: {}                                   
:ArgumentTypes: {}                               
:ReturnType: Manual                              
:End:

/*:482*//*484:*/
#line 8714 "MathView3D.w"

:Begin:
:Function: MLSendFrameBuffer                      
:Pattern: MVGetRasterImage[]                      
:Arguments: {}                                    
:ArgumentTypes: {}                                
:ReturnType: Manual                               
:End:

/*:484*//*486:*/
#line 8728 "MathView3D.w"

:Begin:
:Function: MLPolygonReduce                        
:Pattern: mvPolygonReduce[tol_Real,iter_Integer]  
:Arguments: {tol,iter}                            
:ArgumentTypes: {Real,Integer}                    
:ReturnType: Integer                              
:End:

/*:486*//*488:*/
#line 8741 "MathView3D.w"

:Begin:
:Function: MVLoad3Script                             
:Pattern: MVLoad3Script[fname_String]                
:Arguments: {fname}                              
:ArgumentTypes: {String}                         
:ReturnType: Manual                              
:End:

/*:488*//*490:*/
#line 8755 "MathView3D.w"

:Begin:
:Function: MVLoadDXF                             
:Pattern: MVLoadDXF[fname_String]                
:Arguments: {fname}                              
:ArgumentTypes: {String}                         
:ReturnType: Manual                              
:End:

/*:490*//*492:*/
#line 8768 "MathView3D.w"

:Begin:
:Function: MVLoadOFF                             
:Pattern: MVLoadOFF[fname_String]                
:Arguments: {fname}                              
:ArgumentTypes: {String}                         
:ReturnType: Manual                              
:End:

/*:492*//*494:*/
#line 8781 "MathView3D.w"

:Begin:
:Function: mvWriteOFF                             
:Pattern: mvWriteOFF[fname_String]                
:Arguments: {fname}                              
:ArgumentTypes: {String}                         
:ReturnType: Manual                              
:End:

/*:494*//*496:*/
#line 8794 "MathView3D.w"

:Begin:
:Function: mlRawWritePNG                             
:Pattern: mvRawWritePNG[fname_String,gamma_Real,inter_Integer] 
:Arguments: {fname,gamma,inter}                      
:ArgumentTypes: {String,Real,Integer}                
:ReturnType: Manual                                  
:End:

/*:496*//*498:*/
#line 8807 "MathView3D.w"

:Begin:
:Function: mlRawWriteDepthBuffer                           
:Pattern: mvRawWriteDepthPNG[fname_String,fmt_Integer,     
            gamma_Real,inter_Integer,scale_Real]           
:Arguments: {fname,fmt,gamma,inter,scale}                  
:ArgumentTypes: {String,Integer,Real,Integer,Real}         
:ReturnType: Manual                                        
:End:

/*:498*//*500:*/
#line 8821 "MathView3D.w"

:Begin:
:Function: mlRawWriteSIRDS                               
:Pattern: mvRawWriteSIRDS[fname_String,                  
            dpiOut_Integer,trueWidth_Real,               
            eyeSeparation_Real,mu_Real,osample_Integer,scale_Real] 
:Arguments: {fname,dpiOut,trueWidth,eyeSeparation,mu,osample,scale}
:ArgumentTypes: {String,Integer,Real,Real,Real,Integer,Real}       
:ReturnType: Manual                                      
:End:

/*:500*//*502:*/
#line 8836 "MathView3D.w"

:Begin:
:Function: mlLoadPNGTexture                          
:Pattern: mvRawLoadPNGTexture[fname_String,gamma_Real]         
:Arguments: {fname,gamma}                            
:ArgumentTypes: {String,Real}                        
:ReturnType: Manual                                  
:End:

/*:502*//*504:*/
#line 8849 "MathView3D.w"

:Begin:
:Function: mlGraphicsToTexture                       
:Pattern: mvRawPutMathTexture[clip_Integer,dispstr_String]     
:Arguments: {clip,dispstr}                           
:ArgumentTypes: {Integer,Manual}                     
:ReturnType: Manual                                  
:End:

/*:504*//*506:*/
#line 8862 "MathView3D.w"

:Begin:
:Function: mlRasterToTexture                           
:Pattern: mvRawPutRasterTexture[                       
            bytec_Integer,w_Integer,h_Integer,data_]   
:Arguments: {bytec,w,h,data}                           
:ArgumentTypes: {Integer,Integer,Integer,RealList}     
:ReturnType: Manual                                    
:End:

/*:506*//*508:*/
#line 8876 "MathView3D.w"

:Begin: 
:Function: mvSendTextureInfo                             
:Pattern: MVGetTextureInformation[]                      
:Arguments: {}                                           
:ArgumentTypes:{}                                        
:ReturnType: Manual                                      
:End: 

/*:508*//*510:*/
#line 8890 "MathView3D.w"

:Begin:
:Function: mlMapTexture                              
:Pattern: mvRawMapTexture[gid_Integer,texid_Integer, 
            maptype_Integer,                         
            move_List /; 3==Length[move],            
            axis_List /; 3==Length[axis],            
            angle_Real,                              
            scale_List /; 3==Length[scale],          
            scaleU_Real,scaleV_Real                  
           ]                                         
:Arguments: {gid,texid,maptype,                      
   move,axis,angle,scale,scaleU,scaleV}              
:ArgumentTypes: {Integer,Integer,Integer,            
  RealList,RealList,Real,RealList,Real,Real}         
:ReturnType: Manual                                  
:End:

/*:510*//*512:*/
#line 8912 "MathView3D.w"

:Begin:
:Function: mvRotate3D                             
:Pattern: mvRotate3D[what_Integer,axis_?VectorQ,th_Real,cntr_?VectorQ,update_Integer] 
:Arguments: {what,axis,th,cntr,update}                       
:ArgumentTypes: {Integer,RealList,Real,RealList,Integer}     
:ReturnType: Manual                                          
:End:

/*:512*//*513:*/
#line 8921 "MathView3D.w"

:Begin:
:Function: mvTranslate3D                          
:Pattern: mvTranslate3D[what_Integer,nx_Real,ny_Real,nz_Real,update_Integer]  
:Arguments: {what,nx,ny,nz,update}                     
:ArgumentTypes: {Integer,Real,Real,Real,Integer}          
:ReturnType: Manual                               
:End:

/*:513*//*515:*/
#line 8934 "MathView3D.w"

/*516:*/
#line 8942 "MathView3D.w"

:Evaluate:
  StylePrint[StringJoin["MathGL3d installed, Version 2.2\\[NewLine]",
   "for ",$OperatingSystem," system ", $System,"\\[NewLine]",
   "\\[Copyright] by Jens-Peer Kuska 1997,98,99\\[NewLine]",
   "with MathLink from Wolfram Research."],"Text"];

/*:516*//*517:*/
#line 8950 "MathView3D.w"

:Evaluate:Needs["Graphics`ThreeScript`"]

/*:517*//*518:*/
#line 8960 "MathView3D.w"

:Evaluate:Begin["Graphics`ThreeScript`Private`"]

:Evaluate:TS0[file_,SurfaceColor[g_RGBColor,___]]:=
  (WriteName[file,"surfacecolor"];
	 WriteTriple[file,Apply[List,g]])

:Evaluate:TS0[file_,SurfaceColor[g_GrayLevel,___]]:=
  (WriteName[file,"surfacecolor"];
	 WriteTriple[file,{g,g,g}])

:Evaluate:Clear[SurfaceBounding]

:Evaluate:SurfaceBounding[file_,
  g_List, { {xmin_,xmax_}, {ymin_,ymax_},{zmin_, zmax_}}] :=
  Block[{},
   WriteName[file, "boundingbox"] ;
   WriteTriple[file, {xmin, ymin, zmin}] ;
   WriteTriple[file, {xmax, ymax, zmax}]
  ]

:Evaluate:SurfaceBounding[file_, g_List, _Symbol] :=
        SurfaceBounding[file, g,PlotRange /. FullOptions[g]]


:Evaluate:End[]

/*:518*//*519:*/
#line 8988 "MathView3D.w"

:Evaluate:
  If[Head[$MVTexturePath]=!=List,
   $MVTexturePath={"."}
  ]
:Evaluate:
  $MVTexturePath::usage=
  "$MVTexturePath is the global list of texture directories."

/*:519*//*520:*/
#line 9000 "MathView3D.w"

:Evaluate:Unprotect["MathView3D`*"];

/*:520*/
#line 8935 "MathView3D.w"

/*440:*/
#line 8441 "MathView3D.w"

:Evaluate:
  MViewPoolSize::usage="MViewPoolSize returns the pool size of mathview3d."

/*:440*//*442:*/
#line 8454 "MathView3D.w"

:Evaluate:MVPutGraphics3D::usage="MVPutGraphics3D[lst] is internal used."

/*:442*//*444:*/
#line 8466 "MathView3D.w"

:Evaluate:MVClear::usage="MVClear[] clears the MathGL3d window."

/*:444*//*446:*/
#line 8478 "MathView3D.w"

:Evaluate:MVReset::usage="MVReset[] reset the view parameter in the MathGL3d window."

/*:446*//*448:*/
#line 8490 "MathView3D.w"

:Evaluate:MVShowWindow::usage=
   "MVShowWindow[] show the MathGL3d window if it was hidden."

/*:448*//*450:*/
#line 8503 "MathView3D.w"

:Evaluate:MVHideWindow::usage="MVHideWindow[] hides the MathGL3d window."

/*:450*//*452:*/
#line 8515 "MathView3D.w"

:Evaluate:MVIconifyWindow::usage="MVIconifyWindow[] iconifys the MathGL3d window."

/*:452*//*454:*/
#line 8527 "MathView3D.w"

:Evaluate:MVPopWindow::usage=
   "MVPopWindow[] pops up the MathGL3d window."

/*:454*//*456:*/
#line 8540 "MathView3D.w"

:Evaluate:MVPushWindow::usage=
   "MVPushWindow[] pushs the MathGL3d window down."


/*:456*//*458:*/
#line 8554 "MathView3D.w"



/*:458*//*460:*/
#line 8566 "MathView3D.w"

:Evaluate:MVZoom::usage="MVZoom[value] zooms in (value<1) or out (value>1)."

/*:460*//*463:*/
#line 8583 "MathView3D.w"

:Evaluate:MVSpin::usage="MVSpin[angle,axis] starts spinning of the object
  in the MathGL3d window around the axis {x,y,z} with angle increment angle."

/*:463*//*465:*/
#line 8596 "MathView3D.w"

:Evaluate:MVReshapeWindow::usage="MVReshape[w,h] sets a new width w and height for 
  the MathGL3d window."

/*:465*//*467:*/
#line 8609 "MathView3D.w"

:Evaluate:MVGetPlotRange::usage="MVGetPlotRange[] returns the plot range used by MathGL3d."

/*:467*//*469:*/
#line 8621 "MathView3D.w"

:Evaluate:MVGetViewPoint::usage="MVGetViewPoint[] returns the view point used by MathGL3d."

/*:469*//*471:*/
#line 8633 "MathView3D.w"

:Evaluate:mvRawGetLightSources::usage="mvRawGetLightSources[] is used internal by MathGL3d."

/*:471*//*473:*/
#line 8645 "MathView3D.w"

:Evaluate:mvRawWritePOVRay::usage="mvRawWritePOVRay[] is used internaly by MathGL3d."

/*:473*//*475:*/
#line 8658 "MathView3D.w"

:Evaluate:mvWrite3DMF::usage="mvWrite3DMF[] is used internaly by MathGL3d."

/*:475*//*477:*/
#line 8671 "MathView3D.w"

:Evaluate:mvWriteDXF::usage="mvWriteDXF[] is used internaly by MathGL3d."

/*:477*//*479:*/
#line 8683 "MathView3D.w"

:Evaluate:mvWritePostScript::usage=
  "mvWritePostScript[] is used internaly by MathGL3d."

/*:479*//*481:*/
#line 8697 "MathView3D.w"

:Evaluate:mvRawWriteVRML::usage=
  "mvRawWriteVRML[] is used internaly by MathGL3d."

/*:481*//*483:*/
#line 8710 "MathView3D.w"

:Evaluate:MVGetGraphics3D::usage=
  "MVGetGraphics3D[] will return the graphics data from MathGL3d."

/*:483*//*485:*/
#line 8723 "MathView3D.w"

:Evaluate:MVGetRasterImage::usage=
  "MVGetRasterImage[] will return RasterArray[] image of the rendered scene."


/*:485*//*487:*/
#line 8737 "MathView3D.w"

:Evaluate:mvPolygonReduce::usage=
  "mvPolygonReduce[tol,iter] will reduce the number of polygons in the scene."

/*:487*//*489:*/
#line 8750 "MathView3D.w"

:Evaluate:MVLoad3Script::usage=
  "MVLoad3Script[tsfile] will read a ThreeScript file into the MathGL3d window."


/*:489*//*491:*/
#line 8764 "MathView3D.w"

:Evaluate:MVLoadDXF::usage=
  "MVLoadDXF[dxffile] will read a DXF file into the MathGL3d window."

/*:491*//*493:*/
#line 8777 "MathView3D.w"

:Evaluate:MVLoadOFF::usage=
  "MVLoadOFF[dxffile] will read a OFF file into the MathGL3d window."

/*:493*//*495:*/
#line 8790 "MathView3D.w"

:Evaluate:mvWriteOFF::usage=
  "mvWriteOFF[dxffile] is used internal."

/*:495*//*497:*/
#line 8803 "MathView3D.w"

:Evaluate:mvRawWritePNG::usage=
  "mvRawWritePNG[pngfile,gm,interlace] is used internal."

/*:497*//*499:*/
#line 8817 "MathView3D.w"

:Evaluate:mvRawWriteDepthPNG::usage=
   "mvRawWriteDepthPNG[pngfile,fmt,gm,interlace,scale] is used internal."

/*:499*//*501:*/
#line 8832 "MathView3D.w"

:Evaluate:mvRawWriteSIRDS::usage=
  "mvRawWriteSIRDS[pngfile,dpi,twidth,eyes,mu] is used internal."

/*:501*//*503:*/
#line 8845 "MathView3D.w"

:Evaluate:mvRawLoadPNGTexture::usage=
  "mvRawLoadPNGTexture[pngfile,gm] is used internal."

/*:503*//*505:*/
#line 8858 "MathView3D.w"

:Evaluate:mvRawPutMathTexture::usage=
   "mvRawPutMathTexture[clip,displaystring] is used internal."

/*:505*//*507:*/
#line 8872 "MathView3D.w"

:Evaluate:mvRawPutRasterTexture::usage=
  "mvRawPutRasterTexture[bc,w,h,l] is used internal."

/*:507*//*509:*/
#line 8885 "MathView3D.w"

:Evaluate:MVGetTextureInformation::usage=
  "MVGetTextureInformation[] returns the names and parameters\n
 of the textures used by MathGL3d."

/*:509*//*511:*/
#line 8908 "MathView3D.w"

:Evaluate:mvRawMapTexture::usage=
  "mvRawMapTexture[gid,texid,maptype,t,rx,ra,s,su,sv] is used internal."

/*:511*//*514:*/
#line 8930 "MathView3D.w"

:Evaluate:mvRotate3D::usage=
  "mvRotate3D[] is used internal."

/*:514*//*698:*/
#line 11496 "MathView3D.w"

:Evaluate:MVScene::usage=
  "MVScene[] is used as reference for the data in the scene for texture mapping."

:Evaluate:MVMesh3D::usage=
  "MVMesh3D[id] is used as reference for mesh number id texture mapping."

:Evaluate:MVTexture2D::usage=
  "MVTexture2D[id] is used as reference for texture number id texture mapping."

:Evaluate:
  Map[
    (#::usage=StringJoin[ToString[#]," is used as texture type identifyer."]) &,
    {MVLuminantTexture,MVLuminantAlphaTexture,MVRGBTexture,MVRGBAlphaTexture}
   ]

/*:698*/
#line 8936 "MathView3D.w"

/*521:*/
#line 9003 "MathView3D.w"

:Evaluate:
   BeginPackage["MathView3D`",{"Utilities`FilterOptions`"}]

/*524:*/
#line 9035 "MathView3D.w"

:Evaluate: MVClose::usage="MVClose[] will uninstall the MathGL3d program."

/*:524*//*526:*/
#line 9044 "MathView3D.w"

:Evaluate:MeshGraphics3D::usage="MeshGraphics3D[] is a new graphics object,
  repesenting a rectangular mesh of points."

/*:526*//*548:*/
#line 9485 "MathView3D.w"

:Evaluate:
  MVPutTexture::usage=
  "MVPutTexture[tex, opt...] will load a texture into MathGL3d.\n
 Possible textures are PNG files, Mathematica graphics and RasterArrays.\n
 The MVTextureSize option works only for Mathematica graphics textures\n
 and the MVDisplayGamma option only for PNG files. The MVClipBorderColor
 option works only for Mathematica graphics."

:Evaluate:
  MVTextureSize::usage=
  "MVTextureSize is an option for MVPutTexture when a Mathematica graphics
 is send to MathGL3d. The texture size is scaled inside the OpenGL viewer
 if the texture width and height are not powers of 2. The default size is
 64 x 64."

:Evaluate:
  MVClipBorderColor::usage=
  "MVClipBorderColor is an option for MVPutTexture when a Mathematica graphics
 is send to MathGL3d. The Display[] and DisplayString[] functions add a small
 border of 2-3 pixels to the graphics this option is used to remove this border.
 The default setting is True."

/*:548*//*554:*/
#line 9648 "MathView3D.w"

:Evaluate:
  MVApplyTexture::usage=
  "MVApplyTexture[target, opt...] will map a texture given by the
 MVTexture option to the target. The target may be the scene or a
 mesh inside the MathGL3d window. Possible options are:\n
 MVTexture                 a texture object\n
 MVTextureMapType          the type of the texture mapping\n
 MVTranslateTextureTarget  the translation vector to move the traget\n
 MVRotateTextureTarget     axis and angle to rotate the target\n
 MVScaleTextureTarget      the scaling of the target\n
 MVScaleTexture            the scaling of the texture."

/*:554*//*555:*/
#line 9661 "MathView3D.w"

:Evaluate:
  Map[
    (#::usage=StringJoin[
                ToString[#],
                " is an option for MVApplyTexture and MVShow3D."]) &,
    {MVTexture,MVTextureMapType,MVTranslateTextureTarget,
     MVRotateTextureTarget,MVScaleTextureTarget,MVScaleTexture}
   ]

/*:555*//*561:*/
#line 9730 "MathView3D.w"

:Evaluate:
  MVWorldLightSources::usage=
  "MVWorldLightSources[graph3d] calculates the light sources with positions
  in the coordinate system of the ViewPoint of the graphics 3d object graph3d."


/*:561*//*564:*/
#line 9754 "MathView3D.w"

:Evaluate:
  MVDisplayLightSources::usage=
  "MVDisplayLightSources[lightsources] will calculate the light source
  position for Mathematica's LightSource option from positions
  in the coordinate system of the ViewPoint.\n
  MVDisplayLightSources[ls,graph3d] will return the light sources using
  the options of the graphics 3d object graph3d."

/*:564*//*567:*/
#line 9772 "MathView3D.w"

:Evaluate:MVGetLightSources::usage=
 "MVGetLightSources[] returns the list of light sources used by MathGL3d."

/*:567*//*577:*/
#line 9961 "MathView3D.w"

:Evaluate:
  MVShow3D::usage=
    "MVShow3D[graph3d, opt...] will transfer a Graphics3D object to MathGL3d."
:Evaluate:
 MVLineTubeSize::usage="MVLineTubeSize is an option for MVShow3D
 and MVWritePOVRay. The number given determines the absolute radius
 of the tubes arround lines."
:Evaluate:
 MVTubeSegments::usage="MVTubeSegments is an option for MVShow3D
 and MVWritePOVRay. The integer determines the number of segments
 for the tube approximation."
:Evaluate:
 MVTubeAngle::usage="MVTubeAngle is an option for MVShow3D
 and MVWritePOVRay. The angle (in degree) is used to determine if
 continued lines can smoothly continued with tubes."
:Evaluate:
 MVPointSphereSize::usage="MVPointSphereSize is an option for MVShow3D
 and MVWritePOVRay. The number gives the absolute size of spheres
 for point drawing."
:Evaluate:
 MVPolygonShading::usage="MVPolygonShading is an option for MVShow3D
 and MVWritePOVRay. The values MVSmooth (default), MVFlat and
 MVWireFrame are possible."
:Evaluate:
   MVSmooth::usage="MVSmooth is a value for MVPolygonShading."
:Evaluate:
   MVFlat::usage="MVFlat is a value for MVPolygonShading."
:Evaluate:
   MVWireFrame::usage="MVWireFrame is a value for MVPolygonShading."
:Evaluate:
   MVGrayBackground::usage="MVGrayBackground is a option for MVShow3D."
:Evaluate:
  MVNewScene::usage="MVNewScene is an option for MVShow3D[], MVNewScene->True
  will clear the data give in previous MVShow3D[] commands,MVNewScene->False
  will add the new data to the scene."

/*:577*//*579:*/
#line 10008 "MathView3D.w"

:Evaluate:MVPasteGraphics::usage=
  "MVPasteGraphics[opts] paste the bitmap image in the MathGL3d window to Mathematica."


/*:579*//*584:*/
#line 10047 "MathView3D.w"

:Evaluate:
 MVWritePOVRay::usage="MVWritePOVRay[fname, opt..] writes the current
 object in the MathGL3d window to the POVRay 3.0 input file fname.
 MVWritePOVRay[fname, graph3d, opt..] transfers the Graphics3D graph3d
 to MathGL3d and savs it as a POVRay file."

/*:584*//*585:*/
#line 10054 "MathView3D.w"

:Evaluate:
 MVWriteDXF::usage="MVWriteDXF[fname,opt..] writes the current object
 in the MathGL3d window to a DXF file fname.\n
 MVWriteDXF[fname,graph3d,opt..] transfer the Graphics3D to MathGL3d
 and save a DXF file. Lines will only exported as as tubes, points as
 spheres."

/*:585*//*592:*/
#line 10146 "MathView3D.w"

:Evaluate:MVWritePNG::usage=
 "MVWritePNG[file,opt] write the image of the MathGL3d window to  \n
 a portable network graphics file. Valid options are MVDisplayGamma\n
 for gamma correction and MVInterlacePNG for interlacing the image.\n
 MVWritePNG[file,gr3d,opts] copys the 3d graphics to the MathGL3d\n
 window and saves the png image."

:Evaluate:MVDisplayGamma::usage=
  "MVDisplayGamma is a option for MVWritePNG[]."

:Evaluate:MVInterlacePNG::usage=
  "MVInterlacePNG is a option for MVWritePNG[] and MVPutTexture[]."

/*:592*//*594:*/
#line 10181 "MathView3D.w"

:Evaluate:MVWrite3DMF::usage=
 "MVWrite3DMF[qdfile,opts] writes the 3d data in the viewer window
 into a QuickDraw3D metafile qdfile. Options are MVQuickDrawLights
 (True or False) to write the lights of the scene and
 MVQuickDrawViewPoint (True or False) to save the view point of the
 scene."

/*:594*//*595:*/
#line 10189 "MathView3D.w"

:Evaluate:MVQuickDrawLights::usage=
  "MVQuickDrawLights is a option for MVWrite3DMF."

/*:595*//*596:*/
#line 10193 "MathView3D.w"

:Evaluate:MVQuickDrawViewPoint::usage=
  "MVQuickDrawViewPoint is a option for MVWrite3DMF."


/*:596*//*598:*/
#line 10230 "MathView3D.w"

:Evaluate:MVWriteVRML::usage=
 "MVWriteVRML[vrfile,opts] writes the 3d data in the viewer window
 into a VRML 2.0 file. MVWriteVRML[vrfile,graph3d,opts] sends
 the graphics to the viewer window and saves the VRML file.\n
 Options are VRMLHeadLight (True or False) to use a head light and
 VRMLExtrusion (True or False) to save thick lines as extrusions
 (True) or expanded polygons (False)."


/*:598*//*599:*/
#line 10240 "MathView3D.w"

:Evaluate:VRMLHeadLight::usage=
  "VRMLHeadLight is an option for MVWriteVRML[]."

/*:599*//*600:*/
#line 10244 "MathView3D.w"

:Evaluate:VRMLExtrusion::usage=
  "VRMLExtrusion is an option for MVWriteVRML[]."

/*:600*//*601:*/
#line 10248 "MathView3D.w"

:Evaluate:MVTextureFilePrefix::usage=
  "MVTextureFilePrefix is an option for MVWriteVRML[].
  The textures are stored as png-files with file names that begin with
  the string given in this option."

/*:601*//*620:*/
#line 10512 "MathView3D.w"

:Evaluate: MVPolygonReduce::usage=
  "MVPolygonReduce[graph3d,tol] will pass the data to MathGL3d and
  perform a polygon reduction. MVPolygonReduce[tol] will simplify the
  number of polygones in the current MVScene[]. The option ReductionLoops|
  is used to determine the number of iterations."

/*:620*//*621:*/
#line 10519 "MathView3D.w"

:Evaluate: ReductionLoops::usage=
  "ReductionLoops is an option for MVPolygonReduce.
  It gives the maximum number of polygon reduction interations."


/*:621*//*629:*/
#line 10647 "MathView3D.w"

:Evaluate: RotationCenter::usage=
  "RotationCenter is an option for MVRotate3D[], it specify a 3d
 point as center of the rotation."

/*:629*//*630:*/
#line 10652 "MathView3D.w"

:Evaluate: MVObject::usage=
 "MVObject is an option for MVRotate3D[] and MVTranslate3D[]. It specify the object
 that will be transformed. Possible values are All, MVScene3D[] and a MVMesh3D[] object."
/*:630*//*631:*/
#line 10656 "MathView3D.w"


:Evaluate: MVUpdateDisplay::usage=
 "MVUpdateDisplay is an option for MVRotate3D[] and MVTranslate3D[]. If True the new
 object positions are show, if False the motion of the objects is not finshed."

/*:631*//*632:*/
#line 10662 "MathView3D.w"

:Evaluate: MVRotate3D::usage=
  "MVRotate3D[graph3d,vec,angle] will rotate the graphics 3d object arround vec by angle."

/*:632*//*633:*/
#line 10666 "MathView3D.w"

:Evaluate: MVTranslate3D::usage=
  "MVTranslate3D[graph3d,move] will translate the graphics 3d object by move."

/*:633*//*634:*/
#line 10670 "MathView3D.w"

:Evaluate:MVCreatePalette::usage=
  "MVCreatePalette[] will create a Mathematica 3.0 palette for MathGL3d."

/*:634*//*638:*/
#line 10697 "MathView3D.w"

:Evaluate:MVWriteEPS::usage=
 "MVWriteEPS[file] will write a encapsulate PostScript file from the current
 contens of the MathGL3d window. MVWriteEPS[file,graph3d,opt] will clear the
 window pass graph3d with options opt to MathGL3d and write a PostScript file."

/*:638*//*639:*/
#line 10703 "MathView3D.w"

:Evaluate:EPSBitmap::usage=
 "EPSBitmap is an option vor MVWriteEPS[].
 With EPSBitmap->True the bitmap image is saved instead of the
 resolution independent image."

/*:639*//*642:*/
#line 10719 "MathView3D.w"

:Evaluate:MVWriteOFF::usage=
 "MVWriteOFF[file] will write a OFF file in Geomviews 3d input file format.
 MVWriteOFF[file,graph3d,opt] will clear the window and pass graph3d to
 MathGL3d and write a OFF file form the data."

/*:642*//*644:*/
#line 10741 "MathView3D.w"

:Evaluate:ExpandTextureFileName::usage=
  "ExpandTextureFileName[string] will search texture
 names with `name` in the $MVTexturePath variable."

/*:644*//*712:*/
#line 11682 "MathView3D.w"

:Evaluate:
  Map[
    (#::usage=StringJoin[ToString[#]," is used as texture mapping identifyer."]) &,
    {MVPlaneMapping,MVCylinderMapping,MVSphereMapping,
     MVTorusMapping,MVMeshUVMapping,MVMeshVUMapping}
   ]

/*:712*/
#line 9007 "MathView3D.w"


:Evaluate: Begin["`Private`"]
:Evaluate: SetAttributes[MVShow3D,ReadProtected]
:Evaluate: SetAttributes[MVWritePOVRay,ReadProtected]
:Evaluate: SetAttributes[MVWriteDXF,ReadProtected]
:Evaluate: SetAttributes[MVClose,ReadProtected]
:Evaluate: SetAttributes[MVCreatePalette,ReadProtected]
:Evaluate: SetAttributes[MVWriteEPS,ReadProtected]
:Evaluate: SetAttributes[MVWritePNG,ReadProtected]
:Evaluate: SetAttributes[MVWrite3DMF,ReadProtected]
:Evaluate: SetAttributes[MVWriteVRML,ReadProtected]
:Evaluate: SetAttributes[MVPolygonReduce,ReadProtected]
:Evaluate: SetAttributes[MVPutTexture,ReadProtected]
:Evaluate: SetAttributes[MVApplyTexture,ReadProtected]


/*:521*/
#line 8937 "MathView3D.w"

/*523:*/
#line 9031 "MathView3D.w"

:Evaluate:MVClose[]:= Uninstall /@
  Select[Links[],StringMatchQ[First[#],"*mathview3d*"] &]

/*:523*//*527:*/
#line 9049 "MathView3D.w"

:Evaluate:Format[_MeshGraphics3D]:="-MeshGraphics3D-"

/*:527*//*528:*/
#line 9054 "MathView3D.w"

:Evaluate:makeMesh[vl_List,col_List]:=
   Module[{l = vl, l1 = Map[RotateLeft, vl], mesh},
     mesh = {l, l1, RotateLeft[l1], RotateLeft[l]};
     mesh = Map[Drop[#, -1]&, mesh, {1}];
     mesh = Map[Drop[#, -1]&, mesh, {2}];
     mesh=Map[
           Polygon[Reverse[#]] &,
             Transpose[
               Flatten[#,1] & /@
                mesh
              ]
          ] //Flatten;
     If[col==={}, Return[mesh]];
     If[(Drop[Dimensions[vl],-1]-{1,1})==Dimensions[col],
       Transpose[{Flatten[col],mesh}]
      ]
    ]

/*:528*//*529:*/
#line 9076 "MathView3D.w"

:Evaluate:MeshGraphics3D /:
   Graphics3D[MeshGraphics3D[n_,m_,data_,color_,{opts___}]]:=
     Graphics3D[makeMesh[data,color],{FilterOptions[Graphics3D,opts]}]
:Evaluate:MeshGraphics3D /:
   Show[m_MeshGraphics3D,opts___?OptionQ]:=(Show[Graphics3D[m],opts];m)
:Evaluate:Unprotect[Show];
:Evaluate:Show[a___,m_MeshGraphics3D,b___]:=
    Show[a,Graphics3D[m],b]
:Evaluate:Protect[Show];
:Evaluate:MeshGraphics3D /:
   Display[file_,m_MeshGraphics3D,args___]:=
     (Display[file,Graphics3D[m],args];m)
:Evaluate:MeshGraphics3D /:
   DisplayString[m_MeshGraphics3D,args___]:=
     (DisplayString[Graphics3D[m],args];m)

:Evaluate: Unprotect[PlotRange]

:Evaluate:
  PlotRange[MeshGraphics3D[arg__]]:=PlotRange /. Last[{arg}]

:Evaluate:
  PlotRange[ms:{_MeshGraphics3D..}]:=
    {Min[First[#]], Max[Last[#]]} & /@
      Transpose /@ Transpose[PlotRange /@ ms]

:Evaluate: Protect[PlotRange]

:Evaluate:
  MeshGraphics3D /: FullOptions[MeshGraphics3D[arg__]]:=
    Module[{opt},
      opt=Last[{arg}];
      Join[
        opt,
        Select[
          Options[Graphics3D],
          FreeQ[opt,First[#]] &
         ]
       ]
     ]

/*:529*//*530:*/
#line 9120 "MathView3D.w"

:Evaluate:makePlotPoints[plotp_]:=		
   Switch[plotp,                  		
     Automatic, {15,15},          		
     _Integer, {plotp, plotp},    		
    {_Integer,_Integer}, plotp,   		
    _, {15,15}                    		
   ]                              		
:Evaluate:addDefault[fromFun_,optName_,opt_List]:=
   Module[{val},
     If[FreeQ[First /@ opt,optName], (* Then *)
       val=optName /. Options[fromFun];
       Append[opt,optName->val], (* Else *)
       opt
      ]
    ]
:Evaluate:protected=Unprotect[ParametricPlot3D];

/*:530*//*531:*/
#line 9143 "MathView3D.w"

:Evaluate:ParametricPlot3D[f:{{_,_,__}..},
            ur:{u_,u0_,u1_},vr:{v_,v0_,v1_},opts___]:=
     Module[{gr,ds},
       ds=DisplayFunction /. {opts} /. Options[ParametricPlot3D];
       gr=ParametricPlot3D[
            #,ur,vr,DisplayFunction->Identity,opts] &/@ f;
       Show[
         Graphics3D[#,FilterOptions[Graphics3D,opts]] & /@ gr,
         DisplayFunction->ds
        ];
       gr
      ]

/*:531*//*532:*/
#line 9158 "MathView3D.w"

:Evaluate:								
   ParametricPlot3D[f:{_,_,__}/; !MatrixQ[f],ur:{u_,u0_,u1_},vr:{v_,v0_,v1_},opts___]:=
     Module[{pltpoints,ff,colorf,cfun,i,j,comp,
             nu,nv,du,dv,data,cdata,prange,options={opts},mgraph},
       pltpoints=PlotPoints /. {opts} /. Options[ParametricPlot3D];
       {nu,nv}=makePlotPoints[pltpoints];
       If[Length[f]==3,(* Then no colors *)
         ff=f;
         colorf={},(* Else split the colors *)
         ff=Drop[f,-1];
         colorf=Last[f]
        ];
       comp=Compiled /. {opts} /. Options[ParametricPlot3D];
       If[comp,
         cfun=Compile[
                {{u,_Real},{v,_Real}},
                  Evaluate[N[ff]]],
         cfun=Function[{u,v},Evaluate[ff]]
        ];
	      du=(u1-u0)/(nu-1);
	      dv=(v1-v0)/(nv-1);
       Off[CompiledFunction::ccr];
       Off[CompiledFunction::cfex];
       data=Table[cfun[N[u],N[v]],
              {u,u0,u1+10*$MachineEpsilon,du},
              {v,v0,v1+10*$MachineEpsilon,dv}];
       data=data /. a_Complex :> Abs[a];
       On[CompiledFunction::ccr];
       On[CompiledFunction::cfex];
       If[colorf=!={},
         colorf=Function[{u,v},Evaluate[colorf]];
         cdata=N[Table[
                 colorf[N[u],N[v]],
                 {u,u0+du/2,u1,du},{v,v0+dv/2,v1,dv}]],
         cdata={}
        ];
       prange={Min[Chop[#]],Max[Chop[#]]} & /@
          Transpose[Flatten[data,1]];
       PrependTo[options,PlotRange->prange];
       options=addDefault[ParametricPlot3D,ViewPoint,options];
       options=addDefault[ParametricPlot3D,ViewVertical,options];
       options=addDefault[ParametricPlot3D,ViewCenter,options];
       options=addDefault[ParametricPlot3D,LightSources,options];
       options=addDefault[ParametricPlot3D,AmbientLight,options];
         mgraph=MeshGraphics3D[
           Sequence @@ Drop[Dimensions[#],-1],
           #,cdata,options
          ] &[data];
       Show[mgraph]
      ]
:Evaluate:Protect /@ protected;

/*:532*//*533:*/
#line 9214 "MathView3D.w"

:Evaluate:surfToMesh[s_SurfaceGraphics]:=
   Module[{mesh=Transpose[First[s]],sopt=Last[s],
           u0,u1,v0,v1,nu,nv,mr,
           du,dv,meshopt,prange},
     mr=MeshRange /. sopt;
     If[Automatic===mr,  (* Then *)
       {{u0,u1},{v0,v1}}= {1,#} & /@ Dimensions[mesh],
     (* Else *)
       {{u0,u1},{v0,v1}}=MeshRange /. sopt
      ];
     {nu,nv}=Dimensions[mesh];
     du=(u1-u0)/(nu-1);
     dv=(v1-v0)/(nv-1);
     mesh=MapIndexed[Append[{u0,v0}+(#2-{1,1})*{du,dv},#1] &,mesh,{2}];
     meshopt=sopt;    (* add the default PlotRange *)
     If[FreeQ[meshopt,BoxRatios], 
       AppendTo[meshopt,BoxRatios->{1,1,1}],
       meshopt=meshopt /. Rule[BoxRatios,_] :> (BoxRatios->{1,1,1});
       ];
     If[MemberQ[{Automatic,All},(PlotRange /. meshopt)],
       prange=Transpose[
                {{u0,v0,Min[#]},{u1,v1,Max[#]}} &[Flatten[First[s]]]
               ];
       meshopt=meshopt /. Rule[PlotRange,_]:>Rule[PlotRange,prange]
      ];
     If[MatchQ[s,SurfaceGraphics[_List,_List]],
       MeshGraphics3D[nu,nv,mesh,{},meshopt],
       MeshGraphics3D[nu,nv,mesh,Transpose[s[[2]]],meshopt]
      ]
    ]

/*:533*//*534:*/
#line 9250 "MathView3D.w"

:Evaluate:GetNotebookDirectory[] :=
  If[$Notebooks,
    GetNotebookDirectory[EvaluationNotebook[]],
    Directory[] <> $PathnameSeparator
   ]

:Evaluate:GetNotebookDirectory[nb_NotebookObject] /; $Notebooks := 
  Module[{fname},
    fname = "FileName" /. 
            NotebookInformation[nb] /. 
            FrontEnd`FileName[dir_, ___] :>  
              ToFileName[ dir];  
    If["FileName" == fname, 
      fname = Directory[] <> $PathnameSeparator 
     ];
    fname 
   ]

:Evaluate:MakeSaveFilename[str_String] := 
  If[StringMatchQ[str, "`*"], 
    GetNotebookDirectory[] <> StringDrop[str, 1], 
    str  
   ]
/*:534*//*535:*/
#line 9277 "MathView3D.w"

:Evaluate:
   hueToRGB[h_?NumericQ,0,b_:1] /; h<0.0 :=RGBColor[b,b,b]

:Evaluate:
   hueToRGB[h_?NumericQ,s_:1,v_:1]:=
	Module[{hh,i,f,p,q,t},
	  hh=Mod[h,1.0];
	  If[hh<0.0, hh+=1.0];
        hh*=6.0;
        i=Floor[hh];
        f=hh-i;
        p=v*(1.0-s);
        q=v*(1.0-s*f);
        t=v*(1.0-s*(1.0-f));
        Switch[i,
	    0, RGBColor[v,t,p],
	    1, RGBColor[q,v,p],
	    2, RGBColor[p,v,t],
	    3, RGBColor[p,q,v],
	    4, RGBColor[t,p,v],
	    5, RGBColor[v,p,q],
          _, RGBColor[0,0,0]
	   ]
	 ]

/*:535*//*536:*/
#line 9304 "MathView3D.w"

:Evaluate: prepMeshColors[{}]:={0,0,{}}
:Evaluate:
   prepMeshColors[l_] /;
        MatrixQ[
          l,
          MemberQ[
            {SurfaceColor,RGBColor,Hue,GrayLevel},
            Head[#]
           ] &]:=
     Module[{trans=Dimensions[l],carr},
       carr=l /. SurfaceColor[c_,___]:> c;
       carr=carr /.
         {Hue[a__]:>hueToRGB[a],
          GrayLevel[a_]:>RGBColor[a,a,a]};
       AppendTo[
         trans,
         Flatten[carr /. RGBColor ->List]
      ]
     ]

/*:536*//*538:*/
#line 9331 "MathView3D.w"

:Evaluate:
  addReplaceOption[Graphics[data_],option_,newval_]:=
    Graphics[data,{option->newval}]
:Evaluate:
  addReplaceOption[Graphics[data_,opts__],option_,newval_] /;
      MemberQ[First /@ {opts},option]:=
    Graphics[data,(opts /. (option -> _) :> (option->newval))]
:Evaluate:
  addReplaceOption[Graphics[data_,opts__],option_,newval_]:=
    Graphics[data,Append[opts,option->newval]]


/*:538*//*539:*/
#line 9347 "MathView3D.w"

:Evaluate:
  cleanMVLink[]:=
    Module[{getlnk},
      getlnk=Part[
               ToExpression[
                 ToString[
	             Definition[Global`MViewPoolSize],
                   InputForm
                  ],
                 InputForm,
                 Hold
                ],
               1,2,1
              ];
      If[LinkReadyQ[getlnk],LinkRead[getlnk]]
     ]

/*:539*//*540:*/
#line 9373 "MathView3D.w"

:Evaluate:Options[MVPutTexture]={
  MVDisplayGamma->1,
  MVTextureSize->Automatic,
  MVClipBorderColor->True};

:Evaluate:
  makeTextureSize[size_]:=
    Switch[size,
      Automatic, {64,64},
      _Integer, {size,size},
      {_Integer,_Integer}, size,
      _,  {32,32}
     ]

:Evaluate:MVPutTexture::noreal="MVDisplayGamma must evaluate to a real number.";

/*:540*//*541:*/
#line 9391 "MathView3D.w"

:Evaluate:
  MVPutTexture[pngfile_String,opts___?OptionQ]:=
    Module[{texfile=pngfile,gamma},
      If[StrinMatchQ[pngfile,"*`"] || StringMatchQ[pngfile,"`*"],
        texfile=ExpandTextureFileName[pngfile],
        texfile=pngfile
       ];
      gamma=N[MVDisplayGamma /. {opts} /. Options[MVPutTexture]];
      If[!NumberQ[gamma],
        Message[MVPutTexture::noreal];
        gamma=1.0;
       ];
      Global`mvRawLoadPNGTexture[texfile,gamma]
     ]

/*:541*//*542:*/
#line 9410 "MathView3D.w"

:Evaluate:If[$VersionNumber>3,(* Then *)
  makeBitmapString[gr_,tsize_]:=
    ExportString[gr,"PBM",ImageSize->tsize],(* Else *)
  makeBitmapString[gr_,tsize_]:=
    DisplayString[gr,"PBM",ImageSize->tsize]
 ]

:Evaluate:
  MVPutTexture[gr_Graphics,opts___?OptionQ]:=
    Module[{tsize,dstr,ggr,clip},
      tsize=MVTextureSize /. {opts} /. Options[MVPutTexture];
      tsize=makeTextureSize[tsize];
      clip=MVClipBorderColor /. {opts} /. Options[MVPutTexture];
      If[TrueQ[clip], clip=1, clip=0];
      ggr=addReplaceOption[
               addReplaceOption[gr,ImageSize,tsize],
               AspectRatio,
               tsize[[2]]/tsize[[1]]
              ];
     dstr=makeBitmapString[ggr,tsize];
     Global`mvRawPutMathTexture[clip,dstr]
    ]

/*:542*//*543:*/
#line 9435 "MathView3D.w"

:Evaluate:
  MVPutTexture[
    gr:(_DensityGraphics|_ContourGraphics|_SurfaceGraphics |_Graphics3D),
    opts___?OptionQ
   ]:=MVPutTexture[Graphics[gr],opts]

:Evaluate:
  MVPutTexture[gr_MeshGraphics3D,opts___?OptionQ]:=
    MVPutTexture[Graphics[Graphics3D[gr]],opts]

/*:543*//*544:*/
#line 9447 "MathView3D.w"

:Evaluate:
  MVPutTexture[ra:{{GrayLevel[_]..}..},opts___]:=
    Global`mvRawPutRasterTexture[
      1,
      Sequence @@(Reverse[Dimensions[ra]]),
      N[Flatten[ra /. GrayLevel -> List]] /. 0-> 0.0
     ]

/*:544*//*545:*/
#line 9457 "MathView3D.w"

:Evaluate:
  MVPutTexture[ra:{{RGBColor[__]..}..},opts___]:=
    Global`mvRawPutRasterTexture[
      3,
      Sequence @@(Reverse[Dimensions[ra]]),
      N[Flatten[ra /. RGBColor -> List]] /. 0->0.0
     ]

/*:545*//*546:*/
#line 9467 "MathView3D.w"

:Evaluate:
  MVPutTexture[ra:{{Hue[__]..}..},opts___]:=
    Global`mvRawPutRasterTexture[
      3,
      Sequence @@(Reverse[Dimensions[ra]]),
      N[Flatten[ra /.
        Hue[a__]:> List @@ hueToRGB[a]]] /. 0->0.0
     ]

/*:546*//*547:*/
#line 9480 "MathView3D.w"

:Evaluate:
  MVPutTexture[RasterArray[ra_,___],opts___]:=
    MVPutTexture[ra /. Hue[a__]:> hueToRGB[a],opts]

/*:547*//*549:*/
#line 9513 "MathView3D.w"

:Evaluate:
 Options[MVApplyTexture]={
  MVTexture->None,
  MVTextureMapType->MVPlaneMapping,
  MVTranslateTextureTarget->Automatic,
  MVRotateTextureTarget->Automatic,
  MVScaleTextureTarget->Automatic,
  MVScaleTexture->Automatic
 };
:Evaluate:
 MVApplyTexture::nodeftex=
  "Texture `1` is not defined inside of MathGL3d.";
:Evaluate:
 MVApplyTexture::invalidtex=
  "MVTexture must be a PNG file, a graphic object or a RasterArray.";
:Evaluate:
 MVApplyTexture::notex=
  "Can't define the texture in MathGL3d.";
:Evaluate:
 MVApplyTexture::unknowmap=
  "MVTextureMapType must be one of `1`, use MVPlaneMapping.";
:Evaluate:
 MVApplyTexture::invmove=
  "MVTranslateTextureTarget must be a list of 3 numbers.";
:Evaluate:
 MVApplyTexture::invrot=
  "MVRotateTextureTarget must be a list of a vector and an angle.";
:Evaluate:
 MVApplyTexture::invscale=
  "MVScaleTextureTarget must be a list of 3 numbers > 0.";
:Evaluate:
 MVApplyTexture::invtscale=
  "MVScaleTexture must be a list of two numbers >0.";

/*:549*//*550:*/
#line 9549 "MathView3D.w"

:Evaluate:
 allgraph={
   Graphics,ContourGraphics,
   DensityGraphics,Graphics3D,
   SurfaceGraphics,MeshGraphics3D,
   RasterArray};

/*:550*//*551:*/
#line 9560 "MathView3D.w"

:Evaluate:
 MVApplyTexture[Global`MVScene3D[],args___]:=MVApplyTexture[0,args]
:Evaluate:
 MVApplyTexture[Global`MVMesh3D[i_Integer],args___]:=MVApplyTexture[i,args]
:Evaluate:
 MVApplyTexture[target_Integer,MVTexture2D[texid_Integer],opts___?OptionQ]:=
   MVApplyTexture[target,texid,opts]

/*:551*//*552:*/
#line 9573 "MathView3D.w"

:Evaluate:
 MVApplyTexture[target_Integer,opts__?OptionQ]:=
   Module[{tex,texid,lst},
     tex=MVTexture /. {opts} /. Options[MVApplyTexture];
     If[tex===None,Return[]];
     If[Head[tex]===Global`MVTexture2D,
       tlst=First /@ MVGetTextureInformation[];
       If[MemberQ[tlst,tex],
         Message[MVApplyTexture::nodeftex,tex];
         Return[];
        ];
       MVApplyTexture[target,First[tex],opts]
      ]
     If[!StringQ[tex] && !MemberQ[allgraph,Head[tex]],
       Message[MVApplyTexture::invalidtex];
       Return[];
      ];
     cleanMVLink[];
     tex=MVPutTexture[tex,FilterOptions[MVPutTexture,opts]];
     If[tex===$Failed,
       Message[MVApplyTexture::notex];
       Return[];
      ];
     texid=First[First[tex]];
       (* get the texture number from {MVTexture2D[id],..} *)
     MVApplyTexture[target,texid,opts]
    ]

/*:552*//*553:*/
#line 9605 "MathView3D.w"

:Evaluate:
 MVApplyTexture[target_Integer,texid_Integer,opts___?OptionQ]:=
  Module[{prange,maptype,move,rot,alpha,scale,scaletex},
    maptype=MVTextureMapType /. {opts} /. Options[MVApplyTexture];
    maptype=maptype /. textureMapToInt;
    If[!IntegerQ[maptype],
      Message[MVApplyTexture::unknowmap,
        First /@ textureMapToInt];
      maptype=If[0==target,2 (* plane *),128 (* mesh *)];
     ];
    move=N[MVTranslateTextureTarget /. {opts} /.Options[MVApplyTexture]] /. 0->0.0;
    rot=N[MVRotateTextureTarget /. {opts} /.Options[MVApplyTexture]] /. 0->0.0;
    scale=N[MVScaleTextureTarget /. {opts} /.Options[MVApplyTexture]] /. 0->0.0;
    scaletex=N[MVScaleTexture /. {opts} /.Options[MVApplyTexture]] /. 0->0.0;
    prange=Global`MVGetPlotRange[];
    If[Automatic===move,move=0.5*(Plus @@ prange)];
    If[Automatic===rot,rot={{0.,0.,1.0},0.0}];
    If[Automatic===scale,
      scale= If[#!=0.0,1./#,#] & /@ (prange[[2]]-prange[[1]])
     ];
    If[Automatic===scaletex,scaletex={1.0,1.0}];
    If[!MatchQ[move,{_Real,_Real,_Real}],
      Message[MVApplyTexture::invmove];
      Return[];
     ];
    If[!MatchQ[rot,{{_Real,_Real,_Real},_Real}],
      Message[MVApplyTexture::invrot];
      Return[];
     ];
    If[!MatchQ[scale,{_Real,_Real,_Real}],
      Message[MVApplyTexture::invscale];
      Return[];
     ];
    If[!MatchQ[scaletex,{_Real,_Real}],
      Message[MVApplyTexture::invtscale];
      Return[];
     ];
    Global`mvRawMapTexture[target,texid,maptype,move,
      Sequence @@ rot,scale,
      Sequence @@ scaletex]
  ]

/*:553*//*557:*/
#line 9677 "MathView3D.w"

:Evaluate:normalizeVec[v_]:=
  Module[{mag,nv=v},
    mag=Sqrt[N[Dot[#,#]]] & [v];
    If[mag>0.0,
	nv/=mag,
	nv]
   ]

/*:557*//*558:*/
#line 9688 "MathView3D.w"

:Evaluate:homogenTranslate[t_List]:=
  Transpose[ReplacePart[IdentityMatrix[4],Append[t,1.],4]]

/*:558*//*559:*/
#line 9695 "MathView3D.w"

:Evaluate:MVLookAtMatrix[viewpnt_List,viewvert_List,viewcntr_]:=
  Module[{vc,x,y,z},
    If[Automatic===viewcntr,
      vc={0.5,0.5,0.5},
      vc=viewcntr
     ];
    z=normalizeVec[viewpnt-vc];
    y=normalizeVec[viewvert];
    x=normalizeVec[Cross[y,z]];
    y=normalizeVec[Cross[z,x]];
    Dot[{Append[x,0.],
	   Append[y,0.],
	   Append[z,0.],{0.,0.,0.,1.0}},
	  homogenTranslate[-0.5*viewpnt]]
  ]

/*:559*//*560:*/
#line 9716 "MathView3D.w"

:Evaluate:
  MVWorldLightSources[
    gr:(_Graphics3D | _SurfaceGraphics | _MeshGraphics3D)]:=
  Module[{ls,vp,vv,vc,opt,hmat},
    opt=FullOptions[gr];
    {ls,vp,vv,vc}=
      {LightSources,ViewPoint,
       ViewVertical,ViewCenter} /. opt;
    hmat=Inverse[MVLookAtMatrix[vp,vv,vc]];
    ls /. {v_?VectorQ,col_} :>
	{Drop[Dot[hmat,Append[v,1.] ],-1],col}
  ]

/*:560*//*562:*/
#line 9740 "MathView3D.w"

:Evaluate:MVDisplayLightSources[ls:{{_List,_}..}]:=
  MVDisplayLightSources[ls,Graphics3D[{}]]

/*:562*//*563:*/
#line 9744 "MathView3D.w"

:Evaluate:MVDisplayLightSources[ls:{{_List,_}..},gr_]:=
  Module[{vp,vv,vc,opt,hmat},
	opt=FullOptions[gr];
	{vp,vv,vc}={ViewPoint,ViewVertical,ViewCenter} /. opt;
	hmat=MVLookAtMatrix[vp,vv,vc];
	ls /. {v_?VectorQ,col_} :>
	  {Drop[Dot[hmat,Append[v,1.] ],-1],col}
   ]

/*:563*//*565:*/
#line 9763 "MathView3D.w"

:Evaluate:MVGetLightSources[]:=
  MVDisplayLightSources[Global`mvRawGetLightSources[]]

/*:565*//*566:*/
#line 9767 "MathView3D.w"

:Evaluate:MVGetLightSources[
  gr:(_Graphics3D | _SurfaceGraphics | _MeshGraphics3D)]:=
  MVDisplayLightSources[Global`mvRawGetLightSources[],gr]

/*:566*//*569:*/
#line 9783 "MathView3D.w"

:Evaluate:
  Options[MVShow3D]={MVNewScene->False,
   MVLineTubeSize->0.01,
   MVTubeSegments->12,
   MVTubeAngle->30,
   MVPointSphereSize-> 0.01,
   MVPolygonShading -> MVSmooth,
   MVGrayBackground->False}

:Evaluate:
   mvShade={MVSmooth,MVFlat,MVWireFrame}

:Evaluate:
   mvShadeID[a_]:=Switch[a,
    MVSmooth,       1,
    MVFlat,         2,
    MVWireFrame,    3,
    _,              1]

/*:569*//*570:*/
#line 9803 "MathView3D.w"

:Evaluate:
  MVShow3D::noint="MVTubeSegments must be an integer."
:Evaluate:
  MVShow3D::noreal="`1` must be a number."

:Evaluate:If[$VersionNumber<4,(* Then *)
  colorRules=Dispatch[
   {SurfaceColor[col_,__] :>  SurfaceColor[col],
    Hue[a__]:> hueToRGB[a],
    Thickness[___] :> Sequence[],
    GrayLevel[a_]:> RGBColor[a,a,a],
    EdgeForm[___]:> Sequence[],
    FaceForm[___]:> Sequence[]}], (* Else *)
   colorRules=Dispatch[
   {SurfaceColor[col_,__] :>  SurfaceColor[col],
    h : (_Hue | _GrayLevel | _CMYKColor) :> 
      ToColor[h, RGBColor], 
    Thickness[___] :> Sequence[],
    EdgeForm[___]:> Sequence[],
    FaceForm[___]:> Sequence[]}] 
  ]

:Evaluate:
   convColors[gr_]:=gr //. colorRules 

/*:570*//*571:*/
#line 9829 "MathView3D.w"

:Evaluate: convertNumbersToC=
  {0->0.0,
   a_Complex :> Abs[a],
   Overflow[]->1.0/$MachineEpsilon};

/*:571*//*572:*/
#line 9835 "MathView3D.w"

:Evaluate:makeShow3DArg[gr_]:=
    Block[{args},
      args=N[{
        pltrange,vp,vert,
        Flatten[lpos],Flatten[lcol],
        List @@ ambient,boxratio,psize,
        ltube,
        tangle} ] //. convertNumbersToC;
    PrependTo[args,newscene];
    args=Join[args,{shade,tseg,graybk}];
    If[Head[gr]===Graphics3D,(* Then *)
      AppendTo[
        args,
        convColors[Flatten[N[{First[gr]}] //.
           convertNumbersToC]]
       ],
     (* Else *)
     AppendTo[args,						
       {MeshGraphics3D[gr[[1]],gr[[2]],
          N[Flatten[gr[[3]]]],
          Sequence @@ prepMeshColors[gr[[4]]]
         ] /. convertNumbersToC }
      ]
    ]
  ]

/*:572*//*573:*/
#line 9862 "MathView3D.w"

:Evaluate:mergePlotRange[pltr_]:=
  Module[{plotr=Transpose[pltr],mvpr},
    mvpr=Global`MVGetPlotRange[];
    If[Indeterminate===First[Union[Flatten[mvpr]]],
      Return[plotr],
    (* Else *)
        {Min /@ Transpose[{plotr[[1]],mvpr[[1]]}],
         Max /@ Transpose[{plotr[[2]],mvpr[[2]]}]}
     ]
   ]

/*:573*//*574:*/
#line 9875 "MathView3D.w"

:Evaluate:
   MVShow3D[surf_SurfaceGraphics,opt___]:=
    MVShow3D[surfToMesh[surf],opt]

/*:574*//*575:*/
#line 9880 "MathView3D.w"

:Evaluate:
   MVShow3D[gr:(_Graphics3D | _MeshGraphics3D),opt___]:=
   Block[{vp,vert,pltrange,light,lpos,lcol,a,col,graybk,
          ambient,ltube,psize,shade,args,tseg,tangle,res,newscene,
          boxratio,gropt},
    If[Head[gr]===Graphics3D, (* Then *)
      If[MatchQ[gr,Graphics3D[_,_]],
        gropt=Last[gr],
        gropt={BoxRatios->Automatic}
       ];
      {pltrange,vp,vert,ambient}=
           FullOptions[gr,#] & /@
              {PlotRange,ViewPoint,ViewVertical,AmbientLight} /.
                GrayLevel[x_] :> RGBColor[x,x,x];
       boxratio=BoxRatios /. gropt /. Options[Graphics3D];,
      (* Else *)
      pltrange=PlotRange[gr];
      {vp,vert,light,ambient,boxratio}=
         {ViewPoint,ViewVertical,LightSources,AmbientLight,BoxRatios} /.
            Last[gr] /.
            GrayLevel[x_]:>RGBColor[x,x,x]
     ];
    If[Automatic===boxratio || !(VectorQ[boxratio,NumericQ] && Length[boxratio]==3),
     boxratio={1.,1.,1.},
     boxratio=#/Max[#] &[Abs[N[boxratio]/. 0->0.]];
     ];
    newscene=MVNewScene /. {opt} /. Options[MVShow3D];
    If[TrueQ[newscene],newscene=1,newscene=0];
    light=MVWorldLightSources[gr];
    If[0===newscene,
      pltrange=mergePlotRange[pltrange],
      pltrange=Transpose[pltrange]
     ];
    pltrange=Partition[Flatten[pltrange],3];
    {lpos,lcol}=
           Partition[Flatten[Transpose[light] /.
             RGBColor[args__] :> List[args],1],
             Length[light]];
    ltube=MVLineTubeSize /. {opt} /.Options[MVShow3D];
    If[!NumericQ[ltube],
      Message[MVShow3D::noreal,"MVLineTubeSize"];
      Return[$Failed];
     ];
    tangle=MVTubeAngle /. {opt} /. Options[MVShow3D];
    If[!NumericQ[tangle],
      Message[MVShow3D::noreal,"MVTubeAngle"];
      Return[$Failed];
     ];
    psize=MVPointSphereSize /. {opt} /.Options[MVShow3D];
    If[!NumericQ[psize],
      Message[MVShow3D::noreal,"MVPointSphereSize"];
      Return[$Failed];
     ];
    shade=MVPolygonShading /. {opt}/.Options[MVShow3D];
    shade=mvShadeID[shade];
    tseg=MVTubeSegments /. {opt} /. Options[MVShow3D];
    graybk=MVGrayBackground /. {opt} /. Options[MVShow3D];
    If[TrueQ[graybk], graybk=1, graybk=0];
    args=makeShow3DArg[gr];
    res=Global`MVPutGraphics3D[args];
    If[$Failed=!=res && MemberQ[First /@ {opt},MVTexture] &&
       None=!=(MVTexture /. {opt}), (* Then *)
     res=MVApplyTexture[
           First[res],
           FilterOptions[MVApplyTexture,opt],
           FilterOptions[MVPutTexture,opt]
          ], (* Else *)
     res
    ]
   ]

/*:575*//*576:*/
#line 9952 "MathView3D.w"

:Evaluate:
  MVShow3D[ms:{_MeshGraphics3D..},opts___]:=
    Module[{rangeall},
      rangeall=PlotRange[ms];
        MVShow3D /@ ( ms /. Rule[PlotRange,_] :>
                             Rule[PlotRange,rangeall])
     ]

/*:576*//*578:*/
#line 9998 "MathView3D.w"

:Evaluate:
 MVPasteGraphics[opt___]:=Show[
   Graphics[
    Global`MVGetRasterImage[],
    AspectRatio->Automatic,
    FilterOptions[Graphics,opt]
   ]
  ]

/*:578*//*581:*/
#line 10018 "MathView3D.w"

:Evaluate: Options[MVWritePOVRay]={MVTextureFilePrefix->"mv"}

/*:581*//*582:*/
#line 10025 "MathView3D.w"

:Evaluate:
  MVWritePOVRay[fname_String,opts___Rule]:=
    Module[{filename,texpre},
      texpre=MVTextureFilePrefix /. {opts} /.Options[MVWritePOVRay];
      filename=StringReplace[
                 fname,
                 "/"->$PathnameSeparator];
      filename=MakeSaveFilename[filename];
      Global`mvRawWritePOVRay[filename,texpre<>"s",texpre<>"m"]
     ]

/*:582*//*583:*/
#line 10039 "MathView3D.w"

:Evaluate:
  MVWritePOVRay[fname_String,
    gr:(_Graphics3D|_SurfaceGraphics | _MeshGraphics3D),
    opts___?OptionQ]:=
    (Global`MVClear[];MVShow3D[gr,opts];
    MVWritePOVRay[fname,opts])

/*:583*//*586:*/
#line 10063 "MathView3D.w"

:Evaluate:Options[MVWriteDXF]={MVLineTubeSize->0.01,
		MVTubeSegments->12,
		MVPointSphereSize->0.01,
		MVTubeAngle->30.}

/*:586*//*587:*/
#line 10069 "MathView3D.w"

:Evaluate: MVWriteDXF::noreal="`1` must be a floating point number"

/*:587*//*588:*/
#line 10073 "MathView3D.w"

:Evaluate:MVWriteDXF[fname_String,opt___?OptionQ]:=
	  Module[{ts,tseg,psize,tangle,filename},
	    ts=MVLineTubeSize /. {opt} /. Options[MVWriteDXF];
	    tseg=MVTubeSegments /. {opt} /. Options[MVWriteDXF];
	    psize=MVPointSphereSize /. {opt} /. Options[MVWriteDXF];
	    tangle=MVTubeAngle /. {opt} /. Options[MVWriteDXF];
	    {ts,psize,tangle}=N[{ts,psize,tangle}] /. 0-> 0.0;
	    If[!IntegerQ[tseg],
	      Message[MVShow3D::noint,"MVTubeSegments"];
	      Return[$Failed];
	     ];
	    If[!NumericQ[ts],
	      Message[MVWriteDXF::noreal,"MVLineTubeSize"];
	      Return[$Failed];
	     ];
	    If[!NumericQ[psize],
	      Message[MVWriteDXF::noreal,"MVPointSphereSize"];
	      Return[$Failed];
	     ];
	    If[!NumericQ[tangle],
	      Message[MVWriteDXF::noreal,"MVTubeAngle"];
	      Return[$Failed];
	     ];
	    filename=StringReplace[fname,"/" -> $PathnameSeparator];
      filename=MakeSaveFilename[filename];
	    Global`mvWriteDXF[filename,ts,tseg,psize,tangle]
	   ]

/*:588*//*589:*/
#line 10103 "MathView3D.w"

:Evaluate:MVWriteDXF[fname_String,
	          gr_ /;
                         Head[gr]===Graphics3D ||
                         Head[gr]===SurfaceGraphics ||
                         Head[gr]===MeshGraphics3D,
	            opts___?OptionQ]:=
	          (Global`MVClear[];
	           MVShow3D[gr,opts];MVWriteDXF[fname,opts])

/*:589*//*590:*/
#line 10114 "MathView3D.w"

:Evaluate:Options[MVWritePNG]=
            {MVDisplayGamma->1.0,MVInterlacePNG->False}

:Evaluate:MVWritePNG::noreal="MVDisplayGamma must be an real number."
:Evaluate:MVWritePNG::nobool="MVInterlacePNG must be True or False."

:Evaluate:MVWritePNG[fname_String,opt___?OptionQ]:=
  Module[{interl,gamma,filename},
    gamma=MVDisplayGamma /. {opt} /. Options[MVWritePNG];
    interl=MVInterlacePNG /. {opt} /. Options[MVWritePNG];
    gamma=N[gamma];
    interl= interl /. {True->1,False->0};
    If[!NumericQ[gamma],
	     Message[MVWritePNG::noreal];
	     Return[$Failed];
     ];
    If[!MemberQ[{0,1},interl],
	     Message[MVWritePNG::nobool];
	     Return[$Failed];
     ];
	  filename=StringReplace[fname,"/" -> $PathnameSeparator];
    filename=MakeSaveFilename[filename];
    Global`mvRawWritePNG[filename,gamma,interl]
   ]

/*:590*//*591:*/
#line 10140 "MathView3D.w"

:Evaluate:MVWritePNG[fname_String,
            gr:(_Graphics3D|_SurfaceGraphics | _MeshGraphics3D),
            opts___?OptionQ]:=
	  (MVClear[];MVShow3D[gr,opts];MVWritePNG[fname,opts])

/*:591*//*593:*/
#line 10161 "MathView3D.w"

:Evaluate:Options[MVWrite3DMF]=
   {MVQuickDrawLights->True,MVQuickDrawViewPoint->True};

:Evaluate:MVWrite3DMF[fname_String,opts___?OptionQ]:=
   Module[{wlight,wview,filename},
     wlight=MVQuickDrawLights /. {opts} /. Options[MVWrite3DMF];
     wview=MVQuickDrawViewPoint /. {opts} /. Options[MVWrite3DMF];
     If[TrueQ[wlight],wlight=1,wlight=0];
     If[TrueQ[wview],wview=1,wview=0];
	   filename=StringReplace[fname,"/" -> $PathnameSeparator];
     filename=MakeSaveFilename[filename];
     Global`mvWrite3DMF[filename,wlight,wview]
    ]

:Evaluate:MVWrite3DMF[fname_String,
            gr:(_Graphics3D|_SurfaceGraphics|_MeshGraphics3D),
            opts___?OptionQ]:=
	  (MVClear[];MVShow3D[gr,opts];MVWrite3DMF[fname,opts])

/*:593*//*597:*/
#line 10199 "MathView3D.w"

:Evaluate:Options[MVWriteVRML]=
           {VRMLHeadLight->False,VRMLExtrusion->True,
            MVTextureFilePrefix->"mv"}

:Evaluate:
  MVWriteVRML[fname_String,opts___?OptionQ]:=
    Module[{vrHeadlight,vrUseExtrude,texpre,filename},
	     vrHeadlight=VRMLHeadLight /. {opts} /.
                    Options[MVWriteVRML];
	     vrUseExtrude=VRMLExtrusion /. {opts} /.
                    Options[MVWriteVRML];
	     texpre=MVTextureFilePrefix /. {opts} /.
                    Options[MVWriteVRML];
	     If[TrueQ[vrHeadlight],vrHeadlight=1,vrHeadlight=0];
      If[TrueQ[vrUseExtrude],vrUseExtrude=1,vrUseExtrude=0];
	    filename=StringReplace[fname,"/" -> $PathnameSeparator];
      filename=MakeSaveFilename[filename];
      Global`mvRawWriteVRML[
        filename,vrHeadlight,vrUseExtrude,texpre<>"s",texpre<>"m"]
     ]


:Evaluate:
   MVWriteVRML[fname_String,
     gr:(_Graphics3D|_SurfaceGraphics | _MeshGraphics3D),
     opts___?OptionQ]:=
     (MVClear[];MVShow3D[gr,opts];
        MVWriteVRML[fname,FilterOptions[MVWriteVRML,opts]])


/*:597*//*603:*/
#line 10255 "MathView3D.w"

:Evaluate:
  mvPaletteHead =
   {
    FrameBox[StyleBox[RowBox[{"Actions"}]]],
    FrameBox[StyleBox[RowBox[{"Main"}   ]]],
    FrameBox[StyleBox[RowBox[{"Spin"}   ]]],
    FrameBox[StyleBox[RowBox[{"Window"} ]]]
   };
:Evaluate:
  mvShowButton = ButtonBox[
    StyleBox[
    RowBox[{"MVShow3D","[","\\[Placeholder]","]"}]],
    ButtonData :>
      BoxData[
        RowBox[{"MVShow3D","[","\[Placeholder]",",",
          "MVLineTubeSize","->","0.01",",",
          "MVPolygonShading","->","MVSmooth",",",
          "MVTubeSegments","->","12",",",
          "MVTubeAngle","->","30",",",
          "MVPointSphereSize","->","0.01","]"}]
         ]];

/*:603*//*604:*/
#line 10278 "MathView3D.w"

:Evaluate:
  mvVRMLButton = ButtonBox[
  StyleBox[
    RowBox[{"MVWriteVRML","[","\\[Placeholder]","]"}]],
    ButtonData :>
      BoxData[
        RowBox[{
          "MVWriteVRML","[","\[Placeholder]",",",
          "\[Placeholder]",",",
          "VRMLHeadLight","->","False",",",
          "VRMLExtrusion","->","True",",",
          "MVPolygonShading","->","MVSmooth",
          "]"}]
         ]];

/*:604*//*605:*/
#line 10294 "MathView3D.w"

:Evaluate:
  mv3DMFButton = ButtonBox[
    StyleBox[
      RowBox[{"MVWrite3DMF","[","\\[Placeholder]","]"}]],
      ButtonData :>
      BoxData[
        RowBox[{
          "MVWrite3DMF","[","\[Placeholder]",",",
          "\[Placeholder]",",",
          "MVQuickDrawLights","->","True",",",
          "MVQuickDrawViewPoint","->","True",",",
          "MVPolygonShading","->","MVSmooth",
         "]"}]]];

/*:605*//*606:*/
#line 10309 "MathView3D.w"

:Evaluate:
  mvPNGButton = ButtonBox[
   StyleBox[
     RowBox[{"MVWritePNG","[","\\[Placeholder]","]"}]],
     ButtonData :>
       BoxData[
         RowBox[{
           "MVWritePNG","[","\[Placeholder]",",",
           "\[Placeholder]",",",
           "MVDisplayGamma","->","True",",",
           "MVInterlacePNG","->","True",",",
           "MVPolygonShading","->","MVSmooth",
           "]"}]]];

/*:606*//*607:*/
#line 10324 "MathView3D.w"

:Evaluate:
  mvEPSButton = ButtonBox[
  StyleBox[
    RowBox[{"MVWriteEPS","[","\\[Placeholder]","]"}]],
     ButtonData :>
      BoxData[
       RowBox[{
        "MVWriteEPS","[","\[Placeholder]",",",
        "\[Placeholder]",",",
        "MVPolygonShading","->","MVSmooth",",",
        "MVGrayBackground","->","True",
        "]"}]]];

/*:607*//*608:*/
#line 10338 "MathView3D.w"

:Evaluate:
  mvSPOVButton = ButtonBox[
   StyleBox[
     RowBox[{"Smooth"," ","MVWritePOVRay","[","\\[Placeholder]","]"}]],
      ButtonData :>
       BoxData[
        RowBox[{"MVWritePOVRay", "[", "\\[Placeholder]",",",
         "MVPolygonShading", "->", "MVSmooth", "]"}]]];
:Evaluate:
  mvFPOVButton = ButtonBox[
  StyleBox[
    RowBox[{"Flat"," ","MVWritePOVray","[","\\[Placeholder]","]"}]],
    ButtonData :>
      BoxData[
        RowBox[{"MVWritePOVRay", "[", "\\[Placeholder]", ",",
          "MVPolygonShading", "->", "MVFlat", "]"}]]];

/*:608*//*609:*/
#line 10356 "MathView3D.w"

:Evaluate:
  mvClearButton = ButtonBox[
   StyleBox[
    RowBox[{"Clear", " ", "Window"}]],
    ButtonFunction :> Global`MVClear[],
    ButtonEvaluator -> Automatic];

:Evaluate:
  mvResetButton = ButtonBox[
   StyleBox[RowBox[{"Reset", " ", "view"}]],
   ButtonFunction :> Global`MVReset[],
   ButtonEvaluator -> Automatic];

:Evaluate:
  mvZoomInButton = ButtonBox[
    StyleBox[RowBox[{"Zoom", " ",  "in"}]],
    ButtonFunction :> Global`MVZoom[0.5],
    ButtonEvaluator -> Automatic];

:Evaluate:
  mvZoomOutButton = ButtonBox[
    StyleBox[RowBox[{"Zoom", " ", "out"}]],
    ButtonFunction :> Global`MVZoom[2.0],
    ButtonEvaluator -> Automatic];

/*:609*//*610:*/
#line 10382 "MathView3D.w"

:Evaluate:
  mvSpinXButton =  ButtonBox[
    StyleBox[RowBox[{"X-axis"}]],
     ButtonFunction :>  (
       Global`MVPopWindow[];
       Global`MVSpin[
       N[Pi/10],
       {1.0, 0.0, 0.0}]
      ),ButtonEvaluator->Automatic];

/*:610*//*611:*/
#line 10393 "MathView3D.w"

:Evaluate:
  mvSpinYButton=ButtonBox[
   StyleBox[RowBox[{"Y-axis"}]],
    ButtonFunction :>  (
     Global`MVPopWindow[];
     Global`MVSpin[
     N[Pi/10],
     {0.0, 1.0, 0.0}]
    ),ButtonEvaluator->Automatic];
:Evaluate:
  mvSpinZButton= ButtonBox[
   StyleBox[RowBox[{"Z-axis"}]],
    ButtonFunction :>  (
     Global`MVPopWindow[];
     Global`MVSpin[
     N[Pi/10],
     {0.0, 0.0, 1.0}]
    ), ButtonEvaluator -> Automatic];

/*:611*//*612:*/
#line 10413 "MathView3D.w"

:Evaluate:
  mvSpinHaltButton =
    ButtonBox[
      StyleBox[RowBox[{"Stop"}]],
      ButtonFunction :>  Global`MVStopSpin[],
      ButtonEvaluator -> Automatic];
:Evaluate:
  mvHideWindowButton =
    ButtonBox[
      StyleBox[RowBox[{"Hide"}]],
      ButtonFunction :>  Global`MVHideWindow[],
      ButtonEvaluator -> Automatic];
:Evaluate:
  mvShowWindowButton =
    ButtonBox[
      StyleBox[RowBox[{"Show"}]],
      ButtonFunction :>  Global`MVShowWindow[],
      ButtonEvaluator -> Automatic];

/*:612*//*613:*/
#line 10433 "MathView3D.w"

:Evaluate:
  mvIconifyButton =
    ButtonBox[
      StyleBox[RowBox[{"Iconify"}]],
      ButtonFunction :>  Global`MVIconifyWindow[],
      ButtonEvaluator -> Automatic];
:Evaluate:
  mvQuitButton =
    ButtonBox[
      StyleBox[RowBox[{"Quit"}]],
      ButtonFunction :>  MVClose[],
      ButtonEvaluator -> Automatic];

/*:613*//*614:*/
#line 10447 "MathView3D.w"

:Evaluate:
  MVCreatePalette[] :=
    Module[{pal},
    pal = Cell[
     BoxData[
      GridBox[
       {
        { " ", StyleBox[
         RowBox[{"OpenGL",  "Viewer"}],
         FontFamily -> "Helvetica"
         ], " "," "},
        mvPaletteHead,
        {mvShowButton, mvClearButton, mvSpinXButton, mvShowWindowButton},
        {mvSPOVButton, mvResetButton, mvSpinYButton, mvHideWindowButton},
        {mvFPOVButton, mvZoomInButton, mvSpinZButton, mvIconifyButton},
        {mvEPSButton,  mvZoomOutButton, mvSpinHaltButton, mvQuitButton},
         {FrameBox[
            StyleBox[
              RowBox[{"\\[Copyright]", " ", "1997-99",
                            " ", "by", " ",  "Jens-Peer"," ", "Kuska"}]]],
          mvVRMLButton,   mv3DMFButton,  mvPNGButton}},
          RowSpacings -> 0,
          ColumnSpacings -> 0
         ]],"Input", Active -> True, CellFrame -> False];
        FrontEndExecute[
          {FrontEnd`NotebookApply[FrontEnd`InputNotebook[], pal]}
         ]]

/*:614*//*615:*/
#line 10476 "MathView3D.w"

:Evaluate:
  Options[MVPolygonReduce]={ReductionLoops->256}

/*:615*//*616:*/
#line 10480 "MathView3D.w"

:Evaluate:
  MVPolygonReduce::noint="ReductionLoops must be an integer."

/*:616*//*617:*/
#line 10484 "MathView3D.w"

:Evaluate:
  MVPolygonReduce[tol_?NumericQ,opts___?OptionQ]:=
    Module[{tt,iter},
      iter=ReductionLoops /. {opts} /. Options[MVPolygonReduce];
      tt=N[tol] /. 0->0.0;
      If[!IntegerQ[iter],
        Message[MVPolygonReduce::noint];
        Return[$Failed]
       ];
      Global`mvPolygonReduce[tt,iter]
     ]

/*:617*//*618:*/
#line 10497 "MathView3D.w"

:Evaluate:
  MVPolygonReduce[graph3d_Graphics3D,tol_?NumericQ,opts___]:=
    (Global`MVClear[];
     MVShow3D[graph3d,opts];
     MVPolygonReduce[tol,FilterOptions[MVPolygonReduce,opts]])

/*:618*//*619:*/
#line 10504 "MathView3D.w"

:Evaluate:
  MVPolygonReduce[graph3d:(_SurfaceGraphics | _MeshGraphics3D),
    tol_?NumericQ,opts___]:=
    (Global`MVClear[];
     MVShow3D[Graphics3D[graph3d],opts];
     MVPolygonReduce[tol,FilterOptions[MVPolygonReduce,opts]])

/*:619*//*622:*/
#line 10527 "MathView3D.w"

:Evaluate:
    rotationmatrix[axis_?VectorQ,theta_?NumericQ] :=
	Module[{nx,ny,nz,st,ct,norm},
    norm=Sqrt[Dot[axis,axis]];
    If[norm<$MachineEpsilon, Return[IndentityMatrix[3]] ];
    {nx,ny,nz} = N[axis/norm];
    {ct,st}=N[{Cos[theta],Sin[theta]}];
	   N[{{ nx^2 + (1 - nx^2)*ct,nx*ny*(1 - ct) + nz*st,nx*nz*(1 - ct) - ny*st},
 	     { nx*ny*(1 - ct) - nz*st,ny^2 + (1 - ny^2)*ct,ny*nz*(1 - ct) + nx*st},
	     { nx*nz*(1 - ct) + ny*st,ny*nz*(1 - ct) - nx*st,nz^2 + (1 - nz^2)*ct}}]
	]

/*:622*//*623:*/
#line 10541 "MathView3D.w"

:Evaluate:Options[MVRotate3D]={
  RotationCenter->{0.,0.,0.},
  MVObject->All,
  MVUpdateDisplay->True }

:Evaluate:moveObjToNumber[obj_]:=
  Switch[obj,
     All, 0, 
     Global`MVScene3D[], 1, 
     Global`MVMesh3D[_Integer], First[obj], 
     _, $Failed]

:Evaluate:MVRotate3D::invcntr="RotationCenter must be a vector of 3 numbers."
:Evaluate:MVRotate3D::invtrgt="MVObject must be All, MVScene3D[] or MVMesh3D[_]."
:Evaluate:
    MVRotate3D[gr_Graphics3D,axis_?VectorQ,th_?NumericQ,opts___?OptionQ]:=
	Module[{rmat,h,pts,pnt,cntr},
    cntr=N[RotationCenter /. {opts} /. Options[MVRotate3D]] /. 0->0.;
    If[!VectorQ[cntr,NumericQ] && 3===Length[cntr],
      Message[MVRotate3D::invcntr];
      Return[gr];
     ];
	  rmat=rotationmatrix[axis,th];
    gr /.{ (h:Line | Polygon)[pts_] :> h[cntr+Dot[rmat,#-cntr] & /@ pts],
    Point[pnt_] :> Point[Dot[rmat,pnt]]}
	 ]

/*:623*//*624:*/
#line 10570 "MathView3D.w"

:Evaluate:
   MVRotate3D[MeshGraphics3D[n_,m_,mdata_,rargs__],
     axis_?VectorQ,th_?NumericQ,opts___?OptionQ]:=
     Module[{rmat,rot,plrange,cntr},
       cntr=N[RotationCenter /. {opts} /. Options[MVRotate3D]] /. 0->0.;
       If[!VectorQ[cntr,NumericQ] && 3===Length[cntr],
         Message[MVRotate3D::invcntr];
         Return[MeshGraphics3D[n,m,mdata,rargs]];
        ];
	     rmat=rotationmatrix[axis,th];
       rot=Map[cntr+Dot[rmat,#-cntr] &,mdata,{2}];
       plrange={Min[#],Max[#] } & /@
         Transpose[Flatten[rot,1]];
       MeshGraphics3D[n,m,rot,rargs] /.
         (PlotRange->_):>(PlotRange->plrange)
      ]

/*:624*//*625:*/
#line 10588 "MathView3D.w"

:Evaluate:
  MVRotate3D[axis_ /; VectorQ[axis,NumericQ], th_?NumericQ,opts___?OptionQ]:=
    Module[{cntr,target,update,updisplay},
       cntr=N[RotationCenter /. {opts} /. Options[MVRotate3D]] /. 0->0.;
       If[!VectorQ[cntr,NumericQ] && 3===Length[cntr],
         Message[MVRotate3D::invcntr];
         Return[MeshGraphics3D[n,m,mdata,rargs]];
        ];
      target=MVObject /. {opts} /. Options[MVRotate3D];
      updisplay=MVUpdateDisplay /. {opts} /. Options[MVRotate3D];
      If[All===target, update=1, update=0];
      If[TrueQ[updisplay],update+=2];
      target=moveObjToNumber[target];
      If[$Failed===target,
        Message[MVRotate3D::invtrgt]; 
        Return[$Failed]; 
       ];
      Global`mvRotate3D[target,Sequence @@ (N[{axis,th}] /. 0->0.0), cntr,update]
     ]

/*:625*//*626:*/
#line 10610 "MathView3D.w"

:Evaluate:Options[MVTranslate3D]={MVObject->All,MVUpdateDisplay->True}
:Evaluate:
    MVTranslate3D[gr_Graphics3D,move_?VectorQ]:=
	Module[{h,pts,pnt},
    gr /.{ (h:Line | Polygon)[pts_] :> h[move+# & /@ pts],
	         Point[pnt_] :> Point[move+pnt]}
	 ]

/*:626*//*627:*/
#line 10619 "MathView3D.w"

:Evaluate:
  MVTranslate3D[MeshGraphics3D[n_,m_,mdata_,rargs__],
   move_?VectorQ]:=
   Module[{tran,plrange},
   tran=Map[Plus[move,#] &,mdata,{2}];
   plrange={Min[#],Max[#] } & /@
    Transpose[Flatten[tran,1]];
   MeshGraphics3D[n,m,tran,rargs] /.
     (PlotRange->_):>(PlotRange->plrange)
  ]

/*:627*//*628:*/
#line 10631 "MathView3D.w"

:Evaluate:
   MVTranslate3D[move_ /; VectorQ[move,NumericQ],opts___?OptionQ]:=
     Module[{target,update,udisplay},
      target=MVObject /. {opts} /. Options[MVTranslate3D];
      If[All===target, update=1, update=0];
      target=moveObjToNumber[target];
      updisplay=MVUpdateDisplay /. {opts} /. Options[MVTranslate3D];
      If[TrueQ[updisplay],update+=2];
      If[$Failed===target,
        Message[MVRotate3D::invtrgt]; 
        Return[$Failed]; 
       ];
       Global`mvTranslate3D[target,Sequence @@ (N[move] /. 0->0.0),update]
      ]

/*:628*//*635:*/
#line 10674 "MathView3D.w"

:Evaluate:Options[MVWriteEPS]={EPSBitmap->False}

/*:635*//*636:*/
#line 10677 "MathView3D.w"

:Evaluate:MVWriteEPS[fname_String,opts___]:=
  Module[{wbitmap,filename},
    wbitmap=EPSBitmap /. {opts} /. Options[MVWriteEPS];
    If[TrueQ[wbitmap],
      wbitmap=1,
      wbitmap=0
     ];
	  filename=StringReplace[fname,"/" -> $PathnameSeparator];
    filename=MakeSaveFilename[filename];
    Global`mvWritePostScript[filename,wbitmap]
  ]

/*:636*//*637:*/
#line 10690 "MathView3D.w"

:Evaluate:MVWriteEPS[fname_String,
            gr:(_Graphics3D|_SurfaceGraphics |_MeshGraphics3D),
            opts___?OptionQ]:=
	  (MVClear[];MVShow3D[gr,opts];
     MVWriteEPS[fname,FilterOptions[WVWriteEPS,opts]])

/*:637*//*640:*/
#line 10709 "MathView3D.w"

:Evaluate:MVWriteOFF[fname_String]:=
  Global`mvWriteOFF[MakeSaveFilename[StringReplace[fname,"/"->$PathnameSeparator]]]

/*:640*//*641:*/
#line 10713 "MathView3D.w"

:Evaluate:MVWriteOFF[fname_String,
            gr:(_Graphics3D|_SurfaceGraphics |_MeshGraphics3D),
            opts___?OptionQ]:=
	  (MVClear[];MVShow3D[gr,opts];Global`mvWriteOFF[fname])

/*:641*//*643:*/
#line 10725 "MathView3D.w"

:Evaluate:ExpandTextureFileName[s_String /;
   StringMatchQ[s,"*`"] && !StringMatchQ[s,"`*"]]:=StringDrop[s,-1]<>".png"
:Evaluate:ExpandTextureFileName[s_String /;
   StringMatchQ[s,"`*"]]:=
  Module[{filename},
	   filename=ExpandTextureFileName[StringDrop[s,1]];
    First[
      Select[
        FileNames["*.png",StringDrop[ToFileName[#],-1] & /@
            Global`$MVTexturePath],
        StringMatchQ[#,"*"<>filename] &
       ]
     ]
   ]

/*:643*//*711:*/
#line 11673 "MathView3D.w"

:Evaluate:textureMapToInt={
  MVPlaneMapping->    2,
  MVCylinderMapping-> 4,
  MVSphereMapping->   8,
  MVTorusMapping->   16,
  MVMeshUVMapping-> 128,
  MVMeshVUMapping-> 256 }

/*:711*/
#line 8938 "MathView3D.w"

/*522:*/
#line 9024 "MathView3D.w"

:Evaluate: End[] (* Private *)

:Evaluate: EndPackage[]
:Evaluate: Protect["MathView3D`*"]

/*:522*/
#line 8939 "MathView3D.w"



/*:515*/
#line 109 "MathView3D.w"



/*:1*/
